import { __awaiter, __asyncGenerator, __await } from 'tslib';
import { priorityQueue } from 'async';
import { EventEmitter } from 'events';
import Punycode from 'punycode';
import { Transform, Duplex } from 'readable-stream';
import * as SDP from 'sdp';

const MAX_SEQ = Math.pow(2, 32);
const mod = (v, n) => ((v % n) + n) % n;
class StreamManagement extends EventEmitter {
    constructor() {
        super();
        this.allowResume = true;
        this.lastAck = 0;
        this.handled = 0;
        this.unacked = [];
        this.inboundStarted = false;
        this.outboundStarted = false;
        this.id = undefined;
        this.jid = undefined;
        this.allowResume = true;
        this.started = false;
        this.cacheHandler = () => undefined;
        this._reset();
    }
    get started() {
        return this.outboundStarted && this.inboundStarted;
    }
    set started(value) {
        if (!value) {
            this.outboundStarted = false;
            this.inboundStarted = false;
        }
    }
    get resumable() {
        return this.started && this.allowResume;
    }
    load(opts) {
        var _a;
        this.id = opts.id;
        this.allowResume = (_a = opts.allowResume) !== null && _a !== void 0 ? _a : true;
        this.handled = opts.handled;
        this.lastAck = opts.lastAck;
        this.unacked = opts.unacked;
        this.emit('prebound', opts.jid);
    }
    cache(handler) {
        this.cacheHandler = handler;
    }
    bind(jid) {
        return __awaiter(this, void 0, void 0, function* () {
            this.jid = jid;
            yield this._cache();
        });
    }
    enable() {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('send', {
                allowResumption: this.allowResume,
                type: 'enable'
            });
        });
    }
    resume() {
        return __awaiter(this, void 0, void 0, function* () {
            this.emit('send', {
                handled: this.handled,
                previousSession: this.id,
                type: 'resume'
            });
        });
    }
    enabled(resp) {
        return __awaiter(this, void 0, void 0, function* () {
            this.id = resp.id;
            this.handled = 0;
            this.inboundStarted = true;
            yield this._cache();
        });
    }
    resumed(resp) {
        return __awaiter(this, void 0, void 0, function* () {
            this.id = resp.previousSession;
            this.inboundStarted = true;
            yield this.process(resp, true);
            yield this._cache();
        });
    }
    failed(resp) {
        return __awaiter(this, void 0, void 0, function* () {
            // Resumption might fail, but the server can still tell us how far
            // the old session progressed.
            yield this.process(resp);
            // We alert that any remaining unacked stanzas failed to send. It has
            // been too long for auto-retrying these to be the right thing to do.
            for (const [kind, stanza] of this.unacked) {
                this.emit('failed', { kind, stanza });
            }
            this._reset();
            yield this._cache();
        });
    }
    ack() {
        this.emit('send', {
            handled: this.handled,
            type: 'ack'
        });
    }
    request() {
        this.emit('send', {
            type: 'request'
        });
    }
    process(ack, resend = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (ack.handled === undefined) {
                return;
            }
            const numAcked = mod(ack.handled - this.lastAck, MAX_SEQ);
            for (let i = 0; i < numAcked && this.unacked.length > 0; i++) {
                const [kind, stanza] = this.unacked.shift();
                this.emit('acked', { kind, stanza });
            }
            this.lastAck = ack.handled;
            if (resend) {
                const resendUnacked = this.unacked;
                this.unacked = [];
                if (resendUnacked.length) {
                    this.emit('begin-resend');
                    for (const [kind, stanza] of resendUnacked) {
                        this.emit('resend', { kind, stanza });
                    }
                    this.emit('end-resend');
                }
            }
            yield this._cache();
        });
    }
    track(kind, stanza) {
        return __awaiter(this, void 0, void 0, function* () {
            if (kind === 'sm' && (stanza.type === 'enable' || stanza.type === 'resume')) {
                this.handled = 0;
                this.outboundStarted = true;
                yield this._cache();
                return false;
            }
            if (!this.outboundStarted) {
                return false;
            }
            if (kind !== 'message' && kind !== 'presence' && kind !== 'iq') {
                return false;
            }
            this.unacked.push([kind, stanza]);
            yield this._cache();
            return true;
        });
    }
    handle() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.inboundStarted) {
                this.handled = mod(this.handled + 1, MAX_SEQ);
                yield this._cache();
            }
        });
    }
    hibernate() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.resumable) {
                return this.shutdown();
            }
            for (const [kind, stanza] of this.unacked) {
                this.emit('hibernated', { kind, stanza });
            }
        });
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.failed({ type: 'failed' });
        });
    }
    _cache() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.cacheHandler({
                    allowResume: this.allowResume,
                    handled: this.handled,
                    id: this.id,
                    jid: this.jid,
                    lastAck: this.lastAck,
                    unacked: this.unacked
                });
            }
            catch (err) {
                // TODO: Is there a good way to handle this?
                // istanbul ignore next
                console.error('Failed to cache stream state', err);
            }
        });
    }
    _reset() {
        this.id = '';
        this.inboundStarted = false;
        this.outboundStarted = false;
        this.lastAck = 0;
        this.handled = 0;
        this.unacked = [];
    }
}

const TABLE_DATA = {
    'A.1': {
        r: 'hk:if|le:lf|nf:nv|qg:qv|rg:rj|rm:rp|rr:rt|rv:s3|vn:vv|17m:17n|17q:17v|18g:19g|1an:1ao|1cb:1cg|1e5:1ef|1fb:1ff|1fl:1gb|1gd:1gq|1gs:1gu|1hr:1hv|1im:1iv|1ne:1nf|1pd:1pf|1qb:1rv|1ti:280|29q:29r|2ae:2af|2al:2an|2bh:2c0|2cd:2ce|2ch:2ci|2dj:2dl|2dq:2dr|2e5:2e6|2e9:2ea|2ee:2em|2eo:2er|2f4:2f5|2fr:2g1|2g3:2g4|2gb:2ge|2gh:2gi|2hq:2hr|2i3:2i6|2i9:2ia|2ie:2io|2iv:2j5|2jl:2k0|2lq:2lr|2me:2mf|2mh:2mv|2n1:2n5|2ng:2o0|2od:2oe|2oh:2oi|2pk:2pl|2pq:2pr|2q4:2q6|2q9:2qa|2qe:2ql|2qo:2qr|2r2:2r5|2rh:2s1|2sb:2sd|2sm:2so|2t0:2t2|2t5:2t7|2tb:2td|2tq:2tt|2u3:2u5|2ue:2um|2uo:2v6|2vj:300|31q:31t|32e:32k|32n:32v|332:335|33g:341|35q:35t|36e:36k|36n:36t|372:375|37g:381|39q:39t|3a4:3a5|3ae:3am|3ao:3av|3b2:3b5|3bg:3c1|3cn:3cp|3du:3dv|3e7:3e9|3eb:3ee|3f0:3fh|3fl:3g0|3hr:3hu|3is:3k0|3k5:3k6|3kb:3kc|3ke:3kj|3l8:3l9|3lu:3lv|3me:3mf|3mq:3mr|3mu:3nv|3rb:3rg|3sc:3sf|3ud:3ue|3ug:3vv|41j:41l|41q:41v|42q:44v|466:46f|47p:47q|47s:47v|4aq:4au|4d3:4d7|4fq:4fv|4ie:4if|4iu:4iv|4ke:4kf|4lm:4ln|4m6:4m7|4om:4on|4qr:4r0|4rt:4sv|4vl:500|5jn:5jv|5kt:5kv|5nh:5nv|5ol:5ov|5pn:5pv|5qk:5qv|5rk:5rv|5ut:5uv|5va:5vv|60q:60v|63o:63v|65a:7fv|7ks:7kv|7nq:7nv|7om:7on|7ou:7ov|7q6:7q7|7qe:7qf|7ru:7rv|7uk:7ul|7vg:7vh|82j:82m|82o:82u|834:839|83i:83j|84f:84v|85i:86f|87b:87v|89r:89s|8ac:8ai|8c4:8cf|8uf:8vv|917:91v|92b:92v|9gk:9gl|9ju:9jv|9ka:9o0|9oa:9ob|9qj:9ql|9qv:9r0|9sl:9sn|9tv:9uf|9vc:9vf|ao0:bjv|bnk:bnv|bum:bvf|bvs:bvv|c4n:c4o|c80:c84|c9d:c9g|cdo:cff|cgt:cgv|ci4:cig|cjs:cju|cmc:cmf|crn:crq|cuu:cuv|jdm:jfv|17t6:17vv|194d:194f|1967:1avv|1lt4:1lvv|1uhe:1uhf|1ujb:1unv|1uo7:1uoi|1uoo:1uos|1uti:1uui|1va0:1vaf|1vcg:1vch|1ve8:1vef|1vft:1vfv|1vgg:1vgv|1vh4:1vhf|1vi7:1vi8|1vjc:1vjf|1vnt:1vnu|1vtv:1vu1|1vu8:1vu9|1vug:1vuh|1vuo:1vup|1vut:1vuv|1vvf:1vvo|2000:20nv|20p4:20pf|20qb:20vv|2116:2117|212e:3jvv|3k7m:3k7v|3k97:3k99|3keu:3kvv|3l50:3l51|3l53:3l54|3l57:3l58|3l8b:3l8c|3la7:3la9|3ll4:3ll7|3lua:3lud|3m00:3vvt|59mn:5tvv|5ugu:5vvt|6000:7vvt|8000:9vvt|a000:bvvt|c000:dvvt|e000:fvvt|g000:hvvt|i000:jvvt|k000:lvvt|m000:nvvt|o000:pvvt|q000:rvvt|s002:s00v|s040:tvvt',
        s: '9p8|9qc|9qe|9qn|9tg|147|bkq|3l2l|3l4t|3l5d|3l5q|3l5s|3l61|3l64|3l86|3l8l|3l8t|3l9q|3l9v|3la5|3lah|16f|c20|ccf|cnv|cvv|1b0|1c8|1d2|1dq|1h0|1nv|1oe|284|2c4|2d9|2dh|2dt|2eu|2h9|2hh|2hk|2hn|2ht|2it|2k4|2kc|2ke|2ki|2l9|2lh|2lk|2m6|2ma|2o4|2p9|2ph|2qu|2s4|2sh|2sr|2st|2tm|2u9|304|30d|30h|319|31k|325|329|344|34d|34h|359|35k|365|369|36v|384|38d|38h|399|3a9|3c4|3di|3ds|3el|3en|3k3|3k9|3ko|3l0|3l4|3l6|3lc|3lq|3m5|3m7|3q8|3so|3tt|412|418|41b|4g7|4i7|4i9|4in|4ip|4k7|4k9|4lf|4lh|4lv|4m1|4mf|4mn|4nf|4of|4oh|4ov|4q7|h1|5od|5rd|5rh|60f|1upn|1upt|1upv|1uq2|1uq5|1vij|1vj7|1vjl|1vo0|1vv7|20ov|7qo|7qq|7qs|7qu|7tl|7u5|7us|7vl|7vv|sb|sd|s000|t2|97v|uf|9go|9o5'
    },
    'B.1': {
        r: '60b:60d|80b:80d|1vg0:1vgf',
        s: '5d|606|1vnv|830|qf'
    },
    'B.2': {
        m: '5l:ts|6v:3j;3j|9g:39;o7|a9:ls;3e|bv:3j|fg:3a;oc|q5:tp|rq:10;tp|sg:tp;o8;o1|tg:u5;o8;o1|u2:u3|ug:ti|uh:to|ui:u5|uj:ud|uk:ub|ul:u6|um:u0|vg:tq|vh:u1|vi:u3|vl:tl|1c7:1b5;1c2|7km:38;ph|7kn:3k;o8|7ko:3n;oa|7kp:3p;oa|7kq:31;lu|7kr:7j1|7qg:u5;oj|7qi:u5;oj;o0|7qk:u5;oj;o1|7qm:u5;oj;q2|7s0:7o0;tp|7s1:7o1;tp|7s2:7o2;tp|7s3:7o3;tp|7s4:7o4;tp|7s5:7o5;tp|7s6:7o6;tp|7s7:7o7;tp|7s8:7o0;tp|7s9:7o1;tp|7sa:7o2;tp|7sb:7o3;tp|7sc:7o4;tp|7sd:7o5;tp|7se:7o6;tp|7sf:7o7;tp|7sg:7p0;tp|7sh:7p1;tp|7si:7p2;tp|7sj:7p3;tp|7sk:7p4;tp|7sl:7p5;tp|7sm:7p6;tp|7sn:7p7;tp|7so:7p0;tp|7sp:7p1;tp|7sq:7p2;tp|7sr:7p3;tp|7ss:7p4;tp|7st:7p5;tp|7su:7p6;tp|7sv:7p7;tp|7t0:7r0;tp|7t1:7r1;tp|7t2:7r2;tp|7t3:7r3;tp|7t4:7r4;tp|7t5:7r5;tp|7t6:7r6;tp|7t7:7r7;tp|7t8:7r0;tp|7t9:7r1;tp|7ta:7r2;tp|7tb:7r3;tp|7tc:7r4;tp|7td:7r5;tp|7te:7r6;tp|7tf:7r7;tp|7ti:7rg;tp|7tj:th;tp|7tk:tc;tp|7tm:th;q2|7tn:th;q2;tp|7ts:th;tp|7tu:tp|7u2:7rk;tp|7u3:tn;tp|7u4:te;tp|7u6:tn;q2|7u7:tn;q2;tp|7uc:tn;tp|7ui:tp;o8;o0|7uj:tp;o8;o1|7um:tp;q2|7un:tp;o8;q2|7v2:u5;o8;o0|7v3:u5;o8;o1|7v4:u1;oj|7v6:u5;q2|7v7:u5;o8;q2|7vi:7rs;tp|7vj:u9;tp|7vk:ue;tp|7vm:u9;q2|7vn:u9;q2;tp|7vs:u9;tp|858:3i;3j|882:33|883:5g;33|887:ir|889:5g;36|88b:38|88c:38|88d:38|88g:39|88h:39|88i:3c|88l:3e|88m:3e;3f|88p:3g|88q:3h|88r:3i|88s:3i|88t:3i|890:3j;3d|891:3k;35;3c|892:3k;3d|894:3q|898:3q|89c:32|89d:33|89g:35|89h:36|89j:3d|89u:tj|89v:u0|8a5:34|crh:38;3g;31|crj:31;3l|crl:3f;3m|cs0:3g;31|cs1:3e;31|cs2:ts;31|cs3:3d;31|cs4:3b;31|cs5:3b;32|cs6:3d;32|cs7:37;32|csa:3g;36|csb:3e;36|csc:ts;36|csg:38;3q|csh:3b;38;3q|csi:3d;38;3q|csj:37;38;3q|csk:3k;38;3q|ct9:3g;31|cta:3b;3g;31|ctb:3d;3g;31|ctc:37;3g;31|ctk:3g;3m|ctl:3e;3m|ctm:ts;3m|ctn:3d;3m|cto:3b;3m|ctp:3d;3m|ctq:3g;3n|ctr:3e;3n|cts:ts;3n|ctt:3d;3n|ctu:3b;3n|ctv:3d;3n|cu0:3b;u9|cu1:3d;u9|cu3:32;3h|cu6:33;8gl;3b;37|cu7:33;3f;1e|cu8:34;32|cu9:37;3p|cub:38;3g|cud:3b;3b|cue:3b;3d|cun:3g;38|cup:3g;3g;3d|cuq:3g;3i|cus:3j;3m|cut:3n;32|1uo0:36;36|1uo1:36;39|1uo2:36;3c|1uo3:36;36;39|1uo4:36;36;3c|1uo5:3j;3k|1uo6:3j;3k|1uoj:1bk;1bm|1uok:1bk;1b5|1uol:1bk;1bb|1uom:1bu;1bm|1uon:1bk;1bd|3l00:31|3l01:32|3l02:33|3l03:34|3l04:35|3l05:36|3l06:37|3l07:38|3l08:39|3l09:3a|3l0a:3b|3l0b:3c|3l0c:3d|3l0d:3e|3l0e:3f|3l0f:3g|3l0g:3h|3l0h:3i|3l0i:3j|3l0j:3k|3l0k:3l|3l0l:3m|3l0m:3n|3l0n:3o|3l0o:3p|3l0p:3q|3l1k:31|3l1l:32|3l1m:33|3l1n:34|3l1o:35|3l1p:36|3l1q:37|3l1r:38|3l1s:39|3l1t:3a|3l1u:3b|3l1v:3c|3l20:3d|3l21:3e|3l22:3f|3l23:3g|3l24:3h|3l25:3i|3l26:3j|3l27:3k|3l28:3l|3l29:3m|3l2a:3n|3l2b:3o|3l2c:3p|3l2d:3q|3l38:31|3l39:32|3l3a:33|3l3b:34|3l3c:35|3l3d:36|3l3e:37|3l3f:38|3l3g:39|3l3h:3a|3l3i:3b|3l3j:3c|3l3k:3d|3l3l:3e|3l3m:3f|3l3n:3g|3l3o:3h|3l3p:3i|3l3q:3j|3l3r:3k|3l3s:3l|3l3t:3m|3l3u:3n|3l3v:3o|3l40:3p|3l41:3q|3l4s:31|3l4u:33|3l4v:34|3l52:37|3l55:3a|3l56:3b|3l59:3e|3l5a:3f|3l5b:3g|3l5c:3h|3l5e:3j|3l5f:3k|3l5g:3l|3l5h:3m|3l5i:3n|3l5j:3o|3l5k:3p|3l5l:3q|3l6g:31|3l6h:32|3l6i:33|3l6j:34|3l6k:35|3l6l:36|3l6m:37|3l6n:38|3l6o:39|3l6p:3a|3l6q:3b|3l6r:3c|3l6s:3d|3l6t:3e|3l6u:3f|3l6v:3g|3l70:3h|3l71:3i|3l72:3j|3l73:3k|3l74:3l|3l75:3m|3l76:3n|3l77:3o|3l78:3p|3l79:3q|3l84:31|3l85:32|3l87:34|3l88:35|3l89:36|3l8a:37|3l8d:3a|3l8e:3b|3l8f:3c|3l8g:3d|3l8h:3e|3l8i:3f|3l8j:3g|3l8k:3h|3l8m:3j|3l8n:3k|3l8o:3l|3l8p:3m|3l8q:3n|3l8r:3o|3l8s:3p|3l9o:31|3l9p:32|3l9r:34|3l9s:35|3l9t:36|3l9u:37|3la0:39|3la1:3a|3la2:3b|3la3:3c|3la4:3d|3la6:3f|3laa:3j|3lab:3k|3lac:3l|3lad:3m|3lae:3n|3laf:3o|3lag:3p|3lbc:31|3lbd:32|3lbe:33|3lbf:34|3lbg:35|3lbh:36|3lbi:37|3lbj:38|3lbk:39|3lbl:3a|3lbm:3b|3lbn:3c|3lbo:3d|3lbp:3e|3lbq:3f|3lbr:3g|3lbs:3h|3lbt:3i|3lbu:3j|3lbv:3k|3lc0:3l|3lc1:3m|3lc2:3n|3lc3:3o|3lc4:3p|3lc5:3q|3ld0:31|3ld1:32|3ld2:33|3ld3:34|3ld4:35|3ld5:36|3ld6:37|3ld7:38|3ld8:39|3ld9:3a|3lda:3b|3ldb:3c|3ldc:3d|3ldd:3e|3lde:3f|3ldf:3g|3ldg:3h|3ldh:3i|3ldi:3j|3ldj:3k|3ldk:3l|3ldl:3m|3ldm:3n|3ldn:3o|3ldo:3p|3ldp:3q|3lek:31|3lel:32|3lem:33|3len:34|3leo:35|3lep:36|3leq:37|3ler:38|3les:39|3let:3a|3leu:3b|3lev:3c|3lf0:3d|3lf1:3e|3lf2:3f|3lf3:3g|3lf4:3h|3lf5:3i|3lf6:3j|3lf7:3k|3lf8:3l|3lf9:3m|3lfa:3n|3lfb:3o|3lfc:3p|3lfd:3q|3lg8:31|3lg9:32|3lga:33|3lgb:34|3lgc:35|3lgd:36|3lge:37|3lgf:38|3lgg:39|3lgh:3a|3lgi:3b|3lgj:3c|3lgk:3d|3lgl:3e|3lgm:3f|3lgn:3g|3lgo:3h|3lgp:3i|3lgq:3j|3lgr:3k|3lgs:3l|3lgt:3m|3lgu:3n|3lgv:3o|3lh0:3p|3lh1:3q|3lhs:31|3lht:32|3lhu:33|3lhv:34|3li0:35|3li1:36|3li2:37|3li3:38|3li4:39|3li5:3a|3li6:3b|3li7:3c|3li8:3d|3li9:3e|3lia:3f|3lib:3g|3lic:3h|3lid:3i|3lie:3j|3lif:3k|3lig:3l|3lih:3m|3lii:3n|3lij:3o|3lik:3p|3lil:3q|3ljg:31|3ljh:32|3lji:33|3ljj:34|3ljk:35|3ljl:36|3ljm:37|3ljn:38|3ljo:39|3ljp:3a|3ljq:3b|3ljr:3c|3ljs:3d|3ljt:3e|3lju:3f|3ljv:3g|3lk0:3h|3lk1:3i|3lk2:3j|3lk3:3k|3lk4:3l|3lk5:3m|3lk6:3n|3lk7:3o|3lk8:3p|3lk9:3q|3ll8:th|3ll9:ti|3lla:tj|3llb:tk|3llc:tl|3lld:tm|3lle:tn|3llf:to|3llg:tp|3llh:tq|3lli:tr|3llj:ts|3llk:tt|3lll:tu|3llm:tv|3lln:u0|3llo:u1|3llp:to|3llq:u3|3llr:u4|3lls:u5|3llt:u6|3llu:u7|3llv:u8|3lm0:u9|3lmj:u3|3ln2:th|3ln3:ti|3ln4:tj|3ln5:tk|3ln6:tl|3ln7:tm|3ln8:tn|3ln9:to|3lna:tp|3lnb:tq|3lnc:tr|3lnd:ts|3lne:tt|3lnf:tu|3lng:tv|3lnh:u0|3lni:u1|3lnj:to|3lnk:u3|3lnl:u4|3lnm:u5|3lnn:u6|3lno:u7|3lnp:u8|3lnq:u9|3lod:u3|3los:th|3lot:ti|3lou:tj|3lov:tk|3lp0:tl|3lp1:tm|3lp2:tn|3lp3:to|3lp4:tp|3lp5:tq|3lp6:tr|3lp7:ts|3lp8:tt|3lp9:tu|3lpa:tv|3lpb:u0|3lpc:u1|3lpd:to|3lpe:u3|3lpf:u4|3lpg:u5|3lph:u6|3lpi:u7|3lpj:u8|3lpk:u9|3lq7:u3|3lqm:th|3lqn:ti|3lqo:tj|3lqp:tk|3lqq:tl|3lqr:tm|3lqs:tn|3lqt:to|3lqu:tp|3lqv:tq|3lr0:tr|3lr1:ts|3lr2:tt|3lr3:tu|3lr4:tv|3lr5:u0|3lr6:u1|3lr7:to|3lr8:u3|3lr9:u4|3lra:u5|3lrb:u6|3lrc:u7|3lrd:u8|3lre:u9|3ls1:u3|3lsg:th|3lsh:ti|3lsi:tj|3lsj:tk|3lsk:tl|3lsl:tm|3lsm:tn|3lsn:to|3lso:tp|3lsp:tq|3lsq:tr|3lsr:ts|3lss:tt|3lst:tu|3lsu:tv|3lsv:u0|3lt0:u1|3lt1:to|3lt2:u3|3lt3:u4|3lt4:u5|3lt5:u6|3lt6:u7|3lt7:u8|3lt8:u9|3ltr:u3',
        r: '23:2c|2i:2k|2m:2q|60:6m|6o:6u|bo:bp|c1:c2|c6:c7|c9:cb|ce:ch|cj:ck|cm:co|cs:ct|cv:d0|d6:d7|de:df|dh:dj|dn:do|e4:e5|e7:e8|ea:eb|fh:fi|fm:fo|s8:sa|se:sf|sh:t1|t3:tb|100:11f|19h:1am|7o8:7of|7oo:7ot|7p8:7pf|7po:7pv|7q8:7qd|7r8:7rf|7to:7tr|7u8:7ub|7uo:7ur|7v8:7vc|7vo:7vr|89a:89b|8b0:8bf|95m:96f|1vp1:1vpq|2100:2115',
        s: 'v8|va|vc|ve|vk|130|132|134|136|138|13a|13c|13e|13g|13i|13k|13m|13o|13q|13s|13u|140|14a|14c|14e|14g|14i|14k|14m|14o|14q|14s|14u|150|152|154|156|158|15a|15c|15e|15g|15i|15k|15m|15o|15q|15s|15u|161|163|165|167|169|16b|16d|16g|16i|16k|16m|16o|16q|16s|16u|170|172|174|176|178|17a|17c|17e|17g|17i|17k|17o|180|182|184|186|188|18a|18c|18e|80|82|84|86|88|8a|8c|8e|8g|8i|8k|8m|8o|8q|8s|8u|90|92|94|96|98|9a|9c|9e|9i|9k|9m|9p|9r|9t|9v|a1|a3|a5|a7|aa|ac|ae|ag|ai|ak|am|ao|aq|as|au|b0|b2|b4|b6|b8|ba|bc|be|bg|bi|bk|bm|br|bt|c4|d2|d4|d9|dc|dl|ds|ed|ef|eh|ej|el|en|ep|er|eu|f0|f2|f4|f6|f8|fa|fc|fe|fk|fq|fs|fu|g0|g2|g4|g6|g8|ga|gc|ge|gg|gi|gk|gm|go|gq|gs|gu|h0|h2|h4|h6|h8|ha|hc|he|hg|hi|21|22|7g0|7g2|7g4|7g6|7g8|7ga|7gc|7ge|7gg|7gi|2d|7gk|7gm|7go|7gq|7gs|7gu|7h0|7h2|7h4|7h6|7h8|7ha|7hc|7he|7hg|7hi|7hk|7hm|7ho|7hq|7hs|7hu|7i0|7i2|7i4|7i6|7i8|7ia|7ic|7ie|7ig|7ii|7ik|7im|7io|7iq|7is|7iu|7j0|7j2|7j4|7j6|7j8|7ja|7jc|7je|7jg|7ji|7jk|7jm|2e|7jo|7jq|7js|7ju|7k0|7k2|7k4|7k6|7k8|7ka|7kc|7ke|7kg|7ki|7kk|7l0|7l2|7l4|7l6|7l8|7la|7lc|7le|7lg|7li|7lk|7lm|7lo|7lq|7ls|7lu|7m0|7m2|7m4|7m6|7m8|7ma|7mc|7me|7mg|7mi|7mk|7mm|7mo|7mq|2f|7ms|7mu|7n0|7n2|7n4|7n6|7n8|7na|7nc|7ne|7ng|7ni|7nk|7nm|7no|2g|7qp|7qr|7qt|7qv|2h|896|2l|s6|sc|uo|uq|us|uu|v0|v2|v4|v6'
    },
    'B.3': {
        m: '5l:ts|6v:3j;3j|9g:39;o7|a9:ls;3e|bv:3j|fg:3a;oc|q5:tp|sg:tp;o8;o1|tg:u5;o8;o1|u2:u3|ug:ti|uh:to|ul:u6|um:u0|vg:tq|vh:u1|vi:u3|vl:tl|1c7:1b5;1c2|7km:38;ph|7kn:3k;o8|7ko:3n;oa|7kp:3p;oa|7kq:31;lu|7kr:7j1|7qg:u5;oj|7qi:u5;oj;o0|7qk:u5;oj;o1|7qm:u5;oj;q2|7s0:7o0;tp|7s1:7o1;tp|7s2:7o2;tp|7s3:7o3;tp|7s4:7o4;tp|7s5:7o5;tp|7s6:7o6;tp|7s7:7o7;tp|7s8:7o0;tp|7s9:7o1;tp|7sa:7o2;tp|7sb:7o3;tp|7sc:7o4;tp|7sd:7o5;tp|7se:7o6;tp|7sf:7o7;tp|7sg:7p0;tp|7sh:7p1;tp|7si:7p2;tp|7sj:7p3;tp|7sk:7p4;tp|7sl:7p5;tp|7sm:7p6;tp|7sn:7p7;tp|7so:7p0;tp|7sp:7p1;tp|7sq:7p2;tp|7sr:7p3;tp|7ss:7p4;tp|7st:7p5;tp|7su:7p6;tp|7sv:7p7;tp|7t0:7r0;tp|7t1:7r1;tp|7t2:7r2;tp|7t3:7r3;tp|7t4:7r4;tp|7t5:7r5;tp|7t6:7r6;tp|7t7:7r7;tp|7t8:7r0;tp|7t9:7r1;tp|7ta:7r2;tp|7tb:7r3;tp|7tc:7r4;tp|7td:7r5;tp|7te:7r6;tp|7tf:7r7;tp|7ti:7rg;tp|7tj:th;tp|7tk:tc;tp|7tm:th;q2|7tn:th;q2;tp|7ts:th;tp|7tu:tp|7u2:7rk;tp|7u3:tn;tp|7u4:te;tp|7u6:tn;q2|7u7:tn;q2;tp|7uc:tn;tp|7ui:tp;o8;o0|7uj:tp;o8;o1|7um:tp;q2|7un:tp;o8;q2|7v2:u5;o8;o0|7v3:u5;o8;o1|7v4:u1;oj|7v6:u5;q2|7v7:u5;o8;q2|7vi:7rs;tp|7vj:u9;tp|7vk:ue;tp|7vm:u9;q2|7vn:u9;q2;tp|7vs:u9;tp|1uo0:36;36|1uo1:36;39|1uo2:36;3c|1uo3:36;36;39|1uo4:36;36;3c|1uo5:3j;3k|1uo6:3j;3k|1uoj:1bk;1bm|1uok:1bk;1b5|1uol:1bk;1bb|1uom:1bu;1bm|1uon:1bk;1bd',
        r: '23:2c|2i:2k|2m:2q|60:6m|6o:6u|bo:bp|c1:c2|c6:c7|c9:cb|ce:ch|cj:ck|cm:co|cs:ct|cv:d0|d6:d7|de:df|dh:dj|dn:do|e4:e5|e7:e8|ea:eb|fh:fi|fm:fo|s8:sa|se:sf|sh:t1|t3:tb|100:11f|19h:1am|7o8:7of|7oo:7ot|7p8:7pf|7po:7pv|7q8:7qd|7r8:7rf|7to:7tr|7u8:7ub|7uo:7ur|7v8:7vc|7vo:7vr|89a:89b|8b0:8bf|95m:96f|1vp1:1vpq|2100:2115',
        s: 'v8|va|vc|ve|vk|130|132|134|136|138|13a|13c|13e|13g|13i|13k|13m|13o|13q|13s|13u|140|14a|14c|14e|14g|14i|14k|14m|14o|14q|14s|14u|150|152|154|156|158|15a|15c|15e|15g|15i|15k|15m|15o|15q|15s|15u|161|163|165|167|169|16b|16d|16g|16i|16k|16m|16o|16q|16s|16u|170|172|174|176|178|17a|17c|17e|17g|17i|17k|17o|180|182|184|186|188|18a|18c|18e|80|82|84|86|88|8a|8c|8e|8g|8i|8k|8m|8o|8q|8s|8u|90|92|94|96|98|9a|9c|9e|9i|9k|9m|9p|9r|9t|9v|a1|a3|a5|a7|aa|ac|ae|ag|ai|ak|am|ao|aq|as|au|b0|b2|b4|b6|b8|ba|bc|be|bg|bi|bk|bm|br|bt|c4|d2|d4|d9|dc|dl|ds|ed|ef|eh|ej|el|en|ep|er|eu|f0|f2|f4|f6|f8|fa|fc|fe|fk|fq|fs|fu|g0|g2|g4|g6|g8|ga|gc|ge|gg|gi|gk|gm|go|gq|gs|gu|h0|h2|h4|h6|h8|ha|hc|he|hg|hi|21|22|7g0|7g2|7g4|7g6|7g8|7ga|7gc|7ge|7gg|7gi|2d|7gk|7gm|7go|7gq|7gs|7gu|7h0|7h2|7h4|7h6|7h8|7ha|7hc|7he|7hg|7hi|7hk|7hm|7ho|7hq|7hs|7hu|7i0|7i2|7i4|7i6|7i8|7ia|7ic|7ie|7ig|7ii|7ik|7im|7io|7iq|7is|7iu|7j0|7j2|7j4|7j6|7j8|7ja|7jc|7je|7jg|7ji|7jk|7jm|2e|7jo|7jq|7js|7ju|7k0|7k2|7k4|7k6|7k8|7ka|7kc|7ke|7kg|7ki|7kk|7l0|7l2|7l4|7l6|7l8|7la|7lc|7le|7lg|7li|7lk|7lm|7lo|7lq|7ls|7lu|7m0|7m2|7m4|7m6|7m8|7ma|7mc|7me|7mg|7mi|7mk|7mm|7mo|7mq|2f|7ms|7mu|7n0|7n2|7n4|7n6|7n8|7na|7nc|7ne|7ng|7ni|7nk|7nm|7no|2g|7qp|7qr|7qt|7qv|2h|896|2l|s6|sc|uo|uq|us|uu|v0|v2|v4|v6'
    },
    'C.1.1': {
        s: '10'
    },
    'C.1.2': {
        r: '800:80b',
        s: 'c00|50|5k0|81f|82v'
    },
    'C.2.1': {
        r: '0:v',
        s: '3v'
    },
    'C.2.2': {
        r: '40:4v|80c:80d|818:819|830:833|83a:83f|1vvp:1vvs|3kbj:3kbq',
        s: '1mt|1of|60e|1vnv'
    },
    'C.3': {
        r: '1o00:1u7v|u000:vvvt|10000:11vvt'
    },
    'C.4': {
        r: '1veg:1vff|1vvu:1vvv|3vvu:3vvv|5vvu:5vvv|7vvu:7vvv|9vvu:9vvv|bvvu:bvvv|dvvu:dvvv|fvvu:fvvv|hvvu:hvvv|jvvu:jvvv|lvvu:lvvv|nvvu:nvvv|pvvu:pvvv|rvvu:rvvv|tvvu:tvvv|vvvu:vvvv|11vvu:11vvv'
    },
    'C.5': {
        r: '1m00:1nvv'
    },
    'C.6': {
        r: '1vvp:1vvt'
    },
    'C.7': {
        r: 'bvg:bvr'
    },
    'C.8': {
        r: 'q0:q1|80e:80f|81a:81e|83a:83f'
    },
    'C.9': {
        r: 's010:s03v',
        s: 's001'
    },
    'D.1': {
        r: '1eg:1fa|1fg:1fk|1h1:1hq|1i0:1ia|1jd:1jf|1jh:1ml|1n5:1n6|1nq:1nu|1o0:1od|1oi:1pc|1s0:1t5|1uov:1up8|1upa:1upm|1upo:1ups|1uq0:1uq1|1uq3:1uq4|1uq6:1uth|1uuj:1v9t|1vag:1vcf|1vci:1ve7|1vfg:1vfs|1vjg:1vjk|1vjm:1vns',
        s: '1du|1e0|1e3|1gr|1gv|1mt|1og|1th|1uot|1upu|80f'
    },
    'D.2': {
        r: '21:2q|31:3q|60:6m|6o:7m|7o:h0|h2:hj|ig:ld|lg:lo|lr:m1|mg:mh|n0:n4|s8:sa|se:t1|t3:ue|ug:vl|100:142|14a:16e|16g:17l|17o:17p|180:18f|19h:1am|1ap:1av|1b1:1c7|285:29p|29t:2a0|2a9:2ac|2ao:2b1|2b4:2bg|2c2:2c3|2c5:2cc|2cf:2cg|2cj:2d8|2da:2dg|2dm:2dp|2du:2e0|2e7:2e8|2eb:2ec|2es:2et|2ev:2f1|2f6:2fh|2fk:2fq|2g5:2ga|2gf:2gg|2gj:2h8|2ha:2hg|2hi:2hj|2hl:2hm|2ho:2hp|2hu:2i0|2ip:2is|2j6:2jf|2ji:2jk|2k5:2kb|2kf:2kh|2kj:2l8|2la:2lg|2li:2lj|2ll:2lp|2lt:2m0|2mb:2mc|2n6:2nf|2o2:2o3|2o5:2oc|2of:2og|2oj:2p8|2pa:2pg|2pi:2pj|2pm:2pp|2pt:2pu|2q7:2q8|2qb:2qc|2qs:2qt|2qv:2r1|2r6:2rg|2s5:2sa|2se:2sg|2si:2sl|2sp:2sq|2su:2sv|2t3:2t4|2t8:2ta|2te:2tl|2tn:2tp|2tu:2tv|2u1:2u2|2u6:2u8|2ua:2uc|2v7:2vi|301:303|305:30c|30e:30g|30i:318|31a:31j|31l:31p|321:324|330:331|336:33f|342:343|345:34c|34e:34g|34i:358|35a:35j|35l:35p|360:364|367:368|36a:36b|36l:36m|370:371|376:37f|382:383|385:38c|38e:38g|38i:398|39a:39p|39u:3a0|3a6:3a8|3aa:3ac|3b0:3b1|3b6:3bf|3c2:3c3|3c5:3cm|3cq:3dh|3dj:3dr|3e0:3e6|3ef:3eh|3eo:3ev|3fi:3fk|3g1:3hg|3hi:3hj|3i0:3i6|3if:3ir|3k1:3k2|3k7:3k8|3kk:3kn|3kp:3kv|3l1:3l3|3la:3lb|3ld:3lg|3li:3lj|3m0:3m4|3mg:3mp|3ms:3mt|3o0:3on|3oq:3pk|3pu:3q7|3q9:3ra|3s8:3sb|3tu:3u5|3u7:3uc|400:411|413:417|419:41a|420:42n|450:465|46g:47o|480:4ap|4av:4d2|4d8:4fp|4g0:4g6|4g8:4i6|4ia:4id|4ig:4im|4iq:4it|4j0:4k6|4ka:4kd|4kg:4le|4li:4ll|4lo:4lu|4m2:4m5|4m8:4me|4mg:4mm|4mo:4ne|4ng:4oe|4oi:4ol|4oo:4ou|4p0:4q6|4q8:4qq|4r1:4rs|4t0:4vk|501:5jm|5k1:5kq|5l0:5ng|5o0:5oc|5oe:5oh|5p0:5ph|5pl:5pm|5q0:5qh|5r0:5rc|5re:5rg|5s0:5tm|5tu:5u5|5u7:5u8|5uk:5uq|5v0:5v9|60g:60p|610:63n|640:658|7g0:7kr|7l0:7np|7o0:7ol|7oo:7ot|7p0:7q5|7q8:7qd|7qg:7qn|7qv:7rt|7s0:7tk|7tm:7ts|7u2:7u4|7u6:7uc|7ug:7uj|7um:7ur|7v0:7vc|7vi:7vk|7vm:7vs|88a:88j|88p:88t|89a:89d|89f:89h|89j:89p|89t:89v|8a5:8a9|8b0:8c3|8pm:8rq|94s:979|c05:c07|c11:c19|c1h:c1l|c1o:c1s|c21:c4m|c4t:c4v|c51:c7q|c7s:c7v|c85:c9c|c9h:cce|ccg:cdn|cfg:cgs|ch0:ci3|cj0:cjr|cjv:clg|cm0:cmb|cmg:cnu|co0:crm|crr:cut|cv0:cvu|d00:jdl|jg0:17t5|1800:194c|1b00:1lt3|1m00:1uhd|1uhg:1uja|1uo0:1uo6|1uoj:1uon|1vp1:1vpq|1vq1:1vqq|1vr6:1vtu|1vu2:1vu7|1vua:1vuf|1vui:1vun|1vuq:1vus|20o0:20ou|20p0:20p3|20pg:20qa|2100:2115|2118:212d|3k00:3k7l|3k80:3k96|3k9a:3kb6|3kba:3kbi|3kc3:3kc4|3kcc:3kd9|3kde:3ket|3l00:3l2k|3l2m:3l4s|3l4u:3l4v|3l55:3l56|3l59:3l5c|3l5e:3l5p|3l5t:3l60|3l62:3l63|3l65:3l85|3l87:3l8a|3l8d:3l8k|3l8m:3l8s|3l8u:3l9p|3l9r:3l9u|3la0:3la4|3laa:3lag|3lai:3ll3|3ll8:3lu9|4000:59mm|5u00:5ugt|u000:vvvt|10000:11vvt',
        s: '3l52|3l5r|3la6|1c9|5a|5l|5q|283|2ag|2di|2en|2iu|2k3|2kd|2m9|2mg|2n0|2q0|2qn|2s3|2ss|2un|35u|36u|3an|3dt|3k4|3ka|3kd|3l5|3l7|3lt|3m6|3pm|3po|3rv|3s5|3uf|41c|41h|41o|47r|4i8|4io|4k8|4lg|4m0|4og|5us|ne|7qp|7qr|7qt|7tu|80e|83h|83v|882|887|88l|894|896|898|rq|s6|sc|8sl'
    }
};

class Table {
    constructor(name, points) {
        this.singles = new Set();
        this.ranges = [];
        this.mappings = new Map();
        const data = TABLE_DATA[name];
        this.name = name;
        if (data) {
            if (data.s) {
                this.singles = new Set(data.s.split('|').map(s => parseInt(s, 32)));
            }
            if (data.r) {
                this.ranges = data.r.split('|').map(r => {
                    const [start, end] = r.split(':');
                    return [parseInt(start, 32), parseInt(end, 32)];
                });
            }
            if (data.m) {
                this.mappings = new Map(data.m.split('|').map(m => {
                    const [point, mapping] = m.split(':');
                    const mappedPoints = mapping.split(';').map(p => parseInt(p, 32));
                    return [parseInt(point, 32), mappedPoints];
                }));
            }
        }
        else if (points) {
            this.singles = new Set(points);
        }
    }
    contains(codePoint) {
        if (this.singles.has(codePoint)) {
            return true;
        }
        let left = 0;
        let right = this.ranges.length - 1;
        while (left <= right) {
            const pivot = Math.floor((left + right) / 2);
            const range = this.ranges[pivot];
            if (codePoint < range[0]) {
                right = pivot - 1;
                continue;
            }
            if (codePoint > range[1]) {
                left = pivot + 1;
                continue;
            }
            return true;
        }
        return false;
    }
    hasMapping(codePoint) {
        return this.mappings.has(codePoint) || this.contains(codePoint);
    }
    map(codePoint) {
        if (this.contains(codePoint) && !this.mappings.has(codePoint)) {
            return String.fromCodePoint(codePoint).toLowerCase().codePointAt(0);
        }
        return this.mappings.get(codePoint) || null;
    }
}
const A1 = new Table('A.1');
const B1 = new Table('B.1');
const B2 = new Table('B.2');
new Table('B.3');
const C11 = new Table('C.1.1');
const C12 = new Table('C.1.2');
const C21 = new Table('C.2.1');
const C22 = new Table('C.2.2');
const C3 = new Table('C.3');
const C4 = new Table('C.4');
const C5 = new Table('C.5');
const C6 = new Table('C.6');
const C7 = new Table('C.7');
const C8 = new Table('C.8');
const C9 = new Table('C.9');
const D1 = new Table('D.1');
const D2 = new Table('D.2');
// Shortcut some of the simpler table operations
B1.map = () => {
    return null;
};
C11.contains = (codePoint) => codePoint === 32;
C12.map = (codePoint) => {
    return C12.contains(codePoint) ? 32 : null;
};
function prepare$1(profile, allowUnassigned, input = '') {
    const inputCodePoints = Punycode.ucs2.decode(input);
    let mappedCodePoints = [];
    for (const codePoint of inputCodePoints) {
        if (!allowUnassigned && profile.unassigned.contains(codePoint)) {
            throw new Error('Unassigned code point: x' + codePoint.toString(16));
        }
        let hasMapping = false;
        for (const mappingTable of profile.mappings) {
            if (!mappingTable.hasMapping(codePoint)) {
                continue;
            }
            hasMapping = true;
            const mappedPoint = mappingTable.map(codePoint);
            if (!mappedPoint) {
                continue;
            }
            if (Array.isArray(mappedPoint)) {
                mappedCodePoints = mappedCodePoints.concat(mappedPoint);
            }
            else {
                mappedCodePoints.push(mappedPoint);
            }
        }
        if (!hasMapping) {
            mappedCodePoints.push(codePoint);
        }
    }
    let normalizedCodePoints = mappedCodePoints;
    if (profile.normalize) {
        const mappedString = Punycode.ucs2.encode(mappedCodePoints);
        const normalizedString = mappedString.normalize('NFKC');
        normalizedCodePoints = Punycode.ucs2.decode(normalizedString);
    }
    let hasRandALCat = false;
    let hasLCat = false;
    for (const codePoint of normalizedCodePoints) {
        for (const prohibited of profile.prohibited) {
            if (prohibited.contains(codePoint)) {
                throw new Error('Prohibited code point: x' + codePoint.toString(16));
            }
        }
        if (!allowUnassigned && profile.unassigned.contains(codePoint)) {
            // istanbul ignore next
            throw new Error('Prohibited code point: x' + codePoint.toString(16));
        }
        if (profile.bidirectional) {
            hasRandALCat = hasRandALCat || D1.contains(codePoint);
            hasLCat = hasLCat || D2.contains(codePoint);
        }
    }
    if (profile.bidirectional) {
        if (hasRandALCat && hasLCat) {
            throw new Error('String contained both LCat and RandALCat code points');
        }
        if (hasRandALCat &&
            (!D1.contains(normalizedCodePoints[0]) ||
                !D1.contains(normalizedCodePoints[normalizedCodePoints.length - 1]))) {
            throw new Error('String containing RandALCat code points must start and end with RandALCat code points');
        }
    }
    return Punycode.ucs2.encode(normalizedCodePoints);
}
const NamePrepProfile = {
    bidirectional: true,
    mappings: [B1, B2],
    normalize: true,
    prohibited: [C12, C22, C3, C4, C5, C6, C7, C8, C9],
    unassigned: A1
};
function nameprep(str, allowUnassigned = true) {
    return prepare$1(NamePrepProfile, allowUnassigned, str);
}
const NodePrepProhibited = new Table('NodePrepProhibited', [
    0x22,
    0x26,
    0x27,
    0x2f,
    0x3a,
    0x3c,
    0x3e,
    0x40
]);
const NodePrepProfile = {
    bidirectional: true,
    mappings: [B1, B2],
    normalize: true,
    prohibited: [C11, C12, C21, C22, C3, C4, C5, C6, C7, C8, C9, NodePrepProhibited],
    unassigned: A1
};
function nodeprep(str, allowUnassigned = true) {
    return prepare$1(NodePrepProfile, allowUnassigned, str);
}
const ResourcePrepProfile = {
    bidirectional: true,
    mappings: [B1],
    normalize: true,
    prohibited: [C12, C21, C22, C3, C4, C5, C6, C7, C8, C9],
    unassigned: A1
};
function resourceprep(str, allowUnassigned = true) {
    return prepare$1(ResourcePrepProfile, allowUnassigned, str);
}
const SASLPrepProfile = {
    bidirectional: true,
    mappings: [C12, B1],
    normalize: true,
    prohibited: [C12, C21, C22, C3, C4, C5, C6, C7, C8, C9],
    unassigned: A1
};
function saslprep(str, allowUnassigned = false) {
    return prepare$1(SASLPrepProfile, allowUnassigned, str);
}

function escapeLocal(val = '') {
    return val
        .replace(/^\s+|\s+$/g, '')
        .replace(/\\5c/g, '\\5c5c')
        .replace(/\\20/g, '\\5c20')
        .replace(/\\22/g, '\\5c22')
        .replace(/\\26/g, '\\5c26')
        .replace(/\\27/g, '\\5c27')
        .replace(/\\2f/g, '\\5c2f')
        .replace(/\\3a/g, '\\5c3a')
        .replace(/\\3c/g, '\\5c3c')
        .replace(/\\3e/g, '\\5c3e')
        .replace(/\\40/g, '\\5c40')
        .replace(/ /g, '\\20')
        .replace(/"/g, '\\22')
        .replace(/&/g, '\\26')
        .replace(/'/g, '\\27')
        .replace(/\//g, '\\2f')
        .replace(/:/g, '\\3a')
        .replace(/</g, '\\3c')
        .replace(/>/g, '\\3e')
        .replace(/@/g, '\\40');
}
function unescapeLocal(val) {
    return val
        .replace(/\\20/g, ' ')
        .replace(/\\22/g, '"')
        .replace(/\\26/g, '&')
        .replace(/\\27/g, `'`)
        .replace(/\\2f/g, '/')
        .replace(/\\3a/g, ':')
        .replace(/\\3c/g, '<')
        .replace(/\\3e/g, '>')
        .replace(/\\40/g, '@')
        .replace(/\\5c/g, '\\');
}
function prepare(data) {
    let local = data.local || '';
    let domain = data.domain;
    let resource = data.resource || '';
    if (local) {
        local = nodeprep(local);
    }
    if (resource) {
        resource = resourceprep(resource);
    }
    if (domain[domain.length - 1] === '.') {
        domain = domain.slice(0, domain.length - 1);
    }
    domain = nameprep(domain.split('.').map(Punycode.toUnicode).join('.'));
    return {
        domain,
        local,
        resource
    };
}
function create(data, opts = {}) {
    let localPart = data.local;
    if (!opts.escaped) {
        localPart = escapeLocal(data.local);
    }
    const prep = !opts.prepared
        ? prepare({ local: localPart, domain: data.domain, resource: data.resource })
        : data;
    const bareJID = localPart ? `${localPart}@${prep.domain}` : prep.domain;
    if (prep.resource) {
        return `${bareJID}/${prep.resource}`;
    }
    return bareJID;
}
function createFull(bare, resource) {
    if (resource) {
        return `${toBare(bare)}/${resource}`;
    }
    else {
        return toBare(bare);
    }
}
function parse$2(jid = '') {
    let local = '';
    let domain = '';
    let resource = '';
    const resourceStart = jid.indexOf('/');
    if (resourceStart > 0) {
        resource = jid.slice(resourceStart + 1);
        jid = jid.slice(0, resourceStart);
    }
    const localEnd = jid.indexOf('@');
    if (localEnd > 0) {
        local = jid.slice(0, localEnd);
        jid = jid.slice(localEnd + 1);
    }
    domain = jid;
    const prepped = prepare({
        domain,
        local,
        resource
    });
    return {
        bare: create({ local: prepped.local, domain: prepped.domain }, {
            escaped: true,
            prepared: true
        }),
        domain: prepped.domain,
        full: create(prepped, {
            escaped: true,
            prepared: true
        }),
        local: unescapeLocal(prepped.local),
        resource: prepped.resource
    };
}
function allowedResponders(jid1, jid2) {
    const allowed = new Set();
    allowed.add(undefined);
    allowed.add('');
    if (jid1) {
        const split1 = parse$2(jid1);
        allowed.add(split1.full);
        allowed.add(split1.bare);
        allowed.add(split1.domain);
    }
    if (jid2) {
        const split2 = parse$2(jid2);
        allowed.add(split2.domain);
        allowed.add(split2.bare);
        allowed.add(split2.full);
    }
    return allowed;
}
function equal(jid1, jid2) {
    if (!jid1 || !jid2) {
        return false;
    }
    const parsed1 = parse$2(jid1);
    const parsed2 = parse$2(jid2);
    return (parsed1.local === parsed2.local &&
        parsed1.domain === parsed2.domain &&
        parsed1.resource === parsed2.resource);
}
function equalBare(jid1, jid2) {
    if (!jid1 || !jid2) {
        return false;
    }
    const parsed1 = parse$2(jid1);
    const parsed2 = parse$2(jid2);
    return parsed1.local === parsed2.local && parsed1.domain === parsed2.domain;
}
function isFull(jid) {
    const parsed = parse$2(jid);
    return !!parsed.resource;
}
function isBare(jid) {
    return !isFull(jid);
}
function getLocal(jid = '') {
    return parse$2(jid).local;
}
function getDomain(jid = '') {
    return parse$2(jid).domain;
}
function getResource(jid = '') {
    return parse$2(jid).resource;
}
function toBare(jid = '') {
    return parse$2(jid).bare;
}
function parseURI(val) {
    const parsed = new URL(val);
    if (parsed.protocol !== 'xmpp:') {
        throw new Error('Invalid XMPP URI, wrong protocol: ' + parsed.protocol);
    }
    const identity = parsed.hostname
        ? parsed.username
            ? create({
                domain: decodeURIComponent(parsed.hostname),
                local: decodeURIComponent(parsed.username)
            }, {
                escaped: true
            })
            : decodeURIComponent(parsed.hostname)
        : undefined;
    const jid = parse$2(decodeURIComponent(identity ? parsed.pathname.substr(1) : parsed.pathname))
        .full;
    const hasParameters = parsed.search && parsed.search.indexOf(';') >= 1;
    const parameterString = hasParameters
        ? parsed.search.substr(parsed.search.indexOf(';') + 1)
        : '';
    const action = parsed.search
        ? decodeURIComponent(parsed.search.substr(1, hasParameters ? parsed.search.indexOf(';') - 1 : undefined))
        : undefined;
    const params = {};
    for (const token of parameterString.split(';')) {
        const [name, value] = token.split('=').map(decodeURIComponent);
        if (!params[name]) {
            params[name] = value;
        }
        else {
            const existing = params[name];
            if (Array.isArray(existing)) {
                existing.push(value);
            }
            else {
                params[name] = [existing, value];
            }
        }
    }
    return {
        action,
        identity,
        jid,
        parameters: params
    };
}
function toURI(data) {
    const parts = ['xmpp:'];
    const pushJID = (jid, allowResource) => {
        const res = parse$2(jid);
        if (res.local) {
            parts.push(encodeURIComponent(escapeLocal(res.local)));
            parts.push('@');
        }
        parts.push(encodeURIComponent(res.domain));
        if (allowResource && res.resource) {
            parts.push('/');
            parts.push(encodeURIComponent(res.resource));
        }
    };
    if (data.identity) {
        parts.push('//');
        pushJID(data.identity, false);
        if (data.jid) {
            parts.push('/');
        }
    }
    if (data.jid) {
        pushJID(data.jid, true);
    }
    if (data.action) {
        parts.push('?');
        parts.push(encodeURIComponent(data.action));
    }
    for (const [name, values] of Object.entries(data.parameters || {})) {
        for (const val of Array.isArray(values) ? values : [values]) {
            parts.push(';');
            parts.push(encodeURIComponent(name));
            if (val !== undefined) {
                parts.push('=');
                parts.push(encodeURIComponent(val));
            }
        }
    }
    return parts.join('');
}

var JID = /*#__PURE__*/Object.freeze({
    __proto__: null,
    escapeLocal: escapeLocal,
    unescapeLocal: unescapeLocal,
    prepare: prepare,
    create: create,
    createFull: createFull,
    parse: parse$2,
    allowedResponders: allowedResponders,
    equal: equal,
    equalBare: equalBare,
    isFull: isFull,
    isBare: isBare,
    getLocal: getLocal,
    getDomain: getDomain,
    getResource: getResource,
    toBare: toBare,
    parseURI: parseURI,
    toURI: toURI
});

var JXTErrorCondition;
(function (JXTErrorCondition) {
    JXTErrorCondition["NotWellFormed"] = "not-well-formed";
    JXTErrorCondition["RestrictedXML"] = "restricted-xml";
    JXTErrorCondition["AlreadyClosed"] = "already-closed";
    JXTErrorCondition["UnknownRoot"] = "unknown-stream-root";
})(JXTErrorCondition || (JXTErrorCondition = {}));
class JXTError extends Error {
    constructor(opts) {
        super(opts.text);
        this.isJXTError = true;
        this.condition = opts.condition;
        this.text = opts.text;
    }
    static notWellFormed(text) {
        return new JXTError({
            condition: JXTErrorCondition.NotWellFormed,
            text
        });
    }
    static restrictedXML(text) {
        return new JXTError({
            condition: JXTErrorCondition.RestrictedXML,
            text
        });
    }
    static alreadyClosed(text) {
        return new JXTError({
            condition: JXTErrorCondition.AlreadyClosed,
            text
        });
    }
    static unknownRoot(text) {
        return new JXTError({
            condition: JXTErrorCondition.UnknownRoot,
            text
        });
    }
}

const ESCAPE_XML_CHAR = {
    '"': '&quot;',
    '&': '&amp;',
    "'": '&apos;',
    '<': '&lt;',
    '>': '&gt;'
};
const UNESCAPE_XML_CHAR = {
    '&amp;': '&',
    '&apos;': "'",
    '&gt;': '>',
    '&lt;': '<',
    '&quot;': '"'
};
const ESCAPE_SEQUENCE = /&([a-zA-Z0-9]+|#[0-9]+|#x[0-9a-fA-F]+);/g;
const NEED_ESCAPING = /&|<|>|"|'/g;
const NEED_ESCAPING_TEXT = /&|<|>/g;
function escapeXMLReplaceChar(match) {
    return ESCAPE_XML_CHAR[match];
}
function unescapeXMLReplaceChar(match) {
    if (UNESCAPE_XML_CHAR[match]) {
        return UNESCAPE_XML_CHAR[match];
    }
    const hex = match.startsWith('&#x');
    const code = parseInt(match.substring(hex ? 3 : 2, match.length - 1), hex ? 16 : 10);
    if (code === 0x9 ||
        code === 0xa ||
        code === 0xd ||
        (0x20 <= code && code <= 0xd7ff) ||
        (0xe000 <= code && code <= 0xfffd) ||
        (0x10000 <= code && code <= 0x10ffff)) {
        return String.fromCodePoint(code);
    }
    throw JXTError.restrictedXML('Prohibited entity: ' + match);
}
function escapeXML(text) {
    return text.replace(NEED_ESCAPING, escapeXMLReplaceChar);
}
function unescapeXML(text) {
    return text.replace(ESCAPE_SEQUENCE, match => {
        return unescapeXMLReplaceChar(match);
    });
}
function escapeXMLText(text) {
    return text.replace(NEED_ESCAPING_TEXT, escapeXMLReplaceChar);
}
function basicLanguageResolver(available, accept = [], current = '') {
    const avail = new Set(available.map(a => a.toLowerCase()));
    for (let acceptLang of accept.map(a => a.toLowerCase())) {
        if (acceptLang === '*') {
            continue;
        }
        while (acceptLang.length > 0) {
            if (avail.has(acceptLang)) {
                return acceptLang;
            }
            // Remove ending tag
            acceptLang = acceptLang.substring(0, acceptLang.lastIndexOf('-')).toLowerCase();
            // Remove leftover single character tag
            if (acceptLang.lastIndexOf('-') === acceptLang.length - 2) {
                acceptLang = acceptLang.substring(0, acceptLang.lastIndexOf('-'));
            }
        }
    }
    return current;
}

class XMLElement {
    constructor(name, attrs = {}, children = []) {
        this.name = name;
        this.attributes = attrs;
        this.children = [];
        this.optionalNamespaces = {};
        for (const child of children) {
            if (typeof child !== 'string') {
                const xmlChild = new XMLElement(child.name, child.attributes, child.children);
                xmlChild.parent = this;
                this.children.push(xmlChild);
            }
            else {
                this.children.push(child);
            }
        }
    }
    getName() {
        if (this.name.indexOf(':') >= 0) {
            return this.name.substr(this.name.indexOf(':') + 1);
        }
        else {
            return this.name;
        }
    }
    getNamespace() {
        if (this.name.indexOf(':') >= 0) {
            const prefix = this.name.substr(0, this.name.indexOf(':'));
            return this.findNamespaceForPrefix(prefix);
        }
        return this.findNamespaceForPrefix();
    }
    getNamespaceContext() {
        let namespaces = {};
        if (this.parent) {
            namespaces = this.parent.getNamespaceContext();
        }
        for (const [attr, value] of Object.entries(this.attributes)) {
            if (attr.startsWith('xmlns:')) {
                const prefix = attr.substr(6);
                namespaces[value] = prefix;
            }
        }
        return namespaces;
    }
    getDefaultNamespace() {
        if (this.attributes.xmlns) {
            return this.attributes.xmlns;
        }
        if (this.parent) {
            return this.parent.getDefaultNamespace();
        }
        return '';
    }
    getNamespaceRoot(namespace) {
        if (this.parent) {
            const parentRoot = this.parent.getNamespaceRoot(namespace);
            if (parentRoot) {
                return parentRoot;
            }
        }
        for (const [attr, value] of Object.entries(this.attributes)) {
            if (attr.startsWith('xmlns:') && value === namespace) {
                return this;
            }
        }
        if (this.optionalNamespaces[namespace]) {
            return this;
        }
        return undefined;
    }
    getAttribute(name, xmlns) {
        if (!xmlns) {
            return this.attributes[name];
        }
        const namespaces = this.getNamespaceContext();
        if (!namespaces[xmlns]) {
            return undefined;
        }
        return this.attributes[[namespaces[xmlns], name].join(':')];
    }
    getChild(name, xmlns) {
        return this.getChildren(name, xmlns)[0];
    }
    getChildren(name, xmlns) {
        const result = [];
        for (const child of this.children) {
            if (typeof child !== 'string' &&
                child.getName() === name &&
                (!xmlns || child.getNamespace() === xmlns)) {
                result.push(child);
            }
        }
        return result;
    }
    getText() {
        let text = '';
        for (const child of this.children) {
            if (typeof child === 'string') {
                text += child;
            }
        }
        return text;
    }
    appendChild(child) {
        this.children.push(child);
        if (typeof child !== 'string') {
            child.parent = this;
        }
        return child;
    }
    setAttribute(attr, val, force = false) {
        this.attributes[attr] = val || undefined;
        if (val === '' && force) {
            this.attributes[attr] = val;
        }
    }
    addOptionalNamespace(prefix, namespace) {
        this.optionalNamespaces[namespace] = prefix;
    }
    useNamespace(prefix, namespace) {
        if (this.optionalNamespaces[namespace]) {
            prefix = this.optionalNamespaces[namespace];
        }
        this.setAttribute(`xmlns:${prefix}`, namespace);
        return prefix;
    }
    toJSON() {
        const children = this.children
            .map(child => {
            if (typeof child === 'string') {
                return child;
            }
            if (child) {
                return child.toJSON();
            }
        })
            .filter(child => !!child);
        // Strip any undefined/null attributes
        const attrs = {};
        for (const [key, val] of Object.entries(this.attributes)) {
            if (val !== undefined && val !== null) {
                attrs[key] = val;
            }
        }
        return {
            attributes: attrs,
            children,
            name: this.name
        };
    }
    toString() {
        let output = this.openTag(true);
        if (this.children.length) {
            for (const child of this.children) {
                if (typeof child === 'string') {
                    output += escapeXMLText(child);
                }
                else if (child) {
                    output += child.toString();
                }
            }
            output += this.closeTag();
        }
        return output;
    }
    openTag(allowSelfClose = false) {
        let output = '';
        output += `<${this.name}`;
        for (const [key, value] of Object.entries(this.attributes)) {
            if (value !== undefined) {
                output += ` ${key}="${escapeXML(value.toString())}"`;
            }
        }
        if (allowSelfClose && this.children.length === 0) {
            output += '/>';
        }
        else {
            output += '>';
        }
        return output;
    }
    closeTag() {
        return `</${this.name}>`;
    }
    findNamespaceForPrefix(prefix) {
        if (!prefix) {
            if (this.attributes.xmlns) {
                return this.attributes.xmlns;
            }
            else if (this.parent) {
                return this.parent.findNamespaceForPrefix();
            }
        }
        else {
            const attr = 'xmlns:' + prefix;
            if (this.attributes[attr]) {
                return this.attributes[attr];
            }
            else if (this.parent) {
                return this.parent.findNamespaceForPrefix(prefix);
            }
        }
        return '';
    }
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from: xhtml-im.js, Copyright © 2013 Waqas Hussain
 */
const ALLOWED_ELEMENTS = new Set([
    'a',
    'blockquote',
    'br',
    'cite',
    'em',
    'img',
    'li',
    'ol',
    'p',
    'span',
    'strong',
    'ul'
]);
const style = new Set(['style']);
const ALLOWED_ATTRIBUTES = new Map([
    ['a', new Set(['href', 'style'])],
    ['body', new Set(['style', 'xml:lang'])],
    ['blockquote', style],
    ['br', style],
    ['cite', style],
    ['em', style],
    ['img', new Set(['alt', 'height', 'src', 'style', 'width'])],
    ['li', style],
    ['ol', style],
    ['p', style],
    ['span', style],
    ['strong', style],
    ['ul', style]
]);
const CSS_RULES = new Map([
    ['font-style', /normal|italic|oblique|inherit/i],
    ['font-weight', /normal|bold|bolder|lighter|inherit|\d\d\d/i],
    ['text-decoration', /none|underline|overline|line-through|blink|inherit/i]
    // These properties are allowed by XHTML-IM, but really only cause UX issues:
    //  background-color
    //  color
    //  font-family
    //  font-size
    //  margin-left
    //  margin-right
    //  text-align
]);
const sanitizeCSS = (css) => {
    const declarations = `;${css}` // Declarations are ; delimited, not terminated
        .replace(/\/\*[^*]*\*+([^/*][^*]*\*+)*\//g, '') // Strip comments
        .replace(/\/\*.*/, '') // Strip unclosed comments
        .replace(/\\([a-fA-F0-9]{1,6})\s?/, (_, x) => String.fromCharCode(parseInt(x, 16))) // Decode escape sequences
        .match(/;\s*([a-z-]+)\s*:\s*([^;]*[^\s;])\s*/g); // Split into declarations
    const rules = [];
    if (!declarations) {
        return false;
    }
    for (const declaration of declarations) {
        const parts = declaration.match(/^;\s*([a-z-]+)\s*:\s*([^;]*[^\s])\s*$/);
        if (!parts) {
            continue;
        }
        const sanitizer = CSS_RULES.get(parts[1]);
        if (sanitizer) {
            const value = parts[2].match(sanitizer);
            if (value) {
                rules.push(`${parts[1]}:${value[0]}`);
            }
        }
    }
    if (rules.length) {
        return rules.join('');
    }
    return false;
};
const sanitizeURL = (url) => {
    return (!!url.match(/^(https?|xmpp|cid|mailto|ftps?|im|ircs?|sips?|tel|geo|bitcoin|magnet):/i) &&
        url);
};
const sanitizeNumber = (num) => {
    return !!num.match(/^[0-9]*$/) && num;
};
const ATTRIBUTE_SANITIZERS = {
    alt: text => text,
    height: sanitizeNumber,
    href: sanitizeURL,
    src: sanitizeURL,
    style: sanitizeCSS,
    width: sanitizeNumber
};
function stripElement(input) {
    let results = [];
    for (const child of input.children) {
        if (typeof child === 'string') {
            results.push(child);
        }
        else {
            const sanitized = sanitizeInterior(child);
            if (sanitized) {
                if (Array.isArray(sanitized)) {
                    results = results.concat(sanitized);
                }
                else {
                    results.push(sanitized);
                }
            }
        }
    }
    return results;
}
function sanitizeInterior(input) {
    if (typeof input === 'string') {
        return input;
    }
    if (!ALLOWED_ELEMENTS.has(input.name)) {
        if (input.name === 'script') {
            return;
        }
        return stripElement(input);
    }
    const children = input.children
        .map(sanitizeInterior)
        .filter(child => child !== undefined);
    const attributes = {};
    for (const [key, value] of Object.entries(input.attributes)) {
        const allowed = ALLOWED_ATTRIBUTES.get(input.name);
        if (!allowed || !allowed.has(key)) {
            continue;
        }
        if (!value) {
            continue;
        }
        const sanitized = ATTRIBUTE_SANITIZERS[key](value);
        if (!sanitized) {
            continue;
        }
        attributes[key] = sanitized;
    }
    return {
        attributes,
        children,
        name: input.name
    };
}
function sanitizeRoot(input) {
    if (typeof input === 'string') {
        return;
    }
    let children = [];
    for (const child of input.children) {
        if (!child) {
            continue;
        }
        if (typeof child === 'string') {
            children.push(child);
            continue;
        }
        const sanitized = sanitizeInterior(child);
        if (Array.isArray(sanitized)) {
            children = children.concat(sanitized);
        }
        else if (sanitized) {
            children.push(sanitized);
        }
    }
    const attributes = {};
    if (input.name !== 'body') {
        return;
    }
    if (input.attributes.xmlns !== undefined) {
        attributes.xmlns = input.attributes.xmlns;
    }
    if (input.attributes.style) {
        attributes.style = input.attributes.style;
    }
    if (input.attributes['xml:lang'] !== undefined) {
        attributes['xml:lang'] = input.attributes['xml:lang'];
    }
    return {
        attributes,
        children,
        name: 'body'
    };
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from: ltx, Copyright © 2010 Stephan Maka
 */
function isBasicNameStart(c) {
    return ((97 /* a */ <= c && c <= 122 /* z */) ||
        (65 /* A */ <= c && c <= 90 /* Z */) ||
        c === 58 /* Colon */ ||
        c === 95 /* Underscore */);
}
function isExtendedNameStart(c) {
    return ((0xc0 <= c && c <= 0xd6) ||
        (0xd8 <= c && c <= 0xf6) ||
        (0xf8 <= c && c <= 0x2ff) ||
        (0x370 <= c && c <= 0x37d) ||
        (0x37f <= c && c <= 0x1fff) ||
        (0x200c <= c && c <= 0x200d) ||
        (0x2070 <= c && c <= 0x218f) ||
        (0x2c00 <= c && c <= 0x2fef) ||
        (0x3001 <= c && c <= 0xd7ff) ||
        (0xfdf0 <= c && c <= 0xfffd) ||
        (0x10000 <= c && c <= 0xeffff));
}
function isNameStart(c) {
    return isBasicNameStart(c) || isExtendedNameStart(c);
}
function isName(c) {
    return (isBasicNameStart(c) ||
        c === 45 /* Dash */ ||
        c === 46 /* Period */ ||
        (48 /* Zero */ <= c && c <= 57 /* Nine */) ||
        c === 0xb7 ||
        (0x0300 <= c && c <= 0x036f) ||
        (0x203f <= c && c <= 0x2040) ||
        isExtendedNameStart(c));
}
function isWhitespace(c) {
    return (c === 32 /* Space */ ||
        c === 10 /* NewLine */ ||
        c === 13 /* CarriageReturn */ ||
        c === 9 /* Tab */);
}
class Parser extends EventEmitter {
    constructor(opts = {}) {
        super();
        this.allowComments = true;
        this.attributes = {};
        this.state = 34 /* TEXT */;
        this.tagName = '';
        this.haveDeclaration = false;
        this.recordBuffer = [];
        if (opts.allowComments !== undefined) {
            this.allowComments = opts.allowComments;
        }
    }
    write(data) {
        for (const char of data) {
            const c = char.codePointAt(0);
            switch (this.state) {
                case 34 /* TEXT */: {
                    if (c === 60 /* LessThan */) {
                        let text;
                        try {
                            text = unescapeXML(this.endRecord());
                        }
                        catch (err) {
                            this.emit('error', err);
                            return;
                        }
                        if (text) {
                            this.emit('text', text);
                        }
                        this.transition(31 /* TAG_START */);
                        continue;
                    }
                    else {
                        this.record(char);
                        continue;
                    }
                }
                case 31 /* TAG_START */: {
                    if (c === 47 /* Slash */) {
                        this.transition(7 /* CLOSING_TAG_START */);
                        continue;
                    }
                    if (c === 33 /* Exclamation */) {
                        this.transition(24 /* START_INSTRUCTION */);
                        continue;
                    }
                    if (c === 63 /* Question */) {
                        if (this.haveDeclaration) {
                            this.restrictedXML();
                        }
                        this.transition(25 /* START_PROCESSING_INSTRUCTION */);
                        continue;
                    }
                    if (isNameStart(c)) {
                        this.transition(30 /* TAG_NAME */);
                        this.startRecord(char);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 30 /* TAG_NAME */: {
                    if (isName(c)) {
                        this.record(char);
                        continue;
                    }
                    if (isWhitespace(c)) {
                        this.startTag();
                        this.transition(32 /* TAG_WAIT_NAME */);
                        continue;
                    }
                    if (c === 47 /* Slash */) {
                        this.startTag();
                        this.transition(29 /* TAG_END_SLASH */);
                        continue;
                    }
                    if (c === 62 /* GreaterThan */) {
                        this.startTag();
                        this.transition(34 /* TEXT */);
                        this.emit('startElement', this.tagName, this.attributes);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 29 /* TAG_END_SLASH */: {
                    if (c === 62 /* GreaterThan */) {
                        this.emit('startElement', this.tagName, this.attributes);
                        this.emit('endElement', this.tagName);
                        this.transition(34 /* TEXT */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 33 /* TAG */: {
                    if (isWhitespace(c)) {
                        this.transition(32 /* TAG_WAIT_NAME */);
                        continue;
                    }
                    if (c === 47 /* Slash */) {
                        this.transition(29 /* TAG_END_SLASH */);
                        continue;
                    }
                    if (c === 62 /* GreaterThan */) {
                        this.emit('startElement', this.tagName, this.attributes);
                        this.transition(34 /* TEXT */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 32 /* TAG_WAIT_NAME */: {
                    if (isWhitespace(c)) {
                        continue;
                    }
                    if (isNameStart(c)) {
                        this.startRecord(char);
                        this.transition(0 /* ATTR_NAME */);
                        continue;
                    }
                    if (c === 47 /* Slash */) {
                        this.transition(29 /* TAG_END_SLASH */);
                        continue;
                    }
                    if (c === 62 /* GreaterThan */) {
                        this.emit('startElement', this.tagName, this.attributes);
                        this.transition(34 /* TEXT */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 7 /* CLOSING_TAG_START */: {
                    if (isNameStart(c)) {
                        this.startRecord(char);
                        this.transition(6 /* CLOSING_TAG_NAME */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 6 /* CLOSING_TAG_NAME */: {
                    if (isName(c)) {
                        this.record(char);
                        continue;
                    }
                    if (isWhitespace(c)) {
                        this.transition(8 /* CLOSING_TAG */);
                        continue;
                    }
                    if (c === 62 /* GreaterThan */) {
                        const tag = this.endRecord();
                        this.emit('endElement', tag, this.attributes);
                        this.transition(34 /* TEXT */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 8 /* CLOSING_TAG */: {
                    if (isWhitespace(c)) {
                        continue;
                    }
                    if (c === 62 /* GreaterThan */) {
                        const tag = this.endRecord();
                        this.emit('endElement', tag, this.attributes);
                        this.transition(34 /* TEXT */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 0 /* ATTR_NAME */: {
                    if (isName(c)) {
                        this.record(char);
                        continue;
                    }
                    if (c === 61 /* Equal */) {
                        this.addAttribute();
                        this.transition(4 /* ATTR_WAIT_QUOTE */);
                        continue;
                    }
                    if (isWhitespace(c)) {
                        this.addAttribute();
                        this.transition(3 /* ATTR_WAIT_EQ */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 3 /* ATTR_WAIT_EQ */: {
                    if (c === 61 /* Equal */) {
                        this.transition(4 /* ATTR_WAIT_QUOTE */);
                        continue;
                    }
                    if (isWhitespace(c)) {
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 4 /* ATTR_WAIT_QUOTE */: {
                    if (c === 34 /* DoubleQuote */) {
                        this.startRecord();
                        this.transition(1 /* ATTR_QUOTE_DOUBLE */);
                        continue;
                    }
                    if (c === 39 /* SingleQuote */) {
                        this.startRecord();
                        this.transition(2 /* ATTR_QUOTE_SINGLE */);
                        continue;
                    }
                    if (isWhitespace(c)) {
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 1 /* ATTR_QUOTE_DOUBLE */:
                case 2 /* ATTR_QUOTE_SINGLE */: {
                    if ((c === 34 /* DoubleQuote */ && this.state === 1 /* ATTR_QUOTE_DOUBLE */) ||
                        (c === 39 /* SingleQuote */ && this.state === 2 /* ATTR_QUOTE_SINGLE */)) {
                        const value = this.endRecord();
                        this.attributes[this.attributeName] = unescapeXML(value);
                        this.transition(33 /* TAG */);
                        continue;
                    }
                    if (c === 60 /* LessThan */) {
                        return this.notWellFormed();
                    }
                    this.record(char);
                    continue;
                }
                case 24 /* START_INSTRUCTION */: {
                    if (c === 45 /* Dash */) {
                        if (!this.allowComments) {
                            this.restrictedXML();
                        }
                        this.transition(23 /* START_COMMENT_DASH */);
                        continue;
                    }
                    if (c === 91 /* LeftBracket */) {
                        this.transition(21 /* START_CDATA_LB */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 23 /* START_COMMENT_DASH */: {
                    if (c === 45 /* Dash */) {
                        this.transition(14 /* IGNORE_COMMENT */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 14 /* IGNORE_COMMENT */: {
                    if (c === 45 /* Dash */) {
                        this.transition(12 /* END_COMMENT_DASH */);
                    }
                    continue;
                }
                case 12 /* END_COMMENT_DASH */: {
                    if (c === 45 /* Dash */) {
                        this.transition(11 /* END_COMMENT_DASH_DASH */);
                    }
                    else {
                        this.transition(14 /* IGNORE_COMMENT */);
                    }
                    continue;
                }
                case 11 /* END_COMMENT_DASH_DASH */: {
                    if (c === 62 /* GreaterThan */) {
                        this.transition(34 /* TEXT */);
                    }
                    else {
                        this.transition(14 /* IGNORE_COMMENT */);
                    }
                    continue;
                }
                case 25 /* START_PROCESSING_INSTRUCTION */: {
                    if (c === 88 /* X */ || c === 120 /* x */) {
                        this.transition(28 /* START_XML_DECLARATION_X */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 28 /* START_XML_DECLARATION_X */: {
                    if (c === 77 /* M */ || c === 109 /* m */) {
                        this.transition(27 /* START_XML_DECLARATION_X_M */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 27 /* START_XML_DECLARATION_X_M */: {
                    if (c === 76 /* L */ || c === 108 /* l */) {
                        this.transition(26 /* START_XML_DECLARATION_X_M_L */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 26 /* START_XML_DECLARATION_X_M_L */: {
                    if (isWhitespace(c)) {
                        this.haveDeclaration = true;
                        this.transition(15 /* IGNORE_INSTRUCTION */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 13 /* END_XML_DECLARATION_QM */: {
                    if (c === 62 /* GreaterThan */) {
                        this.transition(34 /* TEXT */);
                        continue;
                    }
                    return this.notWellFormed();
                }
                case 15 /* IGNORE_INSTRUCTION */: {
                    if (c === 63 /* Question */) {
                        this.transition(13 /* END_XML_DECLARATION_QM */);
                    }
                    continue;
                }
                case 21 /* START_CDATA_LB */: {
                    this.wait(c, 67 /* C */, 20 /* START_CDATA_LB_C */);
                    continue;
                }
                case 20 /* START_CDATA_LB_C */: {
                    this.wait(c, 68 /* D */, 19 /* START_CDATA_LB_C_D */);
                    continue;
                }
                case 19 /* START_CDATA_LB_C_D */: {
                    this.wait(c, 65 /* A */, 18 /* START_CDATA_LB_C_D_A */);
                    continue;
                }
                case 18 /* START_CDATA_LB_C_D_A */: {
                    this.wait(c, 84 /* T */, 17 /* START_CDATA_LB_C_D_A_T */);
                    continue;
                }
                case 17 /* START_CDATA_LB_C_D_A_T */: {
                    this.wait(c, 65 /* A */, 16 /* START_CDATA_LB_C_D_A_T_A */);
                    continue;
                }
                case 16 /* START_CDATA_LB_C_D_A_T_A */: {
                    this.wait(c, 91 /* LeftBracket */, 5 /* CDATA */);
                    continue;
                }
                case 5 /* CDATA */: {
                    if (c === 93 /* RightBracket */) {
                        this.transition(10 /* END_CDATA_RB */);
                        continue;
                    }
                    this.record(char);
                    continue;
                }
                case 10 /* END_CDATA_RB */: {
                    if (c === 93 /* RightBracket */) {
                        this.transition(9 /* END_CDATA_RB_RB */);
                    }
                    else {
                        this.record(String.fromCodePoint(93 /* RightBracket */));
                        this.record(char);
                        this.transition(5 /* CDATA */);
                    }
                    continue;
                }
                case 9 /* END_CDATA_RB_RB */: {
                    if (c === 62 /* GreaterThan */) {
                        const text = this.endRecord();
                        if (text) {
                            this.emit('text', text);
                        }
                        this.transition(34 /* TEXT */);
                    }
                    else {
                        this.record(String.fromCodePoint(93 /* RightBracket */));
                        this.record(String.fromCodePoint(93 /* RightBracket */));
                        this.record(char);
                        this.transition(5 /* CDATA */);
                    }
                    continue;
                }
            }
        }
    }
    end(data) {
        if (data) {
            this.write(data);
        }
        this.write = () => undefined;
    }
    record(char) {
        this.recordBuffer.push(char);
    }
    startRecord(char) {
        this.recordBuffer = [];
        if (char) {
            this.recordBuffer.push(char);
        }
    }
    endRecord() {
        const data = this.recordBuffer;
        this.recordBuffer = [];
        return data.join('');
    }
    startTag() {
        this.tagName = this.endRecord();
        this.attributes = {};
    }
    addAttribute() {
        const name = this.endRecord();
        if (this.attributes[name] !== undefined) {
            return this.notWellFormed();
        }
        this.attributeName = name;
        this.attributes[name] = '';
    }
    wait(c, nextChar, newState) {
        if (c === nextChar) {
            this.transition(newState);
            return;
        }
        return this.notWellFormed();
    }
    transition(state) {
        this.state = state;
        if (state === 34 /* TEXT */) {
            this.startRecord();
        }
    }
    notWellFormed(msg) {
        this.emit('error', JXTError.notWellFormed(msg));
    }
    restrictedXML(msg) {
        this.emit('error', JXTError.restrictedXML(msg));
    }
}
function parse$1(data, opts = {}) {
    const p = new Parser(opts);
    let result;
    let element;
    let error = null;
    p.on('text', (text) => {
        if (element) {
            element.children.push(text);
        }
    });
    p.on('startElement', (name, attrs) => {
        const child = new XMLElement(name, attrs);
        if (!result) {
            result = child;
        }
        if (!element) {
            element = child;
        }
        else {
            element = element.appendChild(child);
        }
    });
    p.on('endElement', (name) => {
        if (!element) {
            p.emit('error', JXTError.notWellFormed('a'));
        }
        else if (name === element.name) {
            if (element.parent) {
                element = element.parent;
            }
        }
        else {
            p.emit('error', JXTError.notWellFormed('b'));
        }
    });
    p.on('error', (e) => {
        error = e;
    });
    p.write(data);
    p.end();
    if (error) {
        throw error;
    }
    else {
        return result;
    }
}

function createElement(namespace, name, parentNamespace, parent) {
    if (parent) {
        namespace = namespace || parent.getNamespace();
        const root = parent.getNamespaceRoot(namespace);
        if (root) {
            const prefix = root.useNamespace('', namespace);
            name = `${prefix}:${name}`;
        }
    }
    const el = new XMLElement(name);
    if (name.indexOf(':') < 0 && (!parentNamespace || namespace !== parentNamespace)) {
        el.setAttribute('xmlns', namespace);
    }
    return el;
}
function getLang(xml, lang) {
    return (xml.getAttribute('xml:lang') || lang || '').toLowerCase();
}
function getTargetLang(children, context) {
    const availableLanguages = [];
    for (const child of children) {
        availableLanguages.push(getLang(child, context.lang));
    }
    let targetLanguage;
    if (!context.resolveLanguage) {
        targetLanguage = context.lang;
    }
    else {
        targetLanguage = context.resolveLanguage(availableLanguages, context.acceptLanguages || [], context.lang);
    }
    return targetLanguage || '';
}
function findAll(xml, namespace, element, lang) {
    const existing = xml.getChildren(element, namespace);
    const parentLang = getLang(xml);
    if (existing.length) {
        if (lang) {
            return existing.filter(child => {
                const childLang = getLang(child, parentLang);
                if (childLang === lang) {
                    return true;
                }
            });
        }
        else {
            return existing;
        }
    }
    return [];
}
function findOrCreate(xml, namespace, element, lang) {
    namespace = namespace || xml.getNamespace();
    const existing = findAll(xml, namespace, element, lang);
    if (existing.length) {
        return existing[0];
    }
    const created = createElement(namespace, element, xml.getDefaultNamespace(), xml);
    const parentLang = getLang(xml, lang);
    if (lang && parentLang !== lang) {
        created.setAttribute('xml:lang', lang);
    }
    xml.appendChild(created);
    return created;
}
function createAttributeField(opts) {
    return {
        importer(xml) {
            const rawValue = xml.getAttribute(opts.name, opts.namespace);
            if (!rawValue) {
                return opts.dynamicDefault ? opts.dynamicDefault(rawValue) : opts.staticDefault;
            }
            return opts.parseValue(rawValue);
        },
        exporter(xml, value) {
            if (value === undefined || value === opts.staticDefault) {
                return;
            }
            const output = opts.writeValue(value);
            if (!output && !opts.emitEmpty) {
                return;
            }
            if (!opts.namespace || !opts.prefix) {
                xml.setAttribute(opts.name, output, opts.emitEmpty);
            }
            else {
                let prefix;
                const root = xml.getNamespaceRoot(opts.namespace);
                if (root) {
                    prefix = root.useNamespace(opts.prefix, opts.namespace);
                }
                else {
                    const namespaces = xml.getNamespaceContext();
                    if (!namespaces[opts.namespace]) {
                        prefix = xml.useNamespace(opts.prefix, opts.namespace);
                        namespaces[opts.namespace] = prefix;
                    }
                }
                xml.setAttribute(`${prefix}:${opts.name}`, output, opts.emitEmpty);
            }
        }
    };
}
function createAttributeType(parser, createOpts) {
    return (name, defaultValue = undefined, opts = {}) => {
        opts = Object.assign({ staticDefault: defaultValue }, opts);
        return createAttributeField(Object.assign(Object.assign({ name }, parser), (createOpts ? createOpts(opts) : opts)));
    };
}
function createNamespacedAttributeType(parser, createOpts) {
    return (prefix, namespace, name, defaultValue = undefined, opts = {}) => {
        opts = Object.assign({ staticDefault: defaultValue }, opts);
        return createAttributeField(Object.assign(Object.assign({ name,
            namespace,
            prefix }, parser), (createOpts ? createOpts(opts) : opts)));
    };
}
function createChildAttributeField(opts) {
    const converter = opts.converter ||
        createAttributeField(Object.assign(Object.assign({}, opts), { namespace: opts.attributeNamespace }));
    return {
        importer(xml, context) {
            const child = xml.getChild(opts.element, opts.namespace || xml.getNamespace());
            if (!child) {
                return opts.dynamicDefault ? opts.dynamicDefault() : opts.staticDefault;
            }
            return converter.importer(child, context);
        },
        exporter(xml, value, context) {
            if (value !== undefined && value !== opts.staticDefault) {
                const child = findOrCreate(xml, opts.namespace || xml.getNamespace(), opts.element);
                converter.exporter(child, value, context);
            }
        }
    };
}
function createChildAttributeType(parser, createOpts) {
    return (namespace, element, name, defaultValue = undefined, opts = {}) => {
        opts = Object.assign({ staticDefault: defaultValue }, opts);
        return createChildAttributeField(Object.assign(Object.assign({ element,
            name,
            namespace }, parser), (createOpts ? createOpts(opts) : opts)));
    };
}
function createTextField(opts) {
    return {
        importer(xml) {
            const rawValue = xml.getText();
            if (!rawValue) {
                return opts.dynamicDefault ? opts.dynamicDefault(rawValue) : opts.staticDefault;
            }
            return opts.parseValue(rawValue);
        },
        exporter(xml, value) {
            if (!value && opts.emitEmpty) {
                xml.children.push('');
                return;
            }
            if (value === undefined || value === opts.staticDefault) {
                return;
            }
            const output = opts.writeValue(value);
            if (output) {
                xml.children.push(output);
            }
        }
    };
}
function createChildTextField(opts) {
    const converter = createTextField(opts);
    return {
        importer(xml, context) {
            const children = findAll(xml, opts.namespace || xml.getNamespace(), opts.element);
            const targetLanguage = getTargetLang(children, context);
            if (!children.length) {
                return opts.dynamicDefault ? opts.dynamicDefault() : opts.staticDefault;
            }
            if (opts.matchLanguage) {
                for (const child of children) {
                    if (getLang(child, context.lang) === targetLanguage) {
                        return converter.importer(child, context);
                    }
                }
            }
            return converter.importer(children[0], context);
        },
        exporter(xml, value, context) {
            if (!value && opts.emitEmpty) {
                findOrCreate(xml, opts.namespace || xml.getNamespace(), opts.element, opts.matchLanguage ? context.lang : undefined);
                return;
            }
            if (value !== undefined && value !== opts.staticDefault) {
                const child = findOrCreate(xml, opts.namespace || xml.getNamespace(), opts.element, opts.matchLanguage ? context.lang : undefined);
                converter.exporter(child, value, context);
            }
        }
    };
}
const stringParser = {
    parseValue: v => v,
    writeValue: v => v
};
const integerParser = {
    parseValue: v => parseInt(v, 10),
    writeValue: v => v.toString()
};
const floatParser = {
    parseValue: v => parseFloat(v),
    writeValue: v => v.toString()
};
const boolParser = {
    parseValue: v => {
        if (v === 'true' || v === '1') {
            return true;
        }
        if (v === 'false' || v === '0') {
            return false;
        }
        return;
    },
    writeValue: v => (v ? '1' : '0')
};
const dateParser = {
    parseValue: v => new Date(v),
    writeValue: v => (typeof v === 'string' ? v : v.toISOString())
};
const jsonParser = {
    parseValue: v => JSON.parse(v),
    writeValue: v => JSON.stringify(v)
};
const bufferParser = (encoding = 'utf8') => ({
    parseValue: v => {
        if (encoding === 'base64' && v === '=') {
            v = '';
        }
        return Buffer.from(v.trim(), encoding);
    },
    writeValue: v => {
        let data;
        if (typeof v === 'string') {
            data = Buffer.from(v).toString(encoding);
        }
        else if (v) {
            data = v.toString(encoding);
        }
        else {
            data = '';
        }
        if (encoding === 'base64') {
            data = data || '=';
        }
        return data;
    }
});
const tzOffsetParser = {
    parseValue: v => {
        let sign = -1;
        if (v.charAt(0) === '-') {
            sign = 1;
            v = v.slice(1);
        }
        const split = v.split(':');
        const hours = parseInt(split[0], 10);
        const minutes = parseInt(split[1], 10);
        return (hours * 60 + minutes) * sign;
    },
    writeValue: v => {
        if (typeof v === 'string') {
            return v;
        }
        else {
            let formatted = '-';
            if (v < 0) {
                v = -v;
                formatted = '+';
            }
            const hours = v / 60;
            const minutes = v % 60;
            formatted +=
                (hours < 10 ? '0' : '') + hours + ':' + (minutes < 10 ? '0' : '') + minutes;
            return formatted;
        }
    }
};
// ====================================================================
// Field Types
// ====================================================================
const attribute = createAttributeType(stringParser, opts => (Object.assign({ dynamicDefault: opts.emitEmpty ? v => (v === '' ? '' : opts.staticDefault) : undefined }, opts)));
const booleanAttribute = createAttributeType(boolParser);
const integerAttribute = createAttributeType(integerParser);
const floatAttribute = createAttributeType(floatParser);
const dateAttribute = createAttributeType(dateParser);
const namespacedAttribute = createNamespacedAttributeType(stringParser);
const namespacedBooleanAttribute = createNamespacedAttributeType(boolParser);
const namespacedIntegerAttribute = createNamespacedAttributeType(integerParser);
const namespacedFloatAttribute = createNamespacedAttributeType(floatParser);
const namespacedDateAttribute = createNamespacedAttributeType(dateParser);
const childAttribute = createChildAttributeType(stringParser);
const childBooleanAttribute = createChildAttributeType(boolParser);
const childIntegerAttribute = createChildAttributeType(integerParser);
const childFloatAttribute = createChildAttributeType(floatParser);
const childDateAttribute = createChildAttributeType(dateParser);
const text = (defaultValue) => createTextField(Object.assign({ staticDefault: defaultValue }, stringParser));
const textJSON = () => createTextField(Object.assign({}, jsonParser));
const textBuffer = (encoding = 'utf8') => createTextField(Object.assign({}, bufferParser(encoding)));
function languageAttribute() {
    return {
        importer(xml, context) {
            return getLang(xml, context.lang);
        },
        exporter(xml, value, context) {
            if (value && value.toLowerCase() !== context.lang) {
                xml.setAttribute('xml:lang', value);
            }
            else {
                xml.setAttribute('xml:lang', undefined);
            }
        }
    };
}
const childLanguageAttribute = (namespace, element) => createChildAttributeField(Object.assign({ converter: languageAttribute(), element, name: 'xml:lang', namespace }, stringParser));
const childText = (namespace, element, defaultValue, emitEmpty = false) => createChildTextField(Object.assign({ element,
    emitEmpty, matchLanguage: true, namespace, staticDefault: defaultValue }, stringParser));
const childTextBuffer = (namespace, element, encoding = 'utf8') => createChildTextField(Object.assign({ element, matchLanguage: true, namespace }, bufferParser(encoding)));
const childDate = (namespace, element) => createChildTextField(Object.assign({ element,
    namespace }, dateParser));
const childInteger = (namespace, element, defaultValue) => createChildTextField(Object.assign({ element,
    namespace, staticDefault: defaultValue }, integerParser));
const childFloat = (namespace, element, defaultValue) => createChildTextField(Object.assign({ element,
    namespace, staticDefault: defaultValue }, floatParser));
const childJSON = (namespace, element) => createChildTextField(Object.assign({ element,
    namespace }, jsonParser));
function childTimezoneOffset(namespace, element) {
    return createChildTextField(Object.assign({ element,
        namespace, staticDefault: 0 }, tzOffsetParser));
}
function childBoolean(namespace, element) {
    return {
        importer(xml) {
            const child = xml.getChild(element, namespace || xml.getNamespace());
            if (child) {
                return true;
            }
        },
        exporter(xml, value) {
            if (value) {
                findOrCreate(xml, namespace || xml.getNamespace(), element);
            }
        }
    };
}
const deepChildExporter = (path, xml, value) => {
    if (!value) {
        return;
    }
    let current = xml;
    for (const node of path) {
        current = findOrCreate(current, node.namespace || current.getNamespace(), node.element);
    }
    current.children.push(value.toString());
};
function deepChildText(path, defaultValue) {
    return {
        importer(xml) {
            let current = xml;
            for (const node of path) {
                current = current.getChild(node.element, node.namespace || current.getNamespace());
                if (!current) {
                    return defaultValue;
                }
            }
            return current.getText() || defaultValue;
        },
        exporter(xml, value) {
            deepChildExporter(path, xml, value);
        }
    };
}
function deepChildInteger(path, defaultValue) {
    return {
        importer(xml) {
            let current = xml;
            for (const node of path) {
                current = current.getChild(node.element, node.namespace || current.getNamespace());
                if (!current) {
                    return;
                }
            }
            const data = current.getText();
            if (data) {
                return parseInt(data, 10);
            }
            else if (defaultValue) {
                return defaultValue;
            }
        },
        exporter(xml, value) {
            deepChildExporter(path, xml, value);
        }
    };
}
function deepChildBoolean(path) {
    return {
        importer(xml) {
            let current = xml;
            for (const node of path) {
                current = current.getChild(node.element, node.namespace || current.getNamespace());
                if (!current) {
                    return false;
                }
            }
            return true;
        },
        exporter(xml, value) {
            if (!value) {
                return;
            }
            let current = xml;
            for (const node of path) {
                current = findOrCreate(current, node.namespace || current.getNamespace(), node.element);
            }
        }
    };
}
function deepMultipleChildText(path) {
    const finalChild = path.pop();
    return {
        importer(xml, context) {
            let current = xml;
            for (const node of path) {
                current = current.getChild(node.element, node.namespace || current.getNamespace());
                if (!current) {
                    return [];
                }
            }
            const result = [];
            const children = findAll(current, finalChild.namespace || current.getNamespace(), finalChild.element);
            const targetLanguage = getTargetLang(children, context);
            for (const child of children) {
                if (getLang(child, context.lang) === targetLanguage) {
                    result.push(child.getText());
                }
            }
            return result;
        },
        exporter(xml, values, context) {
            if (!values.length) {
                return;
            }
            let current = xml;
            for (const node of path) {
                current = findOrCreate(current, node.namespace || current.getNamespace(), node.element);
            }
            const { namespace, element } = finalChild;
            for (const value of values) {
                const child = createElement(namespace || current.getNamespace(), element, context.namespace, current);
                child.children.push(value);
                current.appendChild(child);
            }
        }
    };
}
function childEnum(namespace, elements, defaultValue) {
    const elementNames = new Map();
    const valueNames = new Map();
    for (const el of elements) {
        if (typeof el === 'string') {
            elementNames.set(el, el);
            valueNames.set(el, el);
        }
        else {
            elementNames.set(el[1], el[0]);
            valueNames.set(el[0], el[1]);
        }
    }
    return {
        importer(xml) {
            for (const child of xml.children) {
                if (typeof child === 'string') {
                    continue;
                }
                else if (child.getNamespace() === (namespace || xml.getNamespace()) &&
                    elementNames.has(child.getName())) {
                    return elementNames.get(child.getName());
                }
            }
            return defaultValue;
        },
        exporter(xml, value) {
            if (valueNames.has(value)) {
                findOrCreate(xml, namespace, valueNames.get(value));
            }
        }
    };
}
function childDoubleEnum(namespace, parentElements, childElements, defaultValue) {
    const parentNames = new Set(parentElements);
    const childNames = new Set(childElements);
    return {
        importer(xml) {
            for (const parent of xml.children) {
                if (typeof parent === 'string') {
                    continue;
                }
                else if (parent.getNamespace() === (namespace || xml.getNamespace()) &&
                    parentNames.has(parent.getName())) {
                    for (const child of parent.children) {
                        if (typeof child === 'string') {
                            continue;
                        }
                        else if (child.getNamespace() === (namespace || xml.getNamespace()) &&
                            childNames.has(child.getName())) {
                            return [parent.getName(), child.getName()];
                        }
                    }
                    return [parent.getName()];
                }
            }
            return defaultValue;
        },
        exporter(xml, value) {
            const parent = findOrCreate(xml, namespace, value[0]);
            if (value[1]) {
                findOrCreate(parent, namespace, value[1]);
            }
        }
    };
}
function multipleChildText(namespace, element) {
    return {
        importer(xml, context) {
            const result = [];
            const children = findAll(xml, namespace || xml.getNamespace(), element);
            const targetLanguage = getTargetLang(children, context);
            for (const child of children) {
                if (getLang(child, context.lang) === targetLanguage) {
                    result.push(child.getText());
                }
            }
            return result;
        },
        exporter(xml, values, context) {
            for (const value of values) {
                const child = createElement(namespace || xml.getNamespace(), element, context.namespace, xml);
                child.children.push(value);
                xml.appendChild(child);
            }
        }
    };
}
function multipleChildAttribute(namespace, element, name) {
    return {
        importer(xml) {
            const result = [];
            const children = xml.getChildren(element, namespace || xml.getNamespace());
            for (const child of children) {
                const childAttr = child.getAttribute(name);
                if (childAttr !== undefined) {
                    result.push(childAttr);
                }
            }
            return result;
        },
        exporter(xml, values, context) {
            for (const value of values) {
                const child = createElement(namespace || xml.getNamespace(), element, context.namespace, xml);
                child.setAttribute(name, value);
                xml.appendChild(child);
            }
        }
    };
}
function multipleChildIntegerAttribute(namespace, element, name) {
    return {
        importer(xml) {
            const result = [];
            const children = xml.getChildren(element, namespace || xml.getNamespace());
            for (const child of children) {
                const childAttr = child.getAttribute(name);
                if (childAttr !== undefined) {
                    result.push(parseInt(childAttr, 10));
                }
            }
            return result;
        },
        exporter(xml, values, context) {
            for (const value of values) {
                const child = createElement(namespace || xml.getNamespace(), element, context.namespace, xml);
                child.setAttribute(name, value.toString());
                xml.appendChild(child);
            }
        }
    };
}
function childAlternateLanguageText(namespace, element) {
    return {
        importer(xml, context) {
            const results = [];
            const children = findAll(xml, namespace || xml.getNamespace(), element);
            const seenLanuages = new Set();
            for (const child of children) {
                const langText = child.getText();
                if (langText) {
                    const lang = getLang(child, context.lang);
                    if (seenLanuages.has(lang)) {
                        continue;
                    }
                    results.push({ lang, value: langText });
                    seenLanuages.add(lang);
                }
            }
            return seenLanuages.size > 0 ? results : undefined;
        },
        exporter(xml, values, context) {
            for (const entry of values) {
                const val = entry.value;
                if (val) {
                    const child = createElement(namespace || xml.getNamespace(), element, context.namespace, xml);
                    if (entry.lang !== context.lang) {
                        child.setAttribute('xml:lang', entry.lang);
                    }
                    child.children.push(val);
                    xml.appendChild(child);
                }
            }
        }
    };
}
function multipleChildAlternateLanguageText(namespace, element) {
    return {
        importer(xml, context) {
            const results = [];
            const langIndex = new Map();
            let hasResults = false;
            const children = findAll(xml, namespace || xml.getNamespace(), element);
            for (const child of children) {
                const langText = child.getText();
                if (langText) {
                    const lang = getLang(child, context.lang);
                    let langResults = langIndex.get(lang);
                    if (!langResults) {
                        langResults = [];
                        langIndex.set(lang, langResults);
                        results.push({ lang, value: langResults });
                    }
                    langResults.push(langText);
                    hasResults = true;
                }
            }
            return hasResults ? results : undefined;
        },
        exporter(xml, values, context) {
            for (const entry of values) {
                for (const val of entry.value) {
                    const child = createElement(namespace || xml.getNamespace(), element, context.namespace, xml);
                    if (entry.lang !== context.lang) {
                        child.setAttribute('xml:lang', entry.lang);
                    }
                    child.children.push(val);
                    xml.appendChild(child);
                }
            }
        }
    };
}
function multipleChildEnum(namespace, elements) {
    const elementNames = new Map();
    const valueNames = new Map();
    for (const el of elements) {
        if (typeof el === 'string') {
            elementNames.set(el, el);
            valueNames.set(el, el);
        }
        else {
            elementNames.set(el[1], el[0]);
            valueNames.set(el[0], el[1]);
        }
    }
    return {
        importer(xml) {
            const results = [];
            for (const child of xml.children) {
                if (typeof child === 'string') {
                    continue;
                }
                else if (child.getNamespace() === (namespace || xml.getNamespace()) &&
                    elementNames.has(child.getName())) {
                    results.push(elementNames.get(child.getName()));
                }
            }
            return results;
        },
        exporter(xml, values) {
            for (const value of values) {
                findOrCreate(xml, namespace, valueNames.get(value));
            }
        }
    };
}
function splicePath(namespace, element, path, multiple = false) {
    return {
        importer(xml, context) {
            const child = xml.getChild(element, namespace || xml.getNamespace());
            if (!child) {
                return;
            }
            const results = [];
            for (const grandChild of child.children) {
                if (typeof grandChild === 'string') {
                    continue;
                }
                if (context.registry.getImportKey(grandChild) === path) {
                    const imported = context.registry.import(grandChild);
                    if (imported) {
                        results.push(imported);
                    }
                }
            }
            return multiple ? results : results[0];
        },
        exporter(xml, data, context) {
            let values = [];
            if (!Array.isArray(data)) {
                values = [data];
            }
            else {
                values = data;
            }
            const children = [];
            for (const value of values) {
                const child = context.registry.export(path, value, Object.assign(Object.assign({}, context), { namespace: namespace || xml.getNamespace() || undefined }));
                if (child) {
                    children.push(child);
                }
            }
            if (children.length) {
                const skipChild = findOrCreate(xml, namespace || xml.getNamespace(), element);
                for (const child of children) {
                    skipChild.appendChild(child);
                }
            }
        }
    };
}
function staticValue(value) {
    return {
        exporter: () => undefined,
        importer: () => value
    };
}
function childRawElement(namespace, element, sanitizer) {
    return {
        importer(xml, context) {
            if (sanitizer && (!context.sanitizers || !context.sanitizers[sanitizer])) {
                return;
            }
            const child = xml.getChild(element, namespace || xml.getNamespace());
            if (child) {
                if (sanitizer) {
                    return context.sanitizers[sanitizer](child.toJSON());
                }
                else {
                    return child.toJSON();
                }
            }
        },
        exporter(xml, value, context) {
            if (typeof value === 'string') {
                const wrapped = parse$1(`<${element} xmlns="${namespace || xml.getNamespace()}">${value}</${element}>`);
                value = wrapped.toJSON();
            }
            if (sanitizer) {
                if (!context.sanitizers || !context.sanitizers[sanitizer]) {
                    return;
                }
                value = context.sanitizers[sanitizer](value);
            }
            if (value) {
                xml.appendChild(new XMLElement(value.name, value.attributes, value.children));
            }
        }
    };
}
function childLanguageRawElement(namespace, element, sanitizer) {
    return {
        importer(xml, context) {
            if (sanitizer && (!context.sanitizers || !context.sanitizers[sanitizer])) {
                return;
            }
            const children = findAll(xml, namespace || xml.getNamespace(), element);
            const targetLanguage = getTargetLang(children, context);
            for (const child of children) {
                if (getLang(child, context.lang) === targetLanguage) {
                    if (sanitizer) {
                        return context.sanitizers[sanitizer](child.toJSON());
                    }
                    else {
                        return child.toJSON();
                    }
                }
            }
            if (children[0]) {
                if (sanitizer) {
                    return context.sanitizers[sanitizer](children[0].toJSON());
                }
                else {
                    return children[0].toJSON();
                }
            }
        },
        exporter(xml, value, context) {
            if (typeof value === 'string') {
                const wrapped = parse$1(`<${element} xmlns="${namespace || xml.getNamespace()}">${value}</${element}>`);
                value = wrapped.toJSON();
            }
            if (value && sanitizer) {
                if (!context.sanitizers || !context.sanitizers[sanitizer]) {
                    return;
                }
                value = context.sanitizers[sanitizer](value);
            }
            if (!value) {
                return;
            }
            const rawElement = findOrCreate(xml, namespace || xml.getNamespace(), element, context.lang);
            for (const child of value.children) {
                if (typeof child === 'string') {
                    rawElement.appendChild(child);
                }
                else if (child) {
                    rawElement.appendChild(new XMLElement(child.name, child.attributes, child.children));
                }
            }
        }
    };
}
function childAlternateLanguageRawElement(namespace, element, sanitizer) {
    return {
        importer(xml, context) {
            if (sanitizer && (!context.sanitizers || !context.sanitizers[sanitizer])) {
                return;
            }
            const results = [];
            const seenLanuages = new Set();
            const children = findAll(xml, namespace || xml.getNamespace(), element);
            for (const child of children) {
                let result = child.toJSON();
                if (sanitizer) {
                    result = context.sanitizers[sanitizer](result);
                }
                if (result) {
                    const lang = getLang(child, context.lang);
                    if (seenLanuages.has(lang)) {
                        continue;
                    }
                    results.push({ lang, value: result });
                    seenLanuages.add(lang);
                }
            }
            return seenLanuages.size > 0 ? results : undefined;
        },
        exporter(xml, values, context) {
            for (const entry of values) {
                let value = entry.value;
                if (typeof value === 'string') {
                    const wrapped = parse$1(`<${element} xmlns="${namespace || xml.getNamespace()}">${value}</${element}>`);
                    value = wrapped.toJSON();
                }
                if (value && sanitizer) {
                    if (!context.sanitizers || !context.sanitizers[sanitizer]) {
                        continue;
                    }
                    value = context.sanitizers[sanitizer](value);
                }
                if (value) {
                    const rawElement = createElement(namespace || xml.getNamespace(), element, context.namespace, xml);
                    xml.appendChild(rawElement);
                    if (entry.lang !== context.lang) {
                        rawElement.setAttribute('xml:lang', entry.lang);
                    }
                    //yjing modify
                    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
                    // @ts-ignore
                    if (entry.value && 'attributes' in entry.value && entry.value.attributes) {
                        if (entry.value.attributes.type) {
                            rawElement.setAttribute('type', entry.value.attributes.type);
                        }
                        if (entry.value.attributes.affiliation) {
                            rawElement.setAttribute('affiliation', entry.value.attributes.affiliation);
                        }
                        if (entry.value.attributes.jid) {
                            rawElement.setAttribute('jid', entry.value.attributes.jid);
                        }
                        if (entry.value.attributes.since) {
                            rawElement.setAttribute('since', entry.value.attributes.since);
                        }
                    }
                    //yjing modify end
                    for (const child of value.children) {
                        if (typeof child === 'string') {
                            rawElement.appendChild(child);
                        }
                        else {
                            rawElement.appendChild(new XMLElement(child.name, child.attributes, child.children));
                        }
                    }
                }
            }
        }
    };
}
function parameterMap(namespace, element, keyName, valueName) {
    return {
        importer(xml, context) {
            const result = {};
            const params = findAll(xml, namespace, element);
            const keyImporter = attribute(keyName).importer;
            const valueImporter = attribute(valueName).importer;
            for (const param of params) {
                result[keyImporter(param, context)] = valueImporter(param, context);
            }
            return result;
        },
        exporter(xml, values, context) {
            const keyExporter = attribute(keyName).exporter;
            const valueExporter = attribute(valueName).exporter;
            const ns = namespace || xml.getNamespace();
            for (const [param, value] of Object.entries(values)) {
                const paramEl = createElement(ns, element, context.namespace, xml);
                keyExporter(paramEl, param, context);
                if (values[param]) {
                    valueExporter(paramEl, value, context);
                }
                xml.appendChild(paramEl);
            }
        }
    };
}

class Translator {
    constructor() {
        this.parents = new Set();
        this.placeholder = false;
        this.typeField = '';
        this.versionField = '';
        this.defaultType = '';
        this.defaultVersion = '';
        this.languageField = 'lang';
        this.typeValues = new Map();
        this.typeOrders = new Map();
        this.importers = new Map();
        this.exporters = new Map();
        this.children = new Map();
        this.childrenIndex = new Map();
        this.implicitChildren = new Set();
        this.contexts = new Map();
    }
    addChild(name, translator, multiple = false, selector, implicit) {
        const child = {
            multiple: multiple || false,
            name,
            selector,
            translator
        };
        const existingChild = this.children.get(name);
        if (!existingChild) {
            child.translator.parents.add(this);
            this.children.set(name, child);
            for (const [xid] of translator.importers) {
                if (!this.implicitChildren.has(xid)) {
                    this.childrenIndex.set(xid, name);
                }
            }
            if (implicit) {
                this.implicitChildren.add(implicit);
            }
            return;
        }
        const existing = existingChild.translator;
        existingChild.multiple = multiple;
        if (selector && existingChild.selector && selector !== existingChild.selector) {
            existingChild.selector = undefined;
        }
        for (const [xid, importer] of translator.importers) {
            const [type, version] = (existing.typeValues.get(xid) || '').split('__v__');
            existing.updateDefinition({
                contexts: translator.contexts,
                element: importer.element,
                exporterOrdering: new Map(),
                exporters: new Map(),
                importerOrdering: importer.fieldOrders,
                importers: importer.fields,
                namespace: importer.namespace,
                optionalNamespaces: new Map(),
                type,
                version
            });
            if (!this.implicitChildren.has(xid)) {
                this.childrenIndex.set(xid, name);
            }
        }
        for (const [exportType, exporter] of translator.exporters) {
            const [type, version] = exportType.split('__v__');
            existing.updateDefinition({
                contexts: translator.contexts,
                element: exporter.element,
                exporterOrdering: exporter.fieldOrders,
                exporters: exporter.fields,
                importerOrdering: new Map(),
                importers: new Map(),
                namespace: exporter.namespace,
                optionalNamespaces: exporter.optionalNamespaces,
                type,
                version
            });
        }
    }
    addContext(path, selector, field, xid, value, implied) {
        if (selector) {
            path = `${path}[${selector}]`;
        }
        let context = this.contexts.get(path);
        if (!context) {
            context = {
                typeField: '',
                versionField: '',
                typeValues: new Map()
            };
        }
        if (implied) {
            context.impliedType = value;
        }
        context.typeField = field || '';
        context.typeValues.set(xid, value);
        this.contexts.set(path, context);
    }
    getChild(name) {
        const child = this.children.get(name);
        if (!child) {
            return;
        }
        return child.translator;
    }
    getImportKey(xml) {
        return this.childrenIndex.get(`{${xml.getNamespace()}}${xml.getName()}`);
    }
    updateDefinition(opts) {
        const xid = `{${opts.namespace}}${opts.element}`;
        const type = opts.type || this.defaultType;
        const version = opts.version || this.defaultVersion;
        const versionType = version ? `${type}__v__${version}` : type;
        const importer = this.importers.get(xid) ||
            {
                element: opts.element,
                fieldOrders: new Map(),
                fields: new Map(),
                namespace: opts.namespace
            };
        for (const [fieldName, fieldImporter] of opts.importers) {
            importer.fields.set(fieldName, fieldImporter);
        }
        for (const [fieldName, order] of opts.importerOrdering) {
            importer.fieldOrders.set(fieldName, order);
        }
        this.importers.set(xid, importer);
        const exporter = this.exporters.get(versionType) ||
            {
                element: opts.element,
                fieldOrders: new Map(),
                fields: new Map(),
                namespace: opts.namespace,
                optionalNamespaces: opts.optionalNamespaces
            };
        for (const [fieldName, fieldExporter] of opts.exporters) {
            exporter.fields.set(fieldName, fieldExporter);
        }
        for (const [name, order] of opts.exporterOrdering) {
            exporter.fieldOrders.set(name, order);
        }
        for (const [prefix, namespace] of opts.optionalNamespaces) {
            exporter.optionalNamespaces.set(prefix, namespace);
        }
        this.exporters.set(versionType, exporter);
        for (const [path, newContext] of opts.contexts) {
            const context = this.contexts.get(path) || {
                impliedType: undefined,
                typeField: newContext.typeField,
                versionField: newContext.versionField,
                typeValues: new Map()
            };
            if (!context.typeField) {
                context.typeField = newContext.typeField;
            }
            if (!context.versionField) {
                context.versionField = newContext.versionField;
            }
            if (!context.impliedType) {
                context.impliedType = newContext.impliedType;
            }
            for (const [xid2, type] of newContext.typeValues) {
                context.typeValues.set(xid2, type);
            }
            this.contexts.set(path, context);
        }
        if (opts.type) {
            this.typeValues.set(xid, versionType);
            if (opts.typeOrder && opts.type) {
                this.typeOrders.set(opts.type, opts.typeOrder);
            }
        }
        else if (this.typeField && !opts.type) {
            for (const [, imp] of this.importers) {
                for (const [fieldName, fieldImporter] of opts.importers) {
                    imp.fields.set(fieldName, fieldImporter);
                }
                for (const [fieldName, order] of opts.importerOrdering) {
                    imp.fieldOrders.set(fieldName, order);
                }
            }
            for (const [, exp] of this.exporters) {
                for (const [fieldName, fieldExporter] of opts.exporters) {
                    exp.fields.set(fieldName, fieldExporter);
                }
                for (const [fieldName, order] of opts.exporterOrdering) {
                    exp.fieldOrders.set(fieldName, order);
                }
            }
        }
    }
    replaceWith(replacement) {
        for (const [a, b] of this.children) {
            replacement.children.set(a, b);
        }
        for (const [a, b] of this.childrenIndex) {
            replacement.childrenIndex.set(a, b);
        }
        for (const [a, b] of this.contexts) {
            replacement.contexts.set(a, b);
        }
        for (const a of this.implicitChildren) {
            replacement.implicitChildren.add(a);
        }
        for (const parent of this.parents) {
            for (const child of parent.children.values()) {
                if (child.translator === this) {
                    child.translator = replacement;
                }
            }
        }
        this.parents = new Set();
    }
    import(xml, parentContext) {
        const xid = `{${xml.getNamespace()}}${xml.getName()}`;
        const output = {};
        const importer = this.importers.get(xid);
        if (!importer) {
            return;
        }
        const versionTypeValue = this.typeValues.get(xid) || '';
        const [typeValue, versionValue] = versionTypeValue.split('__v__');
        const path = parentContext.path || '';
        let implied;
        if (parentContext.pathSelector) {
            implied = this.contexts.get(`${path}[${parentContext.pathSelector}]`);
        }
        if (!implied) {
            implied = this.contexts.get(path);
        }
        if (implied) {
            if (!implied.impliedType) {
                const impliedTypeValue = implied.typeValues.get(xid) || '';
                if (impliedTypeValue) {
                    output[implied.typeField] = impliedTypeValue;
                }
            }
        }
        else if (this.typeField && typeValue && typeValue !== this.defaultType) {
            output[this.typeField] = typeValue;
        }
        if (this.versionField && versionValue && versionValue !== this.defaultVersion) {
            output[this.versionField] = versionValue;
        }
        const context = Object.assign(Object.assign({}, parentContext), { data: output, importer, lang: (xml.getAttribute('xml:lang') || parentContext.lang || '').toLowerCase(), pathSelector: typeValue, translator: this });
        const importFields = [...importer.fieldOrders.entries()].sort((a, b) => a[1] > b[1] ? -1 : a[1] < b[1] ? 1 : 0);
        const preChildren = importFields.filter(field => field[1] >= 0);
        const postChildren = importFields.filter(field => field[1] < 0);
        for (const [fieldName] of preChildren) {
            const importField = importer.fields.get(fieldName);
            context.path = `${parentContext.path}.${fieldName}`;
            const value = importField(xml, context);
            if (value !== null && value !== undefined) {
                output[fieldName] = value;
            }
        }
        for (const child of xml.children) {
            if (typeof child === 'string') {
                continue;
            }
            const childName = `{${child.getNamespace()}}${child.getName()}`;
            const fieldName = this.childrenIndex.get(childName);
            if (!fieldName) {
                continue;
            }
            context.path = `${parentContext.path}.${fieldName}`;
            const { translator, multiple, selector } = this.children.get(fieldName);
            if (!selector || selector === typeValue) {
                const childOutput = translator.import(child, context);
                if (childOutput !== undefined) {
                    if (multiple) {
                        if (!output[fieldName]) {
                            output[fieldName] = [];
                        }
                        output[fieldName].push(childOutput);
                    }
                    else if (!output[fieldName]) {
                        output[fieldName] = childOutput;
                    }
                    else {
                        output[fieldName] = translator.resolveCollision(output[fieldName], childOutput);
                    }
                }
            }
        }
        for (const [fieldName] of postChildren) {
            const importField = importer.fields.get(fieldName);
            context.path = `${parentContext.path}.${fieldName}`;
            const value = importField(xml, context);
            if (value !== null && value !== undefined) {
                output[fieldName] = value;
            }
        }
        return output;
    }
    export(data, parentContext) {
        if (!data) {
            return;
        }
        let exportType = this.defaultType;
        let exportVersion = this.defaultVersion;
        const path = parentContext.path || '';
        let implied;
        if (parentContext.pathSelector) {
            implied = this.contexts.get(`${path}[${parentContext.pathSelector}]`);
        }
        if (!implied) {
            implied = this.contexts.get(path);
        }
        if (implied) {
            exportType = implied.impliedType || data[implied.typeField] || this.defaultType;
        }
        else if (this.typeField) {
            exportType = data[this.typeField] || this.defaultType;
        }
        if (this.versionField) {
            exportVersion = data[this.versionField] || this.defaultVersion;
        }
        //yjing modify
        switch (path) {
            case 'iq.createGroupStanza':
                exportVersion = 'createGroupStanza';
                break;
            case 'iq.groupAffiliationIQ':
                exportVersion = 'groupAffiliationIQ';
                break;
        }
        //yjing modify end
        let exportVersionType = exportVersion ? `${exportType}__v__${exportVersion}` : exportType;
        //yjing modify
        if (data['customFieldName']) {
            exportVersionType = data['customFieldName'];
        }
        //yjing modify end
        const exporter = this.exporters.get(exportVersionType);
        if (!exporter) {
            return;
        }
        const output = createElement(exporter.namespace, exporter.element, parentContext.namespace, parentContext.element);
        if (parentContext.element) {
            output.parent = parentContext.element;
        }
        for (const [prefix, namespace] of exporter.optionalNamespaces) {
            output.addOptionalNamespace(prefix, namespace);
        }
        const context = Object.assign(Object.assign({}, parentContext), { data, element: output, exporter, lang: (data[this.languageField] || parentContext.lang || '').toLowerCase(), namespace: output.getDefaultNamespace(), pathSelector: exportType, translator: this });
        const langExporter = exporter.fields.get(this.languageField);
        if (langExporter) {
            langExporter(output, data[this.languageField], parentContext);
        }
        const keys = Object.keys(data);
        keys.sort((key1, key2) => {
            const a = exporter.fieldOrders.get(key1) || 100000;
            const b = exporter.fieldOrders.get(key2) || 100000;
            return a - b;
        });
        for (const key of keys) {
            if (key === this.languageField) {
                // We've already processed this field
                continue;
            }
            const value = data[key];
            const fieldExporter = exporter.fields.get(key);
            if (fieldExporter) {
                fieldExporter(output, value, context);
                continue;
            }
            const childTranslator = this.children.get(key);
            if (!childTranslator) {
                continue;
            }
            context.path = `${parentContext.path ? parentContext.path + '.' : ''}${key}`;
            const { translator, multiple, selector } = childTranslator;
            if (!selector || selector === exportType) {
                let items;
                if (multiple) {
                    items = value !== null && value !== void 0 ? value : [];
                }
                else {
                    items = [value];
                }
                for (const item of items) {
                    const childOutput = translator.export(item, context);
                    if (childOutput) {
                        output.appendChild(childOutput);
                    }
                }
            }
        }
        return output;
    }
    resolveCollision(existingData, newData) {
        const existingOrder = this.typeOrders.get(existingData[this.typeField] || this.defaultType) || 0;
        const newOrder = this.typeOrders.get(newData[this.typeField] || this.defaultType) || 0;
        return existingOrder <= newOrder ? existingData : newData;
    }
}

class Registry {
    constructor() {
        this.translators = new Map();
        this.root = new Translator();
        this.setLanguageResolver(basicLanguageResolver);
    }
    setLanguageResolver(resolver) {
        this.languageResolver = resolver;
    }
    import(xml, context = { registry: this }) {
        if (!this.hasTranslator(xml.getNamespace(), xml.getName())) {
            return;
        }
        if (!context.acceptLanguages) {
            context.acceptLanguages = [];
        }
        context.acceptLanguages = context.acceptLanguages.map(lang => lang.toLowerCase());
        if (context.lang) {
            context.lang = context.lang.toLowerCase();
        }
        if (!context.resolveLanguage) {
            context.resolveLanguage = this.languageResolver;
        }
        context.path = this.getImportKey(xml);
        if (!context.sanitizers) {
            context.sanitizers = {
                xhtmlim: sanitizeRoot
            };
        }
        const translator = this.getOrCreateTranslator(xml.getNamespace(), xml.getName());
        return translator.import(xml, Object.assign(Object.assign({}, context), { registry: this }));
    }
    export(path, data, context = { registry: this }) {
        if (!context.acceptLanguages) {
            context.acceptLanguages = [];
        }
        context.acceptLanguages = context.acceptLanguages.map(lang => lang.toLowerCase());
        if (context.lang) {
            context.lang = context.lang.toLowerCase();
        }
        if (!context.sanitizers) {
            context.sanitizers = {
                xhtmlim: sanitizeRoot
            };
        }
        context.path = path;
        const fields = path.split('.').filter(item => {
            return item !== '';
        });
        let translator = this.root;
        for (const field of fields) {
            const nextTranslator = translator.getChild(field);
            if (!nextTranslator) {
                return;
            }
            translator = nextTranslator;
        }
        return translator.export(data, Object.assign(Object.assign({}, context), { registry: this }));
    }
    getImportKey(xml, path = '') {
        const root = !path ? this.root : this.walkToTranslator(path.split('.'));
        if (!root) {
            return undefined;
        }
        return root.getImportKey(xml);
    }
    define(defs) {
        if (Array.isArray(defs)) {
            for (const def of defs) {
                if (typeof def === 'object') {
                    this.define(def);
                }
                else {
                    def(this);
                }
            }
            return;
        }
        else if (typeof defs !== 'object') {
            defs(this);
            return;
        }
        const definition = defs;
        definition.aliases = definition.aliases || [];
        if (definition.path && !definition.aliases.includes(definition.path)) {
            definition.aliases.push(definition.path);
        }
        const aliases = definition.aliases
            .map(alias => (typeof alias === 'string' ? { path: alias } : alias))
            .sort((a, b) => {
            const aLen = a.path.split('.').length;
            const bLen = b.path.split('.').length;
            return bLen - aLen;
        });
        let translator;
        if (this.hasTranslator(definition.namespace, definition.element)) {
            // Get existing translator
            translator = this.getOrCreateTranslator(definition.namespace, definition.element);
        }
        if (!translator) {
            let placeholder;
            for (const alias of aliases) {
                const t = this.walkToTranslator(alias.path.split('.'));
                if (t && !t.placeholder) {
                    translator = t;
                    break;
                }
                else if (t) {
                    placeholder = t;
                }
            }
            if (placeholder && !translator) {
                translator = placeholder;
                translator.placeholder = false;
            }
        }
        if (!translator) {
            // Create a new translator
            translator = this.getOrCreateTranslator(definition.namespace, definition.element);
        }
        this.indexTranslator(definition.namespace, definition.element, translator);
        const fields = definition.fields || {};
        const importers = new Map();
        const exporters = new Map();
        const importerOrdering = new Map();
        const exporterOrdering = new Map();
        if (definition.typeField) {
            translator.typeField = definition.typeField;
        }
        if (definition.defaultType) {
            translator.defaultType = definition.defaultType;
        }
        if (definition.versionField) {
            translator.versionField = definition.versionField;
        }
        if (definition.defaultVersion) {
            translator.defaultVersion = definition.defaultVersion;
        }
        if (definition.languageField) {
            translator.languageField = definition.languageField;
        }
        for (const [key, field] of Object.entries(fields)) {
            importers.set(key, field.importer);
            importerOrdering.set(key, field.importOrder || field.order || 0);
            exporters.set(key, field.exporter);
            exporterOrdering.set(key, field.exportOrder || field.order || 0);
        }
        if (definition.childrenExportOrder) {
            for (const [key, order] of Object.entries(definition.childrenExportOrder)) {
                exporterOrdering.set(key, order || 0);
            }
        }
        const optionalNamespaces = new Map();
        for (const [prefix, namespace] of Object.entries(definition.optionalNamespaces || {})) {
            optionalNamespaces.set(prefix, namespace);
        }
        translator.updateDefinition({
            contexts: new Map(),
            element: definition.element,
            exporterOrdering,
            exporters,
            importerOrdering,
            importers,
            namespace: definition.namespace,
            optionalNamespaces,
            type: definition.type,
            version: definition.version,
            typeOrder: definition.typeOrder
        });
        for (const link of aliases) {
            this.alias(definition.namespace, definition.element, link.path, link.multiple, link.selector, link.contextField, definition.type, link.impliedType);
        }
        for (const alias of aliases) {
            const existing = this.walkToTranslator(alias.path.split('.'));
            if (existing && existing !== translator) {
                existing.replaceWith(translator);
            }
        }
    }
    alias(namespace, element, path, multiple = false, selector, contextField, contextType, contextImpliedType = false) {
        const linkedTranslator = this.getOrCreateTranslator(namespace, element);
        linkedTranslator.placeholder = false;
        const keys = path.split('.').filter(key => {
            return key !== '';
        });
        const finalKey = keys.pop();
        const translator = this.walkToTranslator(keys, true);
        const xid = `{${namespace}}${element}`;
        if (contextType && (contextField || contextImpliedType)) {
            linkedTranslator.addContext(path, selector, contextField, xid, contextType, contextImpliedType);
        }
        translator.addChild(finalKey, linkedTranslator, multiple, selector, xid);
    }
    walkToTranslator(path, vivify = false) {
        let translator = this.root;
        for (const key of path) {
            let next = translator.getChild(key);
            if (!next) {
                if (vivify) {
                    next = new Translator();
                    next.placeholder = true;
                    translator.addChild(key, next);
                }
                else {
                    return;
                }
            }
            translator = next;
        }
        return translator;
    }
    hasTranslator(namespace, element) {
        return this.translators.has(`{${namespace}}${element}`);
    }
    getOrCreateTranslator(namespace, element) {
        let translator = this.translators.get(`{${namespace}}${element}`);
        if (!translator) {
            translator = new Translator();
            this.indexTranslator(namespace, element, translator);
        }
        return translator;
    }
    indexTranslator(namespace, element, translator) {
        this.translators.set(`{${namespace}}${element}`, translator);
    }
}

// ================================================================
// RFCS
// ================================================================
// RFC 4287
const NS_ATOM = 'http://www.w3.org/2005/Atom';
// RFC 6120
const NS_BIND = 'urn:ietf:params:xml:ns:xmpp-bind';
const NS_CLIENT = 'jabber:client';
const NS_SASL = 'urn:ietf:params:xml:ns:xmpp-sasl';
const NS_SERVER = 'jabber:server';
const NS_SESSION = 'urn:ietf:params:xml:ns:xmpp-session';
const NS_STANZAS = 'urn:ietf:params:xml:ns:xmpp-stanzas';
const NS_STREAM = 'http://etherx.jabber.org/streams';
const NS_STREAMS = 'urn:ietf:params:xml:ns:xmpp-streams';
const NS_STARTTLS = 'urn:ietf:params:xml:ns:xmpp-tls';
// RFC 6121
const NS_ROSTER = 'jabber:iq:roster';
const NS_ROSTER_VERSIONING = 'urn:xmpp:features:rosterver';
const NS_SUBSCRIPTION_PREAPPROVAL = 'urn:xmpp:features:pre-approval';
// RFC 7395
const NS_FRAMING = 'urn:ietf:params:xml:ns:xmpp-framing';
// ================================================================
// XEPS
// ================================================================
// XEP-0004
const NS_DATAFORM = 'jabber:x:data';
// XEP-0009
const NS_RPC = 'jabber:iq:rpc';
// XEP-0012
const NS_LAST_ACTIVITY = 'jabber:iq:last';
// XEP-0016
const NS_PRIVACY = 'jabber:iq:privacy';
// XEP-0022
const NS_LEGACY_CHAT_EVENTS = 'jabber:x:event';
// XEP-0030
const NS_DISCO_INFO = 'http://jabber.org/protocol/disco#info';
const NS_DISCO_ITEMS = 'http://jabber.org/protocol/disco#items';
// XEP-0033
const NS_ADDRESS = 'http://jabber.org/protocol/address';
// XEP-0045
const NS_MUC = 'http://jabber.org/protocol/muc';
const NS_MUC_ADMIN = 'http://jabber.org/protocol/muc#admin';
const NS_MUC_OWNER = 'http://jabber.org/protocol/muc#owner';
const NS_MUC_USER = 'http://jabber.org/protocol/muc#user';
// XEP-0047
const NS_IBB = 'http://jabber.org/protocol/ibb';
// XEP-0048
const NS_BOOKMARKS = 'storage:bookmarks';
// XEP-0049
const NS_PRIVATE = 'jabber:iq:private';
// XEP-0050
const NS_ADHOC_COMMANDS = 'http://jabber.org/protocol/commands';
// XEP-0054
const NS_VCARD_TEMP = 'vcard-temp';
// XEP-0055
const NS_SEARCH = 'jabber:iq:search';
// XEP-0059
const NS_RSM = 'http://jabber.org/protocol/rsm';
// XEP-0060
const NS_PUBSUB = 'http://jabber.org/protocol/pubsub';
const NS_PUBSUB_ERRORS = 'http://jabber.org/protocol/pubsub#errors';
const NS_PUBSUB_EVENT = 'http://jabber.org/protocol/pubsub#event';
const NS_PUBSUB_OWNER = 'http://jabber.org/protocol/pubsub#owner';
// XEP-0065
const NS_SOCKS5 = 'http://jabber.org/protocol/bytestreams';
// XEP-0066
const NS_OOB = 'jabber:x:oob';
const NS_OOB_TRANSFER = 'jabber:iq:oob';
// XEP-0070
const NS_HTTP_AUTH = 'http://jabber.org/protocol/http-auth';
// XEP-0071
const NS_XHTML = 'http://www.w3.org/1999/xhtml';
const NS_XHTML_IM = 'http://jabber.org/protocol/xhtml-im';
// XEP-0077
const NS_REGISTER = 'jabber:iq:register';
const NS_INBAND_REGISTRATION = 'http://jabber.org/features/iq-register';
// XEP-0079
const NS_AMP = 'http://jabber.org/protocol/amp';
// XEP-0080
const NS_GEOLOC = 'http://jabber.org/protocol/geoloc';
// XEP-0083
const NS_ROSTER_DELIMITER = 'roster:delimiter';
// XEP-0084
const NS_AVATAR_DATA = 'urn:xmpp:avatar:data';
const NS_AVATAR_METADATA = 'urn:xmpp:avatar:metadata';
// XEP-0085
const NS_CHAT_STATES = 'http://jabber.org/protocol/chatstates';
// XEP-0092
const NS_VERSION = 'jabber:iq:version';
// XEP-0107
const NS_MOOD = 'http://jabber.org/protocol/mood';
// XEP-0108
const NS_ACTIVITY = 'http://jabber.org/protocol/activity';
// XEP-0114
const NS_COMPONENT = 'jabber:component:accept';
// XEP-0115
const NS_DISCO_LEGACY_CAPS = 'http://jabber.org/protocol/caps';
// XEP-0118
const NS_TUNE = 'http://jabber.org/protocol/tune';
// XEP-0122
const NS_DATAFORM_VALIDATION = 'http://jabber.org/protocol/xdata-validate';
// XEP-0124
const NS_BOSH = 'http://jabber.org/protocol/httpbind';
// XEP-0131
const NS_SHIM = 'http://jabber.org/protocol/shim';
// XEP-0138
const NS_COMPRESSION_FEATURE = 'http://jabber.org/features/compress';
const NS_COMPRESSION = 'http://jabber.org/protocol/compress';
// XEP-0141
const NS_DATAFORM_LAYOUT = 'http://jabber.org/protocol/xdata-layout';
// XEP-0144
const NS_ROSTER_EXCHANGE = 'http://jabber.org/protocol/rosterx';
// XEP-0145
const NS_ROSTER_NOTES = 'storage:rosternotes';
// XEP-0152
const NS_REACH_0 = 'urn:xmpp:reach:0';
// XEP-0153
const NS_VCARD_TEMP_UPDATE = 'vcard-temp:x:update';
// XEP-0156
const NS_ALT_CONNECTIONS_WEBSOCKET = 'urn:xmpp:alt-connections:websocket';
const NS_ALT_CONNECTIONS_XBOSH = 'urn:xmpp:alt-connections:xbosh';
// XEP-0158
const NS_CAPTCHA = 'urn:xmpp:captcha';
// XEP-0163
const NS_PEP_NOTIFY = (ns) => `${ns}+notify`;
// XEP-0166
const NS_JINGLE_1 = 'urn:xmpp:jingle:1';
const NS_JINGLE_ERRORS_1 = 'urn:xmpp:jingle:errors:1';
// XEP-0167
const NS_JINGLE_RTP_1 = 'urn:xmpp:jingle:apps:rtp:1';
const NS_JINGLE_RTP_ERRORS_1 = 'urn:xmpp:jingle:apps:rtp:errors:1';
const NS_JINGLE_RTP_INFO_1 = 'urn:xmpp:jingle:apps:rtp:info:1';
const NS_JINGLE_RTP_AUDIO = 'urn:xmpp:jingle:apps:rtp:audio';
const NS_JINGLE_RTP_VIDEO = 'urn:xmpp:jingle:apps:rtp:video';
// XEP-0171
const NS_LANG_TRANS = 'urn:xmpp:langtrans';
const NS_LANG_TRANS_ITEMS = 'urn:xmpp:langtrans:items';
// XEP-0172
const NS_NICK = 'http://jabber.org/protocol/nick';
// XEP-0176
const NS_JINGLE_ICE_UDP_1 = 'urn:xmpp:jingle:transports:ice-udp:1';
// XEP-0177
const NS_JINGLE_RAW_UDP_1 = 'urn:xmpp:jingle:transports:raw-udp:1';
// XEP-0184
const NS_RECEIPTS = 'urn:xmpp:receipts';
// XEP-0186
const NS_INVISIBLE_0 = 'urn:xmpp:invisible:0';
// XEP-0191
const NS_BLOCKING = 'urn:xmpp:blocking';
const NS_BLOCKING_ERRORS = 'urn:xmpp:blocking:errors';
// XEP-0198
const NS_SMACKS_3 = 'urn:xmpp:sm:3';
// XEP-0199
const NS_PING = 'urn:xmpp:ping';
// XEP-0202
const NS_TIME = 'urn:xmpp:time';
// XEP-0203
const NS_DELAY = 'urn:xmpp:delay';
// XEP-0206
const NS_BOSH_XMPP = 'urn:xmpp:xbosh';
// XEP-0215
const NS_DISCO_EXTERNAL_1 = 'urn:xmpp:extdisco:1';
const NS_DISCO_EXTERNAL_2 = 'urn:xmpp:extdisco:2';
// XEP-0221
const NS_DATAFORM_MEDIA = 'urn:xmpp:media-element';
// XEP-0224
const NS_ATTENTION_0 = 'urn:xmpp:attention:0';
// XEP-0231
const NS_BOB = 'urn:xmpp:bob';
// XEP-0232
const NS_SOFTWARE_INFO = 'urn:xmpp:dataforms:softwareinfo';
// XEP-0234
const NS_JINGLE_FILE_TRANSFER_3 = 'urn:xmpp:jingle:apps:file-transfer:3';
const NS_JINGLE_FILE_TRANSFER_4 = 'urn:xmpp:jingle:apps:file-transfer:4';
const NS_JINGLE_FILE_TRANSFER_5 = 'urn:xmpp:jingle:apps:file-transfer:5';
// XEP-0247
const NS_JINGLE_XML_0 = 'urn:xmpp:jingle:apps:xmlstream:0';
// XEP-0249
const NS_MUC_DIRECT_INVITE = 'jabber:x:conference';
// XEP-0258
const NS_SEC_LABEL_0 = 'urn:xmpp:sec-label:0';
const NS_SEC_LABEL_CATALOG_2 = 'urn:xmpp:sec-label:catalog:2';
const NS_SEC_LABEL_ESS_0 = 'urn:xmpp:sec-label:ess:0';
// XEP-0260
const NS_JINGLE_SOCKS5_1 = 'urn:xmpp:jingle:transports:s5b:1';
// XEP-0261
const NS_JINGLE_IBB_1 = 'urn:xmpp:jingle:transports:ibb:1';
// XEP-0262
const NS_JINGLE_RTP_ZRTP_1 = 'urn:xmpp:jingle:apps:rtp:zrtp:1';
// XEP-0264
const NS_THUMBS_0 = 'urn:xmpp:thumbs:0';
const NS_THUMBS_1 = 'urn:xmpp:thumbs:1';
// XEP-0276
const NS_DECLOAKING_0 = 'urn:xmpp:decloaking:0';
// XEP-0280
const NS_CARBONS_2 = 'urn:xmpp:carbons:2';
// XEP-0293
const NS_JINGLE_RTP_RTCP_FB_0 = 'urn:xmpp:jingle:apps:rtp:rtcp-fb:0';
// XEP-0294
const NS_JINGLE_RTP_HDREXT_0 = 'urn:xmpp:jingle:apps:rtp:rtp-hdrext:0';
// XEP-0297
const NS_FORWARD_0 = 'urn:xmpp:forward:0';
// XEP-0300
const NS_HASHES_1 = 'urn:xmpp:hashes:1';
const NS_HASHES_2 = 'urn:xmpp:hashes:2';
const NS_HASH_NAME = (name) => `urn:xmpp:hash-function-text-names:${name}`;
// XEP-0301
const NS_RTT_0 = 'urn:xmpp:rtt:0';
// XEP-0307
const NS_MUC_UNIQUE = 'http://jabber.org/protocol/muc#unique';
// XEP-308
const NS_CORRECTION_0 = 'urn:xmpp:message-correct:0';
// XEP-0310
const NS_PSA = 'urn:xmpp:psa';
// XEP-0313
const NS_MAM_TMP = 'urn:xmpp:mam:tmp';
const NS_MAM_0 = 'urn:xmpp:mam:0';
const NS_MAM_1 = 'urn:xmpp:mam:1';
const NS_MAM_2 = 'urn:xmpp:mam:2';
// XEP-0317
const NS_HATS_0 = 'urn:xmpp:hats:0';
// XEP-0319
const NS_IDLE_1 = 'urn:xmpp:idle:1';
// XEP-0320
const NS_JINGLE_DTLS_0 = 'urn:xmpp:jingle:apps:dtls:0';
// XEP-0333
const NS_CHAT_MARKERS_0 = 'urn:xmpp:chat-markers:0';
// XEP-0334
const NS_HINTS = 'urn:xmpp:hints';
// XEP-0335
const NS_JSON_0 = 'urn:xmpp:json:0';
// XEP-0338
const NS_JINGLE_GROUPING_0 = 'urn:xmpp:jingle:apps:grouping:0';
// XEP-0339
const NS_JINGLE_RTP_SSMA_0 = 'urn:xmpp:jingle:apps:rtp:ssma:0';
// XEP-0343
const NS_JINGLE_DTLS_SCTP_1 = 'urn:xmpp:jingle:transports:dtls-sctp:1';
// XEP-0352
const NS_CSI_0 = 'urn:xmpp:csi:0';
// XEP-0353
const NS_JINGLE_MSG_INITIATE_0 = 'urn:xmpp:jingle:jingle-message:0';
// XEP-0355
const NS_DELEGATION_1 = 'urn:xmpp:delegation:1';
// XEP-0357
const NS_PUSH_0 = 'urn:xmpp:push:0';
// XEP-0358
const NS_JINGLE_PUB_1 = 'urn:xmpp:jinglepub:1';
// XEP-0359
const NS_SID_0 = 'urn:xmpp:sid:0';
// XEP-0363
const NS_HTTP_UPLOAD_0 = 'urn:xmpp:http:upload:0';
// XEP-0370
const NS_JINGLE_HTTP_0 = 'urn:xmpp:jingle:transports:http:0';
const NS_JINGLE_HTTP_UPLOAD_0 = 'urn:xmpp:jingle:transports:http:upload:0';
// XEP-0371
const NS_JINGLE_ICE_0 = 'urn:xmpp:jingle:transports:ice:0';
// XEP-0372
const NS_REFERENCE_0 = 'urn:xmpp:reference:0';
// XEP-0380
const NS_EME_0 = 'urn:xmpp:eme:0';
// XEP-0382
const NS_SPOILER_0 = 'urn:xmpp:spoiler:0';
// XEP-0384
const NS_OMEMO_AXOLOTL = 'eu.siacs.conversations.axolotl';
const NS_OMEMO_AXOLOTL_DEVICELIST = 'eu.siacs.conversations.axolotl.devicelist';
const NS_OMEMO_AXOLOTL_BUNDLES = 'eu.siacs.conversations.axolotl.bundles';
// istanbul ignore next
const NS_OMEMO_AXOLOTL_BUNDLE = (deviceId) => `${NS_OMEMO_AXOLOTL_BUNDLES}:${deviceId}`;
// XEP-0432
const NS_JSON_MESSAGE_0 = 'urn:xmpp:json-msg:0';
// ================================================================
// Other Standards
// ================================================================
// Extensible Resource Descriptor (XRD) Version 1.0
// http://docs.oasis-open.org/xri/xrd/v1.0/xrd-1.0.html
const NS_XRD = 'http://docs.oasis-open.org/ns/xri/xrd-1.0';
// ====================================================================
// Not yet standardized
// ====================================================================
const NS_JINGLE_RTP_MSID_0 = 'urn:xmpp:jingle:apps:rtp:msid:0';

var Namespaces = /*#__PURE__*/Object.freeze({
    __proto__: null,
    NS_ATOM: NS_ATOM,
    NS_BIND: NS_BIND,
    NS_CLIENT: NS_CLIENT,
    NS_SASL: NS_SASL,
    NS_SERVER: NS_SERVER,
    NS_SESSION: NS_SESSION,
    NS_STANZAS: NS_STANZAS,
    NS_STREAM: NS_STREAM,
    NS_STREAMS: NS_STREAMS,
    NS_STARTTLS: NS_STARTTLS,
    NS_ROSTER: NS_ROSTER,
    NS_ROSTER_VERSIONING: NS_ROSTER_VERSIONING,
    NS_SUBSCRIPTION_PREAPPROVAL: NS_SUBSCRIPTION_PREAPPROVAL,
    NS_FRAMING: NS_FRAMING,
    NS_DATAFORM: NS_DATAFORM,
    NS_RPC: NS_RPC,
    NS_LAST_ACTIVITY: NS_LAST_ACTIVITY,
    NS_PRIVACY: NS_PRIVACY,
    NS_LEGACY_CHAT_EVENTS: NS_LEGACY_CHAT_EVENTS,
    NS_DISCO_INFO: NS_DISCO_INFO,
    NS_DISCO_ITEMS: NS_DISCO_ITEMS,
    NS_ADDRESS: NS_ADDRESS,
    NS_MUC: NS_MUC,
    NS_MUC_ADMIN: NS_MUC_ADMIN,
    NS_MUC_OWNER: NS_MUC_OWNER,
    NS_MUC_USER: NS_MUC_USER,
    NS_IBB: NS_IBB,
    NS_BOOKMARKS: NS_BOOKMARKS,
    NS_PRIVATE: NS_PRIVATE,
    NS_ADHOC_COMMANDS: NS_ADHOC_COMMANDS,
    NS_VCARD_TEMP: NS_VCARD_TEMP,
    NS_SEARCH: NS_SEARCH,
    NS_RSM: NS_RSM,
    NS_PUBSUB: NS_PUBSUB,
    NS_PUBSUB_ERRORS: NS_PUBSUB_ERRORS,
    NS_PUBSUB_EVENT: NS_PUBSUB_EVENT,
    NS_PUBSUB_OWNER: NS_PUBSUB_OWNER,
    NS_SOCKS5: NS_SOCKS5,
    NS_OOB: NS_OOB,
    NS_OOB_TRANSFER: NS_OOB_TRANSFER,
    NS_HTTP_AUTH: NS_HTTP_AUTH,
    NS_XHTML: NS_XHTML,
    NS_XHTML_IM: NS_XHTML_IM,
    NS_REGISTER: NS_REGISTER,
    NS_INBAND_REGISTRATION: NS_INBAND_REGISTRATION,
    NS_AMP: NS_AMP,
    NS_GEOLOC: NS_GEOLOC,
    NS_ROSTER_DELIMITER: NS_ROSTER_DELIMITER,
    NS_AVATAR_DATA: NS_AVATAR_DATA,
    NS_AVATAR_METADATA: NS_AVATAR_METADATA,
    NS_CHAT_STATES: NS_CHAT_STATES,
    NS_VERSION: NS_VERSION,
    NS_MOOD: NS_MOOD,
    NS_ACTIVITY: NS_ACTIVITY,
    NS_COMPONENT: NS_COMPONENT,
    NS_DISCO_LEGACY_CAPS: NS_DISCO_LEGACY_CAPS,
    NS_TUNE: NS_TUNE,
    NS_DATAFORM_VALIDATION: NS_DATAFORM_VALIDATION,
    NS_BOSH: NS_BOSH,
    NS_SHIM: NS_SHIM,
    NS_COMPRESSION_FEATURE: NS_COMPRESSION_FEATURE,
    NS_COMPRESSION: NS_COMPRESSION,
    NS_DATAFORM_LAYOUT: NS_DATAFORM_LAYOUT,
    NS_ROSTER_EXCHANGE: NS_ROSTER_EXCHANGE,
    NS_ROSTER_NOTES: NS_ROSTER_NOTES,
    NS_REACH_0: NS_REACH_0,
    NS_VCARD_TEMP_UPDATE: NS_VCARD_TEMP_UPDATE,
    NS_ALT_CONNECTIONS_WEBSOCKET: NS_ALT_CONNECTIONS_WEBSOCKET,
    NS_ALT_CONNECTIONS_XBOSH: NS_ALT_CONNECTIONS_XBOSH,
    NS_CAPTCHA: NS_CAPTCHA,
    NS_PEP_NOTIFY: NS_PEP_NOTIFY,
    NS_JINGLE_1: NS_JINGLE_1,
    NS_JINGLE_ERRORS_1: NS_JINGLE_ERRORS_1,
    NS_JINGLE_RTP_1: NS_JINGLE_RTP_1,
    NS_JINGLE_RTP_ERRORS_1: NS_JINGLE_RTP_ERRORS_1,
    NS_JINGLE_RTP_INFO_1: NS_JINGLE_RTP_INFO_1,
    NS_JINGLE_RTP_AUDIO: NS_JINGLE_RTP_AUDIO,
    NS_JINGLE_RTP_VIDEO: NS_JINGLE_RTP_VIDEO,
    NS_LANG_TRANS: NS_LANG_TRANS,
    NS_LANG_TRANS_ITEMS: NS_LANG_TRANS_ITEMS,
    NS_NICK: NS_NICK,
    NS_JINGLE_ICE_UDP_1: NS_JINGLE_ICE_UDP_1,
    NS_JINGLE_RAW_UDP_1: NS_JINGLE_RAW_UDP_1,
    NS_RECEIPTS: NS_RECEIPTS,
    NS_INVISIBLE_0: NS_INVISIBLE_0,
    NS_BLOCKING: NS_BLOCKING,
    NS_BLOCKING_ERRORS: NS_BLOCKING_ERRORS,
    NS_SMACKS_3: NS_SMACKS_3,
    NS_PING: NS_PING,
    NS_TIME: NS_TIME,
    NS_DELAY: NS_DELAY,
    NS_BOSH_XMPP: NS_BOSH_XMPP,
    NS_DISCO_EXTERNAL_1: NS_DISCO_EXTERNAL_1,
    NS_DISCO_EXTERNAL_2: NS_DISCO_EXTERNAL_2,
    NS_DATAFORM_MEDIA: NS_DATAFORM_MEDIA,
    NS_ATTENTION_0: NS_ATTENTION_0,
    NS_BOB: NS_BOB,
    NS_SOFTWARE_INFO: NS_SOFTWARE_INFO,
    NS_JINGLE_FILE_TRANSFER_3: NS_JINGLE_FILE_TRANSFER_3,
    NS_JINGLE_FILE_TRANSFER_4: NS_JINGLE_FILE_TRANSFER_4,
    NS_JINGLE_FILE_TRANSFER_5: NS_JINGLE_FILE_TRANSFER_5,
    NS_JINGLE_XML_0: NS_JINGLE_XML_0,
    NS_MUC_DIRECT_INVITE: NS_MUC_DIRECT_INVITE,
    NS_SEC_LABEL_0: NS_SEC_LABEL_0,
    NS_SEC_LABEL_CATALOG_2: NS_SEC_LABEL_CATALOG_2,
    NS_SEC_LABEL_ESS_0: NS_SEC_LABEL_ESS_0,
    NS_JINGLE_SOCKS5_1: NS_JINGLE_SOCKS5_1,
    NS_JINGLE_IBB_1: NS_JINGLE_IBB_1,
    NS_JINGLE_RTP_ZRTP_1: NS_JINGLE_RTP_ZRTP_1,
    NS_THUMBS_0: NS_THUMBS_0,
    NS_THUMBS_1: NS_THUMBS_1,
    NS_DECLOAKING_0: NS_DECLOAKING_0,
    NS_CARBONS_2: NS_CARBONS_2,
    NS_JINGLE_RTP_RTCP_FB_0: NS_JINGLE_RTP_RTCP_FB_0,
    NS_JINGLE_RTP_HDREXT_0: NS_JINGLE_RTP_HDREXT_0,
    NS_FORWARD_0: NS_FORWARD_0,
    NS_HASHES_1: NS_HASHES_1,
    NS_HASHES_2: NS_HASHES_2,
    NS_HASH_NAME: NS_HASH_NAME,
    NS_RTT_0: NS_RTT_0,
    NS_MUC_UNIQUE: NS_MUC_UNIQUE,
    NS_CORRECTION_0: NS_CORRECTION_0,
    NS_PSA: NS_PSA,
    NS_MAM_TMP: NS_MAM_TMP,
    NS_MAM_0: NS_MAM_0,
    NS_MAM_1: NS_MAM_1,
    NS_MAM_2: NS_MAM_2,
    NS_HATS_0: NS_HATS_0,
    NS_IDLE_1: NS_IDLE_1,
    NS_JINGLE_DTLS_0: NS_JINGLE_DTLS_0,
    NS_CHAT_MARKERS_0: NS_CHAT_MARKERS_0,
    NS_HINTS: NS_HINTS,
    NS_JSON_0: NS_JSON_0,
    NS_JINGLE_GROUPING_0: NS_JINGLE_GROUPING_0,
    NS_JINGLE_RTP_SSMA_0: NS_JINGLE_RTP_SSMA_0,
    NS_JINGLE_DTLS_SCTP_1: NS_JINGLE_DTLS_SCTP_1,
    NS_CSI_0: NS_CSI_0,
    NS_JINGLE_MSG_INITIATE_0: NS_JINGLE_MSG_INITIATE_0,
    NS_DELEGATION_1: NS_DELEGATION_1,
    NS_PUSH_0: NS_PUSH_0,
    NS_JINGLE_PUB_1: NS_JINGLE_PUB_1,
    NS_SID_0: NS_SID_0,
    NS_HTTP_UPLOAD_0: NS_HTTP_UPLOAD_0,
    NS_JINGLE_HTTP_0: NS_JINGLE_HTTP_0,
    NS_JINGLE_HTTP_UPLOAD_0: NS_JINGLE_HTTP_UPLOAD_0,
    NS_JINGLE_ICE_0: NS_JINGLE_ICE_0,
    NS_REFERENCE_0: NS_REFERENCE_0,
    NS_EME_0: NS_EME_0,
    NS_SPOILER_0: NS_SPOILER_0,
    NS_OMEMO_AXOLOTL: NS_OMEMO_AXOLOTL,
    NS_OMEMO_AXOLOTL_DEVICELIST: NS_OMEMO_AXOLOTL_DEVICELIST,
    NS_OMEMO_AXOLOTL_BUNDLES: NS_OMEMO_AXOLOTL_BUNDLES,
    NS_OMEMO_AXOLOTL_BUNDLE: NS_OMEMO_AXOLOTL_BUNDLE,
    NS_JSON_MESSAGE_0: NS_JSON_MESSAGE_0,
    NS_XRD: NS_XRD,
    NS_JINGLE_RTP_MSID_0: NS_JINGLE_RTP_MSID_0
});

// ====================================================================
// Useful XMPP Aliases
// ====================================================================
const JIDAttribute = attribute;
const childJIDAttribute = childAttribute;
const childJID = childText;
// ====================================================================
// XMPP Definition Shortcuts
// ====================================================================
function addAlias(namespace, element, aliases) {
    return {
        aliases: Array.isArray(aliases) ? aliases : [aliases],
        element,
        fields: {},
        namespace
    };
}
function extendMessage(fields) {
    return { element: 'message', fields, namespace: NS_CLIENT };
}
function extendPresence(fields) {
    return { element: 'presence', fields, namespace: NS_CLIENT };
}
function extendIQ(fields) {
    return { element: 'iq', fields, namespace: NS_CLIENT };
}
function extendStreamFeatures(fields) {
    return {
        element: 'features',
        fields,
        namespace: NS_STREAM
    };
}
function extendStanzaError(fields) {
    return {
        element: 'error',
        fields,
        namespace: NS_STANZAS,
        path: 'stanzaError'
    };
}
function pubsubItemContentAliases() {
    return [
        { path: 'pubsubcontent', contextField: 'itemType' },
        { path: 'pubsubitem.content', contextField: 'itemType' },
        { path: 'pubsubeventitem.content', contextField: 'itemType' },
        { path: 'iq.pubsub.publish.items', contextField: 'itemType' }
    ];
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from: ltx, Copyright © 2010 Stephan Maka
 */
class StreamParser extends Transform {
    constructor(opts) {
        super({ objectMode: true });
        this.closedStream = false;
        this.wrappedStream = false;
        this.registry = opts.registry;
        this.acceptLanguages = opts.acceptLanguages || [];
        if (opts.wrappedStream) {
            this.wrappedStream = true;
            this.rootImportKey = opts.rootKey;
        }
        this.parser = new Parser({
            allowComments: opts.allowComments
        });
        this.parser.on('error', err => {
            this.destroy(err);
        });
        this.parser.on('startElement', (name, attributes) => {
            if (this.destroyed) {
                return;
            }
            if (this.closedStream) {
                return this.destroy(JXTError.alreadyClosed());
            }
            const el = new XMLElement(name, attributes);
            const key = this.registry.getImportKey(el);
            if (this.wrappedStream && !this.rootElement) {
                if (this.rootImportKey && key !== this.rootImportKey) {
                    return this.destroy(JXTError.unknownRoot());
                }
                const root = this.registry.import(el, {
                    acceptLanguages: this.acceptLanguages,
                    lang: this.lang
                });
                if (root) {
                    this.rootElement = el;
                    this.push({
                        event: 'stream-start',
                        kind: key,
                        stanza: root,
                        xml: el
                    });
                    return;
                }
                else {
                    return this.destroy(JXTError.notWellFormed());
                }
            }
            if (!this.currentElement) {
                this.currentElement = el;
            }
            else {
                this.currentElement = this.currentElement.appendChild(el);
            }
        });
        this.parser.on('endElement', (name) => {
            if (this.destroyed) {
                return;
            }
            if (this.wrappedStream && !this.currentElement) {
                if (!this.rootElement || name !== this.rootElement.name) {
                    this.closedStream = true;
                    return this.destroy(JXTError.notWellFormed());
                }
                this.closedStream = true;
                this.push({
                    event: 'stream-end',
                    kind: this.rootImportKey,
                    stanza: {},
                    xml: this.rootElement
                });
                return this.end();
            }
            if (!this.currentElement || name !== this.currentElement.name) {
                this.closedStream = true;
                return this.destroy(JXTError.notWellFormed());
            }
            if (this.currentElement.parent) {
                this.currentElement = this.currentElement.parent;
            }
            else {
                if (this.wrappedStream) {
                    this.currentElement.parent = this.rootElement;
                }
                const key = this.registry.getImportKey(this.currentElement);
                const stanza = this.registry.import(this.currentElement, {
                    acceptLanguages: this.acceptLanguages,
                    lang: this.lang
                });
                if (stanza) {
                    this.push({
                        kind: key,
                        stanza,
                        xml: this.currentElement
                    });
                }
                this.currentElement = undefined;
            }
        });
        this.parser.on('text', (text) => {
            if (this.currentElement) {
                this.currentElement.children.push(text);
            }
        });
    }
    _transform(chunk, encoding, done) {
        this.parser.write(chunk.toString());
        done();
    }
}

function define(definitions) {
    return (registry) => {
        registry.define(definitions);
    };
}

var index$3 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    Registry: Registry,
    Translator: Translator,
    XMLElement: XMLElement,
    define: define,
    Parser: Parser,
    parse: parse$1,
    StreamParser: StreamParser,
    escapeXML: escapeXML,
    unescapeXML: unescapeXML,
    escapeXMLText: escapeXMLText,
    basicLanguageResolver: basicLanguageResolver,
    createElement: createElement,
    getLang: getLang,
    getTargetLang: getTargetLang,
    findAll: findAll,
    findOrCreate: findOrCreate,
    attribute: attribute,
    booleanAttribute: booleanAttribute,
    integerAttribute: integerAttribute,
    floatAttribute: floatAttribute,
    dateAttribute: dateAttribute,
    namespacedAttribute: namespacedAttribute,
    namespacedBooleanAttribute: namespacedBooleanAttribute,
    namespacedIntegerAttribute: namespacedIntegerAttribute,
    namespacedFloatAttribute: namespacedFloatAttribute,
    namespacedDateAttribute: namespacedDateAttribute,
    childAttribute: childAttribute,
    childBooleanAttribute: childBooleanAttribute,
    childIntegerAttribute: childIntegerAttribute,
    childFloatAttribute: childFloatAttribute,
    childDateAttribute: childDateAttribute,
    text: text,
    textJSON: textJSON,
    textBuffer: textBuffer,
    languageAttribute: languageAttribute,
    childLanguageAttribute: childLanguageAttribute,
    childText: childText,
    childTextBuffer: childTextBuffer,
    childDate: childDate,
    childInteger: childInteger,
    childFloat: childFloat,
    childJSON: childJSON,
    childTimezoneOffset: childTimezoneOffset,
    childBoolean: childBoolean,
    deepChildText: deepChildText,
    deepChildInteger: deepChildInteger,
    deepChildBoolean: deepChildBoolean,
    deepMultipleChildText: deepMultipleChildText,
    childEnum: childEnum,
    childDoubleEnum: childDoubleEnum,
    multipleChildText: multipleChildText,
    multipleChildAttribute: multipleChildAttribute,
    multipleChildIntegerAttribute: multipleChildIntegerAttribute,
    childAlternateLanguageText: childAlternateLanguageText,
    multipleChildAlternateLanguageText: multipleChildAlternateLanguageText,
    multipleChildEnum: multipleChildEnum,
    splicePath: splicePath,
    staticValue: staticValue,
    childRawElement: childRawElement,
    childLanguageRawElement: childLanguageRawElement,
    childAlternateLanguageRawElement: childAlternateLanguageRawElement,
    parameterMap: parameterMap,
    JIDAttribute: JIDAttribute,
    childJIDAttribute: childJIDAttribute,
    childJID: childJID,
    addAlias: addAlias,
    extendMessage: extendMessage,
    extendPresence: extendPresence,
    extendIQ: extendIQ,
    extendStreamFeatures: extendStreamFeatures,
    extendStanzaError: extendStanzaError,
    pubsubItemContentAliases: pubsubItemContentAliases
});

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from:
 * - hash-base, Copyright (c) 2016 Kirill Fomichev
 * - cipher-base, Copyright (c) 2017 crypto-browserify contributors
 * - create-hash, Copyright (c) 2017 crypto-browserify contributors
 */
class Hash extends Transform {
    constructor(blockSize, finalSize, endian = 'be') {
        super();
        this._block = Buffer.alloc(blockSize);
        this._finalSize = finalSize;
        this._blockSize = blockSize;
        this._bigEndian = endian === 'be';
        this._len = 0;
    }
    _transform(chunk, encoding, callback) {
        let error = null;
        try {
            this.update(chunk, encoding);
        } catch (err) {
            error = err;
        }
        callback(error);
    }
    _flush(callback) {
        let error = null;
        try {
            this.push(this.digest());
        } catch (err) {
            error = err;
        }
        callback(error);
    }
    update(data, enc) {
        if (typeof data === 'string') {
            enc = enc || 'utf8';
            data = Buffer.from(data, enc);
        }
        const block = this._block;
        const blockSize = this._blockSize;
        const length = data.length;
        let accum = this._len;
        for (let offset = 0; offset < length; ) {
            const assigned = accum % blockSize;
            const remainder = Math.min(length - offset, blockSize - assigned);
            for (let i = 0; i < remainder; i++) {
                block[assigned + i] = data[offset + i];
            }
            accum += remainder;
            offset += remainder;
            if (accum % blockSize === 0) {
                this._update(block);
            }
        }
        this._len += length;
        return this;
    }
    digest(enc) {
        const rem = this._len % this._blockSize;
        this._block[rem] = 0x80;
        // zero (rem + 1) trailing bits, where (rem + 1) is the smallest
        // non-negative solution to the equation (length + 1 + (rem + 1)) === finalSize mod blockSize
        this._block.fill(0, rem + 1);
        if (rem >= this._finalSize) {
            this._update(this._block);
            this._block.fill(0);
        }
        const bits = this._len * 8;
        if (bits <= 0xffffffff) {
            if (this._bigEndian) {
                this._block.writeUInt32BE(0, this._blockSize - 8);
                this._block.writeUInt32BE(bits, this._blockSize - 4);
            } else {
                this._block.writeUInt32LE(bits, this._blockSize - 8);
                this._block.writeUInt32LE(0, this._blockSize - 4);
            }
        } else {
            const lowBits = (bits & 0xffffffff) >>> 0;
            const highBits = (bits - lowBits) / 0x100000000;
            if (this._bigEndian) {
                this._block.writeUInt32BE(highBits, this._blockSize - 8);
                this._block.writeUInt32BE(lowBits, this._blockSize - 4);
            } else {
                this._block.writeUInt32LE(lowBits, this._blockSize - 8);
                this._block.writeUInt32LE(highBits, this._blockSize - 4);
            }
        }
        this._update(this._block);
        const hash = this._hash();
        return enc ? hash.toString(enc) : hash;
    }
    _update(block) {
        throw new Error('_update must be implemented by subclass');
    }
    _hash() {
        throw new Error('_update must be implemented by subclass');
    }
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from:
 * - hash-base, Copyright (c) 2016 Kirill Fomichev
 * - cipher-base, Copyright (c) 2017 crypto-browserify contributors
 * - md5.js, Copyright (c) 2016 Kirill Fomichev
 */
function rotl(x, n) {
    return (x << n) | (x >>> (32 - n));
}
function fnF(a, b, c, d, m, k, s) {
    return (rotl((a + ((b & c) | (~b & d)) + m + k) | 0, s) + b) | 0;
}
function fnG(a, b, c, d, m, k, s) {
    return (rotl((a + ((b & d) | (c & ~d)) + m + k) | 0, s) + b) | 0;
}
function fnH(a, b, c, d, m, k, s) {
    return (rotl((a + (b ^ c ^ d) + m + k) | 0, s) + b) | 0;
}
function fnI(a, b, c, d, m, k, s) {
    return (rotl((a + (c ^ (b | ~d)) + m + k) | 0, s) + b) | 0;
}
class MD5 extends Hash {
    constructor() {
        super(64, 56, 'le');
        this._a = 0x67452301;
        this._b = 0xefcdab89;
        this._c = 0x98badcfe;
        this._d = 0x10325476;
        this._m = new Array(16);
    }
    _update(B) {
        const M = this._m;
        for (let i = 0; i < 16; ++i) {
            M[i] = B.readInt32LE(i * 4);
        }
        let a = this._a;
        let b = this._b;
        let c = this._c;
        let d = this._d;
        a = fnF(a, b, c, d, M[0], 0xd76aa478, 7);
        d = fnF(d, a, b, c, M[1], 0xe8c7b756, 12);
        c = fnF(c, d, a, b, M[2], 0x242070db, 17);
        b = fnF(b, c, d, a, M[3], 0xc1bdceee, 22);
        a = fnF(a, b, c, d, M[4], 0xf57c0faf, 7);
        d = fnF(d, a, b, c, M[5], 0x4787c62a, 12);
        c = fnF(c, d, a, b, M[6], 0xa8304613, 17);
        b = fnF(b, c, d, a, M[7], 0xfd469501, 22);
        a = fnF(a, b, c, d, M[8], 0x698098d8, 7);
        d = fnF(d, a, b, c, M[9], 0x8b44f7af, 12);
        c = fnF(c, d, a, b, M[10], 0xffff5bb1, 17);
        b = fnF(b, c, d, a, M[11], 0x895cd7be, 22);
        a = fnF(a, b, c, d, M[12], 0x6b901122, 7);
        d = fnF(d, a, b, c, M[13], 0xfd987193, 12);
        c = fnF(c, d, a, b, M[14], 0xa679438e, 17);
        b = fnF(b, c, d, a, M[15], 0x49b40821, 22);
        a = fnG(a, b, c, d, M[1], 0xf61e2562, 5);
        d = fnG(d, a, b, c, M[6], 0xc040b340, 9);
        c = fnG(c, d, a, b, M[11], 0x265e5a51, 14);
        b = fnG(b, c, d, a, M[0], 0xe9b6c7aa, 20);
        a = fnG(a, b, c, d, M[5], 0xd62f105d, 5);
        d = fnG(d, a, b, c, M[10], 0x02441453, 9);
        c = fnG(c, d, a, b, M[15], 0xd8a1e681, 14);
        b = fnG(b, c, d, a, M[4], 0xe7d3fbc8, 20);
        a = fnG(a, b, c, d, M[9], 0x21e1cde6, 5);
        d = fnG(d, a, b, c, M[14], 0xc33707d6, 9);
        c = fnG(c, d, a, b, M[3], 0xf4d50d87, 14);
        b = fnG(b, c, d, a, M[8], 0x455a14ed, 20);
        a = fnG(a, b, c, d, M[13], 0xa9e3e905, 5);
        d = fnG(d, a, b, c, M[2], 0xfcefa3f8, 9);
        c = fnG(c, d, a, b, M[7], 0x676f02d9, 14);
        b = fnG(b, c, d, a, M[12], 0x8d2a4c8a, 20);
        a = fnH(a, b, c, d, M[5], 0xfffa3942, 4);
        d = fnH(d, a, b, c, M[8], 0x8771f681, 11);
        c = fnH(c, d, a, b, M[11], 0x6d9d6122, 16);
        b = fnH(b, c, d, a, M[14], 0xfde5380c, 23);
        a = fnH(a, b, c, d, M[1], 0xa4beea44, 4);
        d = fnH(d, a, b, c, M[4], 0x4bdecfa9, 11);
        c = fnH(c, d, a, b, M[7], 0xf6bb4b60, 16);
        b = fnH(b, c, d, a, M[10], 0xbebfbc70, 23);
        a = fnH(a, b, c, d, M[13], 0x289b7ec6, 4);
        d = fnH(d, a, b, c, M[0], 0xeaa127fa, 11);
        c = fnH(c, d, a, b, M[3], 0xd4ef3085, 16);
        b = fnH(b, c, d, a, M[6], 0x04881d05, 23);
        a = fnH(a, b, c, d, M[9], 0xd9d4d039, 4);
        d = fnH(d, a, b, c, M[12], 0xe6db99e5, 11);
        c = fnH(c, d, a, b, M[15], 0x1fa27cf8, 16);
        b = fnH(b, c, d, a, M[2], 0xc4ac5665, 23);
        a = fnI(a, b, c, d, M[0], 0xf4292244, 6);
        d = fnI(d, a, b, c, M[7], 0x432aff97, 10);
        c = fnI(c, d, a, b, M[14], 0xab9423a7, 15);
        b = fnI(b, c, d, a, M[5], 0xfc93a039, 21);
        a = fnI(a, b, c, d, M[12], 0x655b59c3, 6);
        d = fnI(d, a, b, c, M[3], 0x8f0ccc92, 10);
        c = fnI(c, d, a, b, M[10], 0xffeff47d, 15);
        b = fnI(b, c, d, a, M[1], 0x85845dd1, 21);
        a = fnI(a, b, c, d, M[8], 0x6fa87e4f, 6);
        d = fnI(d, a, b, c, M[15], 0xfe2ce6e0, 10);
        c = fnI(c, d, a, b, M[6], 0xa3014314, 15);
        b = fnI(b, c, d, a, M[13], 0x4e0811a1, 21);
        a = fnI(a, b, c, d, M[4], 0xf7537e82, 6);
        d = fnI(d, a, b, c, M[11], 0xbd3af235, 10);
        c = fnI(c, d, a, b, M[2], 0x2ad7d2bb, 15);
        b = fnI(b, c, d, a, M[9], 0xeb86d391, 21);
        this._a = (this._a + a) | 0;
        this._b = (this._b + b) | 0;
        this._c = (this._c + c) | 0;
        this._d = (this._d + d) | 0;
    }
    _hash() {
        const H = Buffer.allocUnsafe(16);
        H.writeInt32LE(this._a, 0);
        H.writeInt32LE(this._b, 4);
        H.writeInt32LE(this._c, 8);
        H.writeInt32LE(this._d, 12);
        return H;
    }
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from:
 * - hash-base, Copyright (c) 2016 Kirill Fomichev
 * - cipher-base, Copyright (c) 2017 crypto-browserify contributors
 * - sha.js, Copyright (c) 2013-2018 sha.js contributors
 */
const K = [0x5a827999, 0x6ed9eba1, 0x8f1bbcdc | 0, 0xca62c1d6 | 0];
function rotl1(num) {
    return (num << 1) | (num >>> 31);
}
function rotl5(num) {
    return (num << 5) | (num >>> 27);
}
function rotl30(num) {
    return (num << 30) | (num >>> 2);
}
function ft(s, b, c, d) {
    if (s === 0) {
        return (b & c) | (~b & d);
    }
    if (s === 2) {
        return (b & c) | (b & d) | (c & d);
    }
    return b ^ c ^ d;
}
class Sha1 extends Hash {
    constructor() {
        super(64, 56);
        this._a = 0x67452301;
        this._b = 0xefcdab89;
        this._c = 0x98badcfe;
        this._d = 0x10325476;
        this._e = 0xc3d2e1f0;
        this._w = new Array(80);
    }
    _update(M) {
        const W = this._w;
        let a = this._a | 0;
        let b = this._b | 0;
        let c = this._c | 0;
        let d = this._d | 0;
        let e = this._e | 0;
        let i;
        for (i = 0; i < 16; ++i) {
            W[i] = M.readInt32BE(i * 4);
        }
        for (; i < 80; ++i) {
            W[i] = rotl1(W[i - 3] ^ W[i - 8] ^ W[i - 14] ^ W[i - 16]);
        }
        for (let j = 0; j < 80; ++j) {
            const s = ~~(j / 20);
            const t = (rotl5(a) + ft(s, b, c, d) + e + W[j] + K[s]) | 0;
            e = d;
            d = c;
            c = rotl30(b);
            b = a;
            a = t;
        }
        this._a = (a + this._a) | 0;
        this._b = (b + this._b) | 0;
        this._c = (c + this._c) | 0;
        this._d = (d + this._d) | 0;
        this._e = (e + this._e) | 0;
    }
    _hash() {
        const H = Buffer.allocUnsafe(20);
        H.writeInt32BE(this._a | 0, 0);
        H.writeInt32BE(this._b | 0, 4);
        H.writeInt32BE(this._c | 0, 8);
        H.writeInt32BE(this._d | 0, 12);
        H.writeInt32BE(this._e | 0, 16);
        return H;
    }
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from:
 * - hash-base, Copyright (c) 2016 Kirill Fomichev
 * - cipher-base, Copyright (c) 2017 crypto-browserify contributors
 * - sha.js, Copyright (c) 2013-2018 sha.js contributors
 */
const K$1 = [
    0x428a2f98,
    0x71374491,
    0xb5c0fbcf,
    0xe9b5dba5,
    0x3956c25b,
    0x59f111f1,
    0x923f82a4,
    0xab1c5ed5,
    0xd807aa98,
    0x12835b01,
    0x243185be,
    0x550c7dc3,
    0x72be5d74,
    0x80deb1fe,
    0x9bdc06a7,
    0xc19bf174,
    0xe49b69c1,
    0xefbe4786,
    0x0fc19dc6,
    0x240ca1cc,
    0x2de92c6f,
    0x4a7484aa,
    0x5cb0a9dc,
    0x76f988da,
    0x983e5152,
    0xa831c66d,
    0xb00327c8,
    0xbf597fc7,
    0xc6e00bf3,
    0xd5a79147,
    0x06ca6351,
    0x14292967,
    0x27b70a85,
    0x2e1b2138,
    0x4d2c6dfc,
    0x53380d13,
    0x650a7354,
    0x766a0abb,
    0x81c2c92e,
    0x92722c85,
    0xa2bfe8a1,
    0xa81a664b,
    0xc24b8b70,
    0xc76c51a3,
    0xd192e819,
    0xd6990624,
    0xf40e3585,
    0x106aa070,
    0x19a4c116,
    0x1e376c08,
    0x2748774c,
    0x34b0bcb5,
    0x391c0cb3,
    0x4ed8aa4a,
    0x5b9cca4f,
    0x682e6ff3,
    0x748f82ee,
    0x78a5636f,
    0x84c87814,
    0x8cc70208,
    0x90befffa,
    0xa4506ceb,
    0xbef9a3f7,
    0xc67178f2
];
function ch(x, y, z) {
    return z ^ (x & (y ^ z));
}
function maj(x, y, z) {
    return (x & y) | (z & (x | y));
}
function sigma0(x) {
    return ((x >>> 2) | (x << 30)) ^ ((x >>> 13) | (x << 19)) ^ ((x >>> 22) | (x << 10));
}
function sigma1(x) {
    return ((x >>> 6) | (x << 26)) ^ ((x >>> 11) | (x << 21)) ^ ((x >>> 25) | (x << 7));
}
function gamma0(x) {
    return ((x >>> 7) | (x << 25)) ^ ((x >>> 18) | (x << 14)) ^ (x >>> 3);
}
function gamma1(x) {
    return ((x >>> 17) | (x << 15)) ^ ((x >>> 19) | (x << 13)) ^ (x >>> 10);
}
class Sha256 extends Hash {
    constructor() {
        super(64, 56);
        this._a = 0x6a09e667;
        this._b = 0xbb67ae85;
        this._c = 0x3c6ef372;
        this._d = 0xa54ff53a;
        this._e = 0x510e527f;
        this._f = 0x9b05688c;
        this._g = 0x1f83d9ab;
        this._h = 0x5be0cd19;
        this._w = new Array(64);
    }
    _update(M) {
        const W = this._w;
        let a = this._a | 0;
        let b = this._b | 0;
        let c = this._c | 0;
        let d = this._d | 0;
        let e = this._e | 0;
        let f = this._f | 0;
        let g = this._g | 0;
        let h = this._h | 0;
        let i;
        for (i = 0; i < 16; ++i) {
            W[i] = M.readInt32BE(i * 4);
        }
        for (; i < 64; ++i) {
            W[i] = (gamma1(W[i - 2]) + W[i - 7] + gamma0(W[i - 15]) + W[i - 16]) | 0;
        }
        for (let j = 0; j < 64; ++j) {
            const T1 = (h + sigma1(e) + ch(e, f, g) + K$1[j] + W[j]) | 0;
            const T2 = (sigma0(a) + maj(a, b, c)) | 0;
            h = g;
            g = f;
            f = e;
            e = (d + T1) | 0;
            d = c;
            c = b;
            b = a;
            a = (T1 + T2) | 0;
        }
        this._a = (a + this._a) | 0;
        this._b = (b + this._b) | 0;
        this._c = (c + this._c) | 0;
        this._d = (d + this._d) | 0;
        this._e = (e + this._e) | 0;
        this._f = (f + this._f) | 0;
        this._g = (g + this._g) | 0;
        this._h = (h + this._h) | 0;
    }
    _hash() {
        const H = Buffer.allocUnsafe(32);
        H.writeInt32BE(this._a, 0);
        H.writeInt32BE(this._b, 4);
        H.writeInt32BE(this._c, 8);
        H.writeInt32BE(this._d, 12);
        H.writeInt32BE(this._e, 16);
        H.writeInt32BE(this._f, 20);
        H.writeInt32BE(this._g, 24);
        H.writeInt32BE(this._h, 28);
        return H;
    }
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from:
 * - hash-base, Copyright (c) 2016 Kirill Fomichev
 * - cipher-base, Copyright (c) 2017 crypto-browserify contributors
 * - sha.js, Copyright (c) 2013-2018 sha.js contributors
 */
const K$2 = [
    0x428a2f98,
    0xd728ae22,
    0x71374491,
    0x23ef65cd,
    0xb5c0fbcf,
    0xec4d3b2f,
    0xe9b5dba5,
    0x8189dbbc,
    0x3956c25b,
    0xf348b538,
    0x59f111f1,
    0xb605d019,
    0x923f82a4,
    0xaf194f9b,
    0xab1c5ed5,
    0xda6d8118,
    0xd807aa98,
    0xa3030242,
    0x12835b01,
    0x45706fbe,
    0x243185be,
    0x4ee4b28c,
    0x550c7dc3,
    0xd5ffb4e2,
    0x72be5d74,
    0xf27b896f,
    0x80deb1fe,
    0x3b1696b1,
    0x9bdc06a7,
    0x25c71235,
    0xc19bf174,
    0xcf692694,
    0xe49b69c1,
    0x9ef14ad2,
    0xefbe4786,
    0x384f25e3,
    0x0fc19dc6,
    0x8b8cd5b5,
    0x240ca1cc,
    0x77ac9c65,
    0x2de92c6f,
    0x592b0275,
    0x4a7484aa,
    0x6ea6e483,
    0x5cb0a9dc,
    0xbd41fbd4,
    0x76f988da,
    0x831153b5,
    0x983e5152,
    0xee66dfab,
    0xa831c66d,
    0x2db43210,
    0xb00327c8,
    0x98fb213f,
    0xbf597fc7,
    0xbeef0ee4,
    0xc6e00bf3,
    0x3da88fc2,
    0xd5a79147,
    0x930aa725,
    0x06ca6351,
    0xe003826f,
    0x14292967,
    0x0a0e6e70,
    0x27b70a85,
    0x46d22ffc,
    0x2e1b2138,
    0x5c26c926,
    0x4d2c6dfc,
    0x5ac42aed,
    0x53380d13,
    0x9d95b3df,
    0x650a7354,
    0x8baf63de,
    0x766a0abb,
    0x3c77b2a8,
    0x81c2c92e,
    0x47edaee6,
    0x92722c85,
    0x1482353b,
    0xa2bfe8a1,
    0x4cf10364,
    0xa81a664b,
    0xbc423001,
    0xc24b8b70,
    0xd0f89791,
    0xc76c51a3,
    0x0654be30,
    0xd192e819,
    0xd6ef5218,
    0xd6990624,
    0x5565a910,
    0xf40e3585,
    0x5771202a,
    0x106aa070,
    0x32bbd1b8,
    0x19a4c116,
    0xb8d2d0c8,
    0x1e376c08,
    0x5141ab53,
    0x2748774c,
    0xdf8eeb99,
    0x34b0bcb5,
    0xe19b48a8,
    0x391c0cb3,
    0xc5c95a63,
    0x4ed8aa4a,
    0xe3418acb,
    0x5b9cca4f,
    0x7763e373,
    0x682e6ff3,
    0xd6b2b8a3,
    0x748f82ee,
    0x5defb2fc,
    0x78a5636f,
    0x43172f60,
    0x84c87814,
    0xa1f0ab72,
    0x8cc70208,
    0x1a6439ec,
    0x90befffa,
    0x23631e28,
    0xa4506ceb,
    0xde82bde9,
    0xbef9a3f7,
    0xb2c67915,
    0xc67178f2,
    0xe372532b,
    0xca273ece,
    0xea26619c,
    0xd186b8c7,
    0x21c0c207,
    0xeada7dd6,
    0xcde0eb1e,
    0xf57d4f7f,
    0xee6ed178,
    0x06f067aa,
    0x72176fba,
    0x0a637dc5,
    0xa2c898a6,
    0x113f9804,
    0xbef90dae,
    0x1b710b35,
    0x131c471b,
    0x28db77f5,
    0x23047d84,
    0x32caab7b,
    0x40c72493,
    0x3c9ebe0a,
    0x15c9bebc,
    0x431d67c4,
    0x9c100d4c,
    0x4cc5d4be,
    0xcb3e42b6,
    0x597f299c,
    0xfc657e2a,
    0x5fcb6fab,
    0x3ad6faec,
    0x6c44198c,
    0x4a475817
];
function Ch(x, y, z) {
    return z ^ (x & (y ^ z));
}
function maj$1(x, y, z) {
    return (x & y) | (z & (x | y));
}
function sigma0$1(x, xl) {
    return ((x >>> 28) | (xl << 4)) ^ ((xl >>> 2) | (x << 30)) ^ ((xl >>> 7) | (x << 25));
}
function sigma1$1(x, xl) {
    return ((x >>> 14) | (xl << 18)) ^ ((x >>> 18) | (xl << 14)) ^ ((xl >>> 9) | (x << 23));
}
function Gamma0(x, xl) {
    return ((x >>> 1) | (xl << 31)) ^ ((x >>> 8) | (xl << 24)) ^ (x >>> 7);
}
function Gamma0l(x, xl) {
    return ((x >>> 1) | (xl << 31)) ^ ((x >>> 8) | (xl << 24)) ^ ((x >>> 7) | (xl << 25));
}
function Gamma1(x, xl) {
    return ((x >>> 19) | (xl << 13)) ^ ((xl >>> 29) | (x << 3)) ^ (x >>> 6);
}
function Gamma1l(x, xl) {
    return ((x >>> 19) | (xl << 13)) ^ ((xl >>> 29) | (x << 3)) ^ ((x >>> 6) | (xl << 26));
}
function getCarry(a, b) {
    return a >>> 0 < b >>> 0 ? 1 : 0;
}
class Sha512 extends Hash {
    constructor() {
        super(128, 112);
        this._ah = 0x6a09e667;
        this._bh = 0xbb67ae85;
        this._ch = 0x3c6ef372;
        this._dh = 0xa54ff53a;
        this._eh = 0x510e527f;
        this._fh = 0x9b05688c;
        this._gh = 0x1f83d9ab;
        this._hh = 0x5be0cd19;
        this._al = 0xf3bcc908;
        this._bl = 0x84caa73b;
        this._cl = 0xfe94f82b;
        this._dl = 0x5f1d36f1;
        this._el = 0xade682d1;
        this._fl = 0x2b3e6c1f;
        this._gl = 0xfb41bd6b;
        this._hl = 0x137e2179;
        this._w = new Array(160);
    }
    _update(M) {
        const W = this._w;
        let ah = this._ah | 0;
        let bh = this._bh | 0;
        let ch = this._ch | 0;
        let dh = this._dh | 0;
        let eh = this._eh | 0;
        let fh = this._fh | 0;
        let gh = this._gh | 0;
        let hh = this._hh | 0;
        let al = this._al | 0;
        let bl = this._bl | 0;
        let cl = this._cl | 0;
        let dl = this._dl | 0;
        let el = this._el | 0;
        let fl = this._fl | 0;
        let gl = this._gl | 0;
        let hl = this._hl | 0;
        let Wih;
        let Wil;
        let i = 0;
        for (i = 0; i < 32; i += 2) {
            W[i] = M.readInt32BE(i * 4);
            W[i + 1] = M.readInt32BE(i * 4 + 4);
        }
        for (; i < 160; i += 2) {
            let xh = W[i - 15 * 2];
            let xl = W[i - 15 * 2 + 1];
            const gamma0 = Gamma0(xh, xl);
            const gamma0l = Gamma0l(xl, xh);
            xh = W[i - 2 * 2];
            xl = W[i - 2 * 2 + 1];
            const gamma1 = Gamma1(xh, xl);
            const gamma1l = Gamma1l(xl, xh);
            // W[i] = gamma0 + W[i - 7] + gamma1 + W[i - 16]
            const Wi7h = W[i - 7 * 2];
            const Wi7l = W[i - 7 * 2 + 1];
            const Wi16h = W[i - 16 * 2];
            const Wi16l = W[i - 16 * 2 + 1];
            Wil = (gamma0l + Wi7l) | 0;
            Wih = (gamma0 + Wi7h + getCarry(Wil, gamma0l)) | 0;
            Wil = (Wil + gamma1l) | 0;
            Wih = (Wih + gamma1 + getCarry(Wil, gamma1l)) | 0;
            Wil = (Wil + Wi16l) | 0;
            Wih = (Wih + Wi16h + getCarry(Wil, Wi16l)) | 0;
            W[i] = Wih;
            W[i + 1] = Wil;
        }
        for (let j = 0; j < 160; j += 2) {
            Wih = W[j];
            Wil = W[j + 1];
            const majh = maj$1(ah, bh, ch);
            const majl = maj$1(al, bl, cl);
            const sigma0h = sigma0$1(ah, al);
            const sigma0l = sigma0$1(al, ah);
            const sigma1h = sigma1$1(eh, el);
            const sigma1l = sigma1$1(el, eh);
            // t1 = h + sigma1 + ch + K[j] + W[j]
            const Kih = K$2[j];
            const Kil = K$2[j + 1];
            const chh = Ch(eh, fh, gh);
            const chl = Ch(el, fl, gl);
            let t1l = (hl + sigma1l) | 0;
            let t1h = (hh + sigma1h + getCarry(t1l, hl)) | 0;
            t1l = (t1l + chl) | 0;
            t1h = (t1h + chh + getCarry(t1l, chl)) | 0;
            t1l = (t1l + Kil) | 0;
            t1h = (t1h + Kih + getCarry(t1l, Kil)) | 0;
            t1l = (t1l + Wil) | 0;
            t1h = (t1h + Wih + getCarry(t1l, Wil)) | 0;
            // t2 = sigma0 + maj
            const t2l = (sigma0l + majl) | 0;
            const t2h = (sigma0h + majh + getCarry(t2l, sigma0l)) | 0;
            hh = gh;
            hl = gl;
            gh = fh;
            gl = fl;
            fh = eh;
            fl = el;
            el = (dl + t1l) | 0;
            eh = (dh + t1h + getCarry(el, dl)) | 0;
            dh = ch;
            dl = cl;
            ch = bh;
            cl = bl;
            bh = ah;
            bl = al;
            al = (t1l + t2l) | 0;
            ah = (t1h + t2h + getCarry(al, t1l)) | 0;
        }
        this._al = (this._al + al) | 0;
        this._bl = (this._bl + bl) | 0;
        this._cl = (this._cl + cl) | 0;
        this._dl = (this._dl + dl) | 0;
        this._el = (this._el + el) | 0;
        this._fl = (this._fl + fl) | 0;
        this._gl = (this._gl + gl) | 0;
        this._hl = (this._hl + hl) | 0;
        this._ah = (this._ah + ah + getCarry(this._al, al)) | 0;
        this._bh = (this._bh + bh + getCarry(this._bl, bl)) | 0;
        this._ch = (this._ch + ch + getCarry(this._cl, cl)) | 0;
        this._dh = (this._dh + dh + getCarry(this._dl, dl)) | 0;
        this._eh = (this._eh + eh + getCarry(this._el, el)) | 0;
        this._fh = (this._fh + fh + getCarry(this._fl, fl)) | 0;
        this._gh = (this._gh + gh + getCarry(this._gl, gl)) | 0;
        this._hh = (this._hh + hh + getCarry(this._hl, hl)) | 0;
    }
    _hash() {
        const H = Buffer.allocUnsafe(64);
        function writeInt64BE(h, l, offset) {
            H.writeInt32BE(h, offset);
            H.writeInt32BE(l, offset + 4);
        }
        writeInt64BE(this._ah, this._al, 0);
        writeInt64BE(this._bh, this._bl, 8);
        writeInt64BE(this._ch, this._cl, 16);
        writeInt64BE(this._dh, this._dl, 24);
        writeInt64BE(this._eh, this._el, 32);
        writeInt64BE(this._fh, this._fl, 40);
        writeInt64BE(this._gh, this._gl, 48);
        writeInt64BE(this._hh, this._hl, 56);
        return H;
    }
}

const HASH_IMPLEMENTATIONS = new Map([
    ['md5', MD5],
    ['sha-1', Sha1],
    ['sha-256', Sha256],
    ['sha-512', Sha512],
    ['sha1', Sha1],
    ['sha256', Sha256],
    ['sha512', Sha512]
]);
function createHash(alg) {
    alg = alg.toLowerCase();
    const HashImp = HASH_IMPLEMENTATIONS.get(alg);
    if (HashImp) {
        return new HashImp();
    } else {
        throw new Error('Unsupported hash algorithm: ' + alg);
    }
}

/**
 * This file is derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from:
 * - hash-base, Copyright (c) 2016 Kirill Fomichev
 * - cipher-base, Copyright (c) 2017 crypto-browserify contributors
 * - create-hash, Copyright (c) 2017 crypto-browserify contributors
 * - create-hmac, Copyright (c) 2017 crypto-browserify contributors
 * - randombytes, Copyright (c) 2017 crypto-browserify
 */
const ZEROS = Buffer.alloc(128);
class Hmac extends Transform {
    constructor(alg, key) {
        super();
        if (typeof key === 'string') {
            key = Buffer.from(key);
        }
        const blocksize = alg === 'sha512' ? 128 : 64;
        this._alg = alg;
        if (key.length > blocksize) {
            key = createHash(alg)
                .update(key)
                .digest();
        } else if (key.length < blocksize) {
            key = Buffer.concat([key, ZEROS], blocksize);
        }
        this._ipad = Buffer.alloc(blocksize);
        this._opad = Buffer.alloc(blocksize);
        for (let i = 0; i < blocksize; i++) {
            this._ipad[i] = key[i] ^ 0x36;
            this._opad[i] = key[i] ^ 0x5c;
        }
        this._hash = createHash(alg).update(this._ipad);
    }
    _transform(data, enc, next) {
        let err;
        try {
            this.update(data, enc);
        } catch (e) {
            err = e;
        } finally {
            next(err);
        }
    }
    _flush(done) {
        let err;
        try {
            this.push(this._final());
        } catch (e) {
            err = e;
        }
        done(err);
    }
    _final() {
        const h = this._hash.digest();
        return createHash(this._alg)
            .update(this._opad)
            .update(h)
            .digest();
    }
    update(data, inputEnc) {
        this._hash.update(data, inputEnc);
        return this;
    }
    digest(outputEnc) {
        const outData = this._final() || Buffer.alloc(0);
        if (outputEnc) {
            return outData.toString(outputEnc);
        }
        return outData;
    }
}

let root;
if (typeof window !== 'undefined') {
    root = window;
} else if (typeof global !== 'undefined') {
    root = global;
}
function randomBytes(size) {
    const rawBytes = new Uint8Array(size);
    if (size > 0) {
        (root.crypto || root.msCrypto).getRandomValues(rawBytes);
    }
    return Buffer.from(rawBytes.buffer);
}
function getHashes() {
    return ['sha-1', 'sha-256', 'sha-512', 'md5'];
}
function createHmac(alg, key) {
    return new Hmac(alg.toLowerCase(), key);
}
const nativeFetch = fetch;
const nativeWS = WebSocket;
const nativeRTCPeerConnection = root.RTCPeerConnection;

class SimpleMech {
    constructor(name) {
        this.authenticated = false;
        this.mutuallyAuthenticated = false;
        this.name = name;
    }
    getCacheableCredentials() {
        return null;
    }
    // istanbul ignore next
    processChallenge(_challenge) {
        return;
    }
    processSuccess(_success) {
        this.authenticated = true;
    }
    finalize() {
        const result = {
            authenticated: this.authenticated,
            mutuallyAuthenticated: this.mutuallyAuthenticated
        };
        if (this.errorData) {
            result.errorData = this.errorData;
        }
        return result;
    }
}
class Factory {
    constructor() {
        this.mechanisms = [];
    }
    register(name, constructor, priority) {
        this.mechanisms.push({
            constructor,
            name: name.toUpperCase(),
            priority: priority
        });
        // We want mechanisms with highest priority at the start of the list
        this.mechanisms.sort((a, b) => b.priority - a.priority);
    }
    disable(name) {
        const mechName = name.toUpperCase();
        this.mechanisms = this.mechanisms.filter(mech => mech.name !== mechName);
    }
    createMechanism(names) {
        const availableNames = names.map(name => name.toUpperCase());
        for (const knownMech of this.mechanisms) {
            for (const availableMechName of availableNames) {
                if (availableMechName === knownMech.name) {
                    return new knownMech.constructor(knownMech.name);
                }
            }
        }
        return null;
    }
}
// ====================================================================
// Utility helpers
// ====================================================================
// istanbul ignore next
function createClientNonce(length = 32) {
    return randomBytes(length).toString('hex');
}
// tslint:disable no-bitwise
function XOR(a, b) {
    const res = [];
    for (let i = 0; i < a.length; i++) {
        res.push(a[i] ^ b[i]);
    }
    return Buffer.from(res);
}
// tslint:enable no-bitwise
function H(text, alg) {
    return createHash(alg).update(text).digest();
}
function HMAC(key, msg, alg) {
    return createHmac(alg, key).update(msg).digest();
}
function Hi(text, salt, iterations, alg) {
    let ui1 = HMAC(text, Buffer.concat([salt, Buffer.from('00000001', 'hex')]), alg);
    let ui = ui1;
    for (let i = 0; i < iterations - 1; i++) {
        ui1 = HMAC(text, ui1, alg);
        ui = XOR(ui, ui1);
    }
    return ui;
}
function parse(challenge) {
    const directives = {};
    const tokens = challenge.toString().split(/,(?=(?:[^"]|"[^"]*")*$)/);
    for (let i = 0, len = tokens.length; i < len; i++) {
        const directive = /(\w+)=["]?([^"]+)["]?$/.exec(tokens[i]);
        if (directive) {
            directives[directive[1]] = directive[2];
        }
    }
    return directives;
}
function escapeUsername(name) {
    const escaped = [];
    for (const curr of name) {
        if (curr === ',') {
            escaped.push('=2C');
        }
        else if (curr === '=') {
            escaped.push('=3D');
        }
        else {
            escaped.push(curr);
        }
    }
    return escaped.join('');
}
// ====================================================================
// ANONYMOUS
// ====================================================================
class ANONYMOUS extends SimpleMech {
    getExpectedCredentials() {
        return { optional: ['trace'], required: [] };
    }
    createResponse(credentials) {
        return Buffer.from(credentials.trace || '');
    }
}
// ====================================================================
// EXTERNAL
// ====================================================================
class EXTERNAL extends SimpleMech {
    getExpectedCredentials() {
        return { optional: ['authzid'], required: [] };
    }
    createResponse(credentials) {
        return Buffer.from(credentials.authzid || '');
    }
}
// ====================================================================
// PLAIN
// ====================================================================
class PLAIN extends SimpleMech {
    getExpectedCredentials() {
        return {
            optional: ['authzid'],
            required: ['username', 'password']
        };
    }
    createResponse(credentials) {
        return Buffer.from((credentials.authzid || '') +
            '\x00' +
            credentials.username +
            '\x00' +
            (credentials.password || credentials.token));
    }
}
// ====================================================================
// OAUTHBEARER
// ====================================================================
class OAUTH extends SimpleMech {
    constructor(name) {
        super(name);
        this.failed = false;
        this.name = name;
    }
    getExpectedCredentials() {
        return {
            optional: ['authzid'],
            required: ['token']
        };
    }
    createResponse(credentials) {
        if (this.failed) {
            return Buffer.from('\u0001');
        }
        const gs2header = `n,${escapeUsername(saslprep(credentials.authzid))},`;
        const auth = `auth=Bearer ${credentials.token}\u0001`;
        return Buffer.from(gs2header + '\u0001' + auth + '\u0001', 'utf8');
    }
    processChallenge(challenge) {
        this.failed = true;
        this.errorData = JSON.parse(challenge.toString('utf8'));
    }
}
// ====================================================================
// DIGEST-MD5
// ====================================================================
class DIGEST extends SimpleMech {
    constructor(name) {
        super(name);
        this.providesMutualAuthentication = false;
        this.state = 'INITIAL';
        this.name = name;
    }
    processChallenge(challenge) {
        this.state = 'CHALLENGE';
        const values = parse(challenge);
        this.authenticated = !!values.rspauth;
        this.realm = values.realm;
        this.nonce = values.nonce;
        this.charset = values.charset;
    }
    getExpectedCredentials() {
        return {
            optional: ['authzid', 'clientNonce', 'realm'],
            required: ['host', 'password', 'serviceName', 'serviceType', 'username']
        };
    }
    createResponse(credentials) {
        if (this.state === 'INITIAL' || this.authenticated) {
            return null;
        }
        let uri = credentials.serviceType + '/' + credentials.host;
        if (credentials.serviceName && credentials.host !== credentials.serviceName) {
            uri += '/' + credentials.serviceName;
        }
        const realm = credentials.realm || this.realm || '';
        const cnonce = credentials.clientNonce || createClientNonce(16);
        const nc = '00000001';
        const qop = 'auth';
        let str = '';
        str += 'username="' + credentials.username + '"';
        if (realm) {
            str += ',realm="' + realm + '"';
        }
        str += ',nonce="' + this.nonce + '"';
        str += ',cnonce="' + cnonce + '"';
        str += ',nc=' + nc;
        str += ',qop=' + qop;
        str += ',digest-uri="' + uri + '"';
        const base = createHash('md5')
            .update(credentials.username)
            .update(':')
            .update(realm)
            .update(':')
            .update(credentials.password)
            .digest();
        const ha1 = createHash('md5')
            .update(base)
            .update(':')
            .update(this.nonce)
            .update(':')
            .update(cnonce);
        if (credentials.authzid) {
            ha1.update(':').update(credentials.authzid);
        }
        const dha1 = ha1.digest('hex');
        const ha2 = createHash('md5').update('AUTHENTICATE:').update(uri);
        const dha2 = ha2.digest('hex');
        const digest = createHash('md5')
            .update(dha1)
            .update(':')
            .update(this.nonce)
            .update(':')
            .update(nc)
            .update(':')
            .update(cnonce)
            .update(':')
            .update(qop)
            .update(':')
            .update(dha2)
            .digest('hex');
        str += ',response=' + digest;
        if (this.charset === 'utf-8') {
            str += ',charset=utf-8';
        }
        if (credentials.authzid) {
            str += ',authzid="' + credentials.authzid + '"';
        }
        return Buffer.from(str);
    }
}
// ====================================================================
// SCRAM-SHA-1(-PLUS)
// ====================================================================
class SCRAM {
    constructor(name) {
        this.providesMutualAuthentication = true;
        this.name = name;
        this.state = 'INITIAL';
        this.useChannelBinding = this.name.toLowerCase().endsWith('-plus');
        this.algorithm = this.name.toLowerCase().split('scram-')[1].split('-plus')[0];
    }
    getExpectedCredentials() {
        const optional = ['authzid', 'clientNonce'];
        const required = ['username', 'password'];
        if (this.useChannelBinding) {
            required.push('tlsUnique');
        }
        return {
            optional,
            required
        };
    }
    getCacheableCredentials() {
        return this.cache;
    }
    createResponse(credentials) {
        if (this.state === 'INITIAL') {
            return this.initialResponse(credentials);
        }
        return this.challengeResponse(credentials);
    }
    processChallenge(challenge) {
        const values = parse(challenge);
        this.salt = Buffer.from(values.s || '', 'base64');
        this.iterationCount = parseInt(values.i, 10);
        this.nonce = values.r;
        this.verifier = values.v;
        this.error = values.e;
        this.challenge = challenge;
    }
    processSuccess(success) {
        this.processChallenge(success);
    }
    finalize() {
        if (!this.verifier) {
            return {
                authenticated: false,
                error: this.error,
                mutuallyAuthenticated: false
            };
        }
        if (this.serverSignature.toString('base64') !== this.verifier) {
            return {
                authenticated: false,
                error: 'Mutual authentication failed',
                mutuallyAuthenticated: false
            };
        }
        return {
            authenticated: true,
            mutuallyAuthenticated: true
        };
    }
    initialResponse(credentials) {
        const authzid = escapeUsername(saslprep(credentials.authzid));
        const username = escapeUsername(saslprep(credentials.username));
        this.clientNonce = credentials.clientNonce || createClientNonce();
        let cbindHeader = 'n';
        if (credentials.tlsUnique) {
            if (!this.useChannelBinding) {
                cbindHeader = 'y';
            }
            else {
                cbindHeader = 'p=tls-unique';
            }
        }
        this.gs2Header = Buffer.from(authzid ? `${cbindHeader},a=${authzid},` : `${cbindHeader},,`);
        this.clientFirstMessageBare = Buffer.from(`n=${username},r=${this.clientNonce}`);
        const result = Buffer.concat([this.gs2Header, this.clientFirstMessageBare]);
        this.state = 'CHALLENGE';
        return result;
    }
    challengeResponse(credentials) {
        const CLIENT_KEY = Buffer.from('Client Key');
        const SERVER_KEY = Buffer.from('Server Key');
        const cbindData = Buffer.concat([
            this.gs2Header,
            credentials.tlsUnique || Buffer.from('')
        ]).toString('base64');
        const clientFinalMessageWithoutProof = Buffer.from(`c=${cbindData},r=${this.nonce}`);
        let saltedPassword;
        let clientKey;
        let serverKey;
        // If our cached salt is the same, we can reuse cached credentials to speed
        // up the hashing process.
        const cached = credentials.salt && Buffer.compare(credentials.salt, this.salt) === 0;
        if (cached && credentials.clientKey && credentials.serverKey) {
            clientKey = Buffer.from(credentials.clientKey);
            serverKey = Buffer.from(credentials.serverKey);
        }
        else if (cached && credentials.saltedPassword) {
            saltedPassword = Buffer.from(credentials.saltedPassword);
            clientKey = HMAC(saltedPassword, CLIENT_KEY, this.algorithm);
            serverKey = HMAC(saltedPassword, SERVER_KEY, this.algorithm);
        }
        else {
            saltedPassword = Hi(Buffer.from(saslprep(credentials.password)), this.salt, this.iterationCount, this.algorithm);
            clientKey = HMAC(saltedPassword, CLIENT_KEY, this.algorithm);
            serverKey = HMAC(saltedPassword, SERVER_KEY, this.algorithm);
        }
        const storedKey = H(clientKey, this.algorithm);
        const separator = Buffer.from(',');
        const authMessage = Buffer.concat([
            this.clientFirstMessageBare,
            separator,
            this.challenge,
            separator,
            clientFinalMessageWithoutProof
        ]);
        const clientSignature = HMAC(storedKey, authMessage, this.algorithm);
        const clientProof = XOR(clientKey, clientSignature).toString('base64');
        this.serverSignature = HMAC(serverKey, authMessage, this.algorithm);
        const result = Buffer.concat([
            clientFinalMessageWithoutProof,
            Buffer.from(`,p=${clientProof}`)
        ]);
        this.state = 'FINAL';
        this.cache = {
            clientKey,
            salt: this.salt,
            saltedPassword,
            serverKey
        };
        return result;
    }
}

var index$2 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    SimpleMech: SimpleMech,
    Factory: Factory,
    createClientNonce: createClientNonce,
    XOR: XOR,
    H: H,
    HMAC: HMAC,
    Hi: Hi,
    ANONYMOUS: ANONYMOUS,
    EXTERNAL: EXTERNAL,
    PLAIN: PLAIN,
    OAUTH: OAUTH,
    DIGEST: DIGEST,
    SCRAM: SCRAM
});

function Account (client) {
    client.getAccountInfo = (jid) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            account: {},
            to: jid,
            type: 'get'
        });
        return resp.account;
    });
    client.updateAccount = (jid, data) => {
        return client.sendIQ({
            account: data,
            to: jid,
            type: 'set'
        });
    };
    client.deleteAccount = (jid) => {
        return client.sendIQ({
            account: {
                remove: true
            },
            to: jid,
            type: 'set'
        });
    };
    client.getPrivateData = (key) => __awaiter(this, void 0, void 0, function* () {
        const res = yield client.sendIQ({
            privateStorage: {
                [key]: {}
            },
            type: 'get'
        });
        return res.privateStorage[key];
    });
    client.setPrivateData = (key, value) => __awaiter(this, void 0, void 0, function* () {
        return client.sendIQ({
            privateStorage: {
                [key]: value
            },
            type: 'set'
        });
    });
    client.getVCard = (jid) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            to: jid,
            type: 'get',
            vcard: {
                format: NS_VCARD_TEMP
            }
        });
        return resp.vcard;
    });
    client.publishVCard = (vcard) => __awaiter(this, void 0, void 0, function* () {
        yield client.sendIQ({
            type: 'set',
            vcard
        });
    });
    client.enableNotifications = (jid, node, fieldList = []) => {
        return client.sendIQ({
            push: {
                action: 'enable',
                form: {
                    fields: [
                        {
                            name: 'FORM_TYPE',
                            type: 'hidden',
                            value: 'http://jabber.org/protocol/pubsub#publish-options'
                        },
                        ...fieldList
                    ],
                    type: 'submit'
                },
                jid,
                node
            },
            type: 'set'
        });
    };
    client.disableNotifications = (jid, node) => {
        return client.sendIQ({
            push: {
                action: 'disable',
                jid,
                node
            },
            type: 'set'
        });
    };
}

function Avatar (client) {
    client.disco.addFeature(NS_PEP_NOTIFY(NS_AVATAR_METADATA));
    client.on('pubsub:published', msg => {
        if (msg.pubsub.items.node !== NS_AVATAR_METADATA) {
            return;
        }
        const info = msg.pubsub.items.published[0].content;
        client.emit('avatar', {
            avatars: info.versions || [],
            jid: msg.from,
            source: 'pubsub'
        });
    });
    client.on('presence', pres => {
        if (pres.vcardAvatar && typeof pres.vcardAvatar === 'string') {
            client.emit('avatar', {
                avatars: [
                    {
                        id: pres.vcardAvatar
                    }
                ],
                jid: pres.from,
                source: 'vcard'
            });
        }
    });
    client.publishAvatar = (id, data) => {
        return client.publish('', NS_AVATAR_DATA, {
            data,
            itemType: NS_AVATAR_DATA
        }, id);
    };
    client.useAvatars = (versions, pointers = []) => {
        return client.publish('', NS_AVATAR_METADATA, {
            itemType: NS_AVATAR_METADATA,
            pointers,
            versions
        }, 'current');
    };
    client.getAvatar = (jid, id) => {
        return client.getItem(jid, NS_AVATAR_DATA, id);
    };
}

function Bind (client) {
    client.registerFeature('bind', 300, (features, cb) => __awaiter(this, void 0, void 0, function* () {
        try {
            const resp = yield client.sendIQ({
                bind: {
                    resource: client.config.resource
                },
                type: 'set'
            });
            client.features.negotiated.bind = true;
            client.emit('session:prebind', resp.bind.jid);
            const canStartSession = !features.legacySession ||
                (features.legacySession && features.legacySession.optional);
            if (!client.sessionStarted && canStartSession) {
                client.emit('session:started', client.jid);
            }
            return cb();
        }
        catch (err) {
            console.error(err);
            return cb('disconnect', 'JID binding failed');
        }
    }));
    client.registerFeature('legacySession', 1000, (features, cb) => __awaiter(this, void 0, void 0, function* () {
        if (client.sessionStarted || (features.legacySession && features.legacySession.optional)) {
            client.features.negotiated.session = true;
            return cb();
        }
        try {
            yield client.sendIQ({
                legacySession: true,
                type: 'set'
            });
            client.features.negotiated.session = true;
            if (!client.sessionStarted) {
                client.sessionStarted = true;
                client.emit('session:started', client.jid);
            }
            return cb();
        }
        catch (err) {
            return cb('disconnect', 'Session requeset failed');
        }
    }));
    client.on('session:started', () => {
        client.sessionStarted = true;
    });
    client.on('session:prebind', boundJID => {
        client.jid = boundJID;
        client.emit('session:bound', client.jid);
    });
    client.on('--reset-stream-features', () => {
        client.sessionStarted = false;
        client.features.negotiated.bind = false;
        client.features.negotiated.session = false;
    });
}

function Command (client) {
    client.disco.addFeature(NS_ADHOC_COMMANDS);
    client.disco.addItem({
        name: 'Ad-Hoc Commands',
        node: NS_ADHOC_COMMANDS
    });
    client.getCommands = (jid) => {
        return client.getDiscoItems(jid, NS_ADHOC_COMMANDS);
    };
}

/**
 * Portions of this file are derived from prior work.
 *
 * See NOTICE.md for full license text.
 *
 * Derived from:
 * - uuid, Copyright (c) 2010-2016 Robert Kieffer and other contributors
 */
const bth = [];
for (let i = 0; i < 256; ++i) {
    bth[i] = (i + 0x100).toString(16).substr(1);
}
function timeoutPromise(target, delay, rejectValue = () => undefined) {
    return __awaiter(this, void 0, void 0, function* () {
        let timeoutRef;
        const result = yield Promise.race([
            target,
            new Promise((resolve, reject) => {
                timeoutRef = setTimeout(() => reject(rejectValue()), delay);
            })
        ]);
        clearTimeout(timeoutRef);
        return result;
    });
}
function sleep(time) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise(resolve => {
            setTimeout(() => resolve(), time);
        });
    });
}
function octetCompare(str1, str2) {
    const b1 = typeof str1 === 'string' ? Buffer.from(str1, 'utf8') : str1;
    const b2 = typeof str2 === 'string' ? Buffer.from(str2, 'utf8') : str2;
    return b1.compare(b2);
}
function uuid() {
    const buf = randomBytes(16);
    // Per 4.4, set bits for version and `clock_seq_hi_and_reserved`
    buf[6] = (buf[6] & 0x0f) | 0x40;
    buf[8] = (buf[8] & 0x3f) | 0x80;
    let i = 0;
    return [
        bth[buf[i++]],
        bth[buf[i++]],
        bth[buf[i++]],
        bth[buf[i++]],
        '-',
        bth[buf[i++]],
        bth[buf[i++]],
        '-',
        bth[buf[i++]],
        bth[buf[i++]],
        '-',
        bth[buf[i++]],
        bth[buf[i++]],
        '-',
        bth[buf[i++]],
        bth[buf[i++]],
        bth[buf[i++]],
        bth[buf[i++]],
        bth[buf[i++]],
        bth[buf[i]]
    ].join('');
}
const DATE_FIELDS = new Set([
    'date',
    'expires',
    'httpUploadRetry',
    'idleSince',
    'published',
    'since',
    'stamp',
    'timestamp',
    'updated',
    'utc'
]);
const ISO_DT = /^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}(?:\.\d*)?)(?:Z|((\+|-)([\d|:]*)))?$/;
function reviveData(key, value) {
    if (DATE_FIELDS.has(key) && value && typeof value === 'string' && ISO_DT.test(value)) {
        return new Date(value);
    }
    if (value &&
        typeof value === 'object' &&
        value.type === 'Buffer' &&
        Array.isArray(value.data)) {
        return Buffer.from(value);
    }
    return value;
}

var Utils = /*#__PURE__*/Object.freeze({
    __proto__: null,
    timeoutPromise: timeoutPromise,
    sleep: sleep,
    octetCompare: octetCompare,
    uuid: uuid,
    reviveData: reviveData
});

function checkConnection(client) {
    return __awaiter(this, void 0, void 0, function* () {
        if (client.sm.started) {
            return new Promise(resolve => {
                client.once('stream:management:ack', () => resolve());
                client.sm.request();
            });
        }
        try {
            yield client.ping();
        }
        catch (err) {
            if (err.error && err.error.condition !== 'timeout') {
                return;
            }
            else {
                throw err;
            }
        }
    });
}
function sendCSI(client, type) {
    if (client.features.negotiated.clientStateIndication) {
        client.send('csi', {
            type
        });
    }
}
function Connection (client) {
    client.disco.addFeature(NS_PING);
    client.on('iq:get:ping', iq => {
        client.sendIQResult(iq);
    });
    client.on('--reset-stream-features', () => {
        client._stopKeepAliveInterval();
        client.features.negotiated.streamManagement = false;
        client.features.negotiated.clientStateIndication = false;
    });
    client.markActive = () => sendCSI(client, 'active');
    client.markInactive = () => sendCSI(client, 'inactive');
    client.ping = (jid) => __awaiter(this, void 0, void 0, function* () {
        yield client.sendIQ({
            ping: true,
            to: jid,
            type: 'get'
        });
    });
    client.enableKeepAlive = (opts = {}) => {
        client._keepAliveOptions = opts;
        // Ping every 5 minutes
        const interval = opts.interval || 300;
        // Disconnect if no response in 15 seconds
        const timeout = opts.timeout || client.config.timeout || 15;
        function keepalive() {
            return __awaiter(this, void 0, void 0, function* () {
                if (client.sessionStarted) {
                    try {
                        yield timeoutPromise(checkConnection(client), timeout * 1000);
                    }
                    catch (err) {
                        // Kill the apparently dead connection without closing
                        // the stream itself so we can reconnect and potentially
                        // resume the session.
                        client.emit('stream:error', {
                            condition: 'connection-timeout',
                            text: 'Server did not respond in ' + timeout + ' seconds'
                        });
                        if (client.transport) {
                            client.transport.hasStream = false;
                            client.transport.disconnect(false);
                        }
                    }
                }
            });
        }
        clearInterval(client._keepAliveInterval);
        client._keepAliveInterval = setInterval(keepalive, interval * 1000);
    };
    client._stopKeepAliveInterval = () => {
        if (client._keepAliveInterval) {
            clearInterval(client._keepAliveInterval);
            delete client._keepAliveInterval;
        }
    };
    client.disableKeepAlive = () => {
        delete client._keepAliveOptions;
        client._stopKeepAliveInterval();
    };
    client.on('stream:management:resumed', () => {
        client._keepAliveOptions && client.enableKeepAlive(client._keepAliveOptions);
    });
    client.on('stream:start', () => {
        client._keepAliveOptions && client.enableKeepAlive(client._keepAliveOptions);
    });
    const smacks = (features, done) => __awaiter(this, void 0, void 0, function* () {
        if (!client.config.useStreamManagement) {
            return done();
        }
        const smHandler = (sm) => __awaiter(this, void 0, void 0, function* () {
            switch (sm.type) {
                case 'enabled':
                    yield client.sm.enabled(sm);
                    client.features.negotiated.streamManagement = true;
                    client.off('sm', smHandler);
                    return done();
                case 'resumed':
                    yield client.sm.resumed(sm);
                    client.features.negotiated.streamManagement = true;
                    client.features.negotiated.bind = true;
                    client.sessionStarted = true;
                    client.sessionStarting = false;
                    client.off('sm', smHandler);
                    client.emit('stream:management:resumed', sm);
                    return done('break'); // Halt further processing of stream features
                case 'failed':
                    yield client.sm.failed(sm);
                    client.off('sm', smHandler);
                    client.emit('session:end');
                    done();
            }
        });
        client.on('sm', smHandler);
        if (!client.sm.id) {
            if (client.features.negotiated.bind) {
                yield client.sm.enable();
            }
            else {
                client.off('sm', smHandler);
                done();
            }
        }
        else if (client.sm.id && client.sm.allowResume) {
            yield client.sm.resume();
        }
        else {
            client.off('sm', smHandler);
            done();
        }
    });
    client.registerFeature('streamManagement', 200, smacks);
    client.registerFeature('streamManagement', 500, smacks);
    client.registerFeature('clientStateIndication', 400, (features, cb) => {
        client.features.negotiated.clientStateIndication = true;
        cb();
    });
}

function escape(value) {
    return Buffer.from(value.replace(/</g, '&lt;'), 'utf-8');
}
function encodeIdentities(identities = []) {
    const result = [];
    const existing = new Set();
    for (const { category, type, lang, name } of identities) {
        const encoded = `${category}/${type}/${lang || ''}/${name || ''}`;
        if (existing.has(encoded)) {
            return null;
        }
        existing.add(encoded);
        result.push(escape(encoded));
    }
    result.sort(octetCompare);
    return result;
}
function encodeFeatures(features = []) {
    const result = [];
    const existing = new Set();
    for (const feature of features) {
        if (existing.has(feature)) {
            return null;
        }
        existing.add(feature);
        result.push(escape(feature));
    }
    result.sort(octetCompare);
    return result;
}
function encodeFields(fields = []) {
    const sortedFields = [];
    for (const field of fields) {
        if (field.name === 'FORM_TYPE') {
            continue;
        }
        if (field.rawValues) {
            sortedFields.push({
                name: escape(field.name),
                values: field.rawValues.map(val => escape(val)).sort(octetCompare)
            });
        }
        else if (Array.isArray(field.value)) {
            sortedFields.push({
                name: escape(field.name),
                values: field.value.map(val => escape(val)).sort(octetCompare)
            });
        }
        else if (field.value === true || field.value === false) {
            sortedFields.push({
                name: escape(field.name),
                values: [escape(field.value ? '1' : '0')]
            });
        }
        else {
            sortedFields.push({
                name: escape(field.name),
                values: [escape(field.value || '')]
            });
        }
    }
    sortedFields.sort((a, b) => octetCompare(a.name, b.name));
    const result = [];
    for (const field of sortedFields) {
        result.push(field.name);
        for (const value of field.values) {
            result.push(value);
        }
    }
    return result;
}
function encodeForms(extensions = []) {
    const forms = [];
    const types = new Set();
    for (const form of extensions) {
        let type;
        for (const field of form.fields || []) {
            if (!(field.name === 'FORM_TYPE' && field.type === 'hidden')) {
                continue;
            }
            if (field.rawValues && field.rawValues.length === 1) {
                type = escape(field.rawValues[0]);
                break;
            }
            if (field.value && typeof field.value === 'string') {
                type = escape(field.value);
                break;
            }
        }
        if (!type) {
            continue;
        }
        if (types.has(type.toString())) {
            return null;
        }
        types.add(type.toString());
        forms.push({ type, form });
    }
    forms.sort((a, b) => octetCompare(a.type, b.type));
    const results = [];
    for (const form of forms) {
        results.push(form.type);
        const fields = encodeFields(form.form.fields);
        for (const field of fields) {
            results.push(field);
        }
    }
    return results;
}
function generate(info, hashName) {
    const S = [];
    const separator = Buffer.from('<', 'utf8');
    const append = (b1) => {
        S.push(b1);
        S.push(separator);
    };
    const identities = encodeIdentities(info.identities);
    const features = encodeFeatures(info.features);
    const extensions = encodeForms(info.extensions);
    if (!identities || !features || !extensions) {
        return null;
    }
    for (const id of identities) {
        append(id);
    }
    for (const feature of features) {
        append(feature);
    }
    for (const form of extensions) {
        append(form);
    }
    return createHash(hashName).update(Buffer.concat(S)).digest('base64');
}

class Disco$1 {
    constructor() {
        this.capsAlgorithms = ['sha-1'];
        this.features = new Map();
        this.identities = new Map();
        this.extensions = new Map();
        this.items = new Map();
        this.caps = new Map();
        this.features.set('', new Set());
        this.identities.set('', []);
        this.extensions.set('', []);
    }
    getNodeInfo(node) {
        return {
            extensions: [...(this.extensions.get(node) || [])],
            features: [...(this.features.get(node) || [])],
            identities: [...(this.identities.get(node) || [])]
        };
    }
    addFeature(feature, node = '') {
        if (!this.features.has(node)) {
            this.features.set(node, new Set());
        }
        this.features.get(node).add(feature);
    }
    addIdentity(identity, node = '') {
        if (!this.identities.has(node)) {
            this.identities.set(node, []);
        }
        this.identities.get(node).push(identity);
    }
    addItem(item, node = '') {
        if (!this.items.has(node)) {
            this.items.set(node, []);
        }
        this.items.get(node).push(item);
    }
    addExtension(form, node = '') {
        if (!this.extensions.has(node)) {
            this.extensions.set(node, []);
        }
        this.extensions.get(node).push(form);
    }
    updateCaps(node, algorithms = this.capsAlgorithms) {
        const info = {
            extensions: [...this.extensions.get('')],
            features: [...this.features.get('')],
            identities: [...this.identities.get('')],
            type: 'info'
        };
        for (const algorithm of algorithms) {
            const version = generate(info, algorithm);
            if (!version) {
                this.caps.delete(algorithm);
                continue;
            }
            this.caps.set(algorithm, {
                algorithm,
                node,
                value: version
            });
            const hashedNode = `${node}#${version}`;
            for (const feature of info.features) {
                this.addFeature(feature, hashedNode);
            }
            for (const identity of info.identities) {
                this.addIdentity(identity, hashedNode);
            }
            for (const form of info.extensions) {
                this.addExtension(form, hashedNode);
            }
            this.identities.set(hashedNode, info.identities);
            this.features.set(hashedNode, new Set(info.features));
            this.extensions.set(hashedNode, info.extensions);
        }
        return [...this.caps.values()];
    }
    getCaps() {
        return [...this.caps.values()];
    }
}

function Disco (client) {
    client.disco = new Disco$1();
    client.disco.addFeature(NS_DISCO_INFO);
    client.disco.addFeature(NS_DISCO_ITEMS);
    client.disco.addIdentity({
        category: 'client',
        type: 'web'
    });
    client.registerFeature('caps', 100, (features, done) => {
        const domain = getDomain(client.jid) || client.config.server;
        client.emit('disco:caps', {
            caps: features.legacyCapabilities || [],
            jid: domain
        });
        client.features.negotiated.caps = true;
        done();
    });
    client.getDiscoInfo = (jid, node) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            disco: {
                node,
                type: 'info'
            },
            to: jid,
            type: 'get'
        });
        return Object.assign({ extensions: [], features: [], identities: [] }, resp.disco);
    });
    client.getDiscoItems = (jid, node) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            disco: {
                node,
                type: 'items'
            },
            to: jid,
            type: 'get'
        });
        return Object.assign({ items: [] }, resp.disco);
    });
    client.updateCaps = () => {
        const node = client.config.capsNode || 'https://stanzajs.org';
        return client.disco.updateCaps(node);
    };
    client.getCurrentCaps = () => {
        const caps = client.disco.getCaps();
        if (!caps) {
            return;
        }
        const node = `${caps[0].node}#${caps[0].value}`;
        return {
            info: client.disco.getNodeInfo(node),
            legacyCapabilities: caps
        };
    };
    client.on('presence', pres => {
        if (pres.legacyCapabilities) {
            client.emit('disco:caps', {
                caps: pres.legacyCapabilities,
                jid: pres.from
            });
        }
    });
    client.on('iq:get:disco', iq => {
        const { type, node } = iq.disco;
        if (type === 'info') {
            client.sendIQResult(iq, {
                disco: Object.assign(Object.assign({}, client.disco.getNodeInfo(node || '')), { node, type: 'info' })
            });
        }
        if (type === 'items') {
            client.sendIQResult(iq, {
                disco: {
                    items: client.disco.items.get(node || '') || [],
                    type: 'items'
                }
            });
        }
    });
}

const VERSION$1 = '12.17.1';
// ====================================================================
// Frequently Used Values
// ====================================================================
const NotAuthorized = 'not-authorized';
// ====================================================================
// Named Enum Constants
// ====================================================================
const StreamType = {
    Bosh: NS_BOSH,
    Client: NS_CLIENT,
    Component: NS_COMPONENT,
    Server: NS_SERVER
};
const SASLFailureCondition = {
    AccountDisabled: 'account-disabled',
    CredentialsExpired: 'credentials-expired',
    EncryptionRequired: 'encryption-required',
    IncorrectEncoding: 'incorrect-encoding',
    InvalidAuthzid: 'invalid-authzid',
    InvalidMechanism: 'invalid-mechanism',
    MalformedRequest: 'malformed-request',
    MechanismTooWeak: 'mechanism-too-weak',
    NotAuthorized,
    TemporaryAuthFailure: 'temporary-auth-failure'
};
const StreamErrorCondition = {
    BadFormat: 'bad-format',
    BadNamespacePrefix: 'bad-namespace-prefix',
    Conflict: 'conflict',
    ConnectionTimeout: 'connection-timeout',
    HostGone: 'host-gone',
    HostUnknown: 'host-unknown',
    ImproperAddressing: 'improper-addressing',
    InternalServerError: 'internal-server-error',
    InvalidFrom: 'invalid-from',
    InvalidId: 'invalid-id',
    InvalidNamespace: 'invalid-namespace',
    InvalidXML: 'invalid-xml',
    NotAuthorized,
    NotWellFormed: 'not-well-formed',
    PolicyViolation: 'policy-violation',
    RemoteConnectionFailed: 'remote-connection-failed',
    Reset: 'reset',
    ResourceConstraint: 'resource-constraint',
    RestrictedXML: 'restricted-xml',
    SeeOtherHost: 'see-other-host',
    SystemShutdown: 'system-shutdown',
    UndefinedCondition: 'undefined-condition',
    UnsupportedEncoding: 'unsupported-encoding',
    UnsupportedStanzaType: 'unsupported-stanza-type',
    UnsupportedVersion: 'unsupported-version'
};
const StanzaErrorCondition = {
    BadRequest: 'bad-request',
    Conflict: 'conflict',
    FeatureNotImplemented: 'feature-not-implemented',
    Forbidden: 'forbidden',
    Gone: 'gone',
    InternalServerError: 'internal-server-error',
    ItemNotFound: 'item-not-found',
    JIDMalformed: 'jid-malformed',
    NotAcceptable: 'not-acceptable',
    NotAllowed: 'not-allowed',
    NotAuthorized,
    PolicyViolation: 'policy-violation',
    RecipientUnavailable: 'recipient-unavailable',
    Redirect: 'redirect',
    RegistrationRequired: 'registration-required',
    RemoteServerNotFound: 'remote-server-not-found',
    RemoteServerTimeout: 'remote-server-timeout',
    ResourceConstraint: 'resource-constraint',
    ServiceUnavailable: 'service-unavailable',
    SubscriptionRequired: 'subscription-required',
    UndefinedCondition: 'undefined-condition',
    UnexpectedRequest: 'unexpected-request'
};
const MessageType = {
    Chat: 'chat',
    Error: 'error',
    GroupChat: 'groupchat',
    Headline: 'headline',
    Normal: 'normal'
};
const PresenceType = {
    Available: undefined,
    Error: 'error',
    Probe: 'probe',
    Subscribe: 'subscribe',
    Subscribed: 'subscribed',
    Unavailable: 'unavailable',
    Unsubscribe: 'unsubscribe',
    Unsubscribed: 'unsubscribed'
};
const IQType = {
    Error: 'error',
    Get: 'get',
    Result: 'result',
    Set: 'set'
};
const PresenceShow = {
    Away: 'away',
    Chat: 'chat',
    DoNotDisturb: 'dnd',
    ExtendedAway: 'xa'
};
const RosterSubscription = {
    Both: 'both',
    From: 'from',
    None: 'none',
    ReceivePresenceOnly: 'to',
    Remove: 'remove',
    SendAndReceivePresence: 'both',
    SendPresenceOnly: 'from',
    To: 'to'
};
const DataFormType = {
    Cancel: 'cancel',
    Form: 'form',
    Result: 'result',
    Submit: 'submit'
};
const DataFormFieldType = {
    Boolean: 'boolean',
    Fixed: 'fixed',
    Hidden: 'hidden',
    JID: 'jid-single',
    JIDMultiple: 'jid-multi',
    List: 'list-single',
    ListMultiple: 'list-multi',
    Password: 'text-private',
    Text: 'text-single',
    TextMultiple: 'text-multi',
    TextPrivate: 'text-private'
};
const MUCAffiliation = {
    Admin: 'admin',
    Banned: 'outcast',
    Member: 'member',
    None: 'none',
    Outcast: 'outcast',
    Owner: 'owner'
};
const MUCRole = {
    Moderator: 'moderator',
    None: 'none',
    Participant: 'participant',
    Visitor: 'visitor'
};
const MUCStatusCode = {
    AffiliationChanged: '101',
    AffiliationLost: '321',
    Banned: '301',
    Error: '333',
    Kicked: '307',
    LoggingDisabled: '171',
    LoggingEnabled: '170',
    MembershipLost: '322',
    NickChanged: '303',
    NickChangedByService: '210',
    NonAnonymous: '172',
    NonAnonymousRoom: '100',
    NonPrivacyConfigurationChange: '104',
    RoomCreated: '201',
    SelfPresence: '110',
    SemiAnonymous: '173',
    Shutdown: '332',
    UnavailableMembersListed: '102',
    UnavailableMembersNotListed: '103'
};
const PubsubErrorCondition = {
    ClosedNode: 'closed-node',
    ConfigurationRequired: 'configuration-required',
    InvalidJID: 'invalid-jid',
    InvalidOptions: 'invalid-options',
    InvalidPayload: 'invalid-payload',
    InvalidSubscriptionId: 'invalid-subid',
    ItemForbidden: 'item-forbidden',
    ItemRequired: 'item-required',
    JIDRequired: 'jid-required',
    MaxItemsExceeded: 'max-items-exceeded',
    MaxNodesExceeded: 'max-nodes-exceeded',
    NodeIdRequired: 'nodeid-required',
    NotInRosterGroup: 'not-in-roster-group',
    NotSubscribed: 'not-subscribed',
    PayloadRequired: 'payload-required',
    PayloadTooBig: 'payload-too-big',
    PendingSubscription: 'pending-subscription',
    PresenceSubscriptionRequired: 'presence-subscription-required',
    SubscriptionIdRequired: 'subid-required',
    TooManySubscriptions: 'too-many-subscriptions',
    Unsupported: 'unsupported',
    UnsupportedAccessModel: 'unsupported-access-model'
};
const ChatState = {
    Active: 'active',
    Composing: 'composing',
    Gone: 'gone',
    Inactive: 'inactive',
    Paused: 'paused'
};
const JingleSessionRole = {
    Initiator: 'initiator',
    Responder: 'responder'
};
const JingleApplicationDirection = {
    Inactive: 'inactive',
    Receive: 'recvonly',
    Send: 'sendonly',
    SendReceive: 'sendrecv'
};
const JingleContentSenders = {
    Both: 'both',
    Initiator: 'initiator',
    None: 'none',
    Responder: 'responder'
};
const JingleAction = {
    ContentAccept: 'content-accept',
    ContentAdd: 'content-add',
    ContentModify: 'content-modify',
    ContentReject: 'content-reject',
    ContentRemove: 'content-remove',
    DescriptionInfo: 'description-info',
    SecurityInfo: 'security-info',
    SessionAccept: 'session-accept',
    SessionInfo: 'session-info',
    SessionInitiate: 'session-initiate',
    SessionTerminate: 'session-terminate',
    TransportAccept: 'transport-accept',
    TransportInfo: 'transport-info',
    TransportReject: 'transport-reject',
    TransportReplace: 'transport-replace'
};
const JingleErrorCondition = {
    OutOfOrder: 'out-of-order',
    TieBreak: 'tie-break',
    UnknownContent: 'unknown-content',
    UnknownSession: 'unknown-session',
    UnsupportedInfo: 'unsupported-info'
};
const JingleReasonCondition = {
    AlternativeSession: 'alternative-session',
    Busy: 'busy',
    Cancel: 'cancel',
    ConnectivityError: 'connectivity-error',
    Decline: 'decline',
    Expired: 'expired',
    FailedApplication: 'failed-application',
    FailedTransport: 'failed-transport',
    GeneralError: 'general-error',
    Gone: 'gone',
    IncompatibleParameters: 'incompatible-parameters',
    MediaError: 'media-error',
    SecurityError: 'security-error',
    Success: 'success',
    Timeout: 'timeout',
    UnsupportedApplications: 'unsupported-applications',
    UnsupportedTransports: 'unsupported-transports'
};
// ====================================================================
// Standalone Constants
// ====================================================================
const USER_MOODS = [
    'afraid',
    'amazed',
    'amorous',
    'angry',
    'annoyed',
    'anxious',
    'aroused',
    'ashamed',
    'bored',
    'brave',
    'calm',
    'cautious',
    'cold',
    'confident',
    'confused',
    'contemplative',
    'contented',
    'cranky',
    'crazy',
    'creative',
    'curious',
    'dejected',
    'depressed',
    'disappointed',
    'disgusted',
    'dismayed',
    'distracted',
    'embarrassed',
    'envious',
    'excited',
    'flirtatious',
    'frustrated',
    'grateful',
    'grieving',
    'grumpy',
    'guilty',
    'happy',
    'hopeful',
    'hot',
    'humbled',
    'humiliated',
    'hungry',
    'hurt',
    'impressed',
    'in_awe',
    'in_love',
    'indignant',
    'interested',
    'intoxicated',
    'invincible',
    'jealous',
    'lonely',
    'lost',
    'lucky',
    'mean',
    'moody',
    'nervous',
    'neutral',
    'offended',
    'outraged',
    'playful',
    'proud',
    'relaxed',
    'relieved',
    'remorseful',
    'restless',
    'sad',
    'sarcastic',
    'satisfied',
    'serious',
    'shocked',
    'shy',
    'sick',
    'sleepy',
    'spontaneous',
    'stressed',
    'strong',
    'surprised',
    'thankful',
    'thirsty',
    'tired',
    'undefined',
    'weak',
    'worried'
];
const USER_ACTIVITY_GENERAL = [
    'doing_chores',
    'drinking',
    'eating',
    'exercising',
    'grooming',
    'having_appointment',
    'inactive',
    'relaxing',
    'talking',
    'traveling',
    'undefined',
    'working'
];
const USER_ACTIVITY_SPECIFIC = [
    'at_the_spa',
    'brushing_teeth',
    'buying_groceries',
    'cleaning',
    'coding',
    'commuting',
    'cooking',
    'cycling',
    'cycling',
    'dancing',
    'day_off',
    'doing_maintenance',
    'doing_the_dishes',
    'doing_the_laundry',
    'driving',
    'fishing',
    'gaming',
    'gardening',
    'getting_a_haircut',
    'going_out',
    'hanging_out',
    'having_a_beer',
    'having_a_snack',
    'having_breakfast',
    'having_coffee',
    'having_dinner',
    'having_lunch',
    'having_tea',
    'hiding',
    'hiking',
    'in_a_car',
    'in_a_meeting',
    'in_real_life',
    'jogging',
    'on_a_bus',
    'on_a_plane',
    'on_a_train',
    'on_a_trip',
    'on_the_phone',
    'on_vacation',
    'on_video_phone',
    'other',
    'partying',
    'playing_sports',
    'praying',
    'reading',
    'rehearsing',
    'running',
    'running_an_errand',
    'scheduled_holiday',
    'shaving',
    'shopping',
    'skiing',
    'sleeping',
    'smoking',
    'socializing',
    'studying',
    'sunbathing',
    'swimming',
    'taking_a_bath',
    'taking_a_shower',
    'thinking',
    'walking',
    'walking_the_dog',
    'watching_a_movie',
    'watching_tv',
    'working_out',
    'writing'
];
const JINGLE_INFO = (namespace, name) => `{${namespace}}${name}`;
const JINGLE_INFO_MUTE = JINGLE_INFO(NS_JINGLE_RTP_INFO_1, 'mute');
const JINGLE_INFO_UNMUTE = JINGLE_INFO(NS_JINGLE_RTP_INFO_1, 'unmute');
const JINGLE_INFO_HOLD = JINGLE_INFO(NS_JINGLE_RTP_INFO_1, 'hold');
const JINGLE_INFO_UNHOLD = JINGLE_INFO(NS_JINGLE_RTP_INFO_1, 'unhold');
const JINGLE_INFO_ACTIVE = JINGLE_INFO(NS_JINGLE_RTP_INFO_1, 'active');
const JINGLE_INFO_RINGING = JINGLE_INFO(NS_JINGLE_RTP_INFO_1, 'ringing');
const JINGLE_INFO_CHECKSUM_5 = JINGLE_INFO(NS_JINGLE_FILE_TRANSFER_5, 'checksum');
const JINGLE_INFO_RECEIVED_5 = JINGLE_INFO(NS_JINGLE_FILE_TRANSFER_5, 'received');
// ====================================================================
// Helper Functions
// ====================================================================
function sendersToDirection(role, senders = JingleContentSenders.Both) {
    const isInitiator = role === JingleSessionRole.Initiator;
    switch (senders) {
        case JingleContentSenders.Initiator:
            return isInitiator
                ? JingleApplicationDirection.Send
                : JingleApplicationDirection.Receive;
        case JingleContentSenders.Responder:
            return isInitiator
                ? JingleApplicationDirection.Receive
                : JingleApplicationDirection.Send;
        case JingleContentSenders.Both:
            return JingleApplicationDirection.SendReceive;
    }
    return JingleApplicationDirection.Inactive;
}
function directionToSenders(role, direction = JingleApplicationDirection.SendReceive) {
    const isInitiator = role === JingleSessionRole.Initiator;
    switch (direction) {
        case JingleApplicationDirection.Send:
            return isInitiator ? JingleContentSenders.Initiator : JingleContentSenders.Responder;
        case JingleApplicationDirection.Receive:
            return isInitiator ? JingleContentSenders.Responder : JingleContentSenders.Initiator;
        case JingleApplicationDirection.SendReceive:
            return JingleContentSenders.Both;
    }
    return JingleContentSenders.None;
}

var Constants = /*#__PURE__*/Object.freeze({
    __proto__: null,
    VERSION: VERSION$1,
    StreamType: StreamType,
    SASLFailureCondition: SASLFailureCondition,
    StreamErrorCondition: StreamErrorCondition,
    StanzaErrorCondition: StanzaErrorCondition,
    MessageType: MessageType,
    PresenceType: PresenceType,
    IQType: IQType,
    PresenceShow: PresenceShow,
    RosterSubscription: RosterSubscription,
    DataFormType: DataFormType,
    DataFormFieldType: DataFormFieldType,
    MUCAffiliation: MUCAffiliation,
    MUCRole: MUCRole,
    MUCStatusCode: MUCStatusCode,
    PubsubErrorCondition: PubsubErrorCondition,
    ChatState: ChatState,
    JingleSessionRole: JingleSessionRole,
    JingleApplicationDirection: JingleApplicationDirection,
    JingleContentSenders: JingleContentSenders,
    JingleAction: JingleAction,
    JingleErrorCondition: JingleErrorCondition,
    JingleReasonCondition: JingleReasonCondition,
    USER_MOODS: USER_MOODS,
    USER_ACTIVITY_GENERAL: USER_ACTIVITY_GENERAL,
    USER_ACTIVITY_SPECIFIC: USER_ACTIVITY_SPECIFIC,
    JINGLE_INFO: JINGLE_INFO,
    JINGLE_INFO_MUTE: JINGLE_INFO_MUTE,
    JINGLE_INFO_UNMUTE: JINGLE_INFO_UNMUTE,
    JINGLE_INFO_HOLD: JINGLE_INFO_HOLD,
    JINGLE_INFO_UNHOLD: JINGLE_INFO_UNHOLD,
    JINGLE_INFO_ACTIVE: JINGLE_INFO_ACTIVE,
    JINGLE_INFO_RINGING: JINGLE_INFO_RINGING,
    JINGLE_INFO_CHECKSUM_5: JINGLE_INFO_CHECKSUM_5,
    JINGLE_INFO_RECEIVED_5: JINGLE_INFO_RECEIVED_5,
    sendersToDirection: sendersToDirection,
    directionToSenders: directionToSenders
});

function Entity (client) {
    client.disco.addFeature('jid\\20escaping');
    client.disco.addFeature(NS_DELAY);
    client.disco.addFeature(NS_EME_0);
    client.disco.addFeature(NS_FORWARD_0);
    client.disco.addFeature(NS_HASHES_2);
    client.disco.addFeature(NS_HASHES_1);
    client.disco.addFeature(NS_IDLE_1);
    client.disco.addFeature(NS_JSON_0);
    client.disco.addFeature(NS_OOB);
    client.disco.addFeature(NS_PSA);
    client.disco.addFeature(NS_REFERENCE_0);
    client.disco.addFeature(NS_SHIM);
    client.disco.addFeature(NS_DATAFORM);
    client.disco.addFeature(NS_DATAFORM_MEDIA);
    client.disco.addFeature(NS_DATAFORM_VALIDATION);
    client.disco.addFeature(NS_DATAFORM_LAYOUT);
    const names = getHashes();
    for (const name of names) {
        client.disco.addFeature(NS_HASH_NAME(name));
    }
    client.disco.addFeature(NS_TIME);
    client.disco.addFeature(NS_VERSION);
    client.on('iq:get:softwareVersion', iq => {
        return client.sendIQResult(iq, {
            softwareVersion: client.config.softwareVersion || {
                name: 'stanzajs.org',
                version: VERSION$1
            }
        });
    });
    client.on('iq:get:time', (iq) => {
        const time = new Date();
        client.sendIQResult(iq, {
            time: {
                tzo: time.getTimezoneOffset(),
                utc: time
            }
        });
    });
    client.getSoftwareVersion = (jid) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            softwareVersion: {},
            to: jid,
            type: 'get'
        });
        return resp.softwareVersion;
    });
    client.getTime = (jid) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            time: {},
            to: jid,
            type: 'get'
        });
        return resp.time;
    });
    client.getLastActivity = (jid) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            lastActivity: {},
            to: jid,
            type: 'get'
        });
        return resp.lastActivity;
    });
}

function Features (client) {
    client.features = {
        handlers: Object.create(null),
        negotiated: Object.create(null),
        order: []
    };
    client.registerFeature = function (name, priority, handler) {
        this.features.order.push({
            name,
            priority
        });
        // We want the features with smallest priority values at the start of the list
        this.features.order.sort((a, b) => a.priority - b.priority);
        this.features.handlers[name] = handler.bind(client);
    };
    client.on('features', (features) => __awaiter(this, void 0, void 0, function* () {
        const negotiated = client.features.negotiated;
        const handlers = client.features.handlers;
        const processingOrder = [];
        for (const { name } of client.features.order) {
            if (features[name] && handlers[name] && !negotiated[name]) {
                processingOrder.push(name);
            }
        }
        function processFeature(featureName) {
            return new Promise(resolve => {
                handlers[featureName](features, (command, message) => {
                    if (command) {
                        resolve({ command, message });
                    }
                    else {
                        resolve(null);
                    }
                });
            });
        }
        for (const item of processingOrder) {
            if (negotiated[item]) {
                continue;
            }
            let cmd = '';
            let msg = '';
            try {
                const res = yield processFeature(item);
                if (res) {
                    cmd = res.command;
                    msg = res.message || '';
                }
            }
            catch (err) {
                cmd = 'disconnect';
                msg = err.message;
                console.error(err);
            }
            if (!cmd) {
                continue;
            }
            if (cmd === 'restart' && client.transport) {
                client.transport.restart();
            }
            if (cmd === 'disconnect') {
                client.emit('stream:error', {
                    condition: 'policy-violation',
                    text: 'Failed to negotiate stream features: ' + msg
                });
                client.disconnect();
            }
            return;
        }
    }));
}

function promiseAny(promises) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const errors = yield Promise.all(promises.map(p => {
                return p.then(val => Promise.reject(val), err => Promise.resolve(err));
            }));
            return Promise.reject(errors);
        }
        catch (val) {
            return Promise.resolve(val);
        }
    });
}
function getHostMeta(registry, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        if (typeof opts === 'string') {
            opts = { host: opts };
        }
        const config = Object.assign({ json: true, ssl: true, xrd: true }, opts);
        const scheme = config.ssl ? 'https://' : 'http://';
        return promiseAny([
            nativeFetch(`${scheme}${config.host}/.well-known/host-meta.json`).then((res) => __awaiter(this, void 0, void 0, function* () {
                if (!res.ok) {
                    throw new Error('could-not-fetch-json');
                }
                return res.json();
            })),
            nativeFetch(`${scheme}${config.host}/.well-known/host-meta`).then((res) => __awaiter(this, void 0, void 0, function* () {
                if (!res.ok) {
                    throw new Error('could-not-fetch-xml');
                }
                const data = yield res.text();
                const xml = parse$1(data);
                if (xml) {
                    return registry.import(xml);
                }
            }))
        ]);
    });
}
function HostMeta (client, stanzas) {
    client.discoverBindings = (server) => __awaiter(this, void 0, void 0, function* () {
        try {
            const data = yield getHostMeta(stanzas, server);
            const results = {
                bosh: [],
                websocket: []
            };
            const links = data.links || [];
            for (const link of links) {
                if (link.href && link.rel === NS_ALT_CONNECTIONS_WEBSOCKET) {
                    results.websocket.push(link.href);
                }
                if (link.href && link.rel === NS_ALT_CONNECTIONS_XBOSH) {
                    results.bosh.push(link.href);
                }
            }
            return results;
        }
        catch (err) {
            return {};
        }
    });
}

// ====================================================================
// Import SDP to Intermediary
// ====================================================================
function importFromSDP(sdp) {
    const mediaSections = SDP.getMediaSections(sdp);
    const sessionPart = SDP.getDescription(sdp);
    const session = {
        groups: [],
        media: []
    };
    if (SDP.matchPrefix(sessionPart, 'a=ice-lite').length > 0) {
        session.iceLite = true;
    }
    for (const groupLine of SDP.matchPrefix(sessionPart, 'a=group:')) {
        const parts = groupLine.split(' ');
        const semantics = parts.shift().substr(8);
        session.groups.push({
            mids: parts,
            semantics
        });
    }
    for (const mediaSection of mediaSections) {
        const kind = SDP.getKind(mediaSection);
        const isRejected = SDP.isRejected(mediaSection);
        const mLine = SDP.parseMLine(mediaSection);
        const media = {
            direction: SDP.getDirection(mediaSection, sessionPart),
            kind,
            mid: SDP.getMid(mediaSection),
            protocol: mLine.protocol
            // TODO: what about end-of-candidates?
        };
        if (!isRejected) {
            media.iceParameters = SDP.getIceParameters(mediaSection, sessionPart);
            media.dtlsParameters = SDP.getDtlsParameters(mediaSection, sessionPart);
            media.setup = SDP.matchPrefix(mediaSection, 'a=setup:')[0].substr(8);
            if (session.iceLite) {
                media.iceParameters.iceLite = true;
            }
        }
        if (kind === 'audio' || kind === 'video') {
            media.rtpParameters = SDP.parseRtpParameters(mediaSection);
            media.rtpEncodingParameters = SDP.parseRtpEncodingParameters(mediaSection);
            media.rtcpParameters = SDP.parseRtcpParameters(mediaSection);
            const msid = SDP.parseMsid(mediaSection);
            media.streams = msid ? [msid] : [];
        }
        else if (kind === 'application') {
            media.sctp = SDP.parseSctpDescription(mediaSection);
        }
        media.candidates = SDP.matchPrefix(mediaSection, 'a=candidate:').map(SDP.parseCandidate);
        session.media.push(media);
    }
    return session;
}
// ====================================================================
// Export Intermediary to SDP
// ====================================================================
function exportToSDP(session) {
    const output = [];
    output.push(SDP.writeSessionBoilerplate(session.sessionId, session.sessionVersion), 'a=msid-semantic:WMS *\r\n');
    if (session.iceLite ||
        session.media.filter(m => m.iceParameters && m.iceParameters.iceLite).length > 0) {
        output.push('a=ice-lite\r\n');
    }
    for (const group of session.groups || []) {
        output.push(`a=group:${group.semantics} ${group.mids.join(' ')}\r\n`);
    }
    for (const media of session.media || []) {
        const isRejected = !(media.iceParameters && media.dtlsParameters);
        if (media.kind === 'application' && media.sctp) {
            output.push(SDP.writeSctpDescription(media, media.sctp));
        }
        else if (media.rtpParameters) {
            let mline = SDP.writeRtpDescription(media.kind, media.rtpParameters);
            if (isRejected) {
                mline = mline.replace(`m=${media.kind} 9 `, `m=${media.kind} 0 `);
            }
            output.push(mline);
            output.push(`a=${media.direction || 'sendrecv'}\r\n`);
            for (const stream of media.streams || []) {
                output.push(`a=msid:${stream.stream} ${stream.track}\r\n`);
            }
            if (media.rtcpParameters) {
                output.push(SDP.writeRtcpParameters(media.rtcpParameters));
                if (media.rtcpParameters.cname) {
                    if (media.rtpEncodingParameters && media.rtpEncodingParameters[0].rtx) {
                        const params = media.rtpEncodingParameters[0];
                        output.push(`a=ssrc-group:FID ${params.ssrc} ${params.rtx.ssrc}\r\n`);
                        output.push(`a=ssrc:${params.rtx.ssrc} cname:${media.rtcpParameters.cname}\r\n`);
                    }
                }
            }
        }
        if (media.mid !== undefined) {
            output.push(`a=mid:${media.mid}\r\n`);
        }
        if (media.iceParameters) {
            output.push(SDP.writeIceParameters({
                // Ignoring iceLite, since we already output ice-lite at session level
                usernameFragment: media.iceParameters.usernameFragment,
                password: media.iceParameters.password
            }));
        }
        if (media.dtlsParameters && media.setup) {
            output.push(SDP.writeDtlsParameters(media.dtlsParameters, media.setup));
        }
        if (media.candidates && media.candidates.length) {
            for (const candidate of media.candidates) {
                output.push(`a=${SDP.writeCandidate(candidate)}\r\n`);
            }
        }
    }
    return output.join('');
}

function convertIntermediateToApplication(media, role) {
    const rtp = media.rtpParameters;
    const rtcp = media.rtcpParameters || {};
    const encodingParameters = media.rtpEncodingParameters || [];
    let hasSSRC = false;
    if (encodingParameters && encodingParameters.length) {
        hasSSRC = !!encodingParameters[0].ssrc; // !== false ???
    }
    const application = {
        applicationType: NS_JINGLE_RTP_1,
        codecs: [],
        headerExtensions: [],
        media: media.kind,
        rtcpMux: rtcp.mux,
        rtcpReducedSize: rtcp.reducedSize,
        sourceGroups: [],
        sources: [],
        ssrc: hasSSRC ? encodingParameters[0].ssrc.toString() : undefined,
        streams: []
    };
    for (const ext of rtp.headerExtensions || []) {
        const header = {
            id: ext.id,
            uri: ext.uri
        };
        if (ext.direction && ext.direction !== 'sendrecv') {
            header.senders = directionToSenders(role, ext.direction);
        }
        application.headerExtensions.push(header);
    }
    if (rtcp.ssrc && rtcp.cname) {
        application.sources = [
            {
                parameters: {
                    cname: rtcp.cname
                },
                ssrc: rtcp.ssrc.toString()
            }
        ];
    }
    if (hasSSRC && encodingParameters[0] && encodingParameters[0].rtx) {
        application.sourceGroups = [
            {
                semantics: 'FID',
                sources: [
                    encodingParameters[0].ssrc.toString(),
                    encodingParameters[0].rtx.ssrc.toString()
                ]
            }
        ];
    }
    for (const stream of media.streams || []) {
        application.streams.push({
            id: stream.stream,
            track: stream.track
        });
    }
    for (const codec of rtp.codecs || []) {
        const payload = {
            channels: codec.channels,
            clockRate: codec.clockRate,
            id: codec.payloadType.toString(),
            name: codec.name,
            parameters: codec.parameters,
            rtcpFeedback: codec.rtcpFeedback
        };
        if (codec.maxptime) {
            payload.maxptime = codec.maxptime;
        }
        if (codec.parameters && codec.parameters.ptime) {
            payload.ptime = parseInt(codec.parameters.ptime, 10);
        }
        application.codecs.push(payload);
    }
    return application;
}
function convertIntermediateToCandidate(candidate) {
    let component;
    if (candidate.component === 'rtp') {
        component = 1;
    }
    else if (candidate.component === 'rtcp') {
        component = 2;
    }
    else {
        component = candidate.component;
    }
    return {
        component,
        foundation: candidate.foundation,
        generation: undefined,
        id: undefined,
        ip: candidate.ip,
        network: undefined,
        port: candidate.port,
        priority: candidate.priority,
        protocol: candidate.protocol,
        relatedAddress: candidate.relatedAddress,
        relatedPort: candidate.relatedPort,
        tcpType: candidate.tcpType,
        type: candidate.type
    };
}
function convertCandidateToIntermediate(candidate) {
    return {
        address: candidate.ip,
        component: candidate.component === 1
            ? 'rtp'
            : candidate.component === 2
                ? 'rtcp'
                : candidate.component,
        foundation: candidate.foundation,
        ip: candidate.ip,
        port: candidate.port,
        priority: candidate.priority,
        protocol: candidate.protocol,
        relatedAddress: candidate.relatedAddress,
        relatedPort: candidate.relatedPort,
        tcpType: candidate.tcpType,
        type: candidate.type
    };
}
function convertIntermediateToTransport(media, transportType) {
    const ice = media.iceParameters;
    const dtls = media.dtlsParameters;
    const transport = {
        candidates: [],
        transportType
    };
    if (ice) {
        transport.usernameFragment = ice.usernameFragment;
        transport.password = ice.password;
        if (ice.iceLite) {
            transport.iceLite = true;
        }
    }
    if (dtls) {
        transport.fingerprints = dtls.fingerprints.map(fingerprint => ({
            algorithm: fingerprint.algorithm,
            setup: media.setup,
            value: fingerprint.value
        }));
    }
    if (media.sctp) {
        transport.sctp = media.sctp;
    }
    for (const candidate of media.candidates || []) {
        transport.candidates.push(convertIntermediateToCandidate(candidate));
    }
    return transport;
}
function convertIntermediateToRequest(session, role, transportType) {
    return {
        contents: session.media.map(media => {
            const isRTP = media.kind === 'audio' || media.kind === 'video';
            return {
                application: isRTP
                    ? convertIntermediateToApplication(media, role)
                    : {
                        applicationType: 'datachannel',
                        protocol: media.protocol
                    },
                creator: JingleSessionRole.Initiator,
                name: media.mid,
                senders: directionToSenders(role, media.direction),
                transport: convertIntermediateToTransport(media, transportType)
            };
        }),
        groups: session.groups
            ? session.groups.map(group => ({
                contents: group.mids,
                semantics: group.semantics
            }))
            : []
    };
}
function convertContentToIntermediate(content, role) {
    const application = content.application || {};
    const transport = content.transport;
    const isRTP = application && application.applicationType === NS_JINGLE_RTP_1;
    const media = {
        direction: sendersToDirection(role, content.senders),
        kind: application.media || 'application',
        mid: content.name,
        protocol: isRTP ? 'UDP/TLS/RTP/SAVPF' : 'UDP/DTLS/SCTP'
    };
    if (isRTP) {
        media.rtcpParameters = {
            compound: !application.rtcpReducedSize,
            mux: application.rtcpMux,
            reducedSize: application.rtcpReducedSize
        };
        if (application.sources && application.sources.length) {
            const source = application.sources[0];
            media.rtcpParameters.ssrc = parseInt(source.ssrc, 10);
            if (source.parameters) {
                media.rtcpParameters.cname = source.parameters.cname;
            }
        }
        media.rtpParameters = {
            codecs: [],
            fecMechanisms: [],
            headerExtensions: [],
            rtcp: []
        };
        if (application.streams) {
            media.streams = [];
            for (const stream of application.streams) {
                media.streams.push({
                    stream: stream.id,
                    track: stream.track
                });
            }
        }
        if (application.ssrc) {
            media.rtpEncodingParameters = [
                {
                    ssrc: parseInt(application.ssrc, 10)
                }
            ];
            if (application.sourceGroups && application.sourceGroups.length) {
                const group = application.sourceGroups[0];
                media.rtpEncodingParameters[0].rtx = {
                    // TODO: actually look for a FID one with matching ssrc
                    ssrc: parseInt(group.sources[1], 10)
                };
            }
        }
        let hasRED = false;
        let hasULPFEC = false;
        for (const payload of application.codecs || []) {
            const parameters = payload.parameters || {};
            const rtcpFeedback = [];
            for (const fb of payload.rtcpFeedback || []) {
                rtcpFeedback.push({
                    parameter: fb.parameter,
                    type: fb.type
                });
            }
            if (payload.name === 'red' || payload.name === 'ulpfec') {
                hasRED = hasRED || payload.name === 'red';
                hasULPFEC = hasULPFEC || payload.name === 'ulpfec';
                const fec = payload.name.toUpperCase();
                if (!media.rtpParameters.fecMechanisms.includes(fec)) {
                    media.rtpParameters.fecMechanisms.push(fec);
                }
            }
            media.rtpParameters.codecs.push({
                channels: payload.channels,
                clockRate: payload.clockRate,
                name: payload.name,
                numChannels: payload.channels,
                parameters,
                payloadType: parseInt(payload.id, 10),
                rtcpFeedback
            });
        }
        for (const ext of application.headerExtensions || []) {
            media.rtpParameters.headerExtensions.push({
                direction: sendersToDirection(role, ext.senders || 'both'),
                id: ext.id,
                uri: ext.uri
            });
        }
    }
    if (transport) {
        if (transport.usernameFragment && transport.password) {
            media.iceParameters = {
                password: transport.password,
                usernameFragment: transport.usernameFragment
            };
            if (transport.iceLite) {
                media.iceParameters.iceLite = true;
            }
        }
        if (transport.fingerprints && transport.fingerprints.length) {
            media.dtlsParameters = {
                fingerprints: [],
                role: 'auto'
            };
            for (const fingerprint of transport.fingerprints) {
                media.dtlsParameters.fingerprints.push({
                    algorithm: fingerprint.algorithm,
                    value: fingerprint.value
                });
            }
            if (transport.sctp) {
                media.sctp = transport.sctp;
            }
            media.setup = transport.fingerprints[0].setup;
        }
        media.candidates = (transport.candidates || []).map(convertCandidateToIntermediate);
    }
    return media;
}
function convertRequestToIntermediate(jingle, role) {
    const session = {
        groups: [],
        media: []
    };
    for (const group of jingle.groups || []) {
        session.groups.push({
            mids: group.contents,
            semantics: group.semantics
        });
    }
    for (const content of jingle.contents || []) {
        session.media.push(convertContentToIntermediate(content, role));
    }
    return session;
}

const badRequest = { condition: StanzaErrorCondition.BadRequest };
const unsupportedInfo = {
    condition: StanzaErrorCondition.FeatureNotImplemented,
    jingleError: JingleErrorCondition.UnsupportedInfo,
    type: 'modify'
};
class JingleSession {
    constructor(opts) {
        this.parent = opts.parent;
        this.sid = opts.sid || uuid();
        this.peerID = opts.peerID;
        this.role = opts.initiator ? JingleSessionRole.Initiator : JingleSessionRole.Responder;
        this._sessionState = 'starting';
        this._connectionState = 'starting';
        // We track the intial pending description types in case
        // of the need for a tie-breaker.
        this.pendingApplicationTypes = opts.applicationTypes || [];
        this.pendingAction = undefined;
        // Here is where we'll ensure that all actions are processed
        // in order, even if a particular action requires async handling.
        this.processingQueue = priorityQueue((task, next) => __awaiter(this, void 0, void 0, function* () {
            if (this.state === 'ended') {
                // Don't process anything once the session has been ended
                if (task.type === 'local' && task.reject) {
                    task.reject(new Error('Session ended'));
                }
                if (next) {
                    next();
                }
                return;
            }
            if (task.type === 'local') {
                this._log('debug', 'Processing local action:', task.name);
                try {
                    const res = yield task.handler();
                    task.resolve(res);
                }
                catch (err) {
                    task.reject(err);
                }
                if (next) {
                    next();
                }
                return;
            }
            const { action, changes, cb } = task;
            this._log('debug', 'Processing remote action:', action);
            return new Promise(resolve => {
                const done = (err, result) => {
                    cb(err, result);
                    if (next) {
                        next();
                    }
                    resolve();
                };
                switch (action) {
                    case JingleAction.ContentAccept:
                        return this.onContentAccept(changes, done);
                    case JingleAction.ContentAdd:
                        return this.onContentAdd(changes, done);
                    case JingleAction.ContentModify:
                        return this.onContentModify(changes, done);
                    case JingleAction.ContentReject:
                        return this.onContentReject(changes, done);
                    case JingleAction.ContentRemove:
                        return this.onContentRemove(changes, done);
                    case JingleAction.DescriptionInfo:
                        return this.onDescriptionInfo(changes, done);
                    case JingleAction.SecurityInfo:
                        return this.onSecurityInfo(changes, done);
                    case JingleAction.SessionAccept:
                        return this.onSessionAccept(changes, done);
                    case JingleAction.SessionInfo:
                        return this.onSessionInfo(changes, done);
                    case JingleAction.SessionInitiate:
                        return this.onSessionInitiate(changes, done);
                    case JingleAction.SessionTerminate:
                        return this.onSessionTerminate(changes, done);
                    case JingleAction.TransportAccept:
                        return this.onTransportAccept(changes, done);
                    case JingleAction.TransportInfo:
                        return this.onTransportInfo(changes, done);
                    case JingleAction.TransportReject:
                        return this.onTransportReject(changes, done);
                    case JingleAction.TransportReplace:
                        return this.onTransportReplace(changes, done);
                    default:
                        this._log('error', 'Invalid or unsupported action: ' + action);
                        done({ condition: StanzaErrorCondition.BadRequest });
                }
            });
        }), 1);
    }
    get isInitiator() {
        return this.role === JingleSessionRole.Initiator;
    }
    get peerRole() {
        return this.isInitiator ? JingleSessionRole.Responder : JingleSessionRole.Initiator;
    }
    get state() {
        return this._sessionState;
    }
    set state(value) {
        if (value !== this._sessionState) {
            this._log('info', 'Changing session state to: ' + value);
            this._sessionState = value;
            if (this.parent) {
                this.parent.emit('sessionState', this, value);
            }
        }
    }
    get connectionState() {
        return this._connectionState;
    }
    set connectionState(value) {
        if (value !== this._connectionState) {
            this._log('info', 'Changing connection state to: ' + value);
            this._connectionState = value;
            if (this.parent) {
                this.parent.emit('connectionState', this, value);
            }
        }
    }
    send(action, data) {
        data = data || {};
        data.sid = this.sid;
        data.action = action;
        const requirePending = new Set([
            JingleAction.ContentAccept,
            JingleAction.ContentAdd,
            JingleAction.ContentModify,
            JingleAction.ContentReject,
            JingleAction.ContentRemove,
            JingleAction.SessionAccept,
            JingleAction.SessionInitiate,
            JingleAction.TransportAccept,
            JingleAction.TransportReject,
            JingleAction.TransportReplace
        ]);
        if (requirePending.has(action)) {
            this.pendingAction = action;
        }
        else {
            this.pendingAction = undefined;
        }
        this.parent.signal(this, {
            id: uuid(),
            jingle: data,
            to: this.peerID,
            type: 'set'
        });
    }
    processLocal(name, handler) {
        return new Promise((resolve, reject) => {
            this.processingQueue.push({
                handler,
                name,
                reject,
                resolve,
                type: 'local'
            }, 1 // Process local requests first
            );
        });
    }
    process(action, changes, cb) {
        this.processingQueue.push({
            action,
            cb,
            changes,
            type: 'remote'
        }, 2 // Process remote requests second
        );
    }
    start(_opts, _next) {
        this._log('error', 'Can not start base sessions');
        this.end('unsupported-applications', true);
    }
    accept(_opts, _next) {
        this._log('error', 'Can not accept base sessions');
        this.end('unsupported-applications');
    }
    cancel() {
        this.end('cancel');
    }
    decline() {
        this.end('decline');
    }
    end(reason = 'success', silent = false) {
        this.state = 'ended';
        this.processingQueue.kill();
        if (typeof reason === 'string') {
            reason = {
                condition: reason
            };
        }
        if (!silent) {
            this.send('session-terminate', {
                reason
            });
        }
        this.parent.emit('terminated', this, reason);
        this.parent.forgetSession(this);
    }
    _log(level, message, ...data) {
        if (this.parent) {
            message = this.sid + ': ' + message;
            this.parent.emit('log', level, message, ...data);
            this.parent.emit('log:' + level, message, ...data);
        }
    }
    onSessionInitiate(changes, cb) {
        cb();
    }
    onSessionAccept(changes, cb) {
        cb();
    }
    onSessionTerminate(changes, cb) {
        this.end(changes.reason, true);
        cb();
    }
    // It is mandatory to reply to a session-info action with
    // an unsupported-info error if the info isn't recognized.
    //
    // However, a session-info action with no associated payload
    // is acceptable (works like a ping).
    onSessionInfo(changes, cb) {
        if (!changes.info) {
            cb();
        }
        else {
            cb(unsupportedInfo);
        }
    }
    // It is mandatory to reply to a security-info action with
    // an unsupported-info error if the info isn't recognized.
    onSecurityInfo(changes, cb) {
        cb(unsupportedInfo);
    }
    // It is mandatory to reply to a description-info action with
    // an unsupported-info error if the info isn't recognized.
    onDescriptionInfo(changes, cb) {
        cb(unsupportedInfo);
    }
    // It is mandatory to reply to a transport-info action with
    // an unsupported-info error if the info isn't recognized.
    onTransportInfo(changes, cb) {
        cb(unsupportedInfo);
    }
    // It is mandatory to reply to a content-add action with either
    // a content-accept or content-reject.
    onContentAdd(changes, cb) {
        // Allow ack for the content-add to be sent.
        cb();
        this.send(JingleAction.ContentReject, {
            reason: {
                condition: JingleReasonCondition.FailedApplication,
                text: 'content-add is not supported'
            }
        });
    }
    onContentAccept(changes, cb) {
        cb(badRequest);
    }
    onContentReject(changes, cb) {
        cb(badRequest);
    }
    onContentModify(changes, cb) {
        cb(badRequest);
    }
    onContentRemove(changes, cb) {
        cb(badRequest);
    }
    // It is mandatory to reply to a transport-add action with either
    // a transport-accept or transport-reject.
    onTransportReplace(changes, cb) {
        // Allow ack for the transport-replace be sent.
        cb();
        this.send(JingleAction.TransportReject, {
            reason: {
                condition: JingleReasonCondition.FailedTransport,
                text: 'transport-replace is not supported'
            }
        });
    }
    onTransportAccept(changes, cb) {
        cb(badRequest);
    }
    onTransportReject(changes, cb) {
        cb(badRequest);
    }
}

class ICESession extends JingleSession {
    constructor(opts) {
        super(opts);
        this.bitrateLimit = 0;
        this.candidateBuffer = [];
        this.transportType = NS_JINGLE_ICE_UDP_1;
        this.restartingIce = false;
        this.usingRelay = false;
        this.maxRelayBandwidth = opts.maxRelayBandwidth;
        this.pc = this.parent.createPeerConnection(this, Object.assign(Object.assign({}, opts.config), { iceServers: opts.iceServers }));
        this.pc.oniceconnectionstatechange = () => {
            this.onIceStateChange();
        };
        this.pc.onicecandidate = e => {
            if (e.candidate) {
                this.onIceCandidate(e);
            }
            else {
                this.onIceEndOfCandidates();
            }
        };
        this.restrictRelayBandwidth();
    }
    end(reason = 'success', silent = false) {
        this.pc.close();
        super.end(reason, silent);
    }
    /* actually do an ice restart */
    restartIce() {
        return __awaiter(this, void 0, void 0, function* () {
            // only initiators do an ice-restart to avoid conflicts.
            if (!this.isInitiator) {
                return;
            }
            if (this._maybeRestartingIce !== undefined) {
                clearTimeout(this._maybeRestartingIce);
            }
            this.restartingIce = true;
            try {
                yield this.processLocal('restart-ice', () => __awaiter(this, void 0, void 0, function* () {
                    const offer = yield this.pc.createOffer({ iceRestart: true });
                    // extract new ufrag / pwd, send transport-info with just that.
                    const json = importFromSDP(offer.sdp);
                    this.send(JingleAction.TransportInfo, {
                        contents: json.media.map(media => ({
                            creator: JingleSessionRole.Initiator,
                            name: media.mid,
                            transport: convertIntermediateToTransport(media, this.transportType)
                        })),
                        sid: this.sid
                    });
                    yield this.pc.setLocalDescription(offer);
                }));
            }
            catch (err) {
                this._log('error', 'Could not create WebRTC offer', err);
                this.end(JingleReasonCondition.FailedTransport, true);
            }
        });
    }
    // set the maximum bitrate. Only supported in Chrome and Firefox right now.
    setMaximumBitrate(maximumBitrate) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.maximumBitrate) {
                // potentially take into account bandwidth restrictions due to using TURN.
                maximumBitrate = Math.min(maximumBitrate, this.maximumBitrate);
            }
            this.currentBitrate = maximumBitrate;
            // changes the maximum bandwidth using RTCRtpSender.setParameters.
            const sender = this.pc.getSenders().find(s => !!s.track && s.track.kind === 'video');
            if (!sender || !sender.getParameters) {
                return;
            }
            try {
                yield this.processLocal('set-bitrate', () => __awaiter(this, void 0, void 0, function* () {
                    const parameters = sender.getParameters();
                    if (!parameters.encodings || !parameters.encodings.length) {
                        parameters.encodings = [{}];
                    }
                    if (maximumBitrate === 0) {
                        delete parameters.encodings[0].maxBitrate;
                    }
                    else {
                        parameters.encodings[0].maxBitrate = maximumBitrate;
                    }
                    yield sender.setParameters(parameters);
                }));
            }
            catch (err) {
                this._log('error', 'Set maximumBitrate failed', err);
            }
        });
    }
    // ----------------------------------------------------------------
    // Jingle action handers
    // ----------------------------------------------------------------
    onTransportInfo(changes, cb) {
        return __awaiter(this, void 0, void 0, function* () {
            if (changes.contents &&
                changes.contents[0] &&
                changes.contents[0].transport.gatheringComplete) {
                const candidate = { sdpMid: changes.contents[0].name, candidate: '' };
                try {
                    if (this.pc.signalingState === 'stable') {
                        yield this.pc.addIceCandidate(candidate);
                    }
                    else {
                        this.candidateBuffer.push(candidate);
                    }
                }
                catch (err) {
                    this._log('debug', 'Could not add null end-of-candidate');
                }
                finally {
                    cb();
                }
                return;
            }
            // detect an ice restart.
            if (this.pc.remoteDescription) {
                const remoteDescription = this.pc.remoteDescription;
                const remoteJSON = importFromSDP(remoteDescription.sdp);
                const remoteMedia = remoteJSON.media.find(m => m.mid === changes.contents[0].name);
                const currentUsernameFragment = remoteMedia.iceParameters.usernameFragment;
                const remoteUsernameFragment = changes.contents[0].transport
                    .usernameFragment;
                if (remoteUsernameFragment && currentUsernameFragment !== remoteUsernameFragment) {
                    for (const [idx, content] of changes.contents.entries()) {
                        const transport = content.transport;
                        remoteJSON.media[idx].iceParameters = {
                            password: transport.password,
                            usernameFragment: transport.usernameFragment
                        };
                        remoteJSON.media[idx].candidates = [];
                    }
                    try {
                        yield this.pc.setRemoteDescription({
                            type: remoteDescription.type,
                            sdp: exportToSDP(remoteJSON)
                        });
                        yield this.processBufferedCandidates();
                        if (remoteDescription.type === 'offer') {
                            const answer = yield this.pc.createAnswer();
                            yield this.pc.setLocalDescription(answer);
                            const json = importFromSDP(answer.sdp);
                            this.send(JingleAction.TransportInfo, {
                                contents: json.media.map(media => ({
                                    creator: JingleSessionRole.Initiator,
                                    name: media.mid,
                                    transport: convertIntermediateToTransport(media, this.transportType)
                                })),
                                sid: this.sid
                            });
                        }
                        else {
                            this.restartingIce = false;
                        }
                    }
                    catch (err) {
                        this._log('error', 'Could not do remote ICE restart', err);
                        cb(err);
                        this.end(JingleReasonCondition.FailedTransport);
                        return;
                    }
                }
            }
            const all = (changes.contents || []).map(content => {
                const sdpMid = content.name;
                const results = (content.transport.candidates || []).map((json) => __awaiter(this, void 0, void 0, function* () {
                    const candidate = SDP.writeCandidate(convertCandidateToIntermediate(json));
                    if (this.pc.remoteDescription && this.pc.signalingState === 'stable') {
                        try {
                            yield this.pc.addIceCandidate({ sdpMid, candidate });
                        }
                        catch (err) {
                            this._log('error', 'Could not add ICE candidate', err);
                        }
                    }
                    else {
                        this.candidateBuffer.push({ sdpMid, candidate });
                    }
                }));
                return Promise.all(results);
            });
            try {
                yield Promise.all(all);
                cb();
            }
            catch (err) {
                this._log('error', `Could not process transport-info: ${err}`);
                cb(err);
            }
        });
    }
    onSessionAccept(changes, cb) {
        return __awaiter(this, void 0, void 0, function* () {
            this.state = 'active';
            const json = convertRequestToIntermediate(changes, this.peerRole);
            const sdp = exportToSDP(json);
            try {
                yield this.pc.setRemoteDescription({ type: 'answer', sdp });
                yield this.processBufferedCandidates();
                this.parent.emit('accepted', this, undefined);
                cb();
            }
            catch (err) {
                this._log('error', `Could not process WebRTC answer: ${err}`);
                cb({ condition: 'general-error' });
            }
        });
    }
    onSessionTerminate(changes, cb) {
        this._log('info', 'Terminating session');
        this.pc.close();
        super.end(changes.reason, true);
        cb();
    }
    // ----------------------------------------------------------------
    // ICE action handers
    // ----------------------------------------------------------------
    onIceCandidate(e) {
        if (!e.candidate || !e.candidate.candidate) {
            return;
        }
        const candidate = SDP.parseCandidate(e.candidate.candidate);
        const jingle = {
            contents: [
                {
                    creator: JingleSessionRole.Initiator,
                    name: e.candidate.sdpMid,
                    transport: {
                        candidates: [convertIntermediateToCandidate(candidate)],
                        transportType: this.transportType,
                        usernameFragment: candidate.usernameFragment
                    }
                }
            ]
        };
        this._log('info', 'Discovered new ICE candidate', jingle);
        this.send(JingleAction.TransportInfo, jingle);
    }
    onIceEndOfCandidates() {
        this._log('info', 'ICE end of candidates');
        const json = importFromSDP(this.pc.localDescription.sdp);
        const firstMedia = json.media[0];
        // signal end-of-candidates with our first media mid/ufrag
        this.send(JingleAction.TransportInfo, {
            contents: [
                {
                    creator: JingleSessionRole.Initiator,
                    name: firstMedia.mid,
                    transport: {
                        gatheringComplete: true,
                        transportType: this.transportType,
                        usernameFragment: firstMedia.iceParameters.usernameFragment
                    }
                }
            ]
        });
    }
    onIceStateChange() {
        switch (this.pc.iceConnectionState) {
            case 'checking':
                this.connectionState = 'connecting';
                break;
            case 'completed':
            case 'connected':
                this.connectionState = 'connected';
                break;
            case 'disconnected':
                if (this.pc.signalingState === 'stable') {
                    this.connectionState = 'interrupted';
                }
                else {
                    this.connectionState = 'disconnected';
                }
                if (this.restartingIce) {
                    this.end(JingleReasonCondition.FailedTransport);
                    return;
                }
                this.maybeRestartIce();
                break;
            case 'failed':
                if (this.connectionState === 'failed' || this.restartingIce) {
                    this.end(JingleReasonCondition.FailedTransport);
                    return;
                }
                this.connectionState = 'failed';
                this.restartIce();
                break;
            case 'closed':
                this.connectionState = 'disconnected';
                if (this.restartingIce) {
                    this.end(JingleReasonCondition.FailedTransport);
                }
                else {
                    this.end();
                }
                break;
        }
    }
    processBufferedCandidates() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const candidate of this.candidateBuffer) {
                try {
                    yield this.pc.addIceCandidate(candidate);
                }
                catch (err) {
                    this._log('error', 'Could not add ICE candidate', err);
                }
            }
            this.candidateBuffer = [];
        });
    }
    /* when using TURN, we might want to restrict the bandwidth
     * to the value specified by MAX_RELAY_BANDWIDTH
     * in order to prevent sending excessive traffic through
     * the TURN server.
     */
    restrictRelayBandwidth() {
        this.pc.addEventListener('iceconnectionstatechange', () => __awaiter(this, void 0, void 0, function* () {
            if (this.pc.iceConnectionState !== 'completed' &&
                this.pc.iceConnectionState !== 'connected') {
                return;
            }
            const stats = yield this.pc.getStats();
            let activeCandidatePair;
            stats.forEach(report => {
                if (report.type === 'transport') {
                    activeCandidatePair = stats.get(report.selectedCandidatePairId);
                }
            });
            // Fallback for Firefox.
            if (!activeCandidatePair) {
                stats.forEach(report => {
                    if (report.type === 'candidate-pair' && report.selected) {
                        activeCandidatePair = report;
                    }
                });
            }
            if (!activeCandidatePair) {
                return;
            }
            let isRelay = false;
            let localCandidateType = '';
            let remoteCandidateType = '';
            if (activeCandidatePair.remoteCandidateId) {
                const remoteCandidate = stats.get(activeCandidatePair.remoteCandidateId);
                if (remoteCandidate) {
                    remoteCandidateType = remoteCandidate.candidateType;
                }
            }
            if (activeCandidatePair.localCandidateId) {
                const localCandidate = stats.get(activeCandidatePair.localCandidateId);
                if (localCandidate) {
                    localCandidateType = localCandidate.candidateType;
                }
            }
            if (localCandidateType === 'relay' || remoteCandidateType === 'relay') {
                isRelay = true;
            }
            this.usingRelay = isRelay;
            this.parent.emit('iceConnectionType', this, {
                localCandidateType,
                relayed: isRelay,
                remoteCandidateType
            });
            if (isRelay && this.maxRelayBandwidth !== undefined) {
                this.maximumBitrate = this.maxRelayBandwidth;
                if (this.currentBitrate) {
                    this.setMaximumBitrate(Math.min(this.currentBitrate, this.maximumBitrate));
                }
                else {
                    this.setMaximumBitrate(this.maximumBitrate);
                }
            }
        }));
    }
    /* determine whether an ICE restart is in order
     * when transitioning to disconnected. Strategy is
     * 'wait 2 seconds for things to repair themselves'
     * 'maybe check if bytes are sent/received' by comparing
     *   getStats measurements
     */
    maybeRestartIce() {
        // only initiators do an ice-restart to avoid conflicts.
        if (!this.isInitiator) {
            return;
        }
        if (this._maybeRestartingIce !== undefined) {
            clearTimeout(this._maybeRestartingIce);
        }
        this._maybeRestartingIce = setTimeout(() => {
            this._maybeRestartingIce = undefined;
            if (this.pc.iceConnectionState === 'disconnected') {
                this.restartIce();
            }
        }, 2000);
    }
}

class Sender extends EventEmitter {
    constructor(opts = {}) {
        super();
        this.config = Object.assign({ chunkSize: 16384, hash: 'sha-1' }, opts);
        this.file = undefined;
        this.channel = undefined;
        this.hash = createHash(this.config.hash);
    }
    send(file, channel) {
        if (this.file && this.channel) {
            return;
        }
        this.file = file;
        this.channel = channel;
        this.channel.binaryType = 'arraybuffer';
        const fileReader = new FileReader();
        let offset = 0;
        let pendingRead = false;
        const sliceFile = () => {
            if (pendingRead || offset >= file.size) {
                return;
            }
            pendingRead = true;
            const slice = file.slice(offset, offset + this.config.chunkSize);
            fileReader.readAsArrayBuffer(slice);
        };
        channel.bufferedAmountLowThreshold = 8 * this.config.chunkSize;
        channel.onbufferedamountlow = () => {
            sliceFile();
        };
        fileReader.addEventListener('load', (event) => {
            const data = event.target.result;
            pendingRead = false;
            offset += data.byteLength;
            this.channel.send(data);
            this.hash.update(new Uint8Array(data));
            this.emit('progress', offset, file.size, data);
            if (offset < file.size) {
                if (this.channel.bufferedAmount <= this.channel.bufferedAmountLowThreshold) {
                    sliceFile();
                }
                // Otherwise wait for bufferedamountlow event to trigger reading more data
            }
            else {
                this.emit('progress', file.size, file.size, null);
                this.emit('sentFile', {
                    algorithm: this.config.hash,
                    name: file.name,
                    size: file.size,
                    value: this.hash.digest()
                });
            }
        });
        sliceFile();
    }
}
class Receiver extends EventEmitter {
    constructor(opts = {}) {
        super();
        this.config = Object.assign({ hash: 'sha-1' }, opts);
        this.receiveBuffer = [];
        this.received = 0;
        this.channel = undefined;
        this.hash = createHash(this.config.hash);
    }
    receive(metadata, channel) {
        this.metadata = metadata;
        this.channel = channel;
        this.channel.binaryType = 'arraybuffer';
        this.channel.onmessage = e => {
            const len = e.data.byteLength;
            this.received += len;
            this.receiveBuffer.push(e.data);
            if (e.data) {
                this.hash.update(new Uint8Array(e.data));
            }
            this.emit('progress', this.received, this.metadata.size, e.data);
            if (this.received === this.metadata.size) {
                this.metadata.actualhash = this.hash.digest('hex');
                this.emit('receivedFile', new Blob(this.receiveBuffer), this.metadata);
                this.receiveBuffer = [];
            }
            else if (this.received > this.metadata.size) {
                // FIXME
                console.error('received more than expected, discarding...');
                this.receiveBuffer = []; // just discard...
            }
        };
    }
}
class FileTransferSession extends ICESession {
    constructor(opts) {
        super(opts);
        this.sender = undefined;
        this.receiver = undefined;
        this.file = undefined;
    }
    start(file, next) {
        return __awaiter(this, void 0, void 0, function* () {
            next = next || (() => undefined);
            if (!file || typeof file === 'function') {
                throw new Error('File object required');
            }
            this.state = 'pending';
            this.role = 'initiator';
            this.file = file;
            this.sender = new Sender();
            this.sender.on('progress', (sent, size) => {
                this._log('info', 'Send progress ' + sent + '/' + size);
            });
            this.sender.on('sentFile', meta => {
                this._log('info', 'Sent file', meta.name);
                this.send(JingleAction.SessionInfo, {
                    info: {
                        creator: JingleSessionRole.Initiator,
                        file: {
                            hashes: [
                                {
                                    algorithm: meta.algorithm,
                                    value: meta.value
                                }
                            ]
                        },
                        infoType: JINGLE_INFO_CHECKSUM_5,
                        name: this.contentName
                    }
                });
                this.parent.emit('sentFile', this, meta);
            });
            this.channel = this.pc.createDataChannel('filetransfer', {
                ordered: true
            });
            this.channel.onopen = () => {
                this.sender.send(this.file, this.channel);
            };
            try {
                yield this.processLocal(JingleAction.SessionInitiate, () => __awaiter(this, void 0, void 0, function* () {
                    const offer = yield this.pc.createOffer({
                        offerToReceiveAudio: false,
                        offerToReceiveVideo: false
                    });
                    const json = importFromSDP(offer.sdp);
                    const jingle = convertIntermediateToRequest(json, this.role, this.transportType);
                    this.contentName = jingle.contents[0].name;
                    jingle.sid = this.sid;
                    jingle.action = JingleAction.SessionInitiate;
                    jingle.contents[0].application = {
                        applicationType: NS_JINGLE_FILE_TRANSFER_5,
                        file: {
                            date: file.lastModified ? new Date(file.lastModified) : undefined,
                            hashesUsed: [
                                {
                                    algorithm: 'sha-1'
                                }
                            ],
                            name: file.name,
                            size: file.size
                        }
                    };
                    this.send('session-initiate', jingle);
                    yield this.pc.setLocalDescription(offer);
                }));
                next();
            }
            catch (err) {
                this._log('error', 'Could not create WebRTC offer', err);
                return this.end('failed-application', true);
            }
        });
    }
    accept(next) {
        return __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Accepted incoming session');
            this.role = 'responder';
            this.state = 'active';
            next = next || (() => undefined);
            try {
                yield this.processLocal(JingleAction.SessionAccept, () => __awaiter(this, void 0, void 0, function* () {
                    const answer = yield this.pc.createAnswer();
                    const json = importFromSDP(answer.sdp);
                    const jingle = convertIntermediateToRequest(json, this.role, this.transportType);
                    jingle.sid = this.sid;
                    jingle.action = 'session-accept';
                    for (const content of jingle.contents) {
                        content.creator = 'initiator';
                    }
                    this.contentName = jingle.contents[0].name;
                    this.send('session-accept', jingle);
                    yield this.pc.setLocalDescription(answer);
                    yield this.processBufferedCandidates();
                }));
                next();
            }
            catch (err) {
                this._log('error', 'Could not create WebRTC answer', err);
                this.end('failed-application');
            }
        });
    }
    onSessionInitiate(changes, cb) {
        return __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Initiating incoming session');
            this.role = 'responder';
            this.state = 'pending';
            this.transportType = changes.contents[0].transport.transportType;
            const json = convertRequestToIntermediate(changes, this.peerRole);
            const sdp = exportToSDP(json);
            const desc = changes.contents[0].application;
            const hashes = desc.file.hashesUsed ? desc.file.hashesUsed : desc.file.hashes || [];
            this.receiver = new Receiver({ hash: hashes[0] && hashes[0].algorithm });
            this.receiver.on('progress', (received, size) => {
                this._log('info', 'Receive progress ' + received + '/' + size);
            });
            this.receiver.on('receivedFile', file => {
                this.receivedFile = file;
                this._maybeReceivedFile();
            });
            this.receiver.metadata = desc.file;
            this.pc.addEventListener('datachannel', e => {
                this.channel = e.channel;
                this.receiver.receive(this.receiver.metadata, e.channel);
            });
            try {
                yield this.pc.setRemoteDescription({ type: 'offer', sdp });
                yield this.processBufferedCandidates();
                cb();
            }
            catch (err) {
                this._log('error', 'Could not create WebRTC answer', err);
                cb({ condition: 'general-error' });
            }
        });
    }
    onSessionInfo(changes, cb) {
        const info = changes.info;
        if (!info || !info.file || !info.file.hashes) {
            return;
        }
        this.receiver.metadata.hashes = info.file.hashes;
        if (this.receiver.metadata.actualhash) {
            this._maybeReceivedFile();
        }
        cb();
    }
    _maybeReceivedFile() {
        if (!this.receiver.metadata.hashes || !this.receiver.metadata.hashes.length) {
            // unknown hash, file transfer not completed
            return;
        }
        for (const hash of this.receiver.metadata.hashes || []) {
            if (hash.value && hash.value.toString('hex') === this.receiver.metadata.actualhash) {
                this._log('info', 'File hash matches');
                this.parent.emit('receivedFile', this, this.receivedFile, this.receiver.metadata);
                this.end('success');
                return;
            }
        }
        this._log('error', 'File hash does not match');
        this.end('media-error');
    }
}

function applyStreamsCompatibility(content) {
    const application = content.application;
    /* signal .streams as a=ssrc: msid */
    if (application.streams &&
        application.streams.length &&
        application.sources &&
        application.sources.length) {
        const msid = application.streams[0];
        application.sources[0].parameters.msid = `${msid.id} ${msid.track}`;
        if (application.sourceGroups && application.sourceGroups.length > 0) {
            application.sources.push({
                parameters: {
                    cname: application.sources[0].parameters.cname,
                    msid: `${msid.id} ${msid.track}`
                },
                ssrc: application.sourceGroups[0].sources[1]
            });
        }
    }
}
class MediaSession extends ICESession {
    constructor(opts) {
        super(opts);
        this.includesAudio = false;
        this.includesVideo = false;
        this._ringing = false;
        this.pc.addEventListener('track', (e) => {
            this.onAddTrack(e.track, e.streams[0]);
        });
        if (opts.stream) {
            for (const track of opts.stream.getTracks()) {
                this.addTrack(track, opts.stream);
            }
        }
    }
    get ringing() {
        return this._ringing;
    }
    set ringing(value) {
        if (value !== this._ringing) {
            this._ringing = value;
        }
    }
    get streams() {
        if (this.pc.signalingState !== 'closed') {
            return this.pc.getRemoteStreams();
        }
        return [];
    }
    // ----------------------------------------------------------------
    // Session control methods
    // ----------------------------------------------------------------
    start(opts, next) {
        return __awaiter(this, arguments, void 0, function* () {
            this.state = 'pending';
            if (arguments.length === 1 && typeof opts === 'function') {
                next = opts;
                opts = {};
            }
            next = next || (() => undefined);
            opts = opts || {};
            this.role = 'initiator';
            this.offerOptions = opts;
            try {
                yield this.processLocal(JingleAction.SessionInitiate, () => __awaiter(this, void 0, void 0, function* () {
                    const offer = yield this.pc.createOffer(opts);
                    const json = importFromSDP(offer.sdp);
                    const jingle = convertIntermediateToRequest(json, this.role, this.transportType);
                    jingle.sid = this.sid;
                    jingle.action = JingleAction.SessionInitiate;
                    for (const content of jingle.contents || []) {
                        content.creator = 'initiator';
                        applyStreamsCompatibility(content);
                    }
                    yield this.pc.setLocalDescription(offer);
                    this.send('session-initiate', jingle);
                }));
                next();
            }
            catch (err) {
                this._log('error', 'Could not create WebRTC offer', err);
                this.end('failed-application', true);
            }
        });
    }
    accept(opts, next) {
        return __awaiter(this, arguments, void 0, function* () {
            // support calling with accept(next) or accept(opts, next)
            if (arguments.length === 1 && typeof opts === 'function') {
                next = opts;
                opts = {};
            }
            next = next || (() => undefined);
            opts = opts || {};
            this._log('info', 'Accepted incoming session');
            this.state = 'active';
            this.role = 'responder';
            try {
                yield this.processLocal(JingleAction.SessionAccept, () => __awaiter(this, void 0, void 0, function* () {
                    const answer = yield this.pc.createAnswer(opts);
                    const json = importFromSDP(answer.sdp);
                    const jingle = convertIntermediateToRequest(json, this.role, this.transportType);
                    jingle.sid = this.sid;
                    jingle.action = JingleAction.SessionAccept;
                    for (const content of jingle.contents || []) {
                        content.creator = 'initiator';
                    }
                    yield this.pc.setLocalDescription(answer);
                    yield this.processBufferedCandidates();
                    this.send('session-accept', jingle);
                }));
                next();
            }
            catch (err) {
                this._log('error', 'Could not create WebRTC answer', err);
                this.end('failed-application');
            }
        });
    }
    end(reason = 'success', silent = false) {
        for (const receiver of this.pc.getReceivers()) {
            this.onRemoveTrack(receiver.track);
        }
        super.end(reason, silent);
    }
    ring() {
        return this.processLocal('ring', () => __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Ringing on incoming session');
            this.ringing = true;
            this.send(JingleAction.SessionInfo, {
                info: {
                    infoType: JINGLE_INFO_RINGING
                }
            });
        }));
    }
    mute(creator, name) {
        return this.processLocal('mute', () => __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Muting', name);
            this.send(JingleAction.SessionInfo, {
                info: {
                    creator,
                    infoType: JINGLE_INFO_MUTE,
                    name
                }
            });
        }));
    }
    unmute(creator, name) {
        return this.processLocal('unmute', () => __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Unmuting', name);
            this.send(JingleAction.SessionInfo, {
                info: {
                    creator,
                    infoType: JINGLE_INFO_UNMUTE,
                    name
                }
            });
        }));
    }
    hold() {
        return this.processLocal('hold', () => __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Placing on hold');
            this.send('session-info', {
                info: {
                    infoType: JINGLE_INFO_HOLD
                }
            });
        }));
    }
    resume() {
        return this.processLocal('resume', () => __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Resuming from hold');
            this.send('session-info', {
                info: {
                    infoType: JINGLE_INFO_ACTIVE
                }
            });
        }));
    }
    // ----------------------------------------------------------------
    // Track control methods
    // ----------------------------------------------------------------
    addTrack(track, stream, cb) {
        if (track.kind === 'audio') {
            this.includesAudio = true;
        }
        if (track.kind === 'video') {
            this.includesVideo = true;
        }
        return this.processLocal('addtrack', () => __awaiter(this, void 0, void 0, function* () {
            if (this.pc.addTrack) {
                this.pc.addTrack(track, stream);
            }
            else {
                this.pc.addStream(stream);
            }
            if (cb) {
                cb();
            }
        }));
    }
    removeTrack(sender, cb) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.processLocal('removetrack', () => __awaiter(this, void 0, void 0, function* () {
                this.pc.removeTrack(sender);
                if (cb) {
                    return cb();
                }
            }));
        });
    }
    // ----------------------------------------------------------------
    // Track event handlers
    // ----------------------------------------------------------------
    onAddTrack(track, stream) {
        this._log('info', 'Track added');
        this.parent.emit('peerTrackAdded', this, track, stream);
    }
    onRemoveTrack(track) {
        this._log('info', 'Track removed');
        this.parent.emit('peerTrackRemoved', this, track);
    }
    // ----------------------------------------------------------------
    // Jingle action handers
    // ----------------------------------------------------------------
    onSessionInitiate(changes, cb) {
        return __awaiter(this, void 0, void 0, function* () {
            this._log('info', 'Initiating incoming session');
            this.state = 'pending';
            this.role = 'responder';
            this.transportType = changes.contents[0].transport.transportType;
            const json = convertRequestToIntermediate(changes, this.peerRole);
            for (const media of json.media) {
                if (media.kind === 'audio') {
                    this.includesAudio = true;
                }
                if (media.kind === 'video') {
                    this.includesVideo = true;
                }
                if (!media.streams) {
                    media.streams = [{ stream: 'legacy', track: media.kind }];
                }
            }
            const sdp = exportToSDP(json);
            try {
                yield this.pc.setRemoteDescription({ type: 'offer', sdp });
                yield this.processBufferedCandidates();
                return cb();
            }
            catch (err) {
                this._log('error', 'Could not create WebRTC answer', err);
                return cb({ condition: 'general-error' });
            }
        });
    }
    onSessionTerminate(changes, cb) {
        for (const receiver of this.pc.getReceivers()) {
            this.onRemoveTrack(receiver.track);
        }
        super.onSessionTerminate(changes, cb);
    }
    onSessionInfo(changes, cb) {
        const info = changes.info || { infoType: '' };
        switch (info.infoType) {
            case JINGLE_INFO_RINGING:
                this._log('info', 'Outgoing session is ringing');
                this.ringing = true;
                this.parent.emit('ringing', this);
                return cb();
            case JINGLE_INFO_HOLD:
                this._log('info', 'On hold');
                this.parent.emit('hold', this);
                return cb();
            case JINGLE_INFO_UNHOLD:
            case JINGLE_INFO_ACTIVE:
                this._log('info', 'Resuming from hold');
                this.parent.emit('resumed', this);
                return cb();
            case JINGLE_INFO_MUTE:
                this._log('info', 'Muting', info);
                this.parent.emit('mute', this, info);
                return cb();
            case JINGLE_INFO_UNMUTE:
                this._log('info', 'Unmuting', info);
                this.parent.emit('unmute', this, info);
                return cb();
        }
        return cb();
    }
}

const MAX_RELAY_BANDWIDTH = 768 * 1024; // maximum bandwidth used via TURN.
function isICEServer(val) {
    return !val.type && (val.urls || val.url);
}
class SessionManager extends EventEmitter {
    constructor(conf = {}) {
        super();
        conf = conf || {};
        this.selfID = conf.selfID;
        this.sessions = {};
        this.peers = {};
        this.iceServers = conf.iceServers || [];
        this.prepareSession =
            conf.prepareSession ||
                (opts => {
                    if (!this.config.hasRTCPeerConnection) {
                        return;
                    }
                    if (opts.applicationTypes.indexOf(NS_JINGLE_RTP_1) >= 0) {
                        return new MediaSession(opts);
                    }
                    if (opts.applicationTypes.indexOf(NS_JINGLE_FILE_TRANSFER_5) >= 0) {
                        return new FileTransferSession(opts);
                    }
                });
        this.performTieBreak =
            conf.performTieBreak ||
                ((sess, req) => {
                    const applicationTypes = (req.jingle.contents || []).map(content => {
                        if (content.application) {
                            return content.application.applicationType;
                        }
                    });
                    const intersection = (sess.pendingApplicationTypes || []).filter(appType => applicationTypes.includes(appType));
                    return intersection.length > 0;
                });
        this.createPeerConnection =
            conf.createPeerConnection ||
                ((session, opts) => {
                    if (nativeRTCPeerConnection) {
                        return new nativeRTCPeerConnection(opts);
                    }
                });
        this.config = Object.assign({ debug: false, hasRTCPeerConnection: !!nativeRTCPeerConnection, peerConnectionConfig: {
                bundlePolicy: conf.bundlePolicy || 'balanced',
                iceTransportPolicy: conf.iceTransportPolicy || 'all',
                rtcpMuxPolicy: conf.rtcpMuxPolicy || 'require',
                sdpSemantics: conf.sdpSemantics
            }, peerConnectionConstraints: {
                optional: [{ DtlsSrtpKeyAgreement: true }, { RtpDataChannels: false }]
            } }, conf);
    }
    addICEServer(server) {
        if (typeof server === 'string') {
            this.iceServers.push({ urls: server });
            return;
        }
        if (isICEServer(server)) {
            this.iceServers.push(server);
            return;
        }
        let host = server.host || '';
        if (host.indexOf(':') >= 0) {
            host = `[${host}]`;
        }
        let uri = `${server.type}:${host}`;
        if (server.port) {
            uri += `:${server.port}`;
        }
        if (server.transport) {
            uri += `?transport=${server.transport}`;
        }
        if (server.type === 'turn' || server.type === 'turns') {
            this.iceServers.push({
                credential: server.password,
                urls: [uri],
                username: server.username
            });
        }
        else if (server.type === 'stun' || server.type === 'stuns') {
            this.iceServers.push({ urls: [uri] });
        }
    }
    resetICEServers() {
        this.iceServers = [];
    }
    addSession(session) {
        session.parent = this;
        const sid = session.sid;
        const peer = session.peerID;
        this.sessions[sid] = session;
        if (!this.peers[peer]) {
            this.peers[peer] = [];
        }
        this.peers[peer].push(session);
        this.emit('createdSession', session);
        return session;
    }
    forgetSession(session) {
        const peers = this.peers[session.peerID] || [];
        if (peers.length) {
            peers.splice(peers.indexOf(session), 1);
        }
        delete this.sessions[session.sid];
    }
    createMediaSession(peer, sid, stream) {
        const session = new MediaSession({
            config: this.config.peerConnectionConfig,
            constraints: this.config.peerConnectionConstraints,
            iceServers: this.iceServers,
            initiator: true,
            maxRelayBandwidth: MAX_RELAY_BANDWIDTH,
            parent: this,
            peerID: peer,
            sid,
            stream
        });
        this.addSession(session);
        return session;
    }
    createFileTransferSession(peer, sid) {
        const session = new FileTransferSession({
            config: this.config.peerConnectionConfig,
            constraints: this.config.peerConnectionConstraints,
            iceServers: this.iceServers,
            initiator: true,
            maxRelayBandwidth: MAX_RELAY_BANDWIDTH,
            parent: this,
            peerID: peer,
            sid
        });
        this.addSession(session);
        return session;
    }
    endPeerSessions(peer, reason, silent = false) {
        const sessions = this.peers[peer] || [];
        delete this.peers[peer];
        for (const session of sessions) {
            session.end(reason || 'gone', silent);
        }
    }
    endAllSessions(reason, silent = false) {
        for (const peer of Object.keys(this.peers)) {
            this.endPeerSessions(peer, reason, silent);
        }
    }
    process(req) {
        // Extract the request metadata that we need to verify
        const sid = req.jingle ? req.jingle.sid : undefined;
        let session = sid ? this.sessions[sid] : undefined;
        const rid = req.id;
        const sender = req.from;
        if (!sender) {
            return;
        }
        if (req.type === 'error') {
            this._log('error', 'Received error response', req);
            if (session && req.error && req.error.jingleError === 'unknown-session') {
                return session.end('gone', true);
            }
            const isTieBreak = req.error && req.error.jingleError === 'tie-break';
            if (session && session.state === 'pending' && isTieBreak) {
                return session.end('alternative-session', true);
            }
            else {
                if (session) {
                    session.pendingAction = undefined;
                }
                return;
            }
        }
        if (req.type === 'result') {
            if (session) {
                session.pendingAction = undefined;
            }
            return;
        }
        const action = req.jingle.action;
        const contents = req.jingle.contents || [];
        const applicationTypes = contents.map(content => {
            return content.application ? content.application.applicationType : undefined;
        });
        const transportTypes = contents.map(content => {
            return content.transport ? content.transport.transportType : undefined;
        });
        // Now verify that we are allowed to actually process the
        // requested action
        if (action !== JingleAction.SessionInitiate) {
            // Can't modify a session that we don't have.
            if (!session) {
                if (action === 'session-terminate') {
                    this.emit('send', {
                        id: rid,
                        to: sender,
                        type: 'result'
                    });
                    return;
                }
                this._log('error', 'Unknown session', sid);
                return this._sendError(sender, rid, {
                    condition: 'item-not-found',
                    jingleError: 'unknown-session'
                });
            }
            // Check if someone is trying to hijack a session.
            if (session.peerID !== sender || session.state === 'ended') {
                this._log('error', 'Session has ended, or action has wrong sender');
                return this._sendError(sender, rid, {
                    condition: 'item-not-found',
                    jingleError: 'unknown-session'
                });
            }
            // Can't accept a session twice
            if (action === 'session-accept' && session.state !== 'pending') {
                this._log('error', 'Tried to accept session twice', sid);
                return this._sendError(sender, rid, {
                    condition: 'unexpected-request',
                    jingleError: 'out-of-order'
                });
            }
            // Can't process two requests at once, need to tie break
            if (action !== 'session-terminate' && action === session.pendingAction) {
                this._log('error', 'Tie break during pending request');
                if (session.isInitiator) {
                    return this._sendError(sender, rid, {
                        condition: 'conflict',
                        jingleError: 'tie-break'
                    });
                }
            }
        }
        else if (session) {
            // Don't accept a new session if we already have one.
            if (session.peerID !== sender) {
                this._log('error', 'Duplicate sid from new sender');
                return this._sendError(sender, rid, {
                    condition: 'service-unavailable'
                });
            }
            // Check if we need to have a tie breaker because both parties
            // happened to pick the same random sid.
            if (session.state === 'pending') {
                if (this.selfID &&
                    this.selfID > session.peerID &&
                    this.performTieBreak(session, req)) {
                    this._log('error', 'Tie break new session because of duplicate sids');
                    return this._sendError(sender, rid, {
                        condition: 'conflict',
                        jingleError: 'tie-break'
                    });
                }
            }
            else {
                // The other side is just doing it wrong.
                this._log('error', 'Someone is doing this wrong');
                return this._sendError(sender, rid, {
                    condition: 'unexpected-request',
                    jingleError: 'out-of-order'
                });
            }
        }
        else if (this.peers[sender] && this.peers[sender].length) {
            // Check if we need to have a tie breaker because we already have
            // a different session with this peer that is using the requested
            // content application types.
            for (let i = 0, len = this.peers[sender].length; i < len; i++) {
                const sess = this.peers[sender][i];
                if (sess &&
                    sess.state === 'pending' &&
                    sid &&
                    octetCompare(sess.sid, sid) > 0 &&
                    this.performTieBreak(sess, req)) {
                    this._log('info', 'Tie break session-initiate');
                    return this._sendError(sender, rid, {
                        condition: 'conflict',
                        jingleError: 'tie-break'
                    });
                }
            }
        }
        // We've now weeded out invalid requests, so we can process the action now.
        if (action === 'session-initiate') {
            if (!contents.length) {
                return this._sendError(sender, rid, {
                    condition: 'bad-request'
                });
            }
            session = this._createIncomingSession({
                applicationTypes,
                config: this.config.peerConnectionConfig,
                constraints: this.config.peerConnectionConstraints,
                iceServers: this.iceServers,
                initiator: false,
                parent: this,
                peerID: sender,
                sid,
                transportTypes
            }, req);
        }
        session.process(action, req.jingle, (err) => {
            if (err) {
                this._log('error', 'Could not process request', req, err);
                this._sendError(sender, rid, err);
            }
            else {
                this.emit('send', {
                    id: rid,
                    to: sender,
                    type: 'result'
                });
                // Wait for the initial action to be processed before emitting
                // the session for the user to accept/reject.
                if (action === 'session-initiate') {
                    this.emit('incoming', session);
                }
            }
        });
    }
    signal(session, data) {
        const action = data.jingle && data.jingle.action;
        if (session.isInitiator && action === JingleAction.SessionInitiate) {
            this.emit('outgoing', session);
        }
        this.emit('send', data);
    }
    _createIncomingSession(meta, req) {
        let session;
        if (this.prepareSession) {
            session = this.prepareSession(meta, req);
        }
        // Fallback to a generic session type, which can
        // only be used to end the session.
        if (!session) {
            session = new JingleSession(meta);
        }
        this.addSession(session);
        return session;
    }
    _sendError(to, id, data) {
        if (!data.type) {
            data.type = 'cancel';
        }
        this.emit('send', {
            error: data,
            id,
            to,
            type: 'error'
        });
    }
    _log(level, message, ...args) {
        this.emit('log', level, message, ...args);
        this.emit('log:' + level, message, ...args);
    }
}

var index$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    Session: JingleSession,
    ICESession: ICESession,
    MediaSession: MediaSession,
    FileSession: FileTransferSession,
    SessionManager: SessionManager,
    importFromSDP: importFromSDP,
    exportToSDP: exportToSDP
});

function Jingle (client) {
    const hasNativePeerConnection = !!nativeRTCPeerConnection;
    const defaultConfig = {
        advertiseAudio: hasNativePeerConnection,
        advertiseFileTransfer: hasNativePeerConnection,
        advertiseVideo: hasNativePeerConnection,
        bundlePolicy: 'balanced',
        hasRTCPeerConnection: hasNativePeerConnection,
        iceServers: [{ urls: 'stun:stun.l.google.com:19302' }],
        iceTransportPolicy: 'all',
        rtcpMuxPolicy: 'require',
        trickleIce: true
    };
    const providedConfig = client.config.jingle;
    const config = Object.assign(Object.assign({}, defaultConfig), providedConfig);
    const jingle = (client.jingle = new SessionManager(config));
    const caps = [NS_JINGLE_1];
    if (config.hasRTCPeerConnection) {
        caps.push(NS_JINGLE_ICE_0, NS_JINGLE_ICE_UDP_1, NS_JINGLE_DTLS_SCTP_1, NS_JINGLE_DTLS_0, 'urn:ietf:rfc:5888' // Jingle Grouping Framework
        );
        if (config.trickleIce === false) {
            caps.push('urn:ietf:rfc:3264'); // ICE prefer batched candidates
        }
        if (config.advertiseAudio || config.advertiseVideo) {
            caps.push(NS_JINGLE_RTP_1, NS_JINGLE_RTP_RTCP_FB_0, NS_JINGLE_RTP_HDREXT_0, 'urn:ietf:rfc:5576' // Jingle Source Specific Media Attributes
            );
        }
        if (config.advertiseAudio) {
            caps.push(NS_JINGLE_RTP_AUDIO);
        }
        if (config.advertiseVideo) {
            caps.push(NS_JINGLE_RTP_VIDEO);
        }
        if (config.advertiseFileTransfer) {
            caps.push(NS_JINGLE_FILE_TRANSFER_4, NS_JINGLE_FILE_TRANSFER_5);
        }
    }
    for (const cap of caps) {
        client.disco.addFeature(cap);
    }
    const mappedEvents = [
        'outgoing',
        'incoming',
        'accepted',
        'terminated',
        'ringing',
        'mute',
        'unmute',
        'hold',
        'resumed'
    ];
    for (const event of mappedEvents) {
        jingle.on(event, (session, data) => {
            client.emit(('jingle:' + event), session, data);
        });
    }
    jingle.on('createdSession', data => {
        client.emit('jingle:created', data);
    });
    jingle.on('send', (data) => __awaiter(this, void 0, void 0, function* () {
        try {
            if (data.type === 'set') {
                const resp = yield client.sendIQ(data);
                if (!resp.jingle) {
                    resp.jingle = {};
                }
                resp.jingle.sid = data.jingle.sid;
                jingle.process(resp);
            }
            if (data.type === 'result') {
                client.sendIQResult({ type: 'set', id: data.id, from: data.to }, data);
            }
            if (data.type === 'error') {
                client.sendIQError({ type: 'set', id: data.id, from: data.to }, data);
            }
        }
        catch (err) {
            if (!err.jingle) {
                err.jingle = data.jingle;
            }
            err.jingle.sid = data.jingle.sid;
            jingle.process(err);
        }
    }));
    client.on('session:bound', (jid) => {
        jingle.selfID = jid;
    });
    client.on('iq:set:jingle', (data) => {
        jingle.process(data);
    });
    client.on('unavailable', (pres) => {
        jingle.endPeerSessions(pres.from, undefined, true);
    });
    client.getServices = (jid, type, version) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            externalServices: {
                type,
                version
            },
            to: jid,
            type: 'get'
        });
        const services = resp.externalServices;
        services.services = services.services || [];
        return services;
    });
    client.getServiceCredentials = (jid, host, type, port, version) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            externalServiceCredentials: {
                host,
                port,
                type,
                version
            },
            to: jid,
            type: 'get'
        });
        return resp.externalServiceCredentials;
    });
    client.discoverICEServers = (opts = {}) => __awaiter(this, void 0, void 0, function* () {
        try {
            const resp = yield client.getServices(client.config.server, undefined, opts.version);
            const services = resp.services || [];
            const discovered = [];
            for (const service of services) {
                client.jingle.addICEServer(service);
            }
            return discovered;
        }
        catch (err) {
            return [];
        }
    });
}

function mergeFields(original, updated) {
    const merged = [];
    const mappedUpdates = new Map();
    for (const field of updated) {
        if (field.name) {
            mappedUpdates.set(field.name, field);
        }
    }
    const usedUpdates = new Set();
    // Update any existing fields with new values.
    for (const field of original) {
        if (field.name && mappedUpdates.has(field.name)) {
            merged.push(Object.assign(Object.assign({}, field), mappedUpdates.get(field.name)));
            usedUpdates.add(field.name);
        }
        else {
            merged.push(Object.assign({}, field));
        }
    }
    // Append any brand new fields to the list
    for (const field of updated) {
        if (!field.name || (field.name && !usedUpdates.has(field.name))) {
            merged.push(Object.assign({}, field));
        }
    }
    return merged;
}

var DataForms = /*#__PURE__*/Object.freeze({
    __proto__: null,
    mergeFields: mergeFields
});

function MAM (client) {
    client.getHistorySearchForm = (jid, opts = {}) => __awaiter(this, void 0, void 0, function* () {
        const res = yield client.sendIQ({
            archive: {
                type: 'query',
                version: opts.version
            },
            to: jid,
            type: 'get'
        });
        return res.archive.form;
    });
    client.searchHistory = (jidOrOpts, opts = {}) => __awaiter(this, void 0, void 0, function* () {
        const queryid = client.nextId();
        let jid = '';
        if (typeof jidOrOpts === 'string') {
            jid = jidOrOpts;
        }
        else {
            jid = jidOrOpts.to || '';
            opts = jidOrOpts;
        }
        opts.queryId = queryid;
        const form = opts.form || {};
        form.type = 'submit';
        const fields = form.fields || [];
        const defaultFields = [
            {
                name: 'FORM_TYPE',
                type: 'hidden',
                value: NS_MAM_2
            }
        ];
        if (opts.with) {
            defaultFields.push({
                name: 'with',
                type: 'text-single',
                value: opts.with
            });
        }
        if (opts.start) {
            defaultFields.push({
                name: 'start',
                type: 'text-single',
                value: opts.start.toISOString()
            });
        }
        if (opts.end) {
            defaultFields.push({
                name: 'end',
                type: 'text-single',
                value: opts.end.toISOString()
            });
        }
        form.fields = mergeFields(defaultFields, fields);
        opts.form = form;
        const allowed = allowedResponders(client.jid, jid);
        const results = [];
        const collector = (msg) => {
            if (allowed.has(msg.from) && msg.archive && msg.archive.queryId === queryid) {
                results.push(msg.archive);
            }
        };
        client.on('mam:item', collector);
        try {
            const resp = yield client.sendIQ({
                archive: opts,
                id: queryid,
                to: jid,
                type: 'set'
            });
            return Object.assign(Object.assign({}, resp.archive), { results });
        }
        finally {
            client.off('mam:item', collector);
        }
    });
    client.getHistoryPreferences = () => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            archive: {
                type: 'preferences'
            },
            type: 'get'
        });
        return resp.archive;
    });
    client.setHistoryPreferences = (opts) => {
        return client.sendIQ({
            archive: Object.assign({ type: 'preferences' }, opts),
            type: 'set'
        });
    };
    client.on('message', msg => {
        if (msg.archive) {
            client.emit('mam:item', msg);
        }
    });
}

const ACK_TYPES = new Set(['chat', 'groupchat', 'headline', 'normal']);
const ALLOWED_CHAT_STATE_TYPES = new Set(['chat', 'groupchat', 'normal']);
const MARKER_RANK = new Map([
    ['markable', 0],
    ['received', 1],
    ['displayed', 2],
    ['acknowledged', 3]
]);
const isReceivedCarbon = (msg) => !!msg.carbon && msg.carbon.type === 'received';
const isSentCarbon = (msg) => !!msg.carbon && msg.carbon.type === 'sent';
const isChatState = (msg) => !!msg.chatState;
const isReceiptRequest = (msg) => { var _a; return !!msg.receipt && msg.receipt.type === 'request' && ACK_TYPES.has((_a = msg.type) !== null && _a !== void 0 ? _a : 'normal'); };
const hasRTT = (msg) => !!msg.rtt;
const isCorrection = (msg) => !!msg.replace;
const isMarkable = (msg, marker) => {
    var _a;
    return msg.marker &&
        (MARKER_RANK.get(msg.marker.type) < MARKER_RANK.get(marker) ||
            ((_a = msg.marker) === null || _a === void 0 ? void 0 : _a.type) === 'markable');
};
const isFormsMessage = (msg) => !!msg.forms && msg.forms.length > 0;
function toggleCarbons(client, action) {
    return __awaiter(this, void 0, void 0, function* () {
        yield client.sendIQ({
            carbons: {
                action
            },
            type: 'set'
        });
    });
}
function Messaging (client) {
    client.disco.addFeature(NS_ATTENTION_0);
    client.disco.addFeature(NS_CHAT_MARKERS_0);
    client.disco.addFeature(NS_CHAT_STATES);
    client.disco.addFeature(NS_CORRECTION_0);
    client.disco.addFeature(NS_RECEIPTS);
    client.disco.addFeature(NS_RTT_0);
    client.enableCarbons = () => toggleCarbons(client, 'enable');
    client.disableCarbons = () => toggleCarbons(client, 'disable');
    client.markReceived = (msg) => client.sendMarker(msg, 'received');
    client.markDisplayed = (msg) => client.sendMarker(msg, 'displayed');
    client.markAcknowledged = (msg) => client.sendMarker(msg, 'acknowledged');
    client.getAttention = (jid, opts = {}) => {
        return client.sendMessage(Object.assign(Object.assign({}, opts), { requestingAttention: true, to: jid, type: 'headline' }));
    };
    client.sendMarker = (msg, marker, force) => {
        if (!isMarkable(msg, marker) && !force) {
            return;
        }
        const useStanzaID = client.config.groupchatMarkersUseStanzaID !== false;
        let id = msg.id;
        if (msg.type === 'groupchat' && msg.stanzaIds && useStanzaID) {
            const mucStanzaId = msg.stanzaIds.find(s => equalBare(s.by, msg.from));
            if (mucStanzaId) {
                id = mucStanzaId.id;
            }
        }
        client.sendMessage({
            marker: {
                id,
                type: marker
            },
            to: msg.type === 'groupchat' ? toBare(msg.from) : msg.from,
            type: msg.type
        });
    };
    client.on('message', msg => {
        if (msg.carbon && equalBare(msg.from, client.jid)) {
            const forwardedMessage = msg.carbon.forward.message;
            if (!forwardedMessage.delay) {
                forwardedMessage.delay = msg.carbon.forward.delay || {
                    timestamp: new Date(Date.now())
                };
            }
            if (isReceivedCarbon(msg)) {
                client.emit('carbon:received', msg);
                client.emit('message', forwardedMessage);
            }
            if (isSentCarbon(msg)) {
                client.emit('carbon:sent', msg);
                client.emit('message:sent', forwardedMessage, true);
            }
        }
        if (isFormsMessage(msg)) {
            client.emit('dataform', msg);
        }
        if (msg.requestingAttention) {
            client.emit('attention', msg);
        }
        if (hasRTT(msg)) {
            client.emit('rtt', msg);
        }
        if (isCorrection(msg)) {
            client.emit('replace', msg);
        }
        if (isChatState(msg) && ALLOWED_CHAT_STATE_TYPES.has(msg.type || 'normal')) {
            client.emit('chat:state', msg);
        }
        const sendReceipts = client.config.sendReceipts !== false;
        const sendMUCReceipts = client.config.sendMUCReceipts !== false;
        const sendMarkers = client.config.chatMarkers !== false;
        const useStanzaID = client.config.groupchatMarkersUseStanzaID !== false;
        const isReceipt = isReceiptRequest(msg);
        const isReceivedMarkable = isMarkable(msg, 'received');
        const canSendReceipt = isReceipt && sendReceipts && (msg.type === 'groupchat' ? sendMUCReceipts : true);
        if (canSendReceipt || (sendMarkers && isReceivedMarkable)) {
            const to = msg.type === 'groupchat' ? toBare(msg.from) : msg.from;
            let markerId = msg.id;
            if (msg.type === 'groupchat' && msg.stanzaIds && useStanzaID) {
                const mucStanzaId = msg.stanzaIds.find(s => equalBare(s.by, msg.from));
                if (mucStanzaId) {
                    markerId = mucStanzaId.id;
                }
            }
            client.sendMessage({
                receipt: canSendReceipt
                    ? {
                        id: msg.id,
                        type: 'received'
                    }
                    : undefined,
                marker: sendMarkers && isReceivedMarkable
                    ? {
                        id: markerId,
                        type: 'received'
                    }
                    : undefined,
                to,
                type: msg.type
            });
        }
        if (msg.receipt && msg.receipt.type === 'received') {
            client.emit('receipt', msg);
        }
        if (msg.marker && msg.marker.type !== 'markable') {
            client.emit(`marker:${msg.marker.type}`, msg);
        }
    });
}

function isMUCPresence(pres) {
    return !!pres.muc;
}
function MUC (client) {
    client.disco.addFeature(NS_MUC);
    client.disco.addFeature(NS_MUC_DIRECT_INVITE);
    client.disco.addFeature(NS_HATS_0);
    client.joinedRooms = new Map();
    client.joiningRooms = new Map();
    client.leavingRooms = new Map();
    function rejoinRooms() {
        const oldJoiningRooms = client.joiningRooms;
        client.joiningRooms = new Map();
        for (const [room, nick] of oldJoiningRooms) {
            client.joinRoom(room, nick);
        }
        const oldJoinedRooms = client.joinedRooms;
        client.joinedRooms = new Map();
        for (const [room, nick] of oldJoinedRooms) {
            client.joinRoom(room, nick);
        }
    }
    client.on('session:started', rejoinRooms);
    client.on('message', msg => {
        if (msg.type === 'groupchat' && msg.hasSubject) {
            client.emit('muc:topic', {
                from: msg.from,
                room: toBare(msg.from),
                topic: msg.subject || ''
            });
            return;
        }
        if (!msg.muc) {
            return;
        }
        if (msg.muc.type === 'direct-invite' || (!msg.muc.invite && msg.legacyMUC)) {
            const invite = msg.muc.type === 'direct-invite' ? msg.muc : msg.legacyMUC;
            client.emit('muc:invite', {
                from: msg.from,
                password: invite.password,
                reason: invite.reason,
                room: invite.jid,
                thread: invite.thread,
                type: 'direct'
            });
            return;
        }
        if (msg.muc.invite) {
            client.emit('muc:invite', {
                from: msg.muc.invite[0].from,
                password: msg.muc.password,
                reason: msg.muc.invite[0].reason,
                room: msg.from,
                thread: msg.muc.invite[0].thread,
                type: 'mediated'
            });
            return;
        }
        if (msg.muc.decline) {
            client.emit('muc:declined', {
                from: msg.muc.decline.from,
                reason: msg.muc.decline.reason,
                room: msg.from
            });
            return;
        }
        client.emit('muc:other', msg);
    });
    client.on('presence', pres => {
        const room = toBare(pres.from);
        if (client.joiningRooms.has(room) && pres.type === 'error') {
            client.joiningRooms.delete(room);
            client.emit('muc:failed', pres);
            client.emit('muc:error', pres);
            return;
        }
        if (!isMUCPresence(pres)) {
            return;
        }
        const isSelf = pres.muc.statusCodes && pres.muc.statusCodes.indexOf(MUCStatusCode.SelfPresence) >= 0;
        const isNickChange = pres.muc.statusCodes && pres.muc.statusCodes.indexOf(MUCStatusCode.NickChanged) >= 0;
        if (pres.type === 'error') {
            client.emit('muc:error', pres);
            return;
        }
        if (pres.type === 'unavailable') {
            client.emit('muc:unavailable', pres);
            if (isSelf) {
                if (isNickChange) {
                    client.joinedRooms.set(room, pres.muc.nick);
                }
                else {
                    client.emit('muc:leave', pres);
                    client.joinedRooms.delete(room);
                    client.leavingRooms.delete(room);
                }
            }
            if (pres.muc.destroy) {
                client.emit('muc:destroyed', {
                    newRoom: pres.muc.destroy.jid,
                    password: pres.muc.destroy.password,
                    reason: pres.muc.destroy.reason,
                    room
                });
            }
            return;
        }
        client.emit('muc:available', pres);
        const isJoin = client.joiningRooms.has(room) || !client.joinedRooms.has(room);
        if (isSelf) {
            client.joinedRooms.set(room, getResource(pres.from));
            if (isJoin) {
                client.joiningRooms.delete(room);
                client.emit('muc:join', pres);
            }
        }
    });
    client.joinRoom = (room, nick, opts = {}) => __awaiter(this, void 0, void 0, function* () {
        room = toBare(room);
        client.joiningRooms.set(room, nick || '');
        if (!nick) {
            try {
                nick = yield client.getReservedNick(room);
                client.joiningRooms.set(room, nick);
            }
            catch (err) {
                throw new Error('Room nick required');
            }
        }
        return new Promise((resolve, reject) => {
            function joined(pres) {
                if (equalBare(pres.from, room)) {
                    client.off('muc:join', joined);
                    client.off('muc:failed', failed);
                    resolve(pres);
                }
            }
            function failed(pres) {
                if (equalBare(pres.from, room)) {
                    client.off('muc:join', joined);
                    client.off('muc:failed', failed);
                    reject(pres);
                }
            }
            client.on('muc:join', joined);
            client.on('muc:failed', failed);
            client.sendPresence(Object.assign(Object.assign({}, opts), { muc: Object.assign(Object.assign({}, opts.muc), { type: 'join' }), to: createFull(room, nick) }));
        });
    });
    client.leaveRoom = (room, nick, opts = {}) => {
        room = toBare(room);
        nick = nick || client.joinedRooms.get(room);
        client.leavingRooms.set(room, nick);
        return new Promise((resolve, reject) => {
            const id = opts.id || uuid();
            const allowed = allowedResponders(room);
            function leave(pres) {
                if (equalBare(pres.from, room)) {
                    client.off('muc:leave', leave);
                    client.off('presence:error', leaveError);
                    resolve(pres);
                }
            }
            function leaveError(pres) {
                if (pres.id === id && allowed.has(pres.from)) {
                    if (!client.joinedRooms.has(room)) {
                        client.leavingRooms.delete(room);
                    }
                    client.off('muc:leave', leave);
                    client.off('presence:error', leaveError);
                    reject(pres);
                }
            }
            client.on('muc:leave', leave);
            client.on('presence:error', leaveError);
            client.sendPresence(Object.assign(Object.assign({}, opts), { id, to: createFull(room, nick), type: 'unavailable' }));
        });
    };
    client.ban = (room, occupantRealJID, reason) => {
        return client.setRoomAffiliation(room, occupantRealJID, 'outcast', reason);
    };
    client.kick = (room, nick, reason) => {
        return client.setRoomRole(room, nick, 'none', reason);
    };
    client.invite = (room, opts = []) => {
        if (!Array.isArray(opts)) {
            opts = [opts];
        }
        client.sendMessage({
            muc: {
                invite: opts,
                type: 'info'
            },
            to: room
        });
    };
    client.directInvite = (room, to, opts = {}) => {
        client.sendMessage({
            muc: Object.assign(Object.assign({}, opts), { jid: room, type: 'direct-invite' }),
            to
        });
    };
    client.declineInvite = (room, sender, reason) => {
        client.sendMessage({
            muc: {
                decline: {
                    reason,
                    to: sender
                },
                type: 'info'
            },
            to: room
        });
    };
    client.changeNick = (room, nick) => {
        const id = uuid();
        const newJID = createFull(room, nick);
        const allowed = allowedResponders(room);
        return new Promise((resolve, reject) => {
            function success(pres) {
                if (!allowed.has(toBare(pres.from))) {
                    return;
                }
                if (!pres.muc.statusCodes ||
                    !pres.muc.statusCodes.includes(MUCStatusCode.SelfPresence)) {
                    return;
                }
                client.off('muc:available', success);
                client.off(`presence:id:${id}`, errorOrNoChange);
                resolve(pres);
            }
            function errorOrNoChange(pres) {
                if (!allowed.has(toBare(pres.from)) || pres.id !== id) {
                    return;
                }
                client.off('muc:available', success);
                client.off(`presence:id:${id}`, errorOrNoChange);
                if (pres.type === 'error') {
                    reject(pres);
                }
                else {
                    resolve(pres);
                }
            }
            client.on('muc:available', success);
            client.on(`presence:id:${id}`, errorOrNoChange);
            client.sendPresence({
                id,
                to: newJID
            });
        });
    };
    client.setSubject = (room, subject) => {
        client.sendMessage({
            subject,
            to: room,
            type: 'groupchat'
        });
    };
    client.getReservedNick = (room) => __awaiter(this, void 0, void 0, function* () {
        try {
            const info = yield client.getDiscoInfo(room, 'x-roomuser-item');
            const identity = info.identities[0];
            if (identity.name) {
                return identity.name;
            }
            else {
                throw new Error('No nickname reserved');
            }
        }
        catch (err) {
            throw new Error('No nickname reserved');
        }
    });
    client.requestRoomVoice = (room) => {
        client.sendMessage({
            forms: [
                {
                    fields: [
                        {
                            name: 'FORM_TYPE',
                            type: 'hidden',
                            value: 'http://jabber.org/protocol/muc#request'
                        },
                        {
                            name: 'muc#role',
                            type: 'text-single',
                            value: 'participant'
                        }
                    ],
                    type: 'submit'
                }
            ],
            to: room
        });
    };
    client.setRoomAffiliation = (room, occupantRealJID, affiliation, reason) => {
        return client.sendIQ({
            muc: {
                type: 'user-list',
                users: [
                    {
                        affiliation,
                        jid: occupantRealJID,
                        reason
                    }
                ]
            },
            to: room,
            type: 'set'
        });
    };
    client.setRoomRole = (room, nick, role, reason) => {
        return client.sendIQ({
            muc: {
                type: 'user-list',
                users: [
                    {
                        nick,
                        reason,
                        role
                    }
                ]
            },
            to: room,
            type: 'set'
        });
    };
    client.getRoomMembers = (room, opts = { affiliation: 'member' }) => {
        return client.sendIQ({
            muc: {
                type: 'user-list',
                users: [opts]
            },
            to: room,
            type: 'get'
        });
    };
    client.getRoomConfig = (room) => __awaiter(this, void 0, void 0, function* () {
        const result = yield client.sendIQ({
            muc: {
                type: 'configure'
            },
            to: room,
            type: 'get'
        });
        if (!result.muc.form) {
            throw new Error('No configuration form returned');
        }
        return result.muc.form;
    });
    client.configureRoom = (room, form = {}) => {
        return client.sendIQ({
            muc: {
                form: Object.assign(Object.assign({}, form), { type: 'submit' }),
                type: 'configure'
            },
            to: room,
            type: 'set'
        });
    };
    client.destroyRoom = (room, opts = {}) => {
        return client.sendIQ({
            muc: {
                destroy: opts,
                type: 'configure'
            },
            to: room,
            type: 'set'
        });
    };
    client.getUniqueRoomName = function (mucHost) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.sendIQ({
                muc: {
                    type: 'unique'
                },
                to: mucHost,
                type: 'get'
            });
            if (!result.muc.name) {
                throw new Error('No unique name returned');
            }
            return result.muc.name;
        });
    };
    client.getBookmarks = () => __awaiter(this, void 0, void 0, function* () {
        const res = yield client.getPrivateData('bookmarks');
        if (!res || !res.rooms) {
            return [];
        }
        return res.rooms;
    });
    client.setBookmarks = (bookmarks) => {
        return client.setPrivateData('bookmarks', {
            rooms: bookmarks
        });
    };
    client.addBookmark = (bookmark) => __awaiter(this, void 0, void 0, function* () {
        const mucs = yield client.getBookmarks();
        const updated = [];
        let updatedExisting = false;
        for (const muc of mucs) {
            if (equalBare(muc.jid, bookmark.jid)) {
                updated.push(Object.assign(Object.assign({}, muc), bookmark));
                updatedExisting = true;
            }
            else {
                updated.push(muc);
            }
        }
        if (!updatedExisting) {
            updated.push(bookmark);
        }
        return client.setBookmarks(updated);
    });
    client.removeBookmark = (jid) => __awaiter(this, void 0, void 0, function* () {
        const existingMucs = yield client.getBookmarks();
        const updated = existingMucs.filter(muc => {
            return !equalBare(muc.jid, jid);
        });
        return client.setBookmarks(updated);
    });
}

function PEP (client) {
    client.disco.addFeature(NS_ACTIVITY);
    client.disco.addFeature(NS_GEOLOC);
    client.disco.addFeature(NS_MOOD);
    client.disco.addFeature(NS_NICK);
    client.disco.addFeature(NS_TUNE);
    client.disco.addFeature(NS_PEP_NOTIFY(NS_ACTIVITY));
    client.disco.addFeature(NS_PEP_NOTIFY(NS_GEOLOC));
    client.disco.addFeature(NS_PEP_NOTIFY(NS_MOOD));
    client.disco.addFeature(NS_PEP_NOTIFY(NS_NICK));
    client.disco.addFeature(NS_PEP_NOTIFY(NS_TUNE));
    client.publishActivity = (data) => {
        return client.publish('', NS_ACTIVITY, Object.assign({ itemType: NS_ACTIVITY }, data));
    };
    client.publishGeoLoc = (data) => {
        return client.publish('', NS_GEOLOC, Object.assign({ itemType: NS_GEOLOC }, data));
    };
    client.publishMood = (mood) => {
        return client.publish('', NS_MOOD, Object.assign({ itemType: NS_MOOD }, mood));
    };
    client.publishNick = (nick) => {
        return client.publish('', NS_NICK, {
            itemType: NS_NICK,
            nick
        });
    };
    client.publishTune = (tune) => {
        return client.publish('', NS_TUNE, Object.assign({ itemType: NS_TUNE }, tune));
    };
    client.on('pubsub:published', msg => {
        const content = msg.pubsub.items.published[0].content;
        switch (msg.pubsub.items.node) {
            case NS_ACTIVITY:
                return client.emit('activity', {
                    activity: content,
                    jid: msg.from
                });
            case NS_GEOLOC:
                return client.emit('geoloc', {
                    geoloc: content,
                    jid: msg.from
                });
            case NS_MOOD:
                return client.emit('mood', {
                    jid: msg.from,
                    mood: content
                });
            case NS_NICK:
                return client.emit('nick', {
                    jid: msg.from,
                    nick: content.nick
                });
            case NS_TUNE:
                return client.emit('tune', {
                    jid: msg.from,
                    tune: msg.pubsub.items.published[0].content
                });
        }
    });
}

function isPubsubMessage(msg) {
    return !!msg.pubsub;
}
function isPubsubPublish(msg) {
    return !!msg.pubsub.items && !!msg.pubsub.items.published;
}
function isPubsubRetract(msg) {
    return !!msg.pubsub.items && !!msg.pubsub.items.retracted;
}
function isPubsubPurge(msg) {
    return msg.pubsub.eventType === 'purge';
}
function isPubsubDelete(msg) {
    return msg.pubsub.eventType === 'delete';
}
function isPubsubSubscription(msg) {
    return msg.pubsub.eventType === 'subscription';
}
function isPubsubConfiguration(msg) {
    return msg.pubsub.eventType === 'configuration';
}
function isPubsubAffiliation(msg) {
    if (!msg.pubsub) {
        return false;
    }
    return (!msg.pubsub.context || msg.pubsub.context === 'user') && !!msg.pubsub.affiliations;
}
function PubSub (client) {
    client.disco.addFeature(`${NS_SHIM}#SubID`, NS_SHIM);
    client.on('message', msg => {
        if (isPubsubAffiliation(msg)) {
            client.emit('pubsub:affiliations', msg);
            return;
        }
        if (!isPubsubMessage(msg)) {
            return;
        }
        client.emit('pubsub:event', msg);
        if (isPubsubPublish(msg)) {
            client.emit('pubsub:published', msg);
            return;
        }
        if (isPubsubRetract(msg)) {
            client.emit('pubsub:retracted', msg);
            return;
        }
        if (isPubsubPurge(msg)) {
            client.emit('pubsub:purged', msg);
            return;
        }
        if (isPubsubDelete(msg)) {
            client.emit('pubsub:deleted', msg);
            return;
        }
        if (isPubsubSubscription(msg)) {
            client.emit('pubsub:subscription', msg);
            return;
        }
        if (isPubsubConfiguration(msg)) {
            client.emit('pubsub:config', msg);
            return;
        }
    });
    client.subscribeToNode = (jid, opts) => __awaiter(this, void 0, void 0, function* () {
        const subscribe = {};
        let form;
        if (typeof opts === 'string') {
            subscribe.node = opts;
            subscribe.jid = toBare(client.jid);
        }
        else {
            subscribe.node = opts.node;
            subscribe.jid = opts.jid || (opts.useBareJID ? toBare(client.jid) : client.jid);
            form = opts.options;
        }
        const resp = yield client.sendIQ({
            pubsub: {
                context: 'user',
                subscribe,
                subscriptionOptions: form ? { form } : undefined
            },
            to: jid,
            type: 'set'
        });
        const sub = resp.pubsub.subscription || {};
        if (resp.pubsub.subscriptionOptions) {
            sub.options = resp.pubsub.subscriptionOptions.form;
        }
        return sub;
    });
    client.unsubscribeFromNode = (jid, opts) => __awaiter(this, void 0, void 0, function* () {
        const unsubscribe = {};
        if (typeof opts === 'string') {
            unsubscribe.node = opts;
            unsubscribe.jid = toBare(client.jid);
        }
        else {
            unsubscribe.node = opts.node;
            unsubscribe.subid = opts.subid;
            unsubscribe.jid = opts.jid || (opts.useBareJID ? toBare(client.jid) : client.jid);
        }
        const resp = yield client.sendIQ({
            pubsub: {
                context: 'user',
                unsubscribe
            },
            to: jid,
            type: 'set'
        });
        if (!resp.pubsub || !resp.pubsub.subscription) {
            return Object.assign(Object.assign({}, unsubscribe), { state: 'none' });
        }
        return resp.pubsub.subscription;
    });
    client.publish = (jid, node, item, id) => {
        return client.sendIQ({
            pubsub: {
                context: 'user',
                publish: {
                    item: {
                        content: item,
                        id
                    },
                    node
                }
            },
            to: jid,
            type: 'set'
        });
    };
    client.getItem = (jid, node, id) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                context: 'user',
                fetch: {
                    items: [{ id }],
                    node
                }
            },
            to: jid,
            type: 'get'
        });
        return resp.pubsub.fetch.items[0];
    });
    client.getItems = (jid, node, opts = {}) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                context: 'user',
                fetch: {
                    max: opts.max,
                    node
                },
                paging: opts
            },
            to: jid,
            type: 'get'
        });
        const result = resp.pubsub.fetch;
        result.paging = resp.pubsub.paging;
        return result;
    });
    client.retract = (jid, node, id, notify) => {
        return client.sendIQ({
            pubsub: {
                context: 'user',
                retract: {
                    id,
                    node,
                    notify
                }
            },
            to: jid,
            type: 'set'
        });
    };
    client.purgeNode = (jid, node) => {
        return client.sendIQ({
            pubsub: {
                context: 'owner',
                purge: node
            },
            to: jid,
            type: 'set'
        });
    };
    client.deleteNode = (jid, node, redirect) => {
        return client.sendIQ({
            pubsub: {
                context: 'owner',
                destroy: {
                    node,
                    redirect
                }
            },
            to: jid,
            type: 'set'
        });
    };
    client.createNode = (jid, node, config) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                configure: config
                    ? {
                        form: config
                    }
                    : undefined,
                context: 'user',
                create: {
                    node
                }
            },
            to: jid,
            type: 'set'
        });
        if (!resp.pubsub || !resp.pubsub.create) {
            return {
                node
            };
        }
        return resp.pubsub.create;
    });
    client.getSubscriptions = (jid, opts = {}) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                context: 'user',
                subscriptions: opts
            },
            to: jid,
            type: 'get'
        });
        return resp.pubsub.subscriptions;
    });
    client.getAffiliations = (jid, node) => {
        return client.sendIQ({
            pubsub: {
                affiliations: {
                    node
                }
            },
            to: jid,
            type: 'get'
        });
    };
    client.getNodeSubscribers = (jid, node, opts = {}) => {
        if (typeof node === 'string') {
            opts.node = node;
        }
        else {
            opts = Object.assign(Object.assign({}, opts), node);
        }
        return client.sendIQ({
            pubsub: {
                context: 'owner',
                subscriptions: opts
            },
            to: jid,
            type: 'get'
        });
    };
    client.updateNodeSubscriptions = (jid, node, delta) => {
        return client.sendIQ({
            pubsub: {
                context: 'owner',
                subscriptions: {
                    items: delta,
                    node
                }
            },
            to: jid,
            type: 'set'
        });
    };
    client.getNodeAffiliations = (jid, node) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                affiliations: {
                    node
                },
                context: 'owner'
            },
            to: jid,
            type: 'get'
        });
        return resp.pubsub.affiliations;
    });
    client.updateNodeAffiliations = (jid, node, items) => {
        return client.sendIQ({
            pubsub: {
                affiliations: {
                    items,
                    node
                },
                context: 'owner'
            },
            to: jid,
            type: 'set'
        });
    };
    client.getNodeConfig = (jid, node) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                configure: {
                    node
                },
                context: 'owner'
            },
            to: jid,
            type: 'get'
        });
        return resp.pubsub.configure.form || {};
    });
    client.getDefaultNodeConfig = (jid) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                context: 'owner',
                defaultConfiguration: {}
            },
            to: jid,
            type: 'get'
        });
        return resp.pubsub.defaultConfiguration.form || {};
    });
    client.configureNode = (jid, node, config) => __awaiter(this, void 0, void 0, function* () {
        return client.sendIQ({
            pubsub: {
                configure: {
                    form: config,
                    node
                },
                context: 'owner'
            },
            to: jid,
            type: 'set'
        });
    });
    client.getDefaultSubscriptionOptions = (jid) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            pubsub: {
                defaultSubscriptionOptions: {}
            },
            to: jid,
            type: 'get'
        });
        return resp.pubsub.defaultSubscriptionOptions.form || {};
    });
}

function Roster (client) {
    client.on('iq:set:roster', iq => {
        const allowed = allowedResponders(client.jid);
        if (!allowed.has(iq.from)) {
            return client.sendIQError(iq, {
                error: {
                    condition: 'service-unavailable',
                    type: 'cancel'
                }
            });
        }
        client.emit('roster:update', iq);
        client.sendIQResult(iq);
    });
    client.on('iq:set:blockList', iq => {
        const allowed = allowedResponders(client.jid);
        if (!allowed.has(iq.from)) {
            return client.sendIQError(iq, {
                error: {
                    condition: 'service-unavailable',
                    type: 'cancel'
                }
            });
        }
        const blockList = iq.blockList;
        client.emit(blockList.action, {
            jids: blockList.jids || []
        });
        client.sendIQResult(iq);
    });
    client.getRoster = () => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            roster: {
                version: client.config.rosterVer
            },
            type: 'get'
        });
        if (resp.roster) {
            const version = resp.roster.version;
            if (version) {
                client.config.rosterVer = version;
                client.emit('roster:ver', version);
            }
            resp.roster.items = resp.roster.items || [];
            return resp.roster;
        }
        else {
            return { items: [] };
        }
    });
    client.updateRosterItem = (item) => __awaiter(this, void 0, void 0, function* () {
        yield client.sendIQ({
            roster: {
                items: [item]
            },
            type: 'set'
        });
    });
    client.removeRosterItem = (jid) => {
        return client.updateRosterItem({ jid, subscription: 'remove' });
    };
    client.subscribe = (jid) => {
        client.sendPresence({ type: 'subscribe', to: jid });
    };
    client.unsubscribe = (jid) => {
        client.sendPresence({ type: 'unsubscribe', to: jid });
    };
    client.acceptSubscription = (jid) => {
        client.sendPresence({ type: 'subscribed', to: jid });
    };
    client.denySubscription = (jid) => {
        client.sendPresence({ type: 'unsubscribed', to: jid });
    };
    client.getBlocked = () => __awaiter(this, void 0, void 0, function* () {
        const result = yield client.sendIQ({
            blockList: {
                action: 'list'
            },
            type: 'get'
        });
        return Object.assign({ jids: [] }, result.blockList);
    });
    function toggleBlock(action, jid) {
        return __awaiter(this, void 0, void 0, function* () {
            yield client.sendIQ({
                blockList: {
                    action,
                    jids: [jid]
                },
                type: 'set'
            });
        });
    }
    client.block = (jid) => __awaiter(this, void 0, void 0, function* () { return toggleBlock('block', jid); });
    client.unblock = (jid) => __awaiter(this, void 0, void 0, function* () { return toggleBlock('unblock', jid); });
    client.goInvisible = (probe = false) => __awaiter(this, void 0, void 0, function* () {
        yield client.sendIQ({
            type: 'set',
            visiblity: {
                probe,
                type: 'invisible'
            }
        });
    });
    client.goVisible = () => __awaiter(this, void 0, void 0, function* () {
        yield client.sendIQ({
            type: 'set',
            visiblity: {
                type: 'visible'
            }
        });
    });
}

function SASL (client) {
    client.registerFeature('sasl', 100, (features, done) => __awaiter(this, void 0, void 0, function* () {
        const mech = client.sasl.createMechanism(features.sasl.mechanisms);
        const saslHandler = (sasl) => __awaiter(this, void 0, void 0, function* () {
            if (!mech) {
                return;
            }
            switch (sasl.type) {
                case 'success': {
                    client.features.negotiated.sasl = true;
                    client.off('sasl', saslHandler);
                    client.emit('auth:success', client.config.credentials);
                    if (client.transport) {
                        client.transport.authenticated = true;
                    }
                    done('restart');
                    return;
                }
                case 'challenge': {
                    mech.processChallenge(sasl.value);
                    try {
                        const credentials = (yield client.getCredentials());
                        const resp = mech.createResponse(credentials);
                        if (resp || resp === '') {
                            client.send('sasl', {
                                type: 'response',
                                value: resp
                            });
                        }
                        else {
                            client.send('sasl', {
                                type: 'response'
                            });
                        }
                        const cacheable = mech.getCacheableCredentials();
                        if (cacheable) {
                            if (!client.config.credentials) {
                                client.config.credentials = {};
                            }
                            client.config.credentials = Object.assign(Object.assign({}, client.config.credentials), cacheable);
                            client.emit('credentials:update', client.config.credentials);
                        }
                    }
                    catch (err) {
                        console.error(err);
                        client.send('sasl', {
                            type: 'abort'
                        });
                    }
                    return;
                }
                case 'failure':
                case 'abort': {
                    client.off('sasl', saslHandler);
                    client.emit('auth:failed');
                    done('disconnect', 'authentication failed');
                    return;
                }
            }
        });
        if (!mech) {
            client.off('sasl', saslHandler);
            client.emit('auth:failed');
            return done('disconnect', 'authentication failed');
        }
        client.on('sasl', saslHandler);
        client.once('--reset-stream-features', () => {
            client.features.negotiated.sasl = false;
            client.off('sasl', saslHandler);
        });
        try {
            const credentials = (yield client.getCredentials());
            client.send('sasl', {
                mechanism: mech.name,
                type: 'auth',
                value: mech.createResponse(credentials)
            });
        }
        catch (err) {
            console.error(err);
            client.send('sasl', {
                type: 'abort'
            });
        }
    }));
}

function Sharing (client) {
    client.disco.addFeature(NS_BOB);
    client.getBits = (jid, cid) => __awaiter(this, void 0, void 0, function* () {
        const result = yield client.sendIQ({
            bits: {
                cid
            },
            to: jid,
            type: 'get'
        });
        return result.bits;
    });
    function getUploadParameters(jid) {
        return __awaiter(this, void 0, void 0, function* () {
            const disco = yield client.getDiscoInfo(jid);
            if (!disco.features || !disco.features.includes(NS_HTTP_UPLOAD_0)) {
                return;
            }
            let maxSize;
            for (const form of disco.extensions || []) {
                const fields = form.fields || [];
                if (fields.some(field => field.name === 'FORM_TYPE' && field.value === NS_HTTP_UPLOAD_0)) {
                    const sizeField = fields.find(field => field.name === 'max-file-size');
                    if (sizeField) {
                        maxSize = parseInt(sizeField.value, 10);
                    }
                    return {
                        jid,
                        maxSize
                    };
                }
            }
        });
    }
    client.getUploadService = (domain = getDomain(client.jid)) => __awaiter(this, void 0, void 0, function* () {
        const domainParameters = yield getUploadParameters(domain);
        if (domainParameters) {
            return domainParameters;
        }
        const disco = yield client.getDiscoItems(domain);
        for (const item of disco.items || []) {
            if (!item.jid) {
                continue;
            }
            const itemParameters = yield getUploadParameters(item.jid);
            if (itemParameters) {
                return itemParameters;
            }
        }
        throw new Error('No upload service discovered on: ' + domain);
    });
    client.getUploadSlot = (uploadService, uploadRequest) => __awaiter(this, void 0, void 0, function* () {
        const resp = yield client.sendIQ({
            httpUpload: Object.assign({ type: 'request' }, uploadRequest),
            to: uploadService,
            type: 'get'
        });
        return resp.httpUpload;
    });
}

function core(client) {
    client.use(Features);
    client.use(Disco);
    client.use(Bind);
    client.use(Connection);
    client.use(HostMeta);
    client.use(SASL);
}
function Plugins (client) {
    client.use(Account);
    client.use(Messaging);
    client.use(Avatar);
    client.use(Command);
    client.use(Entity);
    client.use(Jingle);
    client.use(MAM);
    client.use(MUC);
    client.use(PEP);
    client.use(PubSub);
    client.use(Roster);
    client.use(Sharing);
}

// ====================================================================
const Protocol$1a = {
    aliases: ['features.legacySession', 'iq.legacySession'],
    element: 'session',
    fields: {
        // draft-cridland-xmpp-session-01
        //
        // The <optional /> child is not yet standardized, but is
        // still widely deployed to reduce client start times.
        optional: childBoolean(null, 'optional')
    },
    namespace: NS_SESSION
};

// ====================================================================
const Protocol$19 = [
    {
        aliases: ['atomentry', ...pubsubItemContentAliases()],
        element: 'entry',
        fields: {
            id: childText(null, 'id'),
            published: childDate(null, 'published'),
            updated: childDate(null, 'updated')
        },
        namespace: NS_ATOM,
        type: NS_ATOM,
        typeField: 'itemType'
    },
    {
        element: 'summary',
        fields: {
            text: text(),
            type: attribute('type', 'text')
        },
        namespace: NS_ATOM,
        path: 'atomentry.summary'
    },
    {
        element: 'title',
        fields: {
            text: text(),
            type: attribute('type', 'text')
        },
        namespace: NS_ATOM,
        path: 'atomentry.title'
    },
    {
        aliases: [{ path: 'atomentry.links', multiple: true }],
        element: 'link',
        fields: {
            href: attribute('href'),
            mediaType: attribute('type'),
            rel: attribute('rel')
        },
        namespace: NS_ATOM
    },
    {
        aliases: [{ path: 'atomentry.authors', multiple: true }],
        element: 'author',
        fields: {
            name: childText(null, 'name'),
            uri: childText(null, 'uri'),
            email: childText(null, 'email')
        },
        namespace: NS_ATOM
    },
    {
        aliases: [{ path: 'atomentry.contributors', multiple: true }],
        element: 'contributor',
        fields: {
            name: childText(null, 'name'),
            uri: childText(null, 'uri'),
            email: childText(null, 'email')
        },
        namespace: NS_ATOM
    },
    {
        aliases: [{ path: 'atomentry.categories', multiple: true }],
        element: 'category',
        fields: {
            term: attribute('term'),
            scheme: attribute('scheme'),
            label: attribute('label')
        },
        namespace: NS_ATOM
    },
    {
        element: 'content',
        fields: {
            text: text(),
            type: attribute('type', 'text')
        },
        namespace: NS_ATOM,
        path: 'atomentry.content'
    },
    {
        element: 'rights',
        fields: {
            text: text(),
            type: attribute('type', 'text')
        },
        namespace: NS_ATOM,
        path: 'atomentry.rights'
    }
];

// ====================================================================
const _Stream = {
    defaultType: 'stream',
    element: 'stream',
    fields: {
        from: attribute('from'),
        id: attribute('id'),
        lang: languageAttribute(),
        to: attribute('to'),
        version: attribute('version')
    },
    namespace: NS_STREAM,
    path: 'stream',
    type: 'stream',
    typeField: 'action'
};
const _StreamFeatures = {
    element: 'features',
    namespace: NS_STREAM,
    path: 'features'
};
const _StreamError = {
    element: 'error',
    fields: {
        alternateLanguageText: childAlternateLanguageText(NS_STREAMS, 'text'),
        condition: childEnum(NS_STREAMS, Object.values(StreamErrorCondition), StreamErrorCondition.UndefinedCondition),
        seeOtherHost: childText(NS_STREAMS, StreamErrorCondition.SeeOtherHost),
        text: childText(NS_STREAMS, 'text')
    },
    namespace: NS_STREAM,
    path: 'streamError'
};
// --------------------------------------------------------------------
const _StanzaError = Object.values(StreamType).map(streamNS => ({
    aliases: ['stanzaError', 'message.error', 'presence.error', 'iq.error'],
    defaultType: NS_CLIENT,
    element: 'error',
    fields: {
        alternateLanguageText: childAlternateLanguageText(NS_STANZAS, 'text'),
        by: JIDAttribute('by'),
        condition: childEnum(NS_STANZAS, Object.values(StanzaErrorCondition), StanzaErrorCondition.UndefinedCondition),
        gone: childText(NS_STANZAS, StanzaErrorCondition.Gone),
        redirect: childText(NS_STANZAS, StanzaErrorCondition.Redirect),
        text: childText(NS_STANZAS, 'text'),
        type: attribute('type')
    },
    namespace: streamNS,
    type: streamNS,
    typeField: 'streamType'
}));
// --------------------------------------------------------------------
const baseIQFields = new Set([
    'from',
    'id',
    'lang',
    'to',
    'type',
    'payloadType',
    'error',
    'streamType'
]);
const _IQ = Object.values(StreamType).map((streamNS) => ({
    childrenExportOrder: {
        error: 200000
    },
    defaultType: NS_CLIENT,
    element: 'iq',
    fields: {
        from: JIDAttribute('from'),
        id: attribute('id'),
        lang: languageAttribute(),
        payloadType: {
            order: -10000,
            importer(xml, context) {
                if (context.data.type !== 'get' &&
                    context.data.type !== 'set') {
                    return;
                }
                const childCount = xml.children.filter(child => typeof child !== 'string')
                    .length;
                if (childCount !== 1) {
                    return 'invalid-payload-count';
                }
                const extensions = Object.keys(context.data).filter(key => !baseIQFields.has(key));
                if (extensions.length !== 1) {
                    return 'unknown-payload';
                }
                return extensions[0];
            }
        },
        to: JIDAttribute('to'),
        type: attribute('type')
    },
    namespace: streamNS,
    path: 'iq',
    type: streamNS,
    typeField: 'streamType'
}));
// --------------------------------------------------------------------
const _Message = Object.values(StreamType).map(streamNS => ({
    childrenExportOrder: {
        error: 200000
    },
    defaultType: NS_CLIENT,
    element: 'message',
    fields: {
        from: JIDAttribute('from'),
        id: attribute('id'),
        lang: languageAttribute(),
        to: JIDAttribute('to')
    },
    namespace: streamNS,
    path: 'message',
    type: streamNS,
    typeField: 'streamType'
}));
// --------------------------------------------------------------------
const _Presence = Object.values(StreamType).map(streamNS => ({
    childrenExportOrder: {
        error: 200000
    },
    defaultType: NS_CLIENT,
    element: 'presence',
    fields: {
        from: JIDAttribute('from'),
        id: attribute('id'),
        lang: languageAttribute(),
        to: JIDAttribute('to')
    },
    namespace: streamNS,
    path: 'presence',
    type: streamNS,
    typeField: 'streamType'
}));
// --------------------------------------------------------------------
const _SASL = [
    {
        element: 'mechanisms',
        fields: {
            mechanisms: multipleChildText(null, 'mechanism')
        },
        namespace: NS_SASL,
        path: 'features.sasl'
    },
    {
        element: 'abort',
        namespace: NS_SASL,
        path: 'sasl',
        type: 'abort',
        typeField: 'type'
    },
    {
        element: 'auth',
        fields: {
            mechanism: attribute('mechanism'),
            value: textBuffer('base64')
        },
        namespace: NS_SASL,
        path: 'sasl',
        type: 'auth',
        typeField: 'type'
    },
    {
        element: 'challenge',
        fields: {
            value: textBuffer('base64')
        },
        namespace: NS_SASL,
        path: 'sasl',
        type: 'challenge',
        typeField: 'type'
    },
    {
        element: 'response',
        fields: {
            value: textBuffer('base64')
        },
        namespace: NS_SASL,
        path: 'sasl',
        type: 'response',
        typeField: 'type'
    },
    {
        element: 'success',
        fields: {
            value: textBuffer('base64')
        },
        namespace: NS_SASL,
        path: 'sasl',
        type: 'success',
        typeField: 'type'
    },
    {
        element: 'failure',
        fields: {
            alternateLanguageText: childAlternateLanguageText(NS_SASL, 'text'),
            condition: childEnum(NS_SASL, Object.values(SASLFailureCondition)),
            text: childText(NS_SASL, 'text')
        },
        namespace: NS_SASL,
        path: 'sasl',
        type: 'failure',
        typeField: 'type'
    }
];
// --------------------------------------------------------------------
const _STARTTLS = [
    {
        aliases: ['features.tls'],
        defaultType: 'start',
        element: 'starttls',
        fields: {
            required: childBoolean(null, 'required')
        },
        namespace: NS_STARTTLS,
        path: 'tls',
        type: 'start',
        typeField: 'type'
    },
    {
        element: 'proceed',
        namespace: NS_STARTTLS,
        path: 'tls',
        type: 'proceed',
        typeField: 'type'
    },
    {
        element: 'failure',
        namespace: NS_STARTTLS,
        path: 'tls',
        type: 'failure',
        typeField: 'type'
    }
];
// --------------------------------------------------------------------
const _Bind = {
    aliases: ['features.bind', 'iq.bind'],
    element: 'bind',
    fields: {
        jid: childText(null, 'jid'),
        resource: childText(null, 'resource')
    },
    namespace: NS_BIND
};
// --------------------------------------------------------------------
const Protocol$18 = [
    _Stream,
    _StreamFeatures,
    _StreamError,
    ..._StanzaError,
    ..._SASL,
    ..._STARTTLS,
    ..._IQ,
    ..._Message,
    ..._Presence,
    _Bind
];

// ====================================================================
const Protocol$17 = [
    extendStreamFeatures({
        rosterPreApproval: childBoolean(NS_SUBSCRIPTION_PREAPPROVAL, 'sub'),
        rosterVersioning: childBoolean(NS_ROSTER_VERSIONING, 'ver')
    }),
    extendMessage({
        alternateLanguageBodies: childAlternateLanguageText(null, 'body'),
        alternateLanguageSubjects: childAlternateLanguageText(null, 'subject'),
        body: childText(null, 'body'),
        hasSubject: childBoolean(null, 'subject'),
        parentThread: childAttribute(null, 'thread', 'parent'),
        subject: childText(null, 'subject'),
        thread: childText(null, 'thread'),
        type: attribute('type')
    }),
    extendPresence({
        alternateLanguageStatuses: childAlternateLanguageText(null, 'status'),
        priority: childInteger(null, 'priority', 0),
        show: childText(null, 'show'),
        status: childText(null, 'status'),
        type: attribute('type')
    }),
    {
        element: 'query',
        fields: {
            version: attribute('ver', undefined, { emitEmpty: true })
        },
        namespace: NS_ROSTER,
        path: 'iq.roster'
    },
    {
        aliases: [{ path: 'iq.roster.items', multiple: true }],
        element: 'item',
        fields: {
            groups: multipleChildText(null, 'group'),
            jid: JIDAttribute('jid'),
            name: attribute('name'),
            pending: attribute('ask'),
            preApproved: booleanAttribute('approved'),
            subscription: attribute('subscription')
        },
        namespace: NS_ROSTER
    }
];

// ====================================================================
const Protocol$16 = [
    {
        element: 'open',
        fields: {
            from: attribute('from'),
            id: attribute('id'),
            lang: languageAttribute(),
            to: attribute('to'),
            version: attribute('version')
        },
        namespace: NS_FRAMING,
        path: 'stream',
        type: 'open'
    },
    {
        element: 'close',
        fields: {
            seeOtherURI: attribute('see-other-uri')
        },
        namespace: NS_FRAMING,
        path: 'stream',
        type: 'close'
    }
];

// ====================================================================
const Protocol$15 = [
    {
        aliases: [{ path: 'message.forms', multiple: true }],
        element: 'x',
        fields: {
            instructions: Object.assign(Object.assign({}, multipleChildText(null, 'instructions')), { exportOrder: 2 }),
            reported: Object.assign(Object.assign({}, splicePath(null, 'reported', 'dataformField', true)), { exportOrder: 3 }),
            title: Object.assign(Object.assign({}, childText(null, 'title')), { exportOrder: 1 }),
            type: attribute('type')
        },
        namespace: NS_DATAFORM,
        optionalNamespaces: {
            xdv: NS_DATAFORM_VALIDATION
        },
        path: 'dataform'
    },
    {
        aliases: [
            { path: 'dataform.fields', multiple: true },
            { path: 'dataform.items.fields', multiple: true }
        ],
        element: 'field',
        fields: {
            description: childText(null, 'desc'),
            label: attribute('label'),
            name: attribute('var'),
            rawValues: Object.assign(Object.assign({}, multipleChildText(null, 'value')), { exporter: () => null }),
            required: childBoolean(null, 'required'),
            type: attribute('type'),
            value: {
                importer(xml, context) {
                    const fieldType = xml.getAttribute('type');
                    const converter = multipleChildText(NS_DATAFORM, 'value');
                    const rawValues = converter.importer(xml, context);
                    const singleValue = rawValues[0];
                    switch (fieldType) {
                        case DataFormFieldType.TextMultiple:
                        case DataFormFieldType.ListMultiple:
                        case DataFormFieldType.JIDMultiple:
                            return rawValues;
                        case DataFormFieldType.Hidden:
                        case DataFormFieldType.Fixed:
                            if (rawValues.length === 1) {
                                return singleValue;
                            }
                            return rawValues;
                        case DataFormFieldType.Boolean:
                            if (singleValue) {
                                return singleValue === '1' || singleValue === 'true';
                            }
                            break;
                        default:
                            return singleValue;
                    }
                },
                exporter(xml, data, context) {
                    const converter = multipleChildText(null, 'value');
                    let outputData = [];
                    const rawData = context.data && context.data.rawValues
                        ? context.data.rawValues[0]
                        : undefined;
                    if (typeof data === 'boolean') {
                        if (rawData === 'true' || rawData === 'false') {
                            outputData = [rawData];
                        }
                        else {
                            outputData = [data ? '1' : '0'];
                        }
                    }
                    else if (!Array.isArray(data)) {
                        outputData = [data.toString()];
                    }
                    else {
                        for (const value of data) {
                            outputData.push(value.toString());
                        }
                    }
                    converter.exporter(xml, outputData, Object.assign({}, context, {
                        namespace: NS_DATAFORM
                    }));
                }
            }
        },
        namespace: NS_DATAFORM,
        path: 'dataformField'
    },
    {
        aliases: [{ path: 'dataform.fields.options', multiple: true }],
        element: 'option',
        fields: {
            label: attribute('label'),
            value: childText(null, 'value')
        },
        namespace: NS_DATAFORM
    },
    {
        aliases: [{ path: 'dataform.items', multiple: true }],
        element: 'item',
        namespace: NS_DATAFORM
    },
    // ----------------------------------------------------------------
    // XEP-0122: Data Forms Validation
    // ----------------------------------------------------------------
    {
        element: 'validate',
        fields: {
            listMax: childIntegerAttribute(null, 'list-range', 'max'),
            listMin: childIntegerAttribute(null, 'list-range', 'min'),
            method: childEnum(null, ['basic', 'open', 'range', 'regex'], 'basic'),
            rangeMax: childAttribute(null, 'range', 'max'),
            rangeMin: childAttribute(null, 'range', 'min'),
            regex: childText(null, 'regex'),
            type: attribute('datatype', 'xs:string')
        },
        namespace: NS_DATAFORM_VALIDATION,
        path: 'dataform.fields.validation'
    }
];

// ====================================================================
const Protocol$14 = [
    {
        aliases: ['presence.legacyLastActivity', 'iq.lastActivity'],
        element: 'query',
        fields: {
            seconds: integerAttribute('seconds'),
            status: text()
        },
        namespace: NS_LAST_ACTIVITY
    }
];

// ====================================================================
const Protocol$13 = [
    {
        element: 'query',
        fields: {
            activeList: childAttribute(null, 'active', 'name'),
            defaultList: childAttribute(null, 'default', 'name')
        },
        namespace: NS_PRIVACY,
        path: 'iq.privacy'
    },
    {
        aliases: [{ path: 'iq.privacy.lists', multiple: true }],
        element: 'list',
        fields: {
            name: attribute('name')
        },
        namespace: NS_PRIVACY
    },
    {
        aliases: [{ path: 'iq.privacy.lists.items', multiple: true }],
        element: 'item',
        fields: {
            action: attribute('action'),
            incomingPresence: childBoolean(null, 'presence-in'),
            iq: childBoolean(null, 'iq'),
            messages: childBoolean(null, 'message'),
            order: integerAttribute('order'),
            outgoingPresence: childBoolean(null, 'presence-out'),
            type: attribute('type'),
            value: attribute('value')
        },
        namespace: NS_PRIVACY
    }
];

// ====================================================================
const Protocol$12 = [
    {
        aliases: ['iq.disco', 'message.disco', 'features.disco'],
        childrenExportOrder: {
            identities: 100
        },
        element: 'query',
        fields: {
            features: multipleChildAttribute(null, 'feature', 'var'),
            node: attribute('node')
        },
        namespace: NS_DISCO_INFO,
        path: 'disco',
        type: 'info',
        typeField: 'type'
    },
    {
        aliases: [{ path: 'disco.identities', selector: 'info', multiple: true }],
        element: 'identity',
        fields: {
            category: attribute('category'),
            lang: languageAttribute(),
            name: attribute('name'),
            type: attribute('type')
        },
        namespace: NS_DISCO_INFO
    },
    {
        aliases: [{ path: 'disco.items', multiple: true, selector: 'items' }],
        element: 'item',
        fields: {
            jid: JIDAttribute('jid'),
            name: attribute('name'),
            node: attribute('node')
        },
        namespace: NS_DISCO_ITEMS
    },
    {
        aliases: [{ path: 'disco.items', multiple: true, selector: 'info' }],
        element: 'item',
        fields: {
            category: JIDAttribute('category'),
            lang: languageAttribute(),
            name: attribute('name'),
            type: attribute('type')
        },
        namespace: NS_DISCO_INFO
    },
    addAlias(NS_DATAFORM, 'x', [
        // XEP-0128
        { path: 'disco.extensions', multiple: true, selector: 'info' }
    ]),
    addAlias(NS_RSM, 'set', [{ path: 'disco.paging', selector: 'items' }]),
    {
        aliases: ['iq.disco', 'message.disco', 'features.disco'],
        element: 'query',
        fields: {
            node: attribute('node')
        },
        namespace: NS_DISCO_ITEMS,
        path: 'disco',
        type: 'items',
        typeField: 'type'
    }
];

// ====================================================================
const Protocol$11 = [
    extendMessage({
        addresses: splicePath(NS_ADDRESS, 'addresses', 'extendedAddress', true)
    }),
    extendPresence({
        addresses: splicePath(NS_ADDRESS, 'addresses', 'extendedAddress', true)
    }),
    {
        element: 'address',
        fields: {
            alternateLanguageDescriptions: childAlternateLanguageText(null, 'desc'),
            delivered: booleanAttribute('delivered'),
            description: attribute('desc'),
            jid: JIDAttribute('jid'),
            node: attribute('node'),
            type: attribute('type'),
            uri: attribute('uri')
        },
        namespace: NS_ADDRESS,
        path: 'extendedAddress'
    }
];

// ====================================================================
const Protocol$10 = [
    addAlias(NS_DATAFORM, 'x', [{ path: 'iq.muc.form', selector: 'configure' }]),
    {
        defaultType: 'info',
        element: 'x',
        fields: {
            password: childText(null, 'password')
        },
        namespace: NS_MUC,
        path: 'presence.muc',
        type: 'join',
        typeField: 'type'
    },
    {
        aliases: [{ path: 'presence.muc.history', selector: 'join' }],
        element: 'history',
        fields: {
            maxCharacters: integerAttribute('maxchars'),
            maxStanzas: integerAttribute('maxstanzas'),
            seconds: integerAttribute('seconds'),
            since: dateAttribute('since')
        },
        namespace: NS_MUC
    },
    {
        aliases: ['presence.muc', 'message.muc'],
        defaultType: 'info',
        element: 'x',
        fields: {
            action: childEnum(null, ['invite', 'decline', 'destroy']),
            actor: splicePath(null, 'item', 'mucactor'),
            affiliation: childAttribute(null, 'item', 'affiliation'),
            jid: childJIDAttribute(null, 'item', 'jid'),
            nick: childAttribute(null, 'item', 'nick'),
            password: childText(null, 'password'),
            reason: deepChildText([
                { namespace: null, element: 'item' },
                { namespace: null, element: 'reason' }
            ]),
            role: childAttribute(null, 'item', 'role'),
            statusCodes: multipleChildAttribute(null, 'status', 'code')
        },
        namespace: NS_MUC_USER,
        type: 'info',
        typeField: 'type',
        typeOrder: 1
    },
    {
        element: 'actor',
        fields: {
            jid: JIDAttribute('jid'),
            nick: attribute('nick')
        },
        namespace: NS_MUC_USER,
        path: 'mucactor'
    },
    {
        element: 'destroy',
        fields: {
            jid: JIDAttribute('jid'),
            password: childText(null, 'password'),
            reason: childText(null, 'reason')
        },
        namespace: NS_MUC_USER,
        path: 'presence.muc.destroy'
    },
    {
        aliases: [{ path: 'message.muc.invite', multiple: true }],
        element: 'invite',
        fields: {
            continue: childBoolean(null, 'continue'),
            from: JIDAttribute('from'),
            reason: childText(null, 'reason'),
            thread: childAttribute(null, 'continue', 'thread'),
            to: JIDAttribute('to')
        },
        namespace: NS_MUC_USER
    },
    {
        element: 'decline',
        fields: {
            from: JIDAttribute('from'),
            reason: childText(null, 'reason'),
            to: JIDAttribute('to')
        },
        namespace: NS_MUC_USER,
        path: 'message.muc',
        type: 'decline'
    },
    {
        element: 'query',
        namespace: NS_MUC_ADMIN,
        path: 'iq.muc',
        type: 'user-list',
        typeField: 'type'
    },
    {
        aliases: [{ path: 'iq.muc.users', multiple: true, selector: 'user-list' }],
        element: 'item',
        fields: {
            affiliation: attribute('affiliation'),
            jid: JIDAttribute('jid'),
            nick: attribute('nick'),
            reason: childText(null, 'reason'),
            role: attribute('role')
        },
        namespace: NS_MUC_ADMIN
    },
    {
        aliases: ['iq.muc.users.actor'],
        element: 'actor',
        fields: {
            jid: JIDAttribute('jid'),
            nick: attribute('nick')
        },
        namespace: NS_MUC_ADMIN
    },
    {
        element: 'query',
        namespace: NS_MUC_OWNER,
        path: 'iq.muc',
        type: 'configure',
        typeField: 'type'
    },
    {
        aliases: [{ path: 'iq.muc.destroy', selector: 'configure' }],
        element: 'destroy',
        fields: {
            jid: JIDAttribute('jid'),
            password: childText(null, 'password'),
            reason: childText(null, 'reason')
        },
        namespace: NS_MUC_OWNER
    },
    // XEP-0249
    {
        element: 'x',
        fields: {
            action: staticValue('invite'),
            continue: booleanAttribute('continue'),
            jid: JIDAttribute('jid'),
            legacyReason: text(),
            password: attribute('password'),
            reason: attribute('reason'),
            thread: attribute('thread')
        },
        namespace: NS_MUC_DIRECT_INVITE,
        path: 'message.muc',
        type: 'direct-invite',
        typeOrder: 2
    },
    // XEP-0307
    {
        element: 'unique',
        fields: {
            name: text()
        },
        namespace: NS_MUC_UNIQUE,
        path: 'iq.muc',
        type: 'unique'
    },
    extendMessage({
        legacyMUC: {
            exporter(xml, value, context) {
                const out = context.registry
                    ? context.registry.export('message.muc', Object.assign(Object.assign({}, value), { type: 'direct-invite' }))
                    : undefined;
                if (out) {
                    xml.appendChild(out);
                }
            },
            exportOrder: 100001,
            importer(xml, context) {
                const mucElement = findAll(xml, NS_MUC_USER, 'x')[0];
                if (!mucElement) {
                    return;
                }
                const confElement = findAll(xml, NS_MUC_DIRECT_INVITE, 'x')[0];
                if (!confElement) {
                    return;
                }
                return context.registry
                    ? context.registry.import(confElement, Object.assign(Object.assign({}, context), { path: 'message' }))
                    : undefined;
            },
            importOrder: -1
        }
    })
];

// ====================================================================
const Protocol$$ = [
    {
        aliases: ['iq.ibb', 'message.ibb'],
        element: 'open',
        fields: {
            ack: {
                importer(xml, context) {
                    const stanza = attribute('stanza', 'iq').importer(xml, context);
                    return stanza !== 'message';
                },
                exporter(xml, data, context) {
                    attribute('stanza').exporter(xml, data ? 'iq' : 'message', context);
                }
            },
            blockSize: integerAttribute('block-size'),
            sid: attribute('sid')
        },
        namespace: NS_IBB,
        type: 'open',
        typeField: 'action'
    },
    {
        aliases: ['iq.ibb', 'message.ibb'],
        element: 'close',
        fields: {
            sid: attribute('sid')
        },
        namespace: NS_IBB,
        type: 'close',
        typeField: 'action'
    },
    {
        aliases: ['iq.ibb', 'message.ibb'],
        element: 'data',
        fields: {
            data: textBuffer('base64'),
            seq: integerAttribute('seq'),
            sid: attribute('sid')
        },
        namespace: NS_IBB,
        type: 'data',
        typeField: 'action'
    }
];

// ====================================================================
const Protocol$_ = [
    {
        aliases: [
            { path: 'bookmarkStorage', impliedType: true },
            { path: 'iq.privateStorage.bookmarks', impliedType: true },
            ...pubsubItemContentAliases()
        ],
        element: 'storage',
        namespace: NS_BOOKMARKS,
        type: NS_BOOKMARKS,
        typeField: 'itemType'
    },
    {
        aliases: [{ path: 'bookmarkStorage.rooms', multiple: true }],
        element: 'conference',
        fields: {
            autoJoin: booleanAttribute('autojoin'),
            jid: JIDAttribute('jid'),
            name: attribute('name'),
            nick: childText(null, 'nick'),
            password: childText(null, 'password')
        },
        namespace: NS_BOOKMARKS
    }
];

// ====================================================================
// tslint:enable
const Protocol$Z = {
    element: 'query',
    namespace: NS_PRIVATE,
    path: 'iq.privateStorage'
};

// ====================================================================
const Protocol$Y = [
    addAlias(NS_DATAFORM, 'x', ['iq.command.form']),
    extendStanzaError({
        commandError: childEnum(NS_ADHOC_COMMANDS, [
            'bad-action',
            'bad-locale',
            'bad-payload',
            'bad-sessionid',
            'malformed-action',
            'session-expired'
        ])
    }),
    {
        element: 'command',
        fields: {
            action: attribute('action'),
            node: attribute('node'),
            sid: attribute('sessionid'),
            status: attribute('status')
        },
        namespace: NS_ADHOC_COMMANDS,
        path: 'iq.command'
    },
    {
        element: 'actions',
        fields: {
            complete: childBoolean(null, 'complete'),
            execute: attribute('execute'),
            next: childBoolean(null, 'next'),
            prev: childBoolean(null, 'prev')
        },
        namespace: NS_ADHOC_COMMANDS,
        path: 'iq.command.availableActions'
    },
    {
        aliases: [{ path: 'iq.command.notes', multiple: true }],
        element: 'note',
        fields: {
            type: attribute('type'),
            value: text()
        },
        namespace: NS_ADHOC_COMMANDS
    }
];

// ====================================================================
const path$1 = 'vcardTemp.records';
function vcardField(element, type) {
    return {
        aliases: [{ multiple: true, path: path$1 }],
        element,
        fields: {
            value: text()
        },
        namespace: NS_VCARD_TEMP,
        type,
        typeField: 'type'
    };
}
const Protocol$X = [
    {
        aliases: [{ path: 'iq.vcard' }],
        defaultType: NS_VCARD_TEMP,
        element: 'vCard',
        fields: {
            fullName: childText(null, 'FN')
        },
        namespace: NS_VCARD_TEMP,
        path: 'vcardTemp',
        type: NS_VCARD_TEMP,
        typeField: 'format'
    },
    {
        element: 'N',
        fields: {
            additional: Object.assign(Object.assign({}, childText(null, 'MIDDLE')), { order: 3 }),
            family: Object.assign(Object.assign({}, childText(null, 'FAMILY')), { order: 1 }),
            given: Object.assign(Object.assign({}, childText(null, 'GIVEN')), { order: 2 }),
            prefix: Object.assign(Object.assign({}, childText(null, 'PREFIX')), { order: 4 }),
            suffix: Object.assign(Object.assign({}, childText(null, 'SUFFIX')), { order: 5 })
        },
        namespace: NS_VCARD_TEMP,
        path: 'vcardTemp.name'
    },
    vcardField('NICKNAME', 'nickname'),
    vcardField('BDAY', 'birthday'),
    vcardField('JABBERID', 'jid'),
    vcardField('TZ', 'timezone'),
    vcardField('TITLE', 'title'),
    vcardField('ROLE', 'role'),
    vcardField('URL', 'url'),
    vcardField('NOTE', 'note'),
    vcardField('SORT-STRING', 'sort'),
    vcardField('UID', 'uid'),
    vcardField('REV', 'revision'),
    vcardField('PRODID', 'productId'),
    vcardField('DESC', 'description'),
    {
        aliases: [{ multiple: true, path: path$1 }],
        element: 'EMAIL',
        fields: {
            preferred: childBoolean(null, 'PREF'),
            types: multipleChildEnum(null, [
                ['home', 'HOME'],
                ['work', 'WORK'],
                ['internet', 'INTERNET']
            ]),
            value: childText(null, 'USERID')
        },
        namespace: NS_VCARD_TEMP,
        type: 'email'
    },
    {
        aliases: [{ path: path$1, multiple: true }],
        element: 'ORG',
        fields: {
            units: Object.assign(Object.assign({}, multipleChildText(null, 'ORGUNIT')), { order: 2 }),
            value: Object.assign(Object.assign({}, childText(null, 'ORGNAME')), { order: 1 })
        },
        namespace: NS_VCARD_TEMP,
        type: 'organization',
        typeField: 'type'
    },
    {
        aliases: [{ multiple: true, path: path$1 }],
        element: 'ADR',
        fields: {
            city: childText(null, 'LOCALITY'),
            code: childText(null, 'PCODE'),
            country: childText(null, 'CTRY'),
            pobox: childText(null, 'POBOX'),
            preferred: childBoolean(null, 'PREF'),
            region: childText(null, 'REGION'),
            street: childText(null, 'STREET'),
            street2: childText(null, 'EXTADD'),
            types: multipleChildEnum(null, [
                ['home', 'HOME'],
                ['work', 'WORK'],
                ['domestic', 'DOM'],
                ['international', 'INTL'],
                ['postal', 'POSTAL'],
                ['parcel', 'PARCEL']
            ])
        },
        namespace: NS_VCARD_TEMP,
        type: 'address',
        typeField: 'type'
    },
    {
        aliases: [{ multiple: true, path: path$1 }],
        element: 'LABEL',
        fields: {
            lines: multipleChildText(null, 'LINE'),
            preferred: childBoolean(null, 'PREF'),
            types: multipleChildEnum(null, [
                ['home', 'HOME'],
                ['work', 'WORK']
            ])
        },
        namespace: NS_VCARD_TEMP,
        type: 'addressLabel',
        typeField: 'type'
    },
    {
        aliases: [{ multiple: true, path: path$1 }],
        element: 'TEL',
        fields: {
            preferred: childBoolean(null, 'PREF'),
            types: multipleChildEnum(null, [
                ['home', 'HOME'],
                ['work', 'WORK'],
                ['cell', 'CELL'],
                ['fax', 'FAX'],
                ['voice', 'VOICE'],
                ['msg', 'MSG']
            ]),
            value: childText(null, 'NUMBER', '', true)
        },
        namespace: NS_VCARD_TEMP,
        type: 'tel',
        typeField: 'type'
    },
    {
        aliases: [{ multiple: true, path: path$1 }],
        element: 'PHOTO',
        fields: {
            data: childText(null, 'BINVAL'),
            mediaType: childText(null, 'TYPE'),
            url: childText(null, 'EXTVAL')
        },
        namespace: NS_VCARD_TEMP,
        type: 'photo',
        typeField: 'type'
    },
    {
        aliases: [{ multiple: true, path: path$1 }],
        element: 'LOGO',
        fields: {
            data: childText(null, 'BINVAL'),
            mediaType: childText(null, 'TYPE'),
            url: childText(null, 'EXTVAL')
        },
        namespace: NS_VCARD_TEMP,
        type: 'logo',
        typeField: 'type'
    },
    {
        aliases: [{ multiple: true, path: path$1 }],
        element: 'CATEGORIES',
        fields: {
            value: multipleChildText(null, 'KEYWORD')
        },
        namespace: NS_VCARD_TEMP,
        type: 'categories',
        typeField: 'type'
    }
];

// ====================================================================
const Protocol$W = [
    addAlias(NS_DATAFORM, 'x', ['iq.search.form']),
    addAlias(NS_RSM, 'set', ['iq.search.paging']),
    {
        element: 'query',
        fields: {
            email: childText(null, 'email'),
            familyName: childText(null, 'last'),
            givenName: childText(null, 'first'),
            instructions: childText(null, 'instructions'),
            nick: childText(null, 'nick')
        },
        namespace: NS_SEARCH,
        path: 'iq.search'
    },
    {
        aliases: [{ path: 'iq.search.items', multiple: true }],
        element: 'item',
        fields: {
            email: childText(null, 'email'),
            familyName: childText(null, 'last'),
            givenName: childText(null, 'first'),
            jid: JIDAttribute('jid'),
            nick: childText(null, 'nick')
        },
        namespace: NS_SEARCH
    }
];

// ====================================================================
const Protocol$V = {
    aliases: ['iq.pubsub.paging'],
    element: 'set',
    fields: {
        after: childText(null, 'after'),
        before: childText(null, 'before'),
        count: childInteger(null, 'count'),
        first: childText(null, 'first'),
        firstIndex: childIntegerAttribute(null, 'first', 'index'),
        index: childInteger(null, 'index'),
        last: childText(null, 'last'),
        max: childInteger(null, 'max')
    },
    namespace: NS_RSM,
    path: 'paging'
};

// ====================================================================
const dateOrPresenceAttribute = (name) => ({
    importer(xml) {
        const data = xml.getAttribute(name);
        if (data === 'presence') {
            return data;
        }
        if (data) {
            return new Date(data);
        }
    },
    exporter(xml, value) {
        let data;
        if (typeof value === 'string') {
            data = value;
        }
        else {
            data = value.toISOString();
        }
        xml.setAttribute(name, data);
    }
});
const SubscriptionFields = {
    configurable: childBoolean(null, 'subscribe-options'),
    configurationRequired: deepChildBoolean([
        { namespace: null, element: 'subscribe-options' },
        { namespace: null, element: 'required' }
    ]),
    jid: JIDAttribute('jid'),
    node: attribute('node'),
    state: attribute('subscription'),
    subid: attribute('subid')
};
const NodeOnlyField = {
    node: attribute('node')
};
const Protocol$U = [
    {
        aliases: ['pubsub', 'iq.pubsub', 'message.pubsub'],
        childrenExportOrder: {
            configure: 0,
            create: 100,
            publish: 100,
            subscribe: 100,
            subscriptionOptions: 0
        },
        defaultType: 'user',
        element: 'pubsub',
        fields: {
            publishOptions: splicePath(null, 'publish-options', 'dataform')
        },
        namespace: NS_PUBSUB,
        type: 'user',
        typeField: 'context'
    },
    {
        aliases: ['pubsub', 'iq.pubsub', 'message.pubsub'],
        defaultType: 'user',
        element: 'pubsub',
        fields: {
            purge: childAttribute(null, 'purge', 'node')
        },
        namespace: NS_PUBSUB_OWNER,
        type: 'owner',
        typeField: 'context'
    },
    addAlias(NS_DATAFORM, 'x', [
        'iq.pubsub.configure.form',
        'iq.pubsub.defaultConfiguration.form',
        'iq.pubsub.defaultSubscriptionOptions.form',
        'iq.pubsub.subscriptionOptions.form',
        'message.pubsub.configuration.form'
    ]),
    addAlias(NS_RSM, 'set', ['iq.pubsub.fetch.paging']),
    extendStanzaError({
        pubsubError: childEnum(NS_PUBSUB_ERRORS, Object.values(PubsubErrorCondition)),
        pubsubUnsupportedFeature: childAttribute(NS_PUBSUB_ERRORS, 'unsupported', 'feature')
    }),
    {
        element: 'subscribe',
        fields: {
            jid: JIDAttribute('jid'),
            node: attribute('node')
        },
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.subscribe'
    },
    {
        element: 'unsubscribe',
        fields: {
            jid: JIDAttribute('jid'),
            node: attribute('node'),
            subid: attribute('subid')
        },
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.unsubscribe'
    },
    {
        element: 'options',
        fields: {
            jid: JIDAttribute('jid'),
            node: attribute('node'),
            subid: attribute('subid')
        },
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.subscriptionOptions'
    },
    {
        aliases: [{ path: 'iq.pubsub.subscriptions', selector: 'user', impliedType: true }],
        element: 'subscriptions',
        fields: {
            jid: JIDAttribute('jid'),
            node: attribute('node')
        },
        namespace: NS_PUBSUB,
        type: 'user'
    },
    {
        aliases: [{ path: 'iq.pubsub.subscriptions', selector: 'owner', impliedType: true }],
        element: 'subscriptions',
        fields: {
            jid: JIDAttribute('jid'),
            node: attribute('node')
        },
        namespace: NS_PUBSUB_OWNER,
        type: 'owner'
    },
    {
        aliases: [
            { path: 'iq.pubsub.subscription', selector: 'owner' },
            {
                impliedType: true,
                multiple: true,
                path: 'iq.pubsub.subscriptions.items',
                selector: 'owner'
            }
        ],
        element: 'subscription',
        fields: SubscriptionFields,
        namespace: NS_PUBSUB
    },
    {
        aliases: [
            { path: 'iq.pubsub.subscription', selector: 'user' },
            {
                impliedType: true,
                multiple: true,
                path: 'iq.pubsub.subscriptions.items',
                selector: 'user'
            }
        ],
        element: 'subscription',
        fields: SubscriptionFields,
        namespace: NS_PUBSUB,
        type: 'user'
    },
    {
        aliases: [
            {
                impliedType: true,
                multiple: true,
                path: 'iq.pubsub.subscriptions.items',
                selector: 'owner'
            }
        ],
        element: 'subscription',
        fields: SubscriptionFields,
        namespace: NS_PUBSUB_OWNER,
        type: 'owner'
    },
    {
        aliases: [
            { path: 'iq.pubsub.affiliations', selector: 'user', impliedType: true },
            { path: 'message.pubsub.affiliations', selector: 'user', impliedType: true }
        ],
        element: 'affiliations',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB,
        type: 'user'
    },
    {
        aliases: [{ path: 'iq.pubsub.affiliations', selector: 'owner', impliedType: true }],
        element: 'affiliations',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB_OWNER,
        type: 'owner'
    },
    {
        aliases: [
            {
                impliedType: true,
                multiple: true,
                path: 'iq.pubsub.affiliations.items',
                selector: 'user'
            },
            {
                impliedType: true,
                multiple: true,
                path: 'message.pubsub.affiliations.items',
                selector: 'user'
            }
        ],
        element: 'affiliation',
        fields: {
            affiliation: attribute('affiliation'),
            jid: JIDAttribute('jid'),
            node: attribute('node')
        },
        namespace: NS_PUBSUB,
        type: 'user'
    },
    {
        aliases: [
            {
                impliedType: true,
                multiple: true,
                path: 'iq.pubsub.affiliations.items',
                selector: 'owner'
            }
        ],
        element: 'affiliation',
        fields: {
            affiliation: attribute('affiliation'),
            jid: JIDAttribute('jid'),
            node: attribute('node')
        },
        namespace: NS_PUBSUB_OWNER,
        type: 'owner'
    },
    {
        element: 'create',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.create'
    },
    {
        aliases: [{ path: 'iq.pubsub.destroy', selector: 'owner' }],
        element: 'delete',
        fields: {
            node: attribute('node'),
            redirect: childAttribute(null, 'redirect', 'uri')
        },
        namespace: NS_PUBSUB_OWNER
    },
    {
        aliases: [{ path: 'iq.pubsub.configure', selector: 'owner', impliedType: true }],
        element: 'configure',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB_OWNER,
        type: 'owner'
    },
    {
        aliases: [{ path: 'iq.pubsub.configure', selector: 'user', impliedType: true }],
        element: 'configure',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB,
        type: 'user'
    },
    {
        element: 'default',
        fields: {
            node: attribute('node')
        },
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.defaultSubscriptionOptions'
    },
    {
        element: 'default',
        fields: {},
        namespace: NS_PUBSUB_OWNER,
        path: 'iq.pubsub.defaultConfiguration'
    },
    {
        element: 'publish',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.publish'
    },
    {
        element: 'retract',
        fields: {
            id: childAttribute(null, 'item', 'id'),
            node: attribute('node'),
            notify: booleanAttribute('notify')
        },
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.retract'
    },
    {
        element: 'items',
        fields: {
            max: integerAttribute('max_items'),
            node: attribute('node')
        },
        namespace: NS_PUBSUB,
        path: 'iq.pubsub.fetch'
    },
    {
        aliases: [
            'pubsubitem',
            'iq.pubsub.publish.item',
            { multiple: true, path: 'iq.pubsub.fetch.items' }
        ],
        element: 'item',
        fields: {
            id: attribute('id'),
            publisher: JIDAttribute('publisher')
        },
        namespace: NS_PUBSUB
    },
    {
        element: 'event',
        fields: {
            eventType: childEnum(null, [
                'purge',
                'delete',
                'subscription',
                'configuration',
                'items'
            ])
        },
        namespace: NS_PUBSUB_EVENT,
        path: 'message.pubsub',
        type: 'event',
        typeField: 'context'
    },
    {
        aliases: [{ path: 'message.pubsub.items.published', multiple: true }],
        element: 'item',
        fields: {
            id: attribute('id'),
            publisher: JIDAttribute('publisher')
        },
        namespace: NS_PUBSUB_EVENT,
        path: 'pubsubeventitem'
    },
    {
        element: 'purge',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB_EVENT,
        path: 'message.pubsub.purge'
    },
    {
        element: 'delete',
        fields: {
            node: attribute('node'),
            redirect: childAttribute(null, 'redirect', 'uri')
        },
        namespace: NS_PUBSUB_EVENT,
        path: 'message.pubsub.delete'
    },
    {
        aliases: [{ path: 'message.pubsub.subscription', selector: 'event', impliedType: true }],
        element: 'subscription',
        fields: {
            expires: dateOrPresenceAttribute('expiry'),
            jid: JIDAttribute('jid'),
            node: attribute('node'),
            subid: attribute('subid'),
            type: attribute('subscription')
        },
        namespace: NS_PUBSUB_EVENT,
        type: 'event'
    },
    {
        element: 'configuration',
        fields: NodeOnlyField,
        namespace: NS_PUBSUB_EVENT,
        path: 'message.pubsub.configuration'
    },
    {
        element: 'items',
        fields: {
            node: attribute('node'),
            retracted: multipleChildAttribute(null, 'retract', 'id')
        },
        namespace: NS_PUBSUB_EVENT,
        path: 'message.pubsub.items'
    }
];

// ====================================================================
const Protocol$T = [
    {
        element: 'query',
        fields: {
            activate: childText(null, 'activate'),
            address: attribute('dstaddr'),
            candidateUsed: childJIDAttribute(null, 'streamhost-used', 'jid'),
            mode: attribute('mode', 'tcp'),
            sid: attribute('sid'),
            udpSuccess: childAttribute(null, 'udpsuccess', 'dstaddr')
        },
        namespace: NS_SOCKS5,
        path: 'iq.socks5'
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.socks5.candidates'
            }
        ],
        element: 'streamhost',
        fields: {
            host: attribute('host'),
            jid: JIDAttribute('jid'),
            port: integerAttribute('port'),
            uri: attribute('uri')
        },
        namespace: NS_SOCKS5
    }
];

// ====================================================================
const Protocol$S = [
    {
        aliases: [{ multiple: true, path: 'message.links' }],
        element: 'x',
        fields: {
            description: childText(null, 'desc'),
            url: childText(null, 'url')
        },
        namespace: NS_OOB
    },
    {
        element: 'query',
        fields: {
            description: childText(null, 'desc'),
            url: childText(null, 'url')
        },
        namespace: NS_OOB_TRANSFER,
        path: 'iq.transferLink'
    }
];

// ====================================================================
const Protocol$R = {
    element: 'html',
    fields: {
        alternateLanguageBodies: childAlternateLanguageRawElement(NS_XHTML, 'body', 'xhtmlim'),
        body: childLanguageRawElement(NS_XHTML, 'body', 'xhtmlim')
    },
    namespace: NS_XHTML_IM,
    path: 'message.html'
};

// ====================================================================
const Protocol$Q = [
    extendStreamFeatures({
        inbandRegistration: childBoolean(NS_INBAND_REGISTRATION, 'register')
    }),
    addAlias(NS_DATAFORM, 'x', ['iq.account.form']),
    addAlias(NS_OOB, 'x', ['iq.account.registrationLink']),
    {
        element: 'query',
        fields: {
            address: childText(null, 'address'),
            date: childDate(null, 'date'),
            email: childText(null, 'email'),
            familyName: childText(null, 'last'),
            fullName: childText(null, 'name'),
            givenName: childText(null, 'first'),
            instructions: childText(null, 'instructions'),
            key: childText(null, 'key'),
            locality: childText(null, 'city'),
            misc: childText(null, 'misc'),
            nick: childText(null, 'nick'),
            password: childText(null, 'password'),
            phone: childText(null, 'phone'),
            postalCode: childText(null, 'zip'),
            region: childText(null, 'state'),
            registered: childBoolean(null, 'registered'),
            remove: childBoolean(null, 'remove'),
            text: childText(null, 'text'),
            uri: childText(null, 'uri'),
            username: childText(null, 'username')
        },
        namespace: NS_REGISTER,
        path: 'iq.account'
    }
];

// ====================================================================
const Protocol$P = {
    aliases: [
        { path: 'message.geoloc', impliedType: true },
        { path: 'dataform.fields.geoloc', impliedType: true },
        ...pubsubItemContentAliases()
    ],
    element: 'geoloc',
    fields: {
        accuracy: childFloat(null, 'accuracy'),
        altitude: childFloat(null, 'alt'),
        altitudeAccuracy: childFloat(null, 'altaccuracy'),
        area: childText(null, 'area'),
        building: childText(null, 'building'),
        country: childText(null, 'country'),
        countryCode: childText(null, 'countrycode'),
        datum: childText(null, 'datum'),
        description: childText(null, 'description'),
        error: childFloat(null, 'error'),
        floor: childText(null, 'floor'),
        heading: childFloat(null, 'bearing'),
        lang: languageAttribute(),
        latitude: childFloat(null, 'lat'),
        locality: childText(null, 'locality'),
        longitude: childFloat(null, 'lon'),
        postalCode: childText(null, 'postalcode'),
        region: childText(null, 'region'),
        room: childText(null, 'room'),
        speed: childFloat(null, 'speed'),
        street: childText(null, 'street'),
        text: childText(null, 'text'),
        timestamp: childDate(null, 'timestamp'),
        tzo: childTimezoneOffset(null, 'tzo'),
        uri: childText(null, 'uri')
    },
    namespace: NS_GEOLOC,
    type: NS_GEOLOC
};

// ====================================================================
const Protocol$O = [
    {
        aliases: pubsubItemContentAliases(),
        element: 'data',
        fields: {
            data: textBuffer('base64')
        },
        namespace: NS_AVATAR_DATA,
        path: 'avatar',
        type: NS_AVATAR_DATA,
        typeField: 'itemType'
    },
    {
        aliases: pubsubItemContentAliases(),
        element: 'metadata',
        namespace: NS_AVATAR_METADATA,
        path: 'avatar',
        type: NS_AVATAR_METADATA,
        typeField: 'itemType'
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'avatar.versions',
                selector: NS_AVATAR_METADATA
            }
        ],
        element: 'info',
        fields: {
            bytes: integerAttribute('bytes'),
            height: integerAttribute('height'),
            id: attribute('id'),
            mediaType: attribute('type'),
            uri: attribute('url'),
            width: integerAttribute('width')
        },
        namespace: NS_AVATAR_METADATA
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'avatar.pointers',
                selector: NS_AVATAR_METADATA
            }
        ],
        element: 'pointer',
        fields: {
            bytes: integerAttribute('bytes'),
            height: integerAttribute('height'),
            id: attribute('id'),
            mediaType: attribute('type'),
            uri: attribute('url'),
            width: integerAttribute('width')
        },
        namespace: NS_AVATAR_METADATA
    }
];

// ====================================================================
var XEP0085 = extendMessage({
    chatState: childEnum(NS_CHAT_STATES, Object.values(ChatState))
});

// ====================================================================
const Protocol$N = {
    element: 'query',
    fields: {
        name: childText(null, 'name'),
        os: childText(null, 'os'),
        version: childText(null, 'version')
    },
    namespace: NS_VERSION,
    path: 'iq.softwareVersion'
};

// ====================================================================
const Protocol$M = {
    aliases: [{ path: 'message.mood', impliedType: true }, ...pubsubItemContentAliases()],
    element: 'mood',
    fields: {
        alternateLanguageText: childAlternateLanguageText(null, 'text'),
        text: childText(null, 'text'),
        value: childEnum(null, USER_MOODS)
    },
    namespace: NS_MOOD,
    type: NS_MOOD
};

// ====================================================================
const Protocol$L = {
    aliases: [{ path: 'activity', impliedType: true }, ...pubsubItemContentAliases()],
    element: 'activity',
    fields: {
        activity: childDoubleEnum(null, USER_ACTIVITY_GENERAL, USER_ACTIVITY_SPECIFIC),
        alternateLanguageText: childAlternateLanguageText(null, 'text'),
        text: childText(null, 'text')
    },
    namespace: NS_ACTIVITY,
    type: NS_ACTIVITY
};

// ====================================================================
const Protocol$K = {
    element: 'handshake',
    fields: {
        value: textBuffer('hex')
    },
    namespace: NS_COMPONENT,
    path: 'handshake'
};

// ====================================================================
const Protocol$J = {
    aliases: [
        { path: 'presence.legacyCapabilities', multiple: true },
        { path: 'features.legacyCapabilities', multiple: true }
    ],
    element: 'c',
    fields: {
        algorithm: attribute('hash'),
        legacy: staticValue(true),
        node: attribute('node'),
        value: attribute('ver')
    },
    namespace: NS_DISCO_LEGACY_CAPS
};

// ====================================================================
const Protocol$I = {
    aliases: [
        {
            impliedType: true,
            path: 'tune'
        },
        ...pubsubItemContentAliases()
    ],
    element: 'tune',
    fields: {
        artist: childText(null, 'artist'),
        length: childInteger(null, 'length'),
        rating: childInteger(null, 'rating'),
        source: childText(null, 'source'),
        title: childText(null, 'title'),
        track: childText(null, 'track'),
        uri: childText(null, 'uri')
    },
    namespace: NS_TUNE,
    type: NS_TUNE
};

// ====================================================================
const aliases = [
    'dataformLayout',
    {
        multiple: true,
        path: 'dataformLayout.contents'
    },
    {
        multiple: true,
        path: 'dataform.layout.contents'
    }
];
const Protocol$H = [
    {
        aliases: [
            {
                multiple: true,
                path: 'dataform.layout'
            }
        ],
        element: 'page',
        fields: {
            label: attribute('label')
        },
        namespace: NS_DATAFORM_LAYOUT
    },
    {
        aliases,
        element: 'section',
        fields: {
            label: attribute('label')
        },
        namespace: NS_DATAFORM_LAYOUT,
        type: 'section',
        typeField: 'type'
    },
    {
        aliases,
        element: 'text',
        fields: {
            value: text()
        },
        namespace: NS_DATAFORM_LAYOUT,
        type: 'text',
        typeField: 'type'
    },
    {
        aliases,
        element: 'fieldref',
        fields: {
            field: attribute('var')
        },
        namespace: NS_DATAFORM_LAYOUT,
        type: 'fieldref',
        typeField: 'type'
    },
    {
        aliases,
        element: 'reportedref',
        namespace: NS_DATAFORM_LAYOUT,
        type: 'reportedref',
        typeField: 'type'
    }
];

// ====================================================================
const Protocol$G = {
    element: 'body',
    fields: {
        acceptMediaTypes: attribute('accept'),
        ack: integerAttribute('ack'),
        authId: attribute('authid'),
        characterSets: attribute('charsets'),
        condition: attribute('condition'),
        from: JIDAttribute('from'),
        key: attribute('key'),
        lang: languageAttribute(),
        maxClientRequests: integerAttribute('requests'),
        maxHoldOpen: integerAttribute('hold'),
        maxInactivityTime: integerAttribute('inactivity'),
        maxSessionPause: integerAttribute('maxpause'),
        maxWaitTime: integerAttribute('wait'),
        mediaType: attribute('content'),
        minPollingInterval: integerAttribute('polling'),
        newKey: attribute('newkey'),
        pauseSession: integerAttribute('pause'),
        report: integerAttribute('report'),
        rid: integerAttribute('rid'),
        route: attribute('string'),
        seeOtherURI: childText(null, 'uri'),
        sid: attribute('sid'),
        stream: attribute('stream'),
        time: integerAttribute('time'),
        to: JIDAttribute('to'),
        type: attribute('type'),
        version: attribute('ver'),
        // XEP-0206
        xmppRestart: namespacedBooleanAttribute('xmpp', NS_BOSH_XMPP, 'restart', undefined, {
            writeValue: (value) => {
                return value ? 'true' : 'false';
            }
        }),
        xmppRestartLogic: namespacedBooleanAttribute('xmpp', NS_BOSH_XMPP, 'restartlogic', undefined, {
            writeValue: (value) => {
                return value ? 'true' : 'false';
            }
        }),
        xmppVersion: namespacedAttribute('xmpp', NS_BOSH_XMPP, 'version')
    },
    namespace: NS_BOSH,
    path: 'bosh'
};

// ====================================================================
const Protocol$F = [
    extendMessage({
        headers: splicePath(NS_SHIM, 'headers', 'header', true)
    }),
    extendPresence({
        headers: splicePath(NS_SHIM, 'headers', 'header', true)
    }),
    {
        element: 'header',
        fields: {
            name: attribute('name'),
            value: text()
        },
        namespace: NS_SHIM,
        path: 'header'
    }
];

// ====================================================================
const Protocol$E = [
    {
        element: 'compression',
        fields: {
            methods: multipleChildText(null, 'method')
        },
        namespace: NS_COMPRESSION_FEATURE,
        path: 'features.compression'
    },
    {
        element: 'compress',
        fields: {
            method: childText(null, 'method')
        },
        namespace: NS_COMPRESSION,
        path: 'compression',
        type: 'start',
        typeField: 'type'
    },
    {
        aliases: ['error.compressionError'],
        element: 'failure',
        fields: {
            condition: childEnum(null, ['unsupported-method', 'setup-failed', 'processing-failed'])
        },
        namespace: NS_COMPRESSION,
        path: 'compression',
        type: 'failure',
        typeField: 'type'
    },
    {
        element: 'compressed',
        namespace: NS_COMPRESSION,
        path: 'compression',
        type: 'success',
        typeField: 'type'
    }
];

// ====================================================================
const Protocol$D = [
    extendMessage({
        rosterExchange: splicePath(NS_ROSTER_EXCHANGE, 'x', 'rosterExchange', true)
    }),
    extendIQ({
        rosterExchange: splicePath(NS_ROSTER_EXCHANGE, 'x', 'rosterExchange', true)
    }),
    {
        element: 'item',
        fields: {
            action: attribute('action'),
            groups: multipleChildText(null, 'group'),
            jid: JIDAttribute('jid'),
            name: attribute('name')
        },
        namespace: NS_ROSTER_EXCHANGE,
        path: 'rosterExchange'
    }
];

// ====================================================================
var XEP0153 = extendPresence({
    vcardAvatar: {
        importer(xml) {
            const update = findAll(xml, NS_VCARD_TEMP_UPDATE, 'x');
            if (!update.length) {
                return;
            }
            const photo = findAll(update[0], NS_VCARD_TEMP_UPDATE, 'photo');
            if (photo.length) {
                return photo[0].getText();
            }
            else {
                return true;
            }
        },
        exporter(xml, value) {
            const update = findOrCreate(xml, NS_VCARD_TEMP_UPDATE, 'x');
            if (value === '') {
                findOrCreate(update, NS_VCARD_TEMP_UPDATE, 'photo');
            }
            else if (value === true) {
                return;
            }
            else if (value) {
                const photo = findOrCreate(update, NS_VCARD_TEMP_UPDATE, 'photo');
                photo.children.push(value);
            }
        }
    }
});

// ====================================================================
const Protocol$C = [
    extendMessage({
        captcha: splicePath(NS_CAPTCHA, 'captcha', 'dataform')
    }),
    extendIQ({
        captcha: splicePath(NS_CAPTCHA, 'captcha', 'dataform')
    })
];

// ====================================================================
const Protocol$B = [
    extendStanzaError({
        jingleError: childEnum(NS_JINGLE_ERRORS_1, Object.values(JingleErrorCondition))
    }),
    {
        element: 'jingle',
        fields: {
            action: attribute('action'),
            initiator: JIDAttribute('initiator'),
            responder: JIDAttribute('responder'),
            sid: attribute('sid')
        },
        namespace: NS_JINGLE_1,
        path: 'iq.jingle'
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents'
            }
        ],
        element: 'content',
        fields: {
            creator: attribute('creator'),
            disposition: attribute('disposition', 'session'),
            name: attribute('name'),
            senders: attribute('senders', 'both')
        },
        namespace: NS_JINGLE_1
    },
    {
        element: 'reason',
        fields: {
            alternativeSession: childText(null, 'alternative-session'),
            condition: childEnum(null, Object.values(JingleReasonCondition)),
            text: childText(null, 'text')
        },
        namespace: NS_JINGLE_1,
        path: 'iq.jingle.reason'
    }
];

// ====================================================================
function rtcpFeedback() {
    return {
        importer(xml, context) {
            let existing = findAll(xml, NS_JINGLE_RTP_RTCP_FB_0, 'rtcp-fb');
            const typeImporter = attribute('type').importer;
            const subtypeImporter = attribute('subtype').importer;
            const valueImporter = attribute('value').importer;
            const result = [];
            for (const child of existing) {
                const type = typeImporter(child, context);
                const parameter = subtypeImporter(child, context);
                result.push(parameter ? { type, parameter } : { type });
            }
            existing = findAll(xml, NS_JINGLE_RTP_RTCP_FB_0, 'rtcp-fb-trr-int');
            for (const child of existing) {
                const parameter = valueImporter(child, context);
                result.push(parameter ? { type: 'trr-int', parameter } : { type: 'trr-int' });
            }
            return result;
        },
        exporter(xml, values, context) {
            const typeExporter = attribute('type').exporter;
            const subtypeExporter = attribute('subtype').exporter;
            const valueExporter = attribute('value').exporter;
            for (const fb of values) {
                let child;
                if (fb.type === 'trr-int') {
                    child = createElement(NS_JINGLE_RTP_RTCP_FB_0, 'rtcp-fb-trr-int', context.namespace, xml);
                    if (fb.parameter) {
                        valueExporter(child, fb.parameter, context);
                    }
                }
                else {
                    child = createElement(NS_JINGLE_RTP_RTCP_FB_0, 'rtcp-fb', context.namespace, xml);
                    typeExporter(child, fb.type, context);
                    if (fb.parameter) {
                        subtypeExporter(child, fb.parameter, context);
                    }
                }
                xml.appendChild(child);
            }
        }
    };
}
const info = 'iq.jingle.info';
const Protocol$A = [
    {
        aliases: ['iq.jingle.contents.application'],
        childrenExportOrder: {
            codecs: 4,
            encryption: 5,
            headerExtensions: 6,
            sourceGroups: 8,
            sources: 7,
            streams: 9
        },
        element: 'description',
        fields: {
            media: attribute('media'),
            rtcpFeedback: Object.assign(Object.assign({}, rtcpFeedback()), { exportOrder: 3 }),
            rtcpMux: Object.assign(Object.assign({}, childBoolean(null, 'rtcp-mux')), { exportOrder: 1 }),
            rtcpReducedSize: Object.assign(Object.assign({}, childBoolean(null, 'rtcp-reduced-size')), { exportOrder: 2 }),
            ssrc: attribute('ssrc')
        },
        namespace: NS_JINGLE_RTP_1,
        optionalNamespaces: {
            rtcpf: NS_JINGLE_RTP_RTCP_FB_0,
            rtph: NS_JINGLE_RTP_HDREXT_0
        },
        type: NS_JINGLE_RTP_1
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents.application.headerExtensions',
                selector: NS_JINGLE_RTP_1
            }
        ],
        element: 'rtp-hdrext',
        fields: {
            id: integerAttribute('id'),
            senders: attribute('senders'),
            uri: attribute('uri')
        },
        namespace: NS_JINGLE_RTP_HDREXT_0
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents.application.codecs',
                selector: NS_JINGLE_RTP_1
            },
            'rtpcodec'
        ],
        element: 'payload-type',
        fields: {
            channels: integerAttribute('channels'),
            clockRate: integerAttribute('clockrate'),
            id: attribute('id'),
            maxptime: integerAttribute('maxptime'),
            name: attribute('name'),
            parameters: parameterMap(NS_JINGLE_RTP_1, 'parameter', 'name', 'value'),
            ptime: integerAttribute('ptime'),
            rtcpFeedback: rtcpFeedback()
        },
        namespace: NS_JINGLE_RTP_1
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents.application.sources',
                selector: NS_JINGLE_RTP_1
            }
        ],
        element: 'source',
        fields: {
            parameters: parameterMap(NS_JINGLE_RTP_SSMA_0, 'parameter', 'name', 'value'),
            ssrc: attribute('ssrc')
        },
        namespace: NS_JINGLE_RTP_SSMA_0
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents.application.sourceGroups',
                selector: NS_JINGLE_RTP_1
            }
        ],
        element: 'ssrc-group',
        fields: {
            semantics: attribute('semantics'),
            sources: multipleChildAttribute(null, 'source', 'ssrc')
        },
        namespace: NS_JINGLE_RTP_SSMA_0
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents.application.streams',
                selector: NS_JINGLE_RTP_1
            }
        ],
        element: 'stream',
        fields: {
            id: attribute('id'),
            track: attribute('track')
        },
        namespace: NS_JINGLE_RTP_MSID_0
    },
    {
        aliases: [{ path: 'iq.jingle.contents.application.encryption', selector: NS_JINGLE_RTP_1 }],
        element: 'encryption',
        fields: {
            required: booleanAttribute('required')
        },
        namespace: NS_JINGLE_RTP_1
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents.application.encryption.sdes',
                selector: NS_JINGLE_RTP_1
            }
        ],
        element: 'crypto',
        fields: {
            cryptoSuite: attribute('crypto-suite'),
            keyParameters: attribute('key-params'),
            sessionParameters: attribute('session-params'),
            tag: integerAttribute('tag')
        },
        namespace: NS_JINGLE_RTP_1
    },
    {
        aliases: [
            {
                path: 'iq.jingle.contents.application.encryption.zrtp',
                selector: NS_JINGLE_RTP_1
            }
        ],
        element: 'zrtp-hash',
        fields: {
            value: textBuffer('hex'),
            version: attribute('version')
        },
        namespace: NS_JINGLE_RTP_1
    },
    {
        element: 'mute',
        fields: {
            creator: attribute('creator'),
            name: attribute('name')
        },
        namespace: NS_JINGLE_RTP_INFO_1,
        path: info,
        type: JINGLE_INFO_MUTE
    },
    {
        element: 'unmute',
        fields: {
            creator: attribute('creator'),
            name: attribute('name')
        },
        namespace: NS_JINGLE_RTP_INFO_1,
        path: info,
        type: JINGLE_INFO_UNMUTE
    },
    {
        element: 'hold',
        namespace: NS_JINGLE_RTP_INFO_1,
        path: info,
        type: JINGLE_INFO_HOLD
    },
    {
        element: 'unhold',
        namespace: NS_JINGLE_RTP_INFO_1,
        path: info,
        type: JINGLE_INFO_UNHOLD
    },
    {
        element: 'active',
        namespace: NS_JINGLE_RTP_INFO_1,
        path: info,
        type: JINGLE_INFO_ACTIVE
    },
    {
        element: 'ringing',
        namespace: NS_JINGLE_RTP_INFO_1,
        path: info,
        type: JINGLE_INFO_RINGING
    }
];

// ====================================================================
const Protocol$z = [
    extendMessage({
        nick: childText(NS_NICK, 'nick')
    }),
    extendPresence({
        nick: childText(NS_NICK, 'nick')
    }),
    {
        aliases: pubsubItemContentAliases(),
        element: 'nick',
        fields: {
            nick: text()
        },
        namespace: NS_NICK,
        type: NS_NICK
    }
];

// ====================================================================
const ice = (transportType) => [
    {
        element: 'transport',
        fields: {
            gatheringComplete: childBoolean(null, 'gathering-complete'),
            password: attribute('pwd'),
            usernameFragment: attribute('ufrag')
        },
        namespace: transportType,
        path: 'iq.jingle.contents.transport',
        type: transportType,
        typeField: 'transportType'
    },
    {
        aliases: [
            {
                impliedType: true,
                path: 'iq.jingle.contents.transport.remoteCandidate',
                selector: transportType
            }
        ],
        element: 'remote-candidate',
        fields: {
            component: integerAttribute('component'),
            ip: attribute('ip'),
            port: integerAttribute('port')
        },
        namespace: transportType,
        type: transportType,
        typeField: 'transportType'
    },
    {
        aliases: [
            {
                impliedType: true,
                multiple: true,
                path: 'iq.jingle.contents.transport.candidates',
                selector: transportType
            }
        ],
        element: 'candidate',
        fields: {
            component: integerAttribute('component'),
            foundation: attribute('foundation'),
            generation: integerAttribute('generation'),
            id: attribute('id'),
            ip: attribute('ip'),
            network: integerAttribute('network'),
            port: integerAttribute('port'),
            priority: integerAttribute('priority'),
            protocol: attribute('protocol'),
            relatedAddress: attribute('rel-addr'),
            relatedPort: attribute('rel-port'),
            tcpType: attribute('tcptype'),
            type: attribute('type')
        },
        namespace: transportType,
        type: transportType,
        typeField: 'transportType'
    }
];
const Protocol$y = [...ice(NS_JINGLE_ICE_0), ...ice(NS_JINGLE_ICE_UDP_1)];

// ====================================================================
const Protocol$x = [
    {
        element: 'transport',
        fields: {
            gatheringComplete: childBoolean(null, 'gathering-complete'),
            password: attribute('pwd'),
            usernameFragment: attribute('ufrag')
        },
        namespace: NS_JINGLE_RAW_UDP_1,
        path: 'iq.jingle.contents.transport',
        type: NS_JINGLE_RAW_UDP_1,
        typeField: 'transportType'
    },
    {
        aliases: [
            {
                impliedType: true,
                multiple: true,
                path: 'iq.jingle.contents.transport.candidates',
                selector: NS_JINGLE_RAW_UDP_1
            }
        ],
        element: 'candidate',
        fields: {
            component: integerAttribute('component'),
            foundation: attribute('foundation'),
            generation: integerAttribute('generation'),
            id: attribute('id'),
            ip: attribute('ip'),
            port: integerAttribute('port'),
            type: attribute('type')
        },
        namespace: NS_JINGLE_RAW_UDP_1,
        type: NS_JINGLE_RAW_UDP_1,
        typeField: 'transportType'
    }
];

// ====================================================================
const Protocol$w = [
    {
        element: 'request',
        namespace: NS_RECEIPTS,
        path: 'message.receipt',
        type: 'request',
        typeField: 'type'
    },
    {
        element: 'received',
        fields: {
            id: attribute('id')
        },
        namespace: NS_RECEIPTS,
        path: 'message.receipt',
        type: 'received',
        typeField: 'type'
    }
];

// ====================================================================
const Protocol$v = [
    {
        element: 'invisible',
        fields: {
            probe: booleanAttribute('probe')
        },
        namespace: NS_INVISIBLE_0,
        path: 'iq.visibility',
        type: 'invisible',
        typeField: 'type'
    },
    {
        element: 'visible',
        namespace: NS_INVISIBLE_0,
        path: 'iq.visibility',
        type: 'visible'
    }
];

// ====================================================================
const Protocol$u = [
    extendStanzaError({
        blocked: childBoolean(NS_BLOCKING_ERRORS, 'blocked')
    }),
    {
        element: 'blocklist',
        fields: {
            jids: multipleChildAttribute(null, 'item', 'jid')
        },
        namespace: NS_BLOCKING,
        path: 'iq.blockList',
        type: 'list',
        typeField: 'action'
    },
    {
        element: 'block',
        fields: {
            jids: multipleChildAttribute(null, 'item', 'jid')
        },
        namespace: NS_BLOCKING,
        path: 'iq.blockList',
        type: 'block',
        typeField: 'action'
    },
    {
        element: 'unblock',
        fields: {
            jids: multipleChildAttribute(null, 'item', 'jid')
        },
        namespace: NS_BLOCKING,
        path: 'iq.blockList',
        type: 'unblock',
        typeField: 'action'
    }
];

// ====================================================================
const Protocol$t = [
    extendStreamFeatures({
        streamManagement: childBoolean(NS_SMACKS_3, 'sm')
    }),
    {
        element: 'a',
        fields: {
            handled: integerAttribute('h')
        },
        namespace: NS_SMACKS_3,
        path: 'sm',
        type: 'ack',
        typeField: 'type'
    },
    {
        element: 'r',
        namespace: NS_SMACKS_3,
        path: 'sm',
        type: 'request',
        typeField: 'type'
    },
    {
        element: 'enable',
        fields: {
            allowResumption: booleanAttribute('resume')
        },
        namespace: NS_SMACKS_3,
        path: 'sm',
        type: 'enable',
        typeField: 'type'
    },
    {
        element: 'enabled',
        fields: {
            id: attribute('id'),
            resume: booleanAttribute('resume')
        },
        namespace: NS_SMACKS_3,
        path: 'sm',
        type: 'enabled',
        typeField: 'type'
    },
    {
        element: 'resume',
        fields: {
            handled: integerAttribute('h'),
            previousSession: attribute('previd')
        },
        namespace: NS_SMACKS_3,
        path: 'sm',
        type: 'resume',
        typeField: 'type'
    },
    {
        element: 'resumed',
        fields: {
            handled: integerAttribute('h'),
            previousSession: attribute('previd')
        },
        namespace: NS_SMACKS_3,
        path: 'sm',
        type: 'resumed',
        typeField: 'type'
    },
    {
        element: 'failed',
        fields: {
            handled: integerAttribute('h')
        },
        namespace: NS_SMACKS_3,
        path: 'sm',
        type: 'failed',
        typeField: 'type'
    }
];

// ====================================================================
var XEP0199 = extendIQ({
    ping: childBoolean(NS_PING, 'ping')
});

// ====================================================================
const Protocol$s = {
    element: 'time',
    fields: {
        tzo: childTimezoneOffset(null, 'tzo'),
        utc: childDate(null, 'utc')
    },
    namespace: NS_TIME,
    path: 'iq.time'
};

// ====================================================================
const Protocol$r = {
    aliases: ['message.delay', 'presence.delay'],
    element: 'delay',
    fields: {
        from: JIDAttribute('from'),
        reason: text(),
        timestamp: dateAttribute('stamp')
    },
    namespace: NS_DELAY
};

// ====================================================================
const versions$1 = {
    '2': NS_DISCO_EXTERNAL_2,
    '1': NS_DISCO_EXTERNAL_1
};
const Protocol$q = [];
for (const [version, namespace] of Object.entries(versions$1)) {
    Protocol$q.push({
        aliases: ['iq.externalServiceCredentials'],
        defaultType: '2',
        element: 'credentials',
        fields: {
            expires: childDateAttribute(null, 'service', 'expires'),
            host: childAttribute(null, 'service', 'host'),
            name: childAttribute(null, 'service', 'name'),
            password: childAttribute(null, 'service', 'password'),
            port: childIntegerAttribute(null, 'service', 'port'),
            restricted: childBooleanAttribute(null, 'service', 'restricted'),
            transport: childAttribute(null, 'service', 'transport'),
            type: childAttribute(null, 'service', 'type'),
            username: childAttribute(null, 'service', 'username')
        },
        namespace,
        type: version,
        typeField: 'version'
    }, {
        aliases: ['iq.externalServices'],
        defaultType: '2',
        element: 'services',
        fields: {
            type: attribute('type')
        },
        namespace,
        type: version,
        typeField: 'version'
    }, {
        aliases: [{ path: 'iq.externalServices.services', multiple: true }],
        defaultType: '2',
        element: 'service',
        fields: {
            expires: dateAttribute('expires'),
            host: attribute('host'),
            name: attribute('name'),
            password: attribute('password'),
            port: integerAttribute('port'),
            restricted: booleanAttribute('restricted'),
            transport: attribute('transport'),
            type: attribute('type'),
            username: attribute('username')
        },
        namespace,
        type: version,
        typeField: 'version'
    });
}

// ====================================================================
const Protocol$p = [
    {
        element: 'media',
        fields: {
            height: integerAttribute('height'),
            width: integerAttribute('width')
        },
        namespace: NS_DATAFORM_MEDIA,
        path: 'dataform.fields.media'
    },
    {
        aliases: [{ multiple: true, path: 'dataform.fields.media.sources' }],
        element: 'uri',
        fields: {
            mediaType: attribute('type'),
            uri: text()
        },
        namespace: NS_DATAFORM_MEDIA
    }
];

// ====================================================================
var XEP0224 = extendMessage({
    requestingAttention: childBoolean(NS_ATTENTION_0, 'attention')
});

// ====================================================================
const Protocol$o = {
    aliases: [
        'iq.bits',
        { path: 'message.bits', multiple: true },
        { path: 'presence.bits', multiple: true },
        { path: 'iq.jingle.bits', multiple: true }
    ],
    element: 'data',
    fields: {
        cid: attribute('cid'),
        data: textBuffer('base64'),
        maxAge: integerAttribute('max-age'),
        mediaType: attribute('type')
    },
    namespace: NS_BOB
};

// ====================================================================
let Protocol$n = [
    addAlias(NS_HASHES_2, 'hash', [
        { path: 'file.hashes', multiple: true },
        { path: 'file.range.hashes', multiple: true }
    ]),
    addAlias(NS_HASHES_1, 'hash', [
        { path: 'file.hashes', multiple: true },
        { path: 'file.range.hashes', multiple: true }
    ]),
    addAlias(NS_HASHES_2, 'hash-used', [{ path: 'file.hashesUsed', multiple: true }]),
    addAlias(NS_THUMBS_1, 'thumbnail', [{ path: 'file.thumbnails', multiple: true }])
];
for (const ftVersion of [NS_JINGLE_FILE_TRANSFER_4, NS_JINGLE_FILE_TRANSFER_5]) {
    Protocol$n = Protocol$n.concat([
        {
            aliases: [
                'file',
                {
                    impliedType: true,
                    path: 'iq.jingle.contents.application.file',
                    selector: ftVersion
                },
                {
                    impliedType: true,
                    path: 'iq.jingle.info.file',
                    selector: `{${ftVersion}}checksum`
                }
            ],
            defaultType: NS_JINGLE_FILE_TRANSFER_5,
            element: 'file',
            fields: {
                date: childDate(null, 'date'),
                description: childText(null, 'desc'),
                mediaType: childText(null, 'media-type'),
                name: childText(null, 'name'),
                size: childInteger(null, 'size')
            },
            namespace: ftVersion,
            type: ftVersion,
            typeField: 'version'
        },
        {
            aliases: [{ impliedType: true, path: 'file.range', selector: ftVersion }],
            defaultType: NS_JINGLE_FILE_TRANSFER_5,
            element: 'range',
            fields: {
                length: integerAttribute('length'),
                offset: integerAttribute('offset', 0)
            },
            namespace: ftVersion,
            type: ftVersion,
            typeField: 'version'
        },
        {
            element: 'description',
            namespace: ftVersion,
            path: 'iq.jingle.contents.application',
            type: ftVersion,
            typeField: 'applicationType'
        },
        {
            element: 'received',
            fields: {
                creator: attribute('creator'),
                name: attribute('name')
            },
            namespace: ftVersion,
            path: 'iq.jingle.info',
            type: `{${ftVersion}}received`,
            typeField: 'infoType'
        },
        {
            element: 'checksum',
            fields: {
                creator: attribute('creator'),
                name: attribute('name')
            },
            namespace: ftVersion,
            path: 'iq.jingle.info',
            type: `{${ftVersion}}checksum`,
            typeField: 'infoType'
        }
    ]);
}
var XEP0234 = Protocol$n;

// ====================================================================
const Protocol$m = {
    element: 'description',
    namespace: NS_JINGLE_XML_0,
    path: 'iq.jingle.contents.application',
    type: NS_JINGLE_XML_0,
    typeField: 'applicationType'
};

// ====================================================================
const Protocol$l = [
    {
        element: 'transport',
        fields: {
            activated: childAttribute(null, 'activated', 'cid'),
            address: attribute('dstaddr'),
            candidateError: childBoolean(null, 'candidate-error'),
            candidateUsed: childAttribute(null, 'candidate-used', 'cid'),
            mode: attribute('mode', 'tcp'),
            proxyError: childBoolean(null, 'proxy-error'),
            sid: attribute('sid')
        },
        namespace: NS_JINGLE_SOCKS5_1,
        path: 'iq.jingle.contents.transport',
        type: NS_JINGLE_SOCKS5_1,
        typeField: 'transportType'
    },
    {
        aliases: [
            {
                multiple: true,
                path: 'iq.jingle.contents.transport.candidates',
                selector: NS_JINGLE_SOCKS5_1
            }
        ],
        element: 'candidate',
        fields: {
            cid: attribute('cid'),
            host: attribute('host'),
            jid: JIDAttribute('jid'),
            port: integerAttribute('port'),
            priority: integerAttribute('priority'),
            type: attribute('type'),
            uri: attribute('uri')
        },
        namespace: NS_JINGLE_SOCKS5_1
    }
];

// ====================================================================
const Protocol$k = {
    element: 'transport',
    fields: {
        ack: {
            importer(xml, context) {
                const stanza = attribute('stanza', 'iq').importer(xml, context);
                return stanza !== 'message';
            },
            exporter(xml, data, context) {
                if (data === false) {
                    attribute('stanza').exporter(xml, 'message', context);
                }
            }
        },
        blockSize: integerAttribute('block-size'),
        sid: attribute('sid')
    },
    namespace: NS_JINGLE_IBB_1,
    path: 'iq.jingle.contents.transport',
    type: NS_JINGLE_IBB_1,
    typeField: 'transportType'
};

// ====================================================================
const Protocol$j = [
    addAlias(NS_BOB, 'data', [{ path: 'iq.jingle.bits', multiple: true }]),
    {
        element: 'thumbnail',
        fields: {
            height: integerAttribute('height'),
            mediaType: attribute('media-type'),
            uri: attribute('uri'),
            width: integerAttribute('width')
        },
        namespace: NS_THUMBS_1,
        path: 'thumbnail'
    }
];

// ====================================================================
const Protocol$i = [
    addAlias(NS_FORWARD_0, 'forwarded', ['message.carbon.forward']),
    {
        element: 'enable',
        namespace: NS_CARBONS_2,
        path: 'iq.carbons',
        type: 'enable',
        typeField: 'action'
    },
    {
        element: 'disable',
        namespace: NS_CARBONS_2,
        path: 'iq.carbons',
        type: 'disable',
        typeField: 'action'
    },
    {
        element: 'sent',
        namespace: NS_CARBONS_2,
        path: 'message.carbon',
        type: 'sent',
        typeField: 'type'
    },
    {
        element: 'received',
        namespace: NS_CARBONS_2,
        path: 'message.carbon',
        type: 'received',
        typeField: 'type'
    }
];

// ====================================================================
const Protocol$h = [
    ...Object.values(StreamType).map(streamNS => addAlias(streamNS, 'message', ['forward.message'])),
    ...Object.values(StreamType).map(streamNS => addAlias(streamNS, 'presence', ['forward.presence'])),
    ...Object.values(StreamType).map(streamNS => addAlias(streamNS, 'iq', ['forward.iq'])),
    addAlias(NS_DELAY, 'delay', ['forward.delay']),
    {
        aliases: ['message.forward'],
        element: 'forwarded',
        namespace: NS_FORWARD_0,
        path: 'forward'
    }
];

// ====================================================================
const Protocol$g = [
    {
        defaultType: '2',
        element: 'hash',
        fields: {
            algorithm: attribute('algo'),
            value: textBuffer('base64'),
            version: staticValue('2')
        },
        namespace: NS_HASHES_2,
        path: 'hash',
        type: '2',
        typeField: 'version'
    },
    {
        element: 'hash-used',
        fields: {
            algorithm: attribute('algo'),
            version: staticValue('2')
        },
        namespace: NS_HASHES_2,
        path: 'hashUsed'
    },
    {
        element: 'hash',
        fields: {
            algorithm: attribute('algo'),
            value: textBuffer('hex'),
            version: staticValue('1')
        },
        namespace: NS_HASHES_1,
        path: 'hash',
        type: '1',
        typeField: 'version'
    }
];

// ====================================================================
const Protocol$f = [
    {
        element: 'rtt',
        fields: {
            event: attribute('event', 'edit'),
            id: attribute('id'),
            seq: integerAttribute('seq')
        },
        namespace: NS_RTT_0,
        path: 'message.rtt'
    },
    {
        aliases: [{ path: 'message.rtt.actions', multiple: true }],
        element: 't',
        fields: {
            position: integerAttribute('p'),
            text: text()
        },
        namespace: NS_RTT_0,
        type: 'insert',
        typeField: 'type'
    },
    {
        aliases: [{ path: 'message.rtt.actions', multiple: true }],
        element: 'e',
        fields: {
            length: integerAttribute('n', 1),
            position: integerAttribute('p')
        },
        namespace: NS_RTT_0,
        type: 'erase',
        typeField: 'type'
    },
    {
        aliases: [{ multiple: true, path: 'message.rtt.actions' }],
        element: 'w',
        fields: {
            duration: integerAttribute('n', 0)
        },
        namespace: NS_RTT_0,
        type: 'wait',
        typeField: 'type'
    }
];

// ====================================================================
var XEP0308 = extendMessage({
    replace: childAttribute(NS_CORRECTION_0, 'replace', 'id')
});

// ====================================================================
const versions = {
    '2': NS_MAM_2,
    '1': NS_MAM_1
};
const Protocol$e = [
    addAlias(NS_DATAFORM, 'x', ['iq.archive.form']),
    addAlias(NS_FORWARD_0, 'forwarded', ['message.archive.item']),
    addAlias(NS_RSM, 'set', ['iq.archive.paging'])
];
for (const [version, namespace] of Object.entries(versions)) {
    Protocol$e.push({
        defaultType: 'query',
        defaultVersion: '2',
        element: 'query',
        fields: {
            node: attribute('node'),
            queryId: attribute('queryid')
        },
        namespace,
        path: 'iq.archive',
        type: 'query',
        typeField: 'type',
        version,
        versionField: 'version'
    }, {
        element: 'fin',
        fields: {
            complete: booleanAttribute('complete'),
            stable: booleanAttribute('stable')
        },
        namespace,
        path: 'iq.archive',
        type: 'result',
        version
    }, {
        element: 'prefs',
        fields: {
            default: attribute('default'),
            always: deepMultipleChildText([
                { namespace: null, element: 'always' },
                { namespace: null, element: 'jid' }
            ]),
            never: deepMultipleChildText([
                { namespace: null, element: 'never' },
                { namespace: null, element: 'jid' }
            ])
        },
        namespace,
        path: 'iq.archive',
        type: 'preferences',
        version
    }, {
        element: 'result',
        defaultType: '2',
        fields: {
            id: attribute('id'),
            queryId: attribute('queryid')
        },
        namespace,
        path: 'message.archive',
        type: version,
        typeField: 'version'
    });
}

// ====================================================================
const Protocol$d = [
    extendPresence({
        hats: splicePath(NS_HATS_0, 'hats', 'hat', true)
    }),
    {
        element: 'hat',
        fields: {
            id: attribute('name'),
            name: attribute('displayName')
        },
        namespace: NS_HATS_0,
        path: 'hat'
    }
];

// ====================================================================
var XEP0319 = extendPresence({
    idleSince: childDate(NS_IDLE_1, 'since')
});

// ====================================================================
const Protocol$c = {
    aliases: [
        {
            multiple: true,
            path: 'iq.jingle.contents.transport.fingerprints',
            selector: NS_JINGLE_ICE_UDP_1
        },
        {
            multiple: true,
            path: 'iq.jingle.contents.transport.fingerprints',
            selector: NS_JINGLE_ICE_0
        },
        {
            multiple: true,
            path: 'iq.jingle.contents.application.encryption.dtls',
            selector: NS_JINGLE_RTP_1
        }
    ],
    element: 'fingerprint',
    fields: {
        algorithm: attribute('hash'),
        setup: attribute('setup'),
        value: text()
    },
    namespace: NS_JINGLE_DTLS_0
};

// ====================================================================
const path = 'message.marker';
const Protocol$b = [
    {
        element: 'markable',
        namespace: NS_CHAT_MARKERS_0,
        path,
        type: 'markable',
        typeField: 'type'
    },
    {
        element: 'received',
        fields: {
            id: attribute('id')
        },
        namespace: NS_CHAT_MARKERS_0,
        path,
        type: 'received'
    },
    {
        element: 'displayed',
        fields: {
            id: attribute('id')
        },
        namespace: NS_CHAT_MARKERS_0,
        path,
        type: 'displayed'
    },
    {
        element: 'acknowledged',
        fields: {
            id: attribute('id')
        },
        namespace: NS_CHAT_MARKERS_0,
        path,
        type: 'acknowledged'
    }
];

// ====================================================================
function processingHints() {
    return {
        importer(xml) {
            const results = {};
            let found = false;
            for (const child of xml.children) {
                /* istanbul ignore next */
                if (typeof child === 'string') {
                    continue;
                }
                if (child.getNamespace() !== NS_HINTS) {
                    continue;
                }
                switch (child.getName()) {
                    case 'no-copy':
                        results.noCopy = true;
                        found = true;
                        break;
                    case 'no-permanent-store':
                        results.noPermanentStore = true;
                        found = true;
                        break;
                    case 'no-store':
                        results.noStore = true;
                        found = true;
                        break;
                    case 'store':
                        results.store = true;
                        found = true;
                        break;
                }
            }
            return found ? results : undefined;
        },
        exporter(xml, value, context) {
            if (value.noCopy) {
                xml.appendChild(createElement(NS_HINTS, 'no-copy', context.namespace, xml));
            }
            if (value.noPermanentStore) {
                xml.appendChild(createElement(NS_HINTS, 'no-permanent-store', context.namespace, xml));
            }
            if (value.noStore) {
                xml.appendChild(createElement(NS_HINTS, 'no-store', context.namespace, xml));
            }
            if (value.store) {
                xml.appendChild(createElement(NS_HINTS, 'store', context.namespace, xml));
            }
        }
    };
}
var XEP0334 = extendMessage({
    processingHints: processingHints()
});

// ====================================================================
const Protocol$a = [
    extendMessage({
        json: childJSON(NS_JSON_0, 'json')
    }),
    {
        aliases: pubsubItemContentAliases(),
        element: 'json',
        fields: {
            json: textJSON()
        },
        namespace: NS_JSON_0,
        type: NS_JSON_0
    },
    {
        aliases: [{ path: 'message.jsonPayloads', multiple: true }],
        element: 'payload',
        fields: {
            type: attribute('datatype'),
            data: childJSON(NS_JSON_0, 'json')
        },
        namespace: NS_JSON_MESSAGE_0
    }
];

// ====================================================================
const Protocol$9 = [
    {
        aliases: [{ path: 'iq.jingle.groups', multiple: true }],
        element: 'group',
        fields: {
            contents: multipleChildAttribute(null, 'content', 'name'),
            semantics: attribute('semantics')
        },
        namespace: NS_JINGLE_GROUPING_0
    }
];

// ====================================================================
const Protocol$8 = {
    aliases: [
        {
            path: 'iq.jingle.contents.transport.sctp',
            selector: NS_JINGLE_ICE_UDP_1
        },
        {
            path: 'iq.jingle.contents.transport.sctp',
            selector: NS_JINGLE_ICE_0
        }
    ],
    element: 'sctpmap',
    fields: {
        port: integerAttribute('number'),
        protocol: attribute('protocol'),
        streams: attribute('streams')
    },
    namespace: NS_JINGLE_DTLS_SCTP_1
};

// ====================================================================
const Protocol$7 = [
    {
        element: 'active',
        namespace: NS_CSI_0,
        path: 'csi',
        type: 'active',
        typeField: 'state'
    },
    {
        element: 'inactive',
        namespace: NS_CSI_0,
        path: 'csi',
        type: 'inactive',
        typeField: 'state'
    }
];

// ====================================================================
const Protocol$6 = [
    addAlias(NS_DATAFORM, 'x', ['iq.push.form', 'pushNotification.form']),
    {
        element: 'enable',
        fields: {
            jid: JIDAttribute('jid'),
            node: attribute('node')
        },
        namespace: NS_PUSH_0,
        path: 'iq.push',
        type: 'enable',
        typeField: 'action'
    },
    {
        element: 'disable',
        fields: {
            jid: JIDAttribute('jid'),
            node: attribute('node')
        },
        namespace: NS_PUSH_0,
        path: 'iq.push',
        type: 'disable',
        typeField: 'action'
    },
    {
        aliases: pubsubItemContentAliases(),
        element: 'notification',
        namespace: NS_PUSH_0,
        path: 'pushNotification',
        type: NS_PUSH_0,
        typeField: 'itemType'
    }
];

// ====================================================================
const Protocol$5 = [
    extendMessage({
        originId: childAttribute(NS_SID_0, 'origin-id', 'id')
    }),
    {
        aliases: [{ path: 'message.stanzaIds', multiple: true }],
        element: 'stanza-id',
        fields: {
            by: JIDAttribute('by'),
            id: attribute('id')
        },
        namespace: NS_SID_0
    }
];

// ====================================================================
const Protocol$4 = [
    extendStanzaError({
        httpUploadError: childEnum(NS_HTTP_UPLOAD_0, ['file-too-large', 'retry']),
        httpUploadMaxFileSize: deepChildInteger([
            { namespace: NS_HTTP_UPLOAD_0, element: 'file-too-large' },
            { namespace: NS_HTTP_UPLOAD_0, element: 'max-file-size' }
        ]),
        httpUploadRetry: childDateAttribute(NS_HTTP_UPLOAD_0, 'retry', 'stamp')
    }),
    {
        element: 'request',
        fields: {
            mediaType: attribute('content-type'),
            name: attribute('filename'),
            size: integerAttribute('size')
        },
        namespace: NS_HTTP_UPLOAD_0,
        path: 'iq.httpUpload',
        type: 'request',
        typeField: 'type'
    },
    {
        element: 'slot',
        fields: {
            download: childAttribute(null, 'get', 'url')
        },
        namespace: NS_HTTP_UPLOAD_0,
        path: 'iq.httpUpload',
        type: 'slot'
    },
    {
        aliases: [{ path: 'iq.httpUpload.upload', selector: 'slot' }],
        element: 'put',
        fields: {
            url: attribute('url')
        },
        namespace: NS_HTTP_UPLOAD_0
    },
    {
        aliases: [{ path: 'iq.httpUpload.upload.headers', multiple: true }],
        element: 'header',
        fields: {
            name: attribute('name'),
            value: text()
        },
        namespace: NS_HTTP_UPLOAD_0
    }
];

// ====================================================================
const Protocol$3 = {
    element: 'encryption',
    fields: {
        id: attribute('namespace'),
        name: attribute('name')
    },
    namespace: NS_EME_0,
    path: 'message.encryptionMethod'
};

// ====================================================================
const Protocol$2 = [
    {
        aliases: ['message.omemo'],
        element: 'encrypted',
        fields: {
            payload: childTextBuffer(null, 'payload', 'base64')
        },
        namespace: NS_OMEMO_AXOLOTL,
        path: 'omemo'
    },
    {
        element: 'header',
        fields: {
            iv: childTextBuffer(null, 'iv', 'base64'),
            sid: integerAttribute('sid')
        },
        namespace: NS_OMEMO_AXOLOTL,
        path: 'omemo.header'
    },
    {
        aliases: [{ path: 'omemo.header.keys', multiple: true }],
        element: 'key',
        fields: {
            preKey: booleanAttribute('prekey'),
            rid: integerAttribute('rid'),
            value: textBuffer('base64')
        },
        namespace: NS_OMEMO_AXOLOTL
    },
    {
        aliases: pubsubItemContentAliases(),
        element: 'list',
        fields: {
            devices: multipleChildIntegerAttribute(null, 'device', 'id')
        },
        namespace: NS_OMEMO_AXOLOTL,
        type: NS_OMEMO_AXOLOTL_DEVICELIST,
        typeField: 'itemType'
    },
    {
        element: 'preKeyPublic',
        fields: {
            id: integerAttribute('preKeyId'),
            value: textBuffer('base64')
        },
        namespace: NS_OMEMO_AXOLOTL,
        path: 'omemoPreKey'
    },
    {
        element: 'signedPreKeyPublic',
        fields: {
            id: integerAttribute('signedPreKeyId'),
            value: textBuffer('base64')
        },
        namespace: NS_OMEMO_AXOLOTL,
        path: 'omemoDevice.signedPreKeyPublic'
    },
    {
        aliases: pubsubItemContentAliases(),
        element: 'bundle',
        fields: {
            identityKey: childTextBuffer(null, 'identityKey', 'base64'),
            preKeys: splicePath(null, 'prekeys', 'omemoPreKey', true),
            signedPreKeySignature: childTextBuffer(null, 'signedPreKeySignature', 'base64')
        },
        namespace: NS_OMEMO_AXOLOTL,
        path: 'omemoDevice',
        type: NS_OMEMO_AXOLOTL_BUNDLES,
        typeField: 'itemType'
    }
];

// ====================================================================
const Protocol$1 = [
    {
        element: 'XRD',
        fields: {
            subject: childText(null, 'Subject')
        },
        namespace: NS_XRD,
        path: 'xrd'
    },
    {
        aliases: [{ path: 'xrd.links', multiple: true }],
        element: 'Link',
        fields: {
            href: attribute('href'),
            rel: attribute('rel'),
            type: attribute('type')
        },
        namespace: NS_XRD
    }
];

// ====================================================================
const Protocol = [
    Protocol$1a,
    Protocol$19,
    Protocol$18,
    Protocol$17,
    Protocol$16,
    Protocol$15,
    Protocol$14,
    Protocol$13,
    Protocol$12,
    Protocol$11,
    Protocol$10,
    Protocol$$,
    Protocol$_,
    Protocol$Z,
    Protocol$Y,
    Protocol$X,
    Protocol$W,
    Protocol$V,
    Protocol$U,
    Protocol$T,
    Protocol$S,
    Protocol$R,
    Protocol$Q,
    Protocol$P,
    Protocol$O,
    XEP0085,
    Protocol$N,
    Protocol$M,
    Protocol$L,
    Protocol$K,
    Protocol$J,
    Protocol$I,
    Protocol$G,
    Protocol$F,
    Protocol$E,
    Protocol$H,
    Protocol$D,
    XEP0153,
    Protocol$C,
    Protocol$B,
    Protocol$A,
    Protocol$z,
    Protocol$y,
    Protocol$x,
    Protocol$w,
    Protocol$v,
    Protocol$u,
    Protocol$t,
    XEP0199,
    Protocol$s,
    Protocol$r,
    Protocol$q,
    Protocol$p,
    XEP0224,
    Protocol$o,
    XEP0234,
    Protocol$m,
    Protocol$l,
    Protocol$k,
    Protocol$j,
    Protocol$i,
    Protocol$h,
    Protocol$g,
    Protocol$f,
    XEP0308,
    Protocol$e,
    Protocol$d,
    XEP0319,
    Protocol$c,
    Protocol$b,
    XEP0334,
    Protocol$a,
    Protocol$9,
    Protocol$8,
    Protocol$7,
    Protocol$6,
    Protocol$5,
    Protocol$4,
    Protocol$3,
    Protocol$2,
    Protocol$1
];

var index = /*#__PURE__*/Object.freeze({
    __proto__: null,
    'default': Protocol
});

class RequestChannel {
    constructor(stream) {
        this.active = false;
        this.maxRetries = 5;
        this.stream = stream;
        this.maxTimeout = 1000 * 1.1 * this.stream.maxWaitTime;
    }
    send(rid, body) {
        return __awaiter(this, void 0, void 0, function* () {
            this.rid = rid;
            this.active = true;
            let attempts = 0;
            while (attempts <= this.maxRetries) {
                attempts += 1;
                try {
                    const res = yield timeoutPromise(nativeFetch(this.stream.url, {
                        body,
                        headers: {
                            'Content-Type': this.stream.contentType
                        },
                        method: 'POST'
                    }), this.maxTimeout, () => new Error('Request timed out'));
                    if (!res.ok) {
                        throw new Error('HTTP Status Error: ' + res.status);
                    }
                    const result = yield res.text();
                    this.active = false;
                    return result;
                }
                catch (err) {
                    if (attempts === 1) {
                        continue;
                    }
                    else if (attempts < this.maxRetries) {
                        const backoff = Math.min(this.maxTimeout, Math.pow(attempts, 2) * 1000);
                        yield sleep(backoff + Math.random() * 1000);
                        continue;
                    }
                    else {
                        this.active = false;
                        throw err;
                    }
                }
            }
            throw new Error('Request failed');
        });
    }
}
class BOSH extends Duplex {
    constructor(client, sm, stanzas) {
        super({
            objectMode: true
        });
        this.rid = Math.floor(Math.random() * 0xffffffff);
        this.sid = '';
        this.maxHoldOpen = 2;
        this.maxWaitTime = 30;
        this.contentType = 'text/xml; charset=utf-8';
        this.channels = [new RequestChannel(this), new RequestChannel(this)];
        this.activeChannelID = 0;
        this.queue = [];
        this.isEnded = false;
        this.client = client;
        this.sm = sm;
        this.stanzas = stanzas;
        this.on('data', e => {
            this.client.emit('stream:data', e.stanza, e.kind);
        });
        this.on('end', () => {
            this.isEnded = true;
            clearTimeout(this.idleTimeout);
            if (this.client.transport === this) {
                this.client.emit('--transport-disconnected');
            }
        });
    }
    _write(chunk, encoding, done) {
        this.queue.push([chunk, done]);
        this.scheduleRequests();
    }
    _writev(chunks, done) {
        this.queue.push([chunks.map(c => c.chunk).join(''), done]);
        this.scheduleRequests();
    }
    _read() {
        return;
    }
    process(result) {
        const parser = new StreamParser({
            acceptLanguages: this.config.acceptLanguages,
            allowComments: false,
            lang: this.config.lang,
            registry: this.stanzas,
            rootKey: 'bosh',
            wrappedStream: true
        });
        parser.on('error', (err) => {
            const streamError = {
                condition: StreamErrorCondition.InvalidXML
            };
            this.client.emit('stream:error', streamError, err);
            this.send('error', streamError);
            return this.disconnect();
        });
        parser.on('data', (e) => {
            if (e.event === 'stream-start') {
                this.stream = e.stanza;
                if (e.stanza.type === 'terminate') {
                    this.hasStream = false;
                    this.rid = undefined;
                    this.sid = undefined;
                    if (!this.isEnded) {
                        this.isEnded = true;
                        this.client.emit('bosh:terminate', e.stanza);
                        this.client.emit('stream:end');
                        this.push(null);
                    }
                }
                else if (!this.hasStream) {
                    this.hasStream = true;
                    this.stream = e.stanza;
                    this.sid = e.stanza.sid || this.sid;
                    this.maxWaitTime = e.stanza.maxWaitTime || this.maxWaitTime;
                    this.client.emit('stream:start', e.stanza);
                }
                return;
            }
            if (!e.event) {
                this.push({ kind: e.kind, stanza: e.stanza });
            }
        });
        this.client.emit('raw', 'incoming', result);
        parser.write(result);
        this.scheduleRequests();
    }
    connect(opts) {
        this.config = opts;
        this.url = opts.url;
        if (opts.rid) {
            this.rid = opts.rid;
        }
        if (opts.sid) {
            this.sid = opts.sid;
        }
        if (opts.wait) {
            this.maxWaitTime = opts.wait;
        }
        if (opts.maxHoldOpen) {
            this.maxHoldOpen = opts.maxHoldOpen;
        }
        if (this.sid) {
            this.hasStream = true;
            this.stream = {};
            this.client.emit('connected');
            this.client.emit('session:prebind', this.config.jid);
            this.client.emit('session:started');
            return;
        }
        this._send({
            lang: opts.lang,
            maxHoldOpen: this.maxHoldOpen,
            maxWaitTime: this.maxWaitTime,
            to: opts.server,
            version: '1.6',
            xmppVersion: '1.0'
        });
    }
    restart() {
        this.hasStream = false;
        this._send({
            to: this.config.server,
            xmppRestart: true
        });
    }
    disconnect(clean = true) {
        if (this.hasStream && clean) {
            this._send({
                type: 'terminate'
            });
        }
        else {
            this.stream = undefined;
            this.sid = undefined;
            this.rid = undefined;
            this.client.emit('--transport-disconnected');
        }
    }
    send(dataOrName, data) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            let output;
            if (data) {
                output = (_a = this.stanzas.export(dataOrName, data)) === null || _a === void 0 ? void 0 : _a.toString();
            }
            if (!output) {
                return;
            }
            return new Promise((resolve, reject) => {
                this.write(output, 'utf8', err => (err ? reject(err) : resolve()));
            });
        });
    }
    get sendingChannel() {
        return this.channels[this.activeChannelID];
    }
    get pollingChannel() {
        return this.channels[this.activeChannelID === 0 ? 1 : 0];
    }
    toggleChannel() {
        this.activeChannelID = this.activeChannelID === 0 ? 1 : 0;
    }
    _send(boshData, payload = '') {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isEnded) {
                return;
            }
            const rid = this.rid++;
            const header = this.stanzas.export('bosh', Object.assign(Object.assign({}, boshData), { rid, sid: this.sid }));
            let body;
            if (payload) {
                body = [header.openTag(), payload, header.closeTag()].join('');
            }
            else {
                body = header.toString();
            }
            this.client.emit('raw', 'outgoing', body);
            this.sendingChannel
                .send(rid, body)
                .then(result => {
                this.process(result);
            })
                .catch(err => {
                this.end(err);
            });
            this.toggleChannel();
        });
    }
    _poll() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isEnded) {
                return;
            }
            const rid = this.rid++;
            const body = this.stanzas
                .export('bosh', {
                rid,
                sid: this.sid
            })
                .toString();
            this.client.emit('raw', 'outgoing', body);
            this.pollingChannel
                .send(rid, body)
                .then(result => {
                this.process(result);
            })
                .catch(err => {
                this.end(err);
            });
        });
    }
    scheduleRequests() {
        clearTimeout(this.idleTimeout);
        this.idleTimeout = setTimeout(() => {
            this.fireRequests();
        }, 10);
    }
    fireRequests() {
        if (this.isEnded) {
            return;
        }
        if (this.queue.length) {
            if (!this.sendingChannel.active) {
                const [data, done] = this.queue.shift();
                this._send({}, data);
                done();
            }
            else {
                this.scheduleRequests();
            }
            return;
        }
        if (this.authenticated && !(this.channels[0].active || this.channels[1].active)) {
            this._poll();
        }
    }
}

const WS_OPEN = 1;
class WSConnection extends Duplex {
    constructor(client, sm, stanzas) {
        super({ objectMode: true });
        this.sm = sm;
        this.stanzas = stanzas;
        this.closing = false;
        this.client = client;
        this.on('data', e => {
            this.client.emit('stream:data', e.stanza, e.kind);
        });
        this.on('error', () => {
            this.end();
        });
        this.on('end', () => {
            if (this.client.transport === this) {
                this.client.emit('--transport-disconnected');
            }
        });
    }
    _read() {
        return;
    }
    _write(chunk, encoding, done) {
        if (!this.socket || this.socket.readyState !== WS_OPEN) {
            return done(new Error('Socket closed'));
        }
        const data = Buffer.from(chunk, 'utf8').toString();
        this.client.emit('raw', 'outgoing', data);
        this.socket.send(data);
        done();
    }
    connect(opts) {
        this.config = opts;
        this.hasStream = false;
        this.closing = false;
        this.parser = new StreamParser({
            acceptLanguages: this.config.acceptLanguages,
            allowComments: false,
            lang: this.config.lang,
            registry: this.stanzas,
            wrappedStream: false
        });
        this.parser.on('data', (e) => {
            const name = e.kind;
            const stanzaObj = e.stanza;
            if (name === 'stream') {
                if (stanzaObj.action === 'open') {
                    this.hasStream = true;
                    this.stream = stanzaObj;
                    return this.client.emit('stream:start', stanzaObj);
                }
                if (stanzaObj.action === 'close') {
                    this.client.emit('stream:end');
                    return this.disconnect();
                }
            }
            this.push({ kind: e.kind, stanza: e.stanza });
        });
        this.parser.on('error', (err) => {
            const streamError = {
                condition: StreamErrorCondition.InvalidXML
            };
            this.client.emit('stream:error', streamError, err);
            this.write(this.stanzas.export('error', streamError).toString());
            return this.disconnect();
        });
        this.socket = new nativeWS(opts.url, 'xmpp');
        this.socket.onopen = () => {
            this.emit('connect');
            this.sm.started = false;
            this.client.emit('connected');
            this.write(this.startHeader());
        };
        this.socket.onmessage = wsMsg => {
            const data = Buffer.from(wsMsg.data, 'utf8').toString();
            this.client.emit('raw', 'incoming', data);
            if (this.parser) {
                this.parser.write(data);
            }
        };
        this.socket.onclose = () => {
            this.push(null);
        };
    }
    disconnect(clean = true) {
        if (this.socket && !this.closing && this.hasStream && clean) {
            this.closing = true;
            this.write(this.closeHeader());
        }
        else {
            this.hasStream = false;
            this.stream = undefined;
            if (this.socket) {
                this.end();
                this.socket.close();
                if (this.client.transport === this) {
                    this.client.emit('--transport-disconnected');
                }
            }
            this.socket = undefined;
        }
    }
    send(dataOrName, data) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            let output;
            if (data) {
                output = (_a = this.stanzas.export(dataOrName, data)) === null || _a === void 0 ? void 0 : _a.toString();
            }
            if (!output) {
                return;
            }
            return new Promise((resolve, reject) => {
                this.write(output, 'utf8', err => (err ? reject(err) : resolve()));
            });
        });
    }
    restart() {
        this.hasStream = false;
        this.write(this.startHeader());
    }
    startHeader() {
        const header = this.stanzas.export('stream', {
            action: 'open',
            lang: this.config.lang,
            to: this.config.server,
            version: '1.0'
        });
        return header.toString();
    }
    closeHeader() {
        const header = this.stanzas.export('stream', {
            action: 'close'
        });
        return header.toString();
    }
}

class Client extends EventEmitter {
    constructor(opts = {}) {
        super();
        this.reconnectAttempts = 0;
        this.setMaxListeners(100);
        // Some EventEmitter shims don't include off()
        this.off = this.removeListener;
        this.updateConfig(opts);
        this.jid = '';
        this.sasl = new Factory();
        this.sasl.register('EXTERNAL', EXTERNAL, 1000);
        this.sasl.register('SCRAM-SHA-256-PLUS', SCRAM, 350);
        this.sasl.register('SCRAM-SHA-256', SCRAM, 300);
        this.sasl.register('SCRAM-SHA-1-PLUS', SCRAM, 250);
        this.sasl.register('SCRAM-SHA-1', SCRAM, 200);
        this.sasl.register('DIGEST-MD5', DIGEST, 100);
        this.sasl.register('OAUTHBEARER', OAUTH, 100);
        this.sasl.register('X-OAUTH2', PLAIN, 50);
        this.sasl.register('PLAIN', PLAIN, 1);
        this.sasl.register('ANONYMOUS', ANONYMOUS, 0);
        this.stanzas = new Registry();
        this.stanzas.define(Protocol);
        this.use(core);
        this.sm = new StreamManagement();
        if (this.config.allowResumption !== undefined) {
            this.sm.allowResume = this.config.allowResumption;
        }
        this.sm.on('prebound', jid => {
            this.jid = jid;
            this.emit('session:bound', jid);
        });
        this.on('session:bound', jid => this.sm.bind(jid));
        this.sm.on('send', sm => this.send('sm', sm));
        this.sm.on('acked', acked => this.emit('stanza:acked', acked));
        this.sm.on('failed', failed => this.emit('stanza:failed', failed));
        this.sm.on('hibernated', data => this.emit('stanza:hibernated', data));
        // We disable outgoing processing while stanza resends are queued up
        // to prevent any interleaving.
        this.sm.on('begin-resend', () => this.outgoingDataQueue.pause());
        this.sm.on('resend', ({ kind, stanza }) => this.send(kind, stanza, true));
        this.sm.on('end-resend', () => this.outgoingDataQueue.resume());
        // Create message:* flavors of stanza:* SM events
        for (const type of ['acked', 'hibernated', 'failed']) {
            this.on(`stanza:${type}`, (data) => {
                if (data.kind === 'message') {
                    this.emit(`message:${type}`, data.stanza);
                    //yjing modify start
                    this.emit(`message:${type}:${data.stanza.id}`, data);
                    //yjing modify end
                }
            });
        }
        this.transports = {
            bosh: BOSH,
            websocket: WSConnection
        };
        this.incomingDataQueue = priorityQueue((task, done) => __awaiter(this, void 0, void 0, function* () {
            const { kind, stanza } = task;
            this.emit(kind, stanza);
            if (stanza.id) {
                this.emit((kind + ':id:' + stanza.id), stanza);
            }
            if (kind === 'message' || kind === 'presence' || kind === 'iq') {
                this.emit('stanza', stanza);
                yield this.sm.handle();
            }
            else if (kind === 'sm') {
                if (stanza.type === 'ack') {
                    yield this.sm.process(stanza);
                    this.emit('stream:management:ack', stanza);
                }
                if (stanza.type === 'request') {
                    this.sm.ack();
                }
            }
            if (done) {
                done();
            }
        }), 1);
        const handleFailedSend = (kind, stanza) => {
            if (['message', 'presence', 'iq'].includes(kind)) {
                if (!this.sm.started || !this.sm.resumable) {
                    this.emit('stanza:failed', {
                        kind,
                        stanza
                    });
                }
                else if (this.sm.resumable && !this.transport) {
                    this.emit('stanza:hibernated', {
                        kind,
                        stanza
                    });
                }
            }
        };
        this.outgoingDataQueue = priorityQueue((task, done) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            const { kind, stanza, replay } = task;
            const ackRequest = replay || (yield this.sm.track(kind, stanza));
            if (kind === 'message') {
                if (replay) {
                    this.emit('message:retry', stanza);
                }
                else {
                    this.emit('message:sent', stanza, false);
                }
            }
            if (this.transport) {
                try {
                    yield this.transport.send(kind, stanza);
                    if (ackRequest) {
                        (_a = this.transport) === null || _a === void 0 ? void 0 : _a.send('sm', { type: 'request' });
                    }
                }
                catch (err) {
                    console.error(err);
                    handleFailedSend(kind, stanza);
                }
            }
            else {
                handleFailedSend(kind, stanza);
            }
            if (done) {
                done();
            }
        }), 1);
        this.on('stream:data', (json, kind) => {
            this.incomingDataQueue.push({
                kind,
                stanza: json
            }, 0);
        });
        this.on('--transport-disconnected', () => __awaiter(this, void 0, void 0, function* () {
            const drains = [];
            if (!this.incomingDataQueue.idle()) {
                drains.push(this.incomingDataQueue.drain());
            }
            if (!this.outgoingDataQueue.idle()) {
                drains.push(this.outgoingDataQueue.drain());
            }
            yield Promise.all(drains);
            yield this.sm.hibernate();
            if (this.transport) {
                delete this.transport;
            }
            this.emit('--reset-stream-features');
            if (!this.sessionTerminating && this.config.autoReconnect) {
                this.reconnectAttempts += 1;
                clearTimeout(this.reconnectTimer);
                this.reconnectTimer = setTimeout(() => {
                    this.connect();
                }, 1000 * Math.min(Math.pow(2, this.reconnectAttempts) + Math.random(), this.config.maxReconnectBackoff || 32));
            }
            this.emit('disconnected');
        }));
        this.on('iq', (iq) => {
            const iqType = iq.type;
            const payloadType = iq.payloadType;
            const iqEvent = 'iq:' + iqType + ':' + payloadType;
            if (iqType === 'get' || iqType === 'set') {
                if (payloadType === 'invalid-payload-count') {
                    return this.sendIQError(iq, {
                        error: {
                            condition: 'bad-request',
                            type: 'modify'
                        }
                    });
                }
                if (payloadType === 'unknown-payload' || this.listenerCount(iqEvent) === 0) {
                    return this.sendIQError(iq, {
                        error: {
                            condition: 'service-unavailable',
                            type: 'cancel'
                        }
                    });
                }
                this.emit(iqEvent, iq);
            }
        });
        this.on('message', msg => {
            const isChat = (msg.alternateLanguageBodies && msg.alternateLanguageBodies.length) ||
                (msg.links && msg.links.length);
            const isMarker = msg.marker && msg.marker.type !== 'markable';
            if (isChat && !isMarker) {
                if (msg.type === 'chat' || msg.type === 'normal') {
                    this.emit('chat', msg);
                }
                else if (msg.type === 'groupchat') {
                    this.emit('groupchat', msg);
                }
            }
            if (msg.type === 'error') {
                this.emit('message:error', msg);
            }
        });
        this.on('presence', (pres) => {
            let presType = pres.type || 'available';
            if (presType === 'error') {
                presType = 'presence:error';
            }
            this.emit(presType, pres);
        });
        this.on('session:started', () => {
            this.sessionStarting = false;
            this.reconnectAttempts = 0;
            if (this.reconnectTimer) {
                clearTimeout(this.reconnectTimer);
            }
        });
    }
    updateConfig(opts = {}) {
        const currConfig = this.config || {};
        this.config = Object.assign(Object.assign({ allowResumption: true, jid: '', transports: {
                bosh: true,
                websocket: true
            }, useStreamManagement: true }, currConfig), opts);
        if (!this.config.server) {
            this.config.server = getDomain(this.config.jid);
        }
        if (this.config.password) {
            this.config.credentials = this.config.credentials || {};
            this.config.credentials.password = this.config.password;
            delete this.config.password;
        }
    }
    get stream() {
        return this.transport ? this.transport.stream : undefined;
    }
    emit(name, ...args) {
        // Continue supporting the most common and useful wildcard events
        const res = super.emit(name, ...args);
        if (name === 'raw') {
            super.emit(`raw:${args[0]}`, args[1]);
            super.emit('raw:*', `raw:${args[0]}`, args[1]);
            super.emit('*', `raw:${args[0]}`, args[1]);
        }
        else {
            super.emit('*', name, ...args);
        }
        return res;
    }
    use(pluginInit) {
        if (typeof pluginInit !== 'function') {
            return;
        }
        pluginInit(this, this.stanzas, this.config);
    }
    nextId() {
        return uuid();
    }
    getCredentials() {
        return __awaiter(this, void 0, void 0, function* () {
            return this._getConfiguredCredentials();
        });
    }
    connect() {
        return __awaiter(this, void 0, void 0, function* () {
            this.sessionTerminating = false;
            this.sessionStarting = true;
            this.emit('--reset-stream-features');
            if (this.transport) {
                this.transport.disconnect(false);
            }
            const transportPref = ['websocket', 'bosh'];
            let endpoints;
            for (const name of transportPref) {
                let conf = this.config.transports[name];
                if (!conf) {
                    continue;
                }
                if (typeof conf === 'string') {
                    conf = { url: conf };
                }
                else if (conf === true) {
                    if (!endpoints) {
                        try {
                            endpoints = yield this.discoverBindings(this.config.server);
                        }
                        catch (err) {
                            console.error(err);
                            continue;
                        }
                    }
                    endpoints[name] = (endpoints[name] || []).filter(url => url.startsWith('wss:') || url.startsWith('https:'));
                    if (!endpoints[name] || !endpoints[name].length) {
                        continue;
                    }
                    conf = { url: endpoints[name][0] };
                }
                this.transport = new this.transports[name](this, this.sm, this.stanzas);
                this.transport.connect(Object.assign({ acceptLanguages: this.config.acceptLanguages || ['en'], jid: this.config.jid, lang: this.config.lang || 'en', server: this.config.server, url: conf.url }, conf));
                return;
            }
            console.error('No endpoints found for the requested transports.');
            this.emit('--transport-disconnected');
        });
    }
    disconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            this.sessionTerminating = true;
            clearTimeout(this.reconnectTimer);
            this.outgoingDataQueue.pause();
            if (this.sessionStarted && !this.sm.started) {
                // Only emit session:end if we had a session, and we aren't using
                // stream management to keep the session alive.
                this.emit('session:end');
            }
            this.emit('--reset-stream-features');
            this.sessionStarted = false;
            if (this.transport) {
                this.transport.disconnect();
            }
            else {
                this.emit('--transport-disconnected');
            }
            this.outgoingDataQueue.resume();
            if (!this.outgoingDataQueue.idle()) {
                yield this.outgoingDataQueue.drain();
            }
            yield this.sm.shutdown();
        });
    }
    send(kind, stanza, replay = false) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.outgoingDataQueue.push({ kind, stanza, replay }, replay ? 0 : 1, err => err ? reject(err) : resolve());
            });
        });
    }
    sendMessage(data) {
        const id = data.id || this.nextId();
        const msg = Object.assign({ id, originId: id }, data);
        //yjing modify start
        const respEvent = 'message:acked:' + id;
        const request = new Promise((resolve) => {
            const handler = (res) => {
                this.off(respEvent, handler);
                resolve(res);
            };
            this.on(respEvent, handler);
        });
        return this.send('message', msg).then(res => {
            const timeout = this.config.timeout || 15;
            console.log(`sendMessage res ==> ${res}`);
            return timeoutPromise(request, timeout * 1000, () => (Object.assign(Object.assign({}, msg), { error: {
                    condition: 'timeout',
                    text: `Request timed out after ${timeout} seconds.`,
                }, type: 'error' })));
        });
        //return msg.id
        //yjing modify end
    }
    sendPresence(data = {}) {
        const pres = Object.assign({ id: this.nextId() }, data);
        this.send('presence', pres);
        return pres.id;
    }
    sendIQ(data) {
        const iq = Object.assign({ id: this.nextId() }, data);
        const allowed = allowedResponders(this.jid, data.to);
        const respEvent = 'iq:id:' + iq.id;
        const request = new Promise((resolve, reject) => {
            const handler = (res) => {
                // Only process result from the correct responder
                if (!allowed.has(res.from)) {
                    return;
                }
                // Only process result or error responses, if the responder
                // happened to send us a request using the same ID value at
                // the same time.
                if (res.type !== 'result' && res.type !== 'error') {
                    return;
                }
                this.off(respEvent, handler);
                if (res.type === 'result') {
                    resolve(res);
                }
                else {
                    reject(res);
                }
            };
            this.on(respEvent, handler);
        });
        this.send('iq', iq);
        const timeout = this.config.timeout || 15;
        return timeoutPromise(request, timeout * 1000, () => (Object.assign(Object.assign({}, iq), { to: undefined, from: undefined, error: {
                condition: 'timeout',
                text: `Request timed out after ${timeout} seconds.`
            }, id: iq.id, type: 'error' })));
    }
    sendIQResult(original, reply) {
        this.send('iq', Object.assign(Object.assign({}, reply), { id: original.id, to: original.from, type: 'result' }));
    }
    sendIQError(original, error) {
        this.send('iq', Object.assign(Object.assign({}, error), { id: original.id, to: original.from, type: 'error' }));
    }
    sendStreamError(error) {
        this.emit('stream:error', error);
        this.send('error', error);
        this.disconnect();
    }
    _getConfiguredCredentials() {
        const creds = this.config.credentials || {};
        const requestedJID = parse$2(this.config.jid || '');
        const username = creds.username || requestedJID.local;
        const server = creds.host || requestedJID.domain;
        return Object.assign({ host: server, password: this.config.password, realm: server, serviceName: server, serviceType: 'xmpp', username }, creds);
    }
}

/**
 * Calculate the erase and insert actions needed to describe the user's edit operation.
 *
 * Based on the code point buffers before and after the edit, we find the single erase
 * and insert actions needed to describe the full change. We are minimizing the number
 * of deltas, not minimizing the number of affected code points.
 *
 * @param oldText The original buffer of Unicode code points before the user's edit action.
 * @param newText The new buffer of Unicode code points after the user's edit action.
 */
function diff(oldText, newText) {
    const oldLen = oldText.length;
    const newLen = newText.length;
    const searchLen = Math.min(oldLen, newLen);
    let prefixSize = 0;
    for (prefixSize = 0; prefixSize < searchLen; prefixSize++) {
        if (oldText[prefixSize] !== newText[prefixSize]) {
            break;
        }
    }
    let suffixSize = 0;
    for (suffixSize = 0; suffixSize < searchLen - prefixSize; suffixSize++) {
        if (oldText[oldLen - suffixSize - 1] !== newText[newLen - suffixSize - 1]) {
            break;
        }
    }
    const matchedSize = prefixSize + suffixSize;
    const events = [];
    if (matchedSize < oldLen) {
        events.push({
            length: oldLen - matchedSize,
            position: oldLen - suffixSize,
            type: 'erase'
        });
    }
    if (matchedSize < newLen) {
        const insertedText = newText.slice(prefixSize, prefixSize + newLen - matchedSize);
        events.push({
            position: prefixSize,
            text: Punycode.ucs2.encode(insertedText),
            type: 'insert'
        });
    }
    return events;
}
/**
 * Class for processing RTT events and providing a renderable string of the resulting text.
 */
class DisplayBuffer {
    constructor(onStateChange, ignoreWaits = false) {
        this.synced = false;
        this.cursorPosition = 0;
        this.ignoreWaits = false;
        this.timeDeficit = 0;
        this.sequenceNumber = 0;
        this.onStateChange =
            onStateChange ||
                function noop() {
                    return;
                };
        this.ignoreWaits = ignoreWaits;
        this.buffer = [];
        this.resetActionQueue();
    }
    /**
     * The encoded Unicode string to display.
     */
    get text() {
        return Punycode.ucs2.encode(this.buffer.slice());
    }
    /**
     * Mark the RTT message as completed and reset state.
     */
    commit() {
        this.resetActionQueue();
    }
    /**
     * Accept an RTT event for processing.
     *
     * A single event can contain multiple edit actions, including
     * wait pauses.
     *
     * Events must be processed in order of their `seq` value in order
     * to stay in sync.
     *
     * @param event {RTTEvent} The RTT event to process.
     */
    process(event) {
        if (event.event === 'cancel' || event.event === 'init') {
            this.resetActionQueue();
            return;
        }
        else if (event.event === 'reset' || event.event === 'new') {
            this.resetActionQueue();
            if (event.seq !== undefined) {
                this.sequenceNumber = event.seq;
            }
        }
        else if (event.seq !== this.sequenceNumber) {
            this.synced = false;
        }
        if (event.actions) {
            const baseTime = Date.now();
            let accumulatedWait = 0;
            for (const action of event.actions) {
                action.baseTime = baseTime + accumulatedWait;
                if (action.type === 'wait') {
                    accumulatedWait += action.duration;
                }
                this.actionQueue.push(action, 0);
            }
        }
        this.sequenceNumber = this.sequenceNumber + 1;
    }
    /**
     * Insert text into the Unicode code point buffer
     *
     * By default, the insertion position is the end of the buffer.
     *
     * @param text The raw text to insert
     * @param position The position to start the insertion
     */
    insert(text = '', position = this.buffer.length) {
        text = text.normalize('NFC');
        const insertedText = Punycode.ucs2.decode(text);
        this.buffer.splice(position, 0, ...insertedText);
        this.cursorPosition = position + insertedText.length;
        this.emitState();
    }
    /**
     * Erase text from the Unicode code point buffer
     *
     * By default, the erased text length is `1`, and the position is the end of the buffer.
     *
     * @param length The number of code points to erase from the buffer, starting at {position} and erasing to the left.
     * @param position The position to start erasing code points. Erasing continues to the left.
     */
    erase(length = 1, position = this.buffer.length) {
        position = Math.max(Math.min(position, this.buffer.length), 0);
        length = Math.max(Math.min(length, this.text.length), 0);
        this.buffer.splice(Math.max(position - length, 0), length);
        this.cursorPosition = Math.max(position - length, 0);
        this.emitState();
    }
    emitState(additional = {}) {
        this.onStateChange(Object.assign({ cursorPosition: this.cursorPosition, synced: this.synced, text: this.text }, additional));
    }
    /**
     * Reset the processing state and queue.
     *
     * Used when 'init', 'new', 'reset', and 'cancel' RTT events are processed.
     */
    resetActionQueue() {
        if (this.actionQueue) {
            this.actionQueue.kill();
        }
        this.sequenceNumber = 0;
        this.synced = true;
        this.buffer = [];
        this.timeDeficit = 0;
        this.actionQueue = priorityQueue((action, done) => {
            const currentTime = Date.now();
            if (action.type === 'insert') {
                this.insert(action.text, action.position);
                return done();
            }
            else if (action.type === 'erase') {
                this.erase(action.length, action.position);
                return done();
            }
            else if (action.type === 'wait') {
                if (this.ignoreWaits) {
                    return done();
                }
                if (action.duration > 700) {
                    action.duration = 700;
                }
                const waitTime = action.duration - (currentTime - action.baseTime) + this.timeDeficit;
                if (waitTime <= 0) {
                    this.timeDeficit = waitTime;
                    return done();
                }
                else {
                    this.timeDeficit = 0;
                    setTimeout(() => done(), waitTime);
                }
            }
            else {
                return done();
            }
        }, 1);
        this.emitState();
    }
}
/**
 * Class for tracking changes in a source text, and generating RTT events based on those changes.
 */
class InputBuffer {
    constructor(onStateChange, ignoreWaits = false) {
        this.resetInterval = 10000;
        this.ignoreWaits = false;
        this.isStarting = false;
        this.isReset = false;
        this.changedBetweenResets = false;
        this.onStateChange =
            onStateChange ||
                function noop() {
                    return;
                };
        this.ignoreWaits = ignoreWaits;
        this.buffer = [];
        this.actionQueue = [];
        this.sequenceNumber = 0;
    }
    get text() {
        return Punycode.ucs2.encode(this.buffer.slice());
    }
    /**
     * Generate action deltas based on the new full state of the source text.
     *
     * The text provided here is the _entire_ source text, not a diff.
     *
     * @param text The new state of the user's text.
     */
    update(text) {
        let actions = [];
        if (text !== undefined) {
            text = text.normalize('NFC');
            const newBuffer = Punycode.ucs2.decode(text);
            actions = diff(this.buffer, newBuffer.slice());
            this.buffer = newBuffer;
            this.emitState();
        }
        const now = Date.now();
        if (this.changedBetweenResets && now - this.lastResetTime > this.resetInterval) {
            this.actionQueue = [];
            this.actionQueue.push({
                position: 0,
                text: this.text,
                type: 'insert'
            });
            this.isReset = true;
            this.lastActionTime = now;
            this.lastResetTime = now;
            this.changedBetweenResets = false;
        }
        else if (actions.length) {
            const wait = now - (this.lastActionTime || now);
            if (wait > 0 && !this.ignoreWaits) {
                this.actionQueue.push({
                    duration: wait,
                    type: 'wait'
                });
            }
            for (const action of actions) {
                this.actionQueue.push(action);
            }
            this.lastActionTime = now;
            this.changedBetweenResets = true;
        }
        else {
            this.lastActionTime = now;
        }
    }
    /**
     * Formally start an RTT session.
     *
     * Generates a random starting event sequence number.
     *
     * @param resetInterval {Milliseconds} Time to wait between using RTT reset events during editing.
     */
    start(resetInterval = this.resetInterval) {
        this.commit();
        this.isStarting = true;
        this.resetInterval = resetInterval;
        this.sequenceNumber = Math.floor(Math.random() * 10000 + 1);
        return {
            event: 'init'
        };
    }
    /**
     * Formally stops the RTT session.
     */
    stop() {
        this.commit();
        return {
            event: 'cancel'
        };
    }
    /**
     * Generate an RTT event based on queued edit actions.
     *
     * The edit actions included in the event are all those made since the last
     * time a diff was requested.
     */
    diff() {
        this.update();
        if (!this.actionQueue.length) {
            return null;
        }
        const event = {
            actions: this.actionQueue,
            seq: this.sequenceNumber++
        };
        if (this.isStarting) {
            event.event = 'new';
            this.isStarting = false;
            this.lastResetTime = Date.now();
        }
        else if (this.isReset) {
            event.event = 'reset';
            this.isReset = false;
        }
        this.actionQueue = [];
        return event;
    }
    /**
     * Reset the RTT session state to prepare for a new message text.
     */
    commit() {
        this.sequenceNumber = 0;
        this.lastActionTime = 0;
        this.actionQueue = [];
        this.buffer = [];
    }
    emitState() {
        this.onStateChange({
            text: this.text
        });
    }
}

var RTT = /*#__PURE__*/Object.freeze({
    __proto__: null,
    diff: diff,
    DisplayBuffer: DisplayBuffer,
    InputBuffer: InputBuffer
});

class ResultSetPager {
    constructor(opts) {
        var _a, _b, _c;
        this.cursor = { first: opts.before, last: opts.after };
        this.query = opts.query;
        this.direction = (_a = opts.direction) !== null && _a !== void 0 ? _a : 'forward';
        this.reverse = (_b = opts.reverse) !== null && _b !== void 0 ? _b : this.direction === 'backward';
        this.pageSize = (_c = opts.pageSize) !== null && _c !== void 0 ? _c : 20;
    }
    [Symbol.asyncIterator]() {
        return __asyncGenerator(this, arguments, function* _a() {
            let currentResults = [];
            do {
                currentResults = yield __await(this.fetchPage());
                for (const item of currentResults) {
                    yield yield __await(item);
                }
            } while (currentResults.length > 0);
        });
    }
    size() {
        return __awaiter(this, void 0, void 0, function* () {
            const { paging } = yield this.query({ max: 0 });
            return paging.count;
        });
    }
    fetchPage() {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const { results, paging } = yield this.query({
                before: this.direction === 'backward' ? (_a = this.cursor.first) !== null && _a !== void 0 ? _a : '' : undefined,
                after: this.direction === 'forward' ? this.cursor.last : undefined,
                max: this.pageSize
            });
            this.cursor = paging;
            if (this.reverse) {
                results.reverse();
            }
            return results;
        });
    }
}
function createPager(opts) {
    return new ResultSetPager(opts);
}

var RSM = /*#__PURE__*/Object.freeze({
    __proto__: null,
    ResultSetPager: ResultSetPager,
    createPager: createPager
});

const VERSION = VERSION$1;
function createClient(opts) {
    const client = new Client(opts);
    client.use(Plugins);
    return client;
}

export { Client, Constants, DataForms, JID, index$3 as JXT, index$1 as Jingle, Namespaces, RSM, RTT, index$2 as SASL, index as Stanzas, Utils, VERSION, core, createClient, getHostMeta };
